#ifndef CGUL_BASE64_H
#define CGUL_BASE64_H

/**
 * Routines to base64 encode and decode as per RFC 3548:
 *
 *     ftp://ftp.rfc-editor.org/in-notes/rfc3548.txt
 *
 * @file
 * @author Paul Serice
 * @brief base64 encode and decode
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_base64</code> instance. */
typedef struct cgul_base64* cgul_base64_t;

/**
 * Create a new cgul_base64 object.  The caller is responsible for
 * freeing the object by calling cgul_base64__delete().  If memory
 * cannot be allocated, <code>NULL</code> will be returned and an
 * exception will be thrown.
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_base64</code> instance
 */
extern CGUL_EXPORT cgul_base64_t
cgul_base64__new(cgul_exception_t* cex);

/**
 * cgul_base64__delete() frees all internally allocated memory.  Do
 * not try to access "b64" after calling this method.
 *
 * @param b64 <code>cgul_base64</code> instance
 */
extern CGUL_EXPORT void
cgul_base64__delete(cgul_base64_t b64);

/**
 * Get the line length that determines when a '\\n' character is to be
 * embedded.  A value of zero means no '\\n' character will be
 * embedded.
 *
 * @param[in] cex c-style exception
 * @param[in] b64 <code>cgul_base64</code> instance
 * @return maximum line length for encoded lines
 */
extern CGUL_EXPORT size_t
cgul_base64__get_line_length_max(cgul_exception_t* cex,
                                 cgul_base64_t b64);

/**
 * Set the line length that determines when a '\\n' character is to be
 * embedded.  A value of zero (which is the default) means no '\\n'
 * character will be embedded.  The line_length_max value only plays a
 * part when encoding binary data into a base64 text stream.
 *
 * @param[in] cex c-style exception
 * @param[in] b64 <code>cgul_base64</code> instance
 * @param[in] line_length_max maximum line length
 */
extern CGUL_EXPORT void
cgul_base64__set_line_length_max(cgul_exception_t* cex,
                                 cgul_base64_t b64,
                                 size_t line_length_max);

/**
 * Return whether this cgul_base64 object is in strict_decode mode.
 * In this mode, the exhortation in RFC 3548 to reject any
 * non-base64-alphabetic character is honored.  This includes '\\n'
 * characters.  Thus, if you want to decode MIME, you want to disable
 * strict_decode mode.  By default strict_decode mode is off.
 *
 * @param[in] cex c-style exception
 * @param[in] b64 <code>cgul_base64</code> instance
 * @return strict decode mode
 */
extern CGUL_EXPORT int
cgul_base64__get_strict_decode(cgul_exception_t* cex,
                               cgul_base64_t b64);

/**
 * Set whether this cgul_base64 object is in strict_decode mode.
 *
 * @param[in] cex c-style exception
 * @param[in] b64 <code>cgul_base64</code> instance
 * @param[in] strict_decode strict decode mode
 */
extern CGUL_EXPORT void
cgul_base64__set_strict_decode(cgul_exception_t* cex,
                               cgul_base64_t b64,
                               int strict_decode);

/**
 * Base64 encode <code>buf_size</code> bytes in <code>buf</code>.
 * Each new call to this function clears the <code>cgul_base64</code>
 * object's internal buffer except for the remainder.  So, you must
 * call <code>cgul_base64__get_encode_conversion()</code> before
 * calling this method again. <p>
 *
 * NOTE: This method is idempotent only if <code>buf_size</code> is a
 * multiple of 3.  The reason is that base64 needs exactly 3 8-bit
 * values to generate exactly 4 6-bit values.  Thus, if you feed
 * cgul_base64__encode() a buffer that has 1 or 2 bytes remaining,
 * those bytes will be internally saved and used the next time you
 * call this method.  Generally, this is the behavior you want when
 * encoding large files. <p>
 *
 * This class is designed to make it possible and efficient to encode
 * large streams of data without using large amounts of memory.
 * Because of this, it works one block at a time remembering any
 * remainder between blocks. <p>
 *
 * The downside of this is that this class is less than optimal for
 * doing one-shot encoding.  Thus, if you want to encode a single,
 * small block of memory, the least you will have to do is the
 * following: <p>
 *
 * <pre>
 *
 *     // Encode the first (and only) block.
 *     cgul_base64__encode(&cex, b64, block, block_size);
 *     s = cgul_base64__get_encode_conversion(&cex, b64, NULL);
 *     cgul_libc_printf(&cex, "%s", s);
 *
 *     // Flush the remainder.
 *     cgul_base64__encode_finish(&cex, b64);
 *     s = cgul_base64__get_encode_conversion(&cex, b64, NULL);
 *     cgul_libc_printf(&cex, "%s", s);
 *
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @param[in] b64 <code>cgul_base64</code> instance
 * @param[in] buf binary buffer to encode
 * @param[in] buf_size size in bytes of <code>buf</code>
 */
extern CGUL_EXPORT void
cgul_base64__encode(cgul_exception_t* cex,
                    cgul_base64_t b64,
                    const void* buf,
                    size_t buf_size);

/**
 * Call this method after you have fed all the data to
 * <code>cgul_base64__encode()</code>.  This method will flush the 1
 * or 2 bytes that possibly remain after calling
 * <code>cgul_base64__encode()</code>.  It will also insert padding at
 * the end of output as required by the base64 specification.  You can
 * retrieve the data by calling
 * <code>cgul_base64__get_encode_conversion()</code> as usual.  You
 * can continue to use <code>cgul_base64__encode()</code> after
 * calling this method (provided this method does not throw an
 * exception).  The internal buffer will simply be empty when you
 * start encoding data again.
 *
 * @param[in,out] cex c-style exception
 * @param[in] b64 <code>cgul_base64</code> instance
 */
extern CGUL_EXPORT void
cgul_base64__encode_finish(cgul_exception_t* cex,
                           cgul_base64_t b64);

/**
 * For efficiency, this method returns a pointer to the
 * <code>cgul_base64</code> object's internal encoding buffer.  (This
 * buffer can be treated as a valid C-style string.)  If
 * <code>buf_size</code> is not <code>NULL</code>, the count of valid
 * bytes (not including the trailing <code>NULL</code>-terminator) in
 * the buffer will be returned. <p>
 *
 * You should call this method immediately after <b>every</b> call to
 * either <code>cgul_base64__encode()</code> or
 * <code>cgul_base64__encode_finish()</code>. <p>
 *
 * The location of the internal buffer in memory can change during
 * each encoding or decoding operation.  Thus, you should always call
 * this method immediately before trying to access the internal
 * buffer.  Do not cache a pointer to the internal buffer.  Also, do
 * not call <code>free()</code> on the pointer that is returned.
 *
 * @param[in] cex c-style exception
 * @param[in] b64 <code>cgul_base64</code> instance
 * @param[out] buf_size size of the internal encoding buffer
 * @return internal encoding buffer
 */
extern CGUL_EXPORT char*
cgul_base64__get_encode_conversion(cgul_exception_t* cex,
                                   cgul_base64_t b64,
                                   size_t* buf_size);

/**
 * Base64 decode <code>buf_size</code> bytes in <code>buf</code>.
 * Each new call to this function clears the <code>cgul_base64</code>
 * object's internal buffer except for the remainder.  So, you must
 * call <code>cgul_base64__get_decode_conversion()</code> before
 * calling this method again.
 *
 * NOTE: This method is idempotent only if <code>buf_size</code> is a
 * multiple of 4.  The reason is that base64 needs exactly 4 6-bit
 * values to generate exactly 3 8-bit values.  Thus, if you feed this
 * method a buffer that has 1, 2, or 3 bytes remaining, those bytes
 * will be internally saved and used the next time you call this
 * method.  Generally, this is the behavior you want when decoding
 * large files.
 *
 * This class is designed to make it possible and efficient to decode
 * large streams of data without using large amounts of memory.
 * Because of this, it works one block at a time remembering any
 * remainder between blocks. <p>
 *
 * The downside of this is that this class is less than optimal for
 * doing one-shot decoding.  Thus, if you want to decode a single,
 * small block of memory, the least you will have to do is the
 * following: <p>
 *
 * <pre>
 *
 *     // Decode the first (and only) block.
 *     cgul_base64__decode(&cex, b64, block, block_size);
 *     s = cgul_base64__get_decode_conversion(&cex, b64, NULL);
 *     cgul_libc_printf(&cex, "%s", s);
 *
 *     // Flush the remainder.
 *     cgul_base64__decode_finish(&cex, b64);
 *     s = cgul_base64__get_decode_conversion(&cex, b64, NULL);
 *     cgul_libc_printf(&cex, "%s", s);
 *
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @param[in] b64 <code>cgul_base64</code> instance
 * @param[in] buf base64 buffer to decode
 * @param[in] buf_size size in bytes of <code>buf</code>
 */
extern CGUL_EXPORT void
cgul_base64__decode(cgul_exception_t* cex,
                    cgul_base64_t b64,
                    const void* buf,
                    size_t buf_size);

/**
 * Call this method after you have fed all the data to
 * <code>cgul_base64__decode()</code>.  This method will flush the 1,
 * 2, or 3 bytes that possibly remain after calling
 * <code>cgul_base64__decode()</code>.  You can retrieve the data by
 * calling <code>cgul_base64__get_decode_conversion()</code> as usual.
 * You can continue to use <code>cgul_base64__decode()</code> after
 * calling this method (provided this method does not throw an
 * exception).  The internal buffer will simply be empty when you
 * start decoding data again.
 *
 * @param[in,out] cex c-style exception
 * @param[in] b64 <code>cgul_base64</code> instance
 */
extern CGUL_EXPORT void
cgul_base64__decode_finish(cgul_exception_t* cex,
                           cgul_base64_t b64);

/**
 * For efficiency, this method returns a pointer to the
 * <code>cgul_base64</code> object's internal decoding buffer.  (This
 * buffer can be treated as a valid C-style string.)  If
 * <code>buf_size</code> is not <code>NULL</code>, the count of valid
 * bytes (not including the trailing <code>NULL</code>-terminator) in
 * the buffer will be returned. <p>
 *
 * You should call this method immediately after <b>every</b> call to
 * either <code>cgul_base64__decode()</code> or
 * <code>cgul_base64__decode_finish()</code>. <p>
 *
 * The location of the internal buffer in memory can change during
 * each encoding or decoding operation.  Thus, you should always call
 * this method immediately before trying to access the internal
 * buffer.  Do not cache a pointer to the internal buffer.  Also, do
 * not call <code>free()</code> on the pointer that is returned.
 *
 * @param[in] cex c-style exception
 * @param[in] b64 <code>cgul_base64</code> instance
 * @param[out] buf_size size of the internal decoding buffer
 * @return internal decoding buffer
 */
extern CGUL_EXPORT char*
cgul_base64__get_decode_conversion(cgul_exception_t* cex,
                                   cgul_base64_t b64,
                                   size_t* buf_size);

CGUL_END_C

#endif  /*  CGUL_BASE64_H  */
