#ifndef CGUL_BCD_H
#define CGUL_BCD_H

/**
 * Basic arbitrary precision arithmetic using binary-coded decimal
 * (BCD).  This is especially useful when you already have a large
 * string of binary digits (perhaps in a text file) that you want to
 * perform a few basic operations on before you write out the new
 * value.  You can do the arithmetic in binary-coded decimal without
 * having the overhead of converting to base-2 in order to do the
 * arithmetic and then back to base-10 in order to write out the
 * result (at least that is the theory).
 *
 * @file
 * @author Paul Serice
 * @brief arbitrary precision binary-coded decimal arithmetic
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_string.h"

CGUL_BEGIN_C

/*
 * Define cgul_bcd_t before including cgul_big_integer.h to resolve
 * circular dependencies.
 */

/** Opaque pointer to a <code>cgul_bcd</code> instance. */
typedef struct cgul_bcd* cgul_bcd_t;

CGUL_END_C

#include "cgul_big_integer.h"

CGUL_BEGIN_C

/**
 * Create a new <code>cgul_bcd</code> instance initialized to
 * positive 0.  The caller is responsible for calling
 * <code>cgul_bcd__delete()</code> on the value returned.  If memory
 * allocation fails, <code>NULL</code> is returned, and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_bcd</code> instance
 */
extern CGUL_EXPORT cgul_bcd_t
cgul_bcd__new(cgul_exception_t* cex);

/**
 * Create a new <code>cgul_bcd</code> instance initialized to
 * <code>value</code>.  The caller is responsible for calling
 * <code>cgul_bcd__delete()</code> on the value returned.  If memory
 * allocation fails, <code>NULL</code> is returned, and an exception
 * is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] value initial value
 * @return new <code>cgul_bcd</code> instance
 */
extern CGUL_EXPORT cgul_bcd_t
cgul_bcd__new_from_bcd(cgul_exception_t* cex,
                       cgul_bcd_t value);

/**
 * Create a new <code>cgul_bcd</code> instance initialized to
 * <code>value</code>.  The caller is responsible for calling
 * <code>cgul_bcd__delete()</code> on the value returned.  If memory
 * allocation fails, <code>NULL</code> is returned, and an exception
 * is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] value initial value
 * @return new <code>cgul_bcd</code> instance
 */
extern CGUL_EXPORT cgul_bcd_t
cgul_bcd__new_from_long(cgul_exception_t* cex,
                        long value);

/**
 * Create a new <code>cgul_bcd</code> instance initialized to
 * <code>value</code>.  The caller is responsible for calling
 * <code>cgul_bcd__delete()</code> on the value returned.  If memory
 * allocation fails, <code>NULL</code> is returned, and an exception
 * is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] value initial value
 * @return new <code>cgul_bcd</code> instance
 */
extern CGUL_EXPORT cgul_bcd_t
cgul_bcd__new_from_unsigned_long(cgul_exception_t* cex,
                                 unsigned long value);

/**
 * Create a new <code>cgul_bcd</code> instance initialized to
 * <code>value</code>.  The caller is responsible for calling
 * <code>cgul_bcd__delete()</code> on the value returned.  If memory
 * allocation fails, <code>NULL</code> is returned, and an exception
 * is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] value initial value
 * @return new <code>cgul_bcd</code> instance
 */
extern CGUL_EXPORT cgul_bcd_t
cgul_bcd__new_from_size_type(cgul_exception_t* cex,
                             size_t value);

/**
 * Create a new <code>cgul_bcd</code> instance initialized from the
 * string <code>s</code>.  The caller is responsible for calling
 * <code>cgul_bcd__delete()</code> on the value returned.  If an error
 * occurs, <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s initial value
 * @return new <code>cgul_bcd</code> instance
 */
extern CGUL_EXPORT cgul_bcd_t
cgul_bcd__new_from_string(cgul_exception_t* cex,
                          const char* s);

/**
 * Delete the <code>cgul_bcd</code> object in <code>bcd</code>.  The
 * client must not use <code>bcd</code> after calling this method.
 *
 * @param[in] bcd <code>cgul_bcd</code> instance
 */
extern CGUL_EXPORT void
cgul_bcd__delete(cgul_bcd_t bcd);

/**
 * Return 1 if <code>bcd</code> is positive.  Return 0 if it is not.
 *
 * @param[in] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @return whether <code>bcd</code> is positive
 */
extern CGUL_EXPORT int
cgul_bcd__is_positive(cgul_exception_t* cex,
                      cgul_bcd_t bcd);

/**
 * Return 1 if <code>bcd</code> is negative.  Return 0 if it is not.
 *
 * @param[in] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @return whether <code>bcd</code> is negative
 */
extern CGUL_EXPORT int
cgul_bcd__is_negative(cgul_exception_t* cex,
                      cgul_bcd_t bcd);

/**
 * Return 1 if <code>bcd</code> is zero.  Return 0 if it is not.
 *
 * @param[in] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @return whether <code>bcd</code> is zero
 */
extern CGUL_EXPORT int
cgul_bcd__is_zero(cgul_exception_t* cex,
                  cgul_bcd_t bcd);

/**
 * Set the value of <code>bcd</code> to zero.
 *
 * @param[in] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 */
extern CGUL_EXPORT void
cgul_bcd__set_to_zero(cgul_exception_t* cex,
                      cgul_bcd_t bcd);

/**
 * Return 1 if <code>bcd</code> is one.  Return 0 if it is not.
 *
 * @param[in] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @return whether <code>bcd</code> is one
 */
extern CGUL_EXPORT int
cgul_bcd__is_one(cgul_exception_t* cex,
                 cgul_bcd_t bcd);

/**
 * Set the value of <code>bcd</code> to one.
 *
 * @param[in] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 */
extern CGUL_EXPORT void
cgul_bcd__set_to_one(cgul_exception_t* cex,
                     cgul_bcd_t bcd);

/**
 * Return whether <code>bcd</code> is even.
 *
 * @param[in] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @return whether <code>bcd</code> is even
 */
extern CGUL_EXPORT int
cgul_bcd__is_even(cgul_exception_t* cex,
                  cgul_bcd_t bcd);

/**
 * Return whether <code>bcd</code> is odd.
 *
 * @param[in] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @return whether <code>bcd</code> is odd
 */
extern CGUL_EXPORT int
cgul_bcd__is_odd(cgul_exception_t* cex,
                 cgul_bcd_t bcd);

/**
 * Assign <code>rhs</code> to <code>lhs</code>.  It is safe for
 * <code>rhs</code> and <code>lhs</code> to point to the same
 * underlying <code>cgul_bcd</code> object.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs right-hand side
 */
extern CGUL_EXPORT void
cgul_bcd__assign(cgul_exception_t* cex,
                 cgul_bcd_t lhs,
                 cgul_bcd_t rhs);

/**
 * Assign long in <code>rhs</code> to <code>lhs</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs right-hand side
 */
extern CGUL_EXPORT void
cgul_bcd__assign_from_long(cgul_exception_t* cex,
                           cgul_bcd_t lhs,
                           long rhs);

/**
 * Assign unsigned long in <code>rhs</code> to <code>lhs</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs right-hand side
 */
extern CGUL_EXPORT void
cgul_bcd__assign_from_unsigned_long(cgul_exception_t* cex,
                                    cgul_bcd_t lhs,
                                    unsigned long rhs);

/**
 * Assign size type in <code>rhs</code> to <code>lhs</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs right-hand side
 */
extern CGUL_EXPORT void
cgul_bcd__assign_from_size_type(cgul_exception_t* cex,
                                cgul_bcd_t lhs,
                                size_t rhs);

/**
 * Assign the string of decimal digits in <code>rhs</code> to
 * <code>lhs</code>.  Only assigning from a string of decimal digits
 * is supported.  If you need to support other common bases, you
 * should use the <code>cgul_big_integer</code> class instead.  If an
 * error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs right-hand side
 * @return whether <code>rhs</code> parsed
 */
extern CGUL_EXPORT void
cgul_bcd__assign_from_string(cgul_exception_t* cex,
                             cgul_bcd_t lhs,
                             const char* rhs);

/**
 * Assign the negative of <code>rhs</code> to <code>lhs</code>.  It is
 * safe and efficient for <code>rhs</code> and <code>lhs</code> to
 * point to the same underlying <code>cgul_bcd</code> object.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs right-hand side
 */
extern CGUL_EXPORT void
cgul_bcd__negate(cgul_exception_t* cex,
                 cgul_bcd_t lhs,
                 cgul_bcd_t rhs);

/**
 * Assign the absolute value of <code>rhs</code> to <code>lhs</code>.
 * It is safe and efficient for <code>rhs</code> and <code>lhs</code>
 * to point to the same underlying <code>cgul_bcd</code> object.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs right-hand side
 */
extern CGUL_EXPORT void
cgul_bcd__abs(cgul_exception_t* cex,
              cgul_bcd_t lhs,
              cgul_bcd_t rhs);

/**
 * Add <code>rhs1</code> to <code>rhs2</code> and store the result in
 * <code>lhs</code>.  It is safe for <code>rhs1</code> or
 * <code>rhs2</code> and <code>lhs</code> to point to the same
 * underlying <code>cgul_bcd</code> object.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs1 first right-hand side value
 * @param[in] rhs2 second right-hand side value
 */
extern CGUL_EXPORT void
cgul_bcd__add(cgul_exception_t* cex,
              cgul_bcd_t lhs,
              cgul_bcd_t rhs1,
              cgul_bcd_t rhs2);

/**
 * Subtract <code>rhs2</code> from <code>rhs1</code> and store the
 * result in <code>lhs</code>.  It is safe for <code>rhs1</code> or
 * <code>rhs2</code> and <code>lhs</code> to point to the same
 * underlying <code>cgul_bcd</code> object.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs1 first right-hand side value
 * @param[in] rhs2 second right-hand side value
 */
extern CGUL_EXPORT void
cgul_bcd__subtract(cgul_exception_t* cex,
                   cgul_bcd_t lhs,
                   cgul_bcd_t rhs1,
                   cgul_bcd_t rhs2);

/**
 * Multiply <code>rhs1</code> with <code>rhs2</code> and store the
 * result in <code>lhs</code>.  It is safe for <code>rhs1</code> or
 * <code>rhs2</code> and <code>lhs</code> to point to the same
 * underlying <code>cgul_bcd</code> object.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs1 first right-hand side value
 * @param[in] rhs2 second right-hand side value
 */
extern CGUL_EXPORT void
cgul_bcd__multiply(cgul_exception_t* cex,
                   cgul_bcd_t lhs,
                   cgul_bcd_t rhs1,
                   cgul_bcd_t rhs2);

/**
 * Divide <code>dividend</code> by <code>divisor</code> and store the
 * result in <code>quotient</code> and <code>remainder</code>.  It is
 * safe for <code>dividend</code> or <code>divisor</code> and
 * <code>quotient</code> or <code>remainder</code> to point to the
 * same underlying <code>cgul_bcd</code> object.  This method throws
 * an exception if division by zero is attempted or if memory cannot
 * be allocated.
 *
 * @param[in,out] cex c-style exception
 * @param[out] quotient quotient
 * @param[out] remainder remainder
 * @param[in] dividend dividend
 * @param[in] divisor divisor
 */
extern CGUL_EXPORT void
cgul_bcd__divide(cgul_exception_t* cex,
                 cgul_bcd_t quotient,
                 cgul_bcd_t remainder,
                 cgul_bcd_t dividend,
                 cgul_bcd_t divisor);

/**
 * This method calculates <code>base</code> raised to the power of
 * <code>exponent</code> and saves the result in <code>lhs</code>.
 * If <code>exponent</code> is negative, an exception is thrown.  If
 * the program runs out of memory, an exception is thrown. <p>
 *
 * This method is not as effecient as
 * <code>cgul_bcd__power_modulo()</code> when the client is only
 * interested in the modulo and not the power. <p>
 *
 * @note For this implementation, pow(0,0) = 1.  This is in agreement
 * with some, but others leave it undefined. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] base base
 * @param[in] exponent exponent
 * @see cgul_bcd__power_modulo()
 */
extern CGUL_EXPORT void
cgul_bcd__power(cgul_exception_t* cex,
                cgul_bcd_t lhs,
                cgul_bcd_t base,
                cgul_bcd_t exponent);

/**
 * This method calculates the modulo of <code>base</code> to the power
 * of <code>exponent</code> using <code>modulus</code> as the
 * modulus.  If <code>exponent</code> is negative, an exception is
 * thrown.  If the program runs out of memory, an exception is
 * thrown. <p>
 *
 * This method is much more effecient than
 * <code>cgul_bcd__power()</code> when the client is only interested
 * in the modulo and not the power because the modulo operation is
 * used at each stage of the calculation to keep the size of the
 * numbers within reason. <p>
 *
 * @note For this implementation, pow(0,0) = 1.  This is in agreement
 * with some, but others leave it undefined. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] base base
 * @param[in] exponent exponent
 * @param[in] modulus modulus
 * @see cgul_bcd__power()
 */
extern CGUL_EXPORT void
cgul_bcd__power_modulo(cgul_exception_t* cex,
                       cgul_bcd_t lhs,
                       cgul_bcd_t base,
                       cgul_bcd_t exponent,
                       cgul_bcd_t modulus);

/**
 * This method uses the Miller-Rabin Primality Test to determine
 * whether <code>bcd</code> is probably prime.  The error per
 * iteration for Miller-Rabin is bound by 1/4.  This method performs
 * <code>count</code> iterations.  The following table shows the
 * theoretical upper bound for the error associated with different
 * values of <code>count</code>: <p>
 *
 * <pre>
 *     count | error
 *     ------------------------
 *        10 | 9.54e-7
 *        25 | 8.88e-16
 *        50 | 7.89e-31
 *       100 | 6.22e-61
 *       500 | 9.33e-302
 *      1000 | 8.71e-603
 * </pre>
 *
 * If an exception is not thrown, this method returns <code>1</code>
 * if <code>bcd</code> is probably prime and <code>0</code> if
 * it is definitely composite. <p>
 *
 * This method throws an exception if <code>bcd</code> is
 * negative.  It also throws an exception if the program runs out of
 * memory.  If an exception is thrown, <code>0</code> is returned. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @param[in] count number of times test is independently executed
 * @return whether <code>bcd</code> is probably prime
 */
extern CGUL_EXPORT int
cgul_bcd__is_probably_prime(cgul_exception_t* cex,
                            cgul_bcd_t bcd,
                            unsigned int count);

/**
 * Shift <code>rhs</code> left by <code>digit_count</code> and put the
 * result in <code>lhs</code>.  It is safe for <code>rhs</code> and
 * <code>lhs</code> to point to the same underlying
 * <code>cgul_bcd</code> object. <p>
 *
 * Note that this method is a base-10 shift that shifts individual
 * decimal digits.  Contrast this method with
 * <code>cgul_big_integer__shift_left()</code> which shifts individual
 * base-2 bits instead. <p>
 *
 * This method can be used as a fast way to multiply by a power of 10. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs right-hand side
 * @param[in] digit_count how far to shift <code>rhs</code>
 */
extern CGUL_EXPORT void
cgul_bcd__shift_left(cgul_exception_t* cex,
                     cgul_bcd_t lhs,
                     cgul_bcd_t rhs,
                     size_t digit_count);

/**
 * Shift <code>rhs</code> right by <code>digit_count</code> and put
 * the result in <code>lhs</code>.  It is safe for <code>rhs</code>
 * and <code>lhs</code> to point to the same underlying
 * <code>cgul_bcd</code> object. <p>
 *
 * Note that this method is a base-10 shift that shifts individual
 * decimal digits.  Contrast this method with
 * <code>cgul_big_integer__shift_right()</code> which shifts
 * individual base-2 bits instead. <p>
 *
 * This method can be used as a fast way to divide by a power of 10. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs right-hand side
 * @param[in] digit_count how far to shift <code>rhs</code>
 */
extern CGUL_EXPORT void
cgul_bcd__shift_right(cgul_exception_t* cex,
                      cgul_bcd_t lhs,
                      cgul_bcd_t rhs,
                      size_t digit_count);

/**
 * Return less than zero if <code>lhs</code> is less than
 * <code>rhs</code>, return zero if <code>lhs</code> equals
 * <code>rhs</code>, and return greater than zero if <code>lhs</code>
 * is greater than <code>rhs</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] lhs left-hand side
 * @param[in] rhs right-hand side
 * @return comparison of <code>lhs</code> and <code>rhs</code>
 */
extern CGUL_EXPORT int
cgul_bcd__compare(cgul_exception_t* cex,
                  cgul_bcd_t lhs,
                  cgul_bcd_t rhs);

/**
 * This method finds the greatest common divisor (GCD) of
 * <code>rhs1</code> and <code>rhs2</code> and saves the result in
 * <code>lhs</code>.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs1 first right-hand side value
 * @param[in] rhs2 second right-hand side value
 */
extern CGUL_EXPORT void
cgul_bcd__gcd(cgul_exception_t* cex,
              cgul_bcd_t lhs,
              cgul_bcd_t rhs1,
              cgul_bcd_t rhs2);

/**
 * This method finds the least common multiple (LCM) of
 * <code>rhs1</code> and <code>rhs2</code> and saves the result
 * in <code>lhs</code>.  The value returned is often used as the
 * lowest common denominator (LCD).  If an error occurs, an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs1 first right-hand side value
 * @param[in] rhs2 second right-hand side value
 */
extern CGUL_EXPORT void
cgul_bcd__lcm(cgul_exception_t* cex,
              cgul_bcd_t lhs,
              cgul_bcd_t rhs1,
              cgul_bcd_t rhs2);

/**
 * Swap the internal data quickly without copying.
 *
 * @param[in] cex c-style exception
 * @param[in] bcd_1 first <code>cgul_bcd</code> instance
 * @param[in] bcd_2 second <code>cgul_bcd</code> instance
 */
extern CGUL_EXPORT void
cgul_bcd__swap(cgul_exception_t* cex,
               cgul_bcd_t bcd_1,
               cgul_bcd_t bcd_2);

/**
 * Convert the value of <code>bcd</code> to a long.  This function
 * returns 1 if the conversion can be accomplished without losing any
 * significant bits; otherwise, it returns 0.
 *
 * @param[in] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @param[out] value converted value
 * @return whether a conversion occurred
 */
extern CGUL_EXPORT int
cgul_bcd__as_long(cgul_exception_t* cex,
                  cgul_bcd_t bcd,
                  long int* value);

/**
 * Convert the value of <code>bcd</code> to an unsigned long.  This
 * function returns 1 if the conversion can be accomplished without
 * losing any significant bits; otherwise, it returns 0.
 *
 * @param[in] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @param[out] value converted value
 * @return whether a conversion occurred
 */
extern CGUL_EXPORT int
cgul_bcd__as_unsigned_long(cgul_exception_t* cex,
                           cgul_bcd_t bcd,
                           unsigned long int* value);

/**
 * This method converts <code>bcd</code> to a
 * <code>cgul_big_integer</code> object and saves the conversion to
 * <code>big_integer</code> which the caller has already instantiated.
 * If an error occurs, an exception is thrown, and
 * <code>big_integer</code> is not altered.
 *
 * @param[in,out] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @param[out] big_integer <code>cgul_big_integer</code> instance
 */
extern CGUL_EXPORT void
cgul_bcd__as_big_integer(cgul_exception_t* cex,
                         cgul_bcd_t bcd,
                         cgul_big_integer_t big_integer);

/**
 * Return a <code>char*</code> object that holds the string
 * representation of <code>bcd</code> as a radix-2 number.  The caller
 * is responsible for calling <code>free()</code> on the pointer
 * returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @return string representation of <code>bcd</code> as binary
 */
extern CGUL_EXPORT char*
cgul_bcd__as_binary(cgul_exception_t* cex,
                    cgul_bcd_t bcd);

/**
 * Append the binary representation of <code>bcd</code> to
 * <code>s</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @param[in] s where to append the string representation
 */
extern CGUL_EXPORT void
cgul_bcd__append_as_binary(cgul_exception_t* cex,
                           cgul_bcd_t bcd,
                           cgul_string_t s);

/**
 * Return a <code>char*</code> that holds the string representation of
 * <code>bcd</code> as a radix-8 number.  The caller is responsible
 * for calling <code>free()</code> on the pointer returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @return string representation of <code>bcd</code> as octal
 */
extern CGUL_EXPORT char*
cgul_bcd__as_octal(cgul_exception_t* cex,
                   cgul_bcd_t bcd);

/**
 * Append the octal representation of <code>bcd</code> to
 * <code>s</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @param[in] s where to append the string representation
 */
extern CGUL_EXPORT void
cgul_bcd__append_as_octal(cgul_exception_t* cex,
                          cgul_bcd_t bcd,
                          cgul_string_t s);

/**
 * Return a <code>char*</code> object that holds the string
 * representation of the decimal digits in <code>bcd</code>.  The
 * caller is responsible for calling <code>free()</code> on the
 * pointer returned. <p>
 *
 * Producing a string in base-10 is extremely efficient because the
 * internal format of this class is binary-coded decimal. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @return string representation of <code>bcd</code> as decimal
 */
extern CGUL_EXPORT char*
cgul_bcd__as_decimal(cgul_exception_t* cex,
                     cgul_bcd_t bcd);

/**
 * Append the decimal digits in <code>bcd</code> to <code>s</code>. <p>
 *
 * Producing a string in base-10 is extremely efficient because the
 * internal format of this class is binary-coded decimal. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @param[in] s where to append the string representation
 */
extern CGUL_EXPORT void
cgul_bcd__append_as_decimal(cgul_exception_t* cex,
                            cgul_bcd_t bcd,
                            cgul_string_t s);

/**
 * Return a <code>char*</code> object that holds the string
 * representation of <code>bcd</code> as a radix-16 number.  The
 * caller is responsible for calling <code>free()</code> on the
 * pointer returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @return string representation of <code>bcd</code> as hexadecimal
 */
extern CGUL_EXPORT char*
cgul_bcd__as_hexadecimal(cgul_exception_t* cex,
                         cgul_bcd_t bcd);

/**
 * Append the hexadecimal representation of <code>bcd</code> to
 * <code>s</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @param[in] s where to append the string representation
 */
extern CGUL_EXPORT void
cgul_bcd__append_as_hexadecimal(cgul_exception_t* cex,
                                cgul_bcd_t bcd,
                                cgul_string_t s);

/**
 * Print the string representation of <code>bcd</code> as a radix-2
 * number on <code>f</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @param[in] f output file
 */
extern CGUL_EXPORT void
cgul_bcd__print_as_binary(cgul_exception_t* cex,
                          cgul_bcd_t bcd,
                          FILE* f);

/**
 * Print the string representation of <code>bcd</code> as a radix-8
 * number on <code>f</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @param[in] f output file
 */
extern CGUL_EXPORT void
cgul_bcd__print_as_octal(cgul_exception_t* cex,
                         cgul_bcd_t bcd,
                         FILE* f);

/**
 * Print the string representation of <code>bcd</code> as a radix-10
 * number on <code>f</code>. <p>
 *
 * Producing a string in base-10 is extremely efficient because the
 * internal format of this class is binary-coded decimal. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @param[in] f output file
 */
extern CGUL_EXPORT void
cgul_bcd__print_as_decimal(cgul_exception_t* cex,
                           cgul_bcd_t bcd,
                           FILE* f);

/**
 * Print the string representation of <code>bcd</code> as a radix-16
 * number on <code>f</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 * @param[in] f output file
 */
extern CGUL_EXPORT void
cgul_bcd__print_as_hexadecimal(cgul_exception_t* cex,
                               cgul_bcd_t bcd,
                               FILE* f);

CGUL_END_C

#endif  /*  CGUL_BCD_H  */
