#ifndef CGUL_BITWISE_TRIE_NODE_H
#define CGUL_BITWISE_TRIE_NODE_H

/**
 * This class implements a bitwise trie node which holds one
 * index/value pair.
 *
 * @file
 * @author Paul Serice
 * @brief bitwise trie node
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_bitwise_trie_node</code> instance. */
typedef struct cgul_bitwise_trie_node* cgul_bitwise_trie_node_t;

/*
 * The constructor and destructor are both private.  Only the
 * <code>cgul_bitwise_trie</code> class should create are destroy its
 * nodes.
 */

/**
 * Return the index stored in this node.  If you want to set the value
 * of the index, there is no method to do so because changing the
 * value of a index requires changing the node's position in the
 * <code>cgul_bitwise_trie</code>.  Thus, if you need to change the
 * value of a index, you should delete the node from the
 * <code>cgul_bitwise_trie</code> and insert a new node with the new
 * value for the index.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_bitwise_trie_node</code> instance
 * @return index
 */
extern CGUL_EXPORT unsigned long int
cgul_bitwise_trie_node__get_index(cgul_exception_t* cex,
                                  cgul_bitwise_trie_node_t n);

/**
 * Return the value stored in this node.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_bitwise_trie_node</code> instance
 * @return value
 */
extern CGUL_EXPORT void*
cgul_bitwise_trie_node__get_value(cgul_exception_t* cex,
                                  cgul_bitwise_trie_node_t n);

/**
 * Set the value stored in this node.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_bitwise_trie_node</code> instance
 * @param[in] value
 */
extern CGUL_EXPORT void
cgul_bitwise_trie_node__set_value(cgul_exception_t* cex,
                                  cgul_bitwise_trie_node_t n,
                                  const void* value);

/*
 * NOTE: get_next() and get_prev() do not exist because the trie is
 * (generally) a sparse data structure with no direct links from the
 * current to the next or previous node.  If the trie is being used
 * like an array, simply increment or decrement your index and search
 * the trie again with the new index value.
 */

/**
 * Return the next older node.  Together with
 * <code>cgul_bitwise_trie_node__get_younger()</code>, this method
 * lets you traverse the tree in chronological order. If there is no
 * older node, this method returns <code>NULL</code>. <p>
 *
 * The following example shows how to iterate over the entire tree in
 * reverse chronological order.  Notice that you have to start with
 * the youngest node when calling
 * <code>cgul_bitwise_trie_node__get_older()</code>:
 *
 * <pre>
 *     cgul_bitwise_trie_node_t n = cgul_bitwise_trie__get_youngest(cex, t);
 *     for ( ; n ; n = cgul_bitwise_trie_node__get_older(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_bitwise_trie_node</code> instance
 * @return older node
 */
extern CGUL_EXPORT cgul_bitwise_trie_node_t
cgul_bitwise_trie_node__get_older(cgul_exception_t* cex,
                                  cgul_bitwise_trie_node_t n);

/**
 * Return the next younger node.  Together with
 * <code>cgul_bitwise_trie_node__get_older()</code>, this method lets
 * you traverse the tree in chronological order. If there is no
 * younger node, this method returns <code>NULL</code>. <p>
 *
 * The following example shows how to iterate over the entire tree in
 * chronological order.  Notice that you have to start with the oldest
 * node when calling
 * <code>cgul_bitwise_trie_node__get_younger()</code>:
 *
 * <pre>
 *     cgul_bitwise_trie_node_t n = cgul_bitwise_trie__get_oldest(cex, t);
 *     for ( ; n ; n = cgul_bitwise_trie_node__get_younger(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_bitwise_trie_node</code> instance
 * @return younger node
 */
extern CGUL_EXPORT cgul_bitwise_trie_node_t
cgul_bitwise_trie_node__get_younger(cgul_exception_t* cex,
                                    cgul_bitwise_trie_node_t n);

CGUL_END_C

#endif  /*  CGUL_BITWISE_TRIE_NODE_H  */
