#ifndef CGUL_BLOCK_READER_MEMORY_CXX_H
#define CGUL_BLOCK_READER_MEMORY_CXX_H

/**
 * C++ wrapper for <code>cgul_block_reader_memory</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_block_reader_memory</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_block_reader_cxx.h"
#include "cgul_block_reader_memory.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_block_reader_memory</code> objects.  The main purpose of
 * this class is to convert the C-style function calls and exception
 * handling in <code>cgul_block_reader_memory</code> into C++-style
 * function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_block_reader_memory</code>
 */
class cgul_block_reader_memory_cxx : public cgul_block_reader_cxx {

public:

    /**
     * Create a new <code>cgul_block_reader_memory</code> instance.
     * After this method returns, <code>open()</code> should be called
     * before calling <code>read_next_block()</code>.  The client is
     * responsible for deleting the object returned.  If an error
     * occurs, an exception is thrown.
     */
    cgul_block_reader_memory_cxx()
    {
        cgul_exception_t local = NULL;
        obj = cgul_block_reader_memory__new(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }
    
    /**
     * This method creates a new instance and calls
     * <code>open()</code> passing it <code>buffer</code> and
     * <code>buffer_size</code>.  This class does not take ownership
     * of <code>buffer</code> so the client is still responsible for
     * freeing <code>buffer</code> if necessary.  The client is also
     * responsible for deleting the object returned.  If an error
     * occurs, an exception is thrown.
     *
     * @param[in] buffer memory buffer
     * @param[in] buffer_size size of <code>buffer</code> in bytes
     */
    cgul_block_reader_memory_cxx(void* buffer,
                                 size_t buffer_size)
    {
        cgul_exception_t local = NULL;
        obj = cgul_block_reader_memory__new_from_memory(&local,
                                                        buffer,
                                                        buffer_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }
    
    /**
     * This virtual method deletes the instance freeing all internally
     * allocated resources.
     */
    virtual
    ~cgul_block_reader_memory_cxx()
    {
        if (obj) {
            cgul_block_reader_memory__delete(obj);
            obj = NULL;
        }
    }
    
    /**
     * Set the input buffer <code>buffer</code> to use when reading
     * blocks of data from memory.  If a memory buffer is already open, it
     * will be closed before this method attempts to open the new memory
     * buffer.  This class does not take ownership of
     * <code>buffer</code>.  Thus, the client is still responsible for
     * freeing <code>buffer</code> if necessary.  If an error occurs, an
     * exception is thrown.
     *
     * @param[in] buffer input memory buffer
     * @param[in] buffer_size size of <code>buffer</code> in bytes
     */
    virtual void
    open(void* buffer,
         size_t buffer_size)
    {
        cgul_exception_t local = NULL;
        cgul_block_reader_memory__open(&local, obj, buffer, buffer_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }
    
    /**
     * Close the memory.  After this method returns,
     * <code>open()</code> should be called before calling any other
     * method.
     */
    virtual void
    close()
    {
        cgul_exception_t local = NULL;
        cgul_block_reader_memory__close(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }
    
    /**
     * Virtual method that reads one block of data <code>block</code> of
     * size <code>block_size</code> bytes from the underlying buffer.  The
     * user is responsible for making sure <code>block</code> points to at
     * least <code>block_size</code> contiguous bytes of memory.  The user
     * is also responsible for making sure <code>block</code> does not
     * overlap with the open buffer from which blocks are read (as
     * required by <code>memcpy()</code>).  This method returns the number
     * of bytes read or <code>0</code> if EOF is reached without error.
     * If an error occurs, <code>0</code> is returned, and an exception is
     * thrown.
     *
     * @param[in] block block
     * @param[in] block_size block size in bytes
     * @return number of bytes read
     */
    virtual size_t
    read_next_block(void* block,
                    size_t block_size)
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_block_reader_memory__read_next_block(&local,
                                                       obj,
                                                       block,
                                                       block_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }
    
    /**
     * Virtual method that returns the presentation name.
     *
     * @return presentation name
     */
    virtual const char*
    get_presentation_name() const
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_block_reader_memory__get_presentation_name(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }
    
    /**
     * Virtual method that sets the presentation name.  The presentation
     * name is primarily used for error reporting and can be set to
     * anything, but it does need to be set again if the input file
     * changes.  If the presentation name is not set, <code>"MEMORY"</code>
     * is returned.  An exception is thrown if an attempt is made to set
     * the presentation name without first setting the input file.
     *
     * @param[in] presentation_name presentation name
     */
    virtual void
    set_presentation_name(const char* presentation_name)
    {
        cgul_exception_t local = NULL;
        cgul_block_reader_memory__set_presentation_name(&local,
                                                        obj,
                                                        presentation_name);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }
    
    /**
     * Return the currently opened buffer from which blocks are read.  The
     * client must not do anything that would invalidate the pointer or
     * corrupt the associated data stream.  This method throws an
     * exception only if a file is not currently open. <p>
     *
     * Because the client is responsible for freeing the buffer if
     * necessary, it can be convenient to call this method just before
     * deleting the object in order to free the pointer returned. <p>
     *
     * @return currently opened file or memory
     */
    virtual void*
    get_buffer() const
    {
        void* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_block_reader_memory__get_buffer(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the underlying <code>cgul_block_reader</code>
     * object. <p>
     *
     * @note <code>get_obj()</code> is similar to this method except
     * it returns a pointer to the most derived class; whereas, this
     * method returns the same pointer but cast to the base class. <p>
     *
     * @return underlying <code>cgul_block_reader</code> object
     */
    virtual cgul_block_reader_t
    as_block_reader() const
    {
        cgul_block_reader_t rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_block_reader_memory__as_block_reader(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the underlying <code>cgul_block_reader_memory</code> object.
     *
     * @return underlying object
     */
    virtual cgul_block_reader_memory_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_block_reader_memory</code> object.
     * This means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_block_reader_memory__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_block_reader_memory_t
    take_obj()
    {
        cgul_block_reader_memory_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_block_reader_memory_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_block_reader_memory__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_block_reader_memory_cxx(const cgul_block_reader_memory_cxx&);

    cgul_block_reader_memory_cxx&
    operator=(const cgul_block_reader_memory_cxx&);

    /* Underlying <code>cgul_block_reader_memory</code> object. */
    cgul_block_reader_memory_t obj;

};

#endif  /*  CGUL_BLOCK_READER_MEMORY_CXX_H  */
