#ifndef CGUL_CHAR_H
#define CGUL_CHAR_H

/**
 * This file defines the basic set of functions related to the
 * <code>char</code> type. <p>
 *
 * @file
 * @author Paul Serice
 * @brief cgul character support
 * @see cgul_hchar
 * @see cgul_wchar
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/**
 * Perform a case-significant, 32-bit, string comparison of
 * <code>s1</code> and <code>s2</code>.  Return -1, 0, or 1 if
 * <code>s1</code> is less than, equal to, or greater than
 * <code>s2</code> respectively. <p>
 *
 * For simplicity, this function compares strings stricly by the
 * ordinal value of each character which should be sufficient, for
 * example, when comparing strings when inserting them as keys into a
 * <code>cgul_rbtree</code>; however, the comparison results are not
 * likely to be the same as what <code>strcoll()</code> would return
 * if used on the UTF-8 version of the strings. <p>
 *
 * @note This function does not take a <code>cgul_exception</code>
 * object as its first parameter in order to make it easier to use the
 * function as the comparison function of <code>cgul_rbtree</code>
 * objects. <p>
 *
 * @param[in] s1 left-hand side
 * @param[in] s2 right-hand side
 * @return -1, 0, or 1 if <code>s1</code> is less than, equal to, or
 * greater than <code>s2</code> respectively
 * @see cgul_string__compare()
 */
extern CGUL_EXPORT int
cgul_char__strcmp(const char* s1,
                  const char* s2);

/**
 * Return whether the character <code>c</code> is considered to be
 * white-space.  To avoid locale dependencies, white-space is defined
 * as any character from the following list: ' ', '\\t', '\\n', '\\r',
 * '\\f', '\\v'
 *
 * @param[in] cex c-style exception
 * @param[in] c character
 * @return whether the character is white-space
 */
extern CGUL_EXPORT int
cgul_char__isspace(cgul_exception_t* cex,
                   char c);

CGUL_END_C

#endif  /*  CGUL_CHAR_H  */
