#ifndef CGUL_CONVSPEC_H
#define CGUL_CONVSPEC_H

/**
 * printf()-style conversion specifier.  This class is used to
 * serialize and deserialize printf()-style conversion specifications.
 * It is used internally by cgul_string__append_sprintf() and
 * cgul_varsub_formatted(), but it can be useful when you need custom,
 * formatted output. <p>
 *
 * @file
 * @author Paul Serice
 * @brief printf()-style conversion specifier
 */

#include "cgul_exception.h"
#include "cgul_string.h"

CGUL_BEGIN_C

/** Opaque pointer to <code>cgul_convspec</code> instance. */
typedef struct cgul_convspec* cgul_convspec_t;

/**
 * Create a new <code>cgul_convspec</code> object that is
 * initialized to the default state.  The caller is responsible for
 * freeing the object by calling
 * <code>cgul_convspec__delete()</code>.  If memory cannot be
 * allocated, <code>NULL</code> is returned, and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_convspec</code> instance
 */
extern CGUL_EXPORT cgul_convspec_t
cgul_convspec__new(cgul_exception_t* cex);

/**
 * This method deletes <code>cs</code> by freeing all internally
 * allocated resources associated with the object.  The caller must
 * not use <code>cs</code> after calling this method.
 *
 * @param[in] cs <code>cgul_convspec</code> instance
 */
extern CGUL_EXPORT void
cgul_convspec__delete(cgul_convspec_t cs);

/**
 * This method performs a deep copy of <code>rhs</code> so that
 * <code>lhs</code> has the same state.  If an error occurs, an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] lhs left-hand side
 * @param[in] rhs right-hand side
 */
extern CGUL_EXPORT void
cgul_convspec__copy(cgul_exception_t* cex,
                    cgul_convspec_t lhs,
                    cgul_convspec_t rhs);

/**
 * This method assumes the conversion specification begins with a '%'
 * character at <code>format[index]</code> and increments
 * <code>index</code> for each character that is consumed such that
 * when this method returns <code>index</code> points to one character
 * beyond the end of the conversion specification.  This is the
 * primary method to use when you want to initialize an instance of
 * this class. <p>
 *
 * Note that serializing an object of this class produces a conversion
 * specification string that is suitable for use as part of a format
 * string that can be passed to <code>printf()</code>.  Thus,
 * deserializing an instance of this class reduces a
 * <code>printf()</code> conversion specification into its component
 * parts. <p>
 *
 * If an error occurs, an exception is thrown, and <code>cs</code>
 * will be in an undefined state. <p>
 *
 * It is important that you call this method correctly.  Because
 * <code>index</code> is a pointer to a variable, you must define and
 * use a variable:
 *
 * <pre>
 *     unsigned long int index = 0;
 *     cgul_convspec__deserialize(cex, cs, "%d", &index);
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in] format format string
 * @param[in,out] index index into <code>format</code>
 */
extern CGUL_EXPORT void
cgul_convspec__deserialize(cgul_exception_t* cex,
                           cgul_convspec_t cs,
                           const char* format,
                           unsigned long int* index);

/**
 * This method serializes <code>cs</code> to a string and appends the
 * result to <code>s</code>.  Note that serializing an object of this
 * class produces a conversion specification string that is suitable
 * for use as part of a format string that can be passed to
 * <code>printf()</code>. <p>
 *
 * If an error occurs, an exception is thrown.  An error can occur if
 * the system runs out of memory while inserting into <code>s</code>.
 * In this case, <code>s</code> will be in an undefined state. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in,out] s string
 */
extern CGUL_EXPORT void
cgul_convspec__serialize(cgul_exception_t* cex,
                         cgul_convspec_t cs,
                         cgul_string_t s);

/**
 * Get whether the '#' flag is present.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @return whether the '#' flag is present
 */
extern CGUL_EXPORT int
cgul_convspec__get_flag_hash(cgul_exception_t* cex,
                             cgul_convspec_t cs);

/**
 * Set whether the '#' flag is present.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in] flag_hash whether the '#' flag is present
 */
extern CGUL_EXPORT void
cgul_convspec__set_flag_hash(cgul_exception_t* cex,
                             cgul_convspec_t cs,
                             int flag_hash);

/**
 * Get whether the '0' flag is present.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @return whether the '0' flag is present
 */
extern CGUL_EXPORT int
cgul_convspec__get_flag_zero(cgul_exception_t* cex,
                             cgul_convspec_t cs);

/**
 * Set whether the '0' flag is present.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in] flag_zero whether the '0' flag is present
 */
extern CGUL_EXPORT void
cgul_convspec__set_flag_zero(cgul_exception_t* cex,
                             cgul_convspec_t cs,
                             int flag_zero);

/**
 * Get whether the '-' flag is present.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @return whether the '-' flag is present
 */
extern CGUL_EXPORT int
cgul_convspec__get_flag_dash(cgul_exception_t* cex,
                             cgul_convspec_t cs);

/**
 * Set whether the '-' flag is present.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in] flag_dash whether the '-' flag is present
 */
extern CGUL_EXPORT void
cgul_convspec__set_flag_dash(cgul_exception_t* cex,
                             cgul_convspec_t cs,
                             int flag_dash);

/**
 * Get whether the ' ' flag is present.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @return whether the ' ' flag is present
 */
extern CGUL_EXPORT int
cgul_convspec__get_flag_space(cgul_exception_t* cex,
                              cgul_convspec_t cs);

/**
 * Set whether the ' ' flag is present.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in] flag_space whether the ' ' flag is present
 */
extern CGUL_EXPORT void
cgul_convspec__set_flag_space(cgul_exception_t* cex,
                              cgul_convspec_t cs,
                              int flag_space);

/**
 * Get whether the '+' flag is present.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @return whether the '+' flag is present
 */
extern CGUL_EXPORT int
cgul_convspec__get_flag_plus(cgul_exception_t* cex,
                             cgul_convspec_t cs);

/**
 * Set whether the '+' flag is present.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in] flag_plus whether the '+' flag is present
 */
extern CGUL_EXPORT void
cgul_convspec__set_flag_plus(cgul_exception_t* cex,
                             cgul_convspec_t cs,
                             int flag_plus);

/**
 * Whether the field width is present.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @return whether the field width is present
 */
extern CGUL_EXPORT int
cgul_convspec__get_is_field_width_present(cgul_exception_t* cex,
                                          cgul_convspec_t cs);

/**
 * Set whether the field width is present.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in] is_field_width_present whether the field width is present
 */
extern CGUL_EXPORT void
cgul_convspec__set_is_field_width_present(cgul_exception_t* cex,
                                          cgul_convspec_t cs,
                                          int is_field_width_present);

/**
 * Whether the field width is on the stack.  This occurs when the
 * conversion specifier uses '*' instead of an actual width.  It makes
 * no sense to call this function if
 * <code>cgul_convspec__get_is_field_width_present()</code>
 * returns false.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @return whether the field width is on the stack
 */
extern CGUL_EXPORT int
cgul_convspec__get_is_field_width_on_stack(cgul_exception_t* cex,
                                           cgul_convspec_t cs);

/**
 * Set whether the field width is on the stack.  This occurs when the
 * conversion specifier uses '*' instead of an actual width.  Calling
 * this method with <code>is_field_width_on_stack</code> set to true
 * does not implicitly mark the field width as present.  You will
 * still need to call
 * <code>cgul_convspec__set_is_field_width_present()</code> if
 * that is the intent.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in] is_field_width_on_stack whether the field width is on stack
 */
extern CGUL_EXPORT void
cgul_convspec__set_is_field_width_on_stack(cgul_exception_t* cex,
                                           cgul_convspec_t cs,
                                           int is_field_width_on_stack);

/**
 * Get the field width.  It makes no sense to call this function if
 * <code>cgul_convspec__get_is_field_width_present()</code>
 * returns false or
 * <code>cgul_convspec__get_is_field_width_on_stack()</code>
 * returns true.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @return field width
 */
extern CGUL_EXPORT int
cgul_convspec__get_field_width(cgul_exception_t* cex,
                               cgul_convspec_t cs);

/**
 * Set the field width.  Calling this method does not implicitly mark
 * the field width as present.  You will still need to call
 * <code>cgul_convspec__set_is_field_width_present()</code> if
 * that is the intent.  It also does not implicitly mark the field
 * width as not being on the stack.  You will still need to call
 * <code>cgul_convspec__set_is_field_width_on_stack()</code> if
 * that is the intent.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in] field_width field width
 */
extern CGUL_EXPORT void
cgul_convspec__set_field_width(cgul_exception_t* cex,
                               cgul_convspec_t cs,
                               int field_width);

/**
 * Whether the precision is present.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @return whether the precision is present
 */
extern CGUL_EXPORT int
cgul_convspec__get_is_precision_present(cgul_exception_t* cex,
                                        cgul_convspec_t cs);

/**
 * Set whether the precision is present.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in] is_precision_present whether the precision is present
 */
extern CGUL_EXPORT void
cgul_convspec__set_is_precision_present(cgul_exception_t* cex,
                                        cgul_convspec_t cs,
                                        int is_precision_present);

/**
 * Whether the precision is on the stack.  This occurs when the
 * conversion specifier uses '*' instead of an actual precision.  It
 * makes no sense to call this function if
 * <code>cgul_convspec__get_is_precision_present()</code>
 * returns false.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @return whether the precision is on the stack
 */
extern CGUL_EXPORT int
cgul_convspec__get_is_precision_on_stack(cgul_exception_t* cex,
                                         cgul_convspec_t cs);

/**
 * Set whether the precision is on the stack.  This occurs when the
 * conversion specifier uses '*' instead of an actual precision.
 * Calling this method with <code>is_precision_on_stack</code> set to
 * true does not implicitly mark the precision as present.  You will
 * still need to call
 * <code>cgul_convspec__set_is_precision_present()</code> if
 * that is the intent.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in] is_precision_on_stack whether the precision is on stack
 */
extern CGUL_EXPORT void
cgul_convspec__set_is_precision_on_stack(cgul_exception_t* cex,
                                         cgul_convspec_t cs,
                                         int is_precision_on_stack);

/**
 * Get the precision.  It makes no sense to call this function if
 * <code>cgul_convspec__get_is_precision_present()</code>
 * returns false or
 * <code>cgul_convspec__get_is_precision_on_stack()</code>
 * returns true.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @return precision
 */
extern CGUL_EXPORT int
cgul_convspec__get_precision(cgul_exception_t* cex,
                             cgul_convspec_t cs);

/**
 * Set the precision.  Calling this method does not implicitly mark
 * the precision as present.  You will still need to call
 * <code>cgul_convspec__set_is_precision_present()</code> if
 * that is the intent.  It also does not implicitly mark the precision
 * as not being on the stack.  You will still need to call
 * <code>cgul_convspec__set_is_precision_on_stack()</code> if
 * that is the intent.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in] precision precision
 */
extern CGUL_EXPORT void
cgul_convspec__set_precision(cgul_exception_t* cex,
                             cgul_convspec_t cs,
                             int precision);

/**
 * Whether the length modifier is present.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @return whether the length modifier is present
 */
extern CGUL_EXPORT int
cgul_convspec__get_is_length_modifier_present(cgul_exception_t* cex,
                                              cgul_convspec_t cs);

/**
 * Set whether the length modifier is present.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in] is_length_modifier_present whether the length modifier
 * is present
 */
extern CGUL_EXPORT void
cgul_convspec__set_is_length_modifier_present(cgul_exception_t* cex,
                                              cgul_convspec_t cs,
                                              int is_length_modifier_present);

/**
 * Get the length modifier.  It makes no sense to call this function
 * if <code>cgul_convspec__get_is_length_modifier_present()</code>
 * returns false.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @return length modifier
 */
extern CGUL_EXPORT const char*
cgul_convspec__get_length_modifier(cgul_exception_t* cex,
                                   cgul_convspec_t cs);

/**
 * Set the length modifier.  Calling this method does not implicitly
 * mark the length modifier as present.  You will still need to call
 * <code>cgul_convspec__set_is_length_modifier_present()</code>
 * if that is the intent.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in] length_modifier length modifier
 */
extern CGUL_EXPORT void
cgul_convspec__set_length_modifier(cgul_exception_t* cex,
                                   cgul_convspec_t cs,
                                   const char* length_modifier);

/**
 * Get the conversion specifier.  For example, this is the 'd'
 * character in the "%d" conversion specification.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @return conversion specifier
 */
extern CGUL_EXPORT char
cgul_convspec__get_cspec(cgul_exception_t* cex,
                         cgul_convspec_t cs);

/**
 * Set the conversion specifier.  For example, this is the 'd'
 * character in the "%d" conversion specification.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in] cspec conversion specifier
 */
extern CGUL_EXPORT void
cgul_convspec__set_cspec(cgul_exception_t* cex,
                         cgul_convspec_t cs,
                         char cspec);

CGUL_END_C

#endif  /*  CGUL_CONVSPEC_H  */
