#ifndef CGUL_CONVSPEC_CXX_H
#define CGUL_CONVSPEC_CXX_H

/**
 * C++ wrapper for <code>cgul_convspec</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_convspec</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_convspec.h"
#include "cgul_string_cxx.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_convspec</code> objects.  The main purpose of this class
 * is to convert the C-style function calls and exception handling in
 * <code>cgul_convspec</code> into C++-style function calls and
 * exception handling.
 *
 * @brief C++ bindings for <code>cgul_convspec</code>
 * @see cgul_convspec_t
 */
class cgul_convspec_cxx {

public:

    /**
     * Default constructor.
     */
    cgul_convspec_cxx()
    {
        cgul_exception_t local = NULL;
        obj = cgul_convspec__new(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Copy constructor.
     */
    cgul_convspec_cxx(const cgul_convspec_cxx& rhs)
    {
        cgul_exception_t local = NULL;
        obj = cgul_convspec__new(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        this->copy(rhs);
    }

    /**
     * Adaptor constructor.
     */
    cgul_convspec_cxx(cgul_convspec_t rhs)
    {
        cgul_exception_t local = NULL;
        obj = cgul_convspec__new(&local);
        cgul_convspec__copy(&local, obj, rhs);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method deletes the object by freeing all internally
     * allocated resources associated with the object.  The caller must
     * not use the object after calling this method.
     */
    virtual
    ~cgul_convspec_cxx()
    {
        if (obj) {
            cgul_convspec__delete(obj);
            obj = NULL;
        }
    }

    /**
     * operator=()
     *
     * @param[in] rhs right-hand side
     * @return <code>*this</code>
     */
    virtual cgul_convspec_cxx&
    operator=(const cgul_convspec_cxx& rhs)
    {
        this->copy(rhs);
        return *this;
    }

    /**
     * operator=()
     *
     * @param[in] rhs right-hand side
     * @return <code>*this</code>
     */
    virtual cgul_convspec_cxx&
    operator=(cgul_convspec_t rhs)
    {
        cgul_exception_t local = NULL;
        cgul_convspec__copy(&local, obj, rhs);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return *this;
    }

    /**
     * This method performs a deep copy of <code>rhs</code> so that
     * <code>this</code> has the same state.  If an error occurs, an
     * exception is thrown.
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    copy(const cgul_convspec_cxx& rhs)
    {
        cgul_exception_t local = NULL;
        cgul_convspec__copy(&local, this->obj, rhs.obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method assumes the conversion specification begins with a '%'
     * character at <code>format[index]</code> and increments
     * <code>index</code> for each character that is consumed such that
     * when this method returns <code>index</code> points to one character
     * beyond the end of the conversion specification.  This is the
     * primary method to use when you want to initialize an instance of
     * this class. <p>
     *
     * Note that serializing an object of this class produces a conversion
     * specification string that is suitable for use as part of a format
     * string that can be passed to <code>printf()</code>.  Thus,
     * deserializing an instance of this class reduces a
     * <code>printf()</code> conversion specification into its component
     * parts. <p>
     *
     * If an error occurs, an exception is thrown, and the object
     * will be in an undefined state. <p>
     *
     * It is important that you call this method correctly.  Because
     * <code>index</code> is a pointer to a variable, you must define and
     * use a variable:
     *
     * <pre>
     *     unsigned long int index = 0;
     *     cs.deserialize("%d", &index);
     * </pre>
     *
     * @param[in] format format string
     * @param[in,out] index index into <code>format</code>
     */
    virtual void
    deserialize(const char* format,
                unsigned long int* index)
    {
        cgul_exception_t local = NULL;
        cgul_convspec__deserialize(&local, obj, format, index);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method serializes the object to a string and appends the
     * result to <code>s</code>.  Note that serializing an object of this
     * class produces a conversion specification string that is suitable
     * for use as part of a format string that can be passed to
     * <code>printf()</code>. <p>
     *
     * If an error occurs, an exception is thrown.  An error can occur if
     * the system runs out of memory while inserting into <code>s</code>.
     * In this case, <code>s</code> will be in an undefined state. <p>
     *
     * @param[in,out] s string
     */
    virtual void
    serialize(cgul_string_cxx& s)
    {
        cgul_exception_t local = NULL;
        cgul_convspec__serialize(&local, obj, s.get_obj());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get whether the '#' flag is present.
     *
     * @return whether the '#' flag is present
     */
    virtual int
    get_flag_hash() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_convspec__get_flag_hash(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set whether the '#' flag is present.
     *
     * @param[in] flag_hash whether the '#' flag is present
     */
    virtual void
    set_flag_hash(int flag_hash)
    {
        cgul_exception_t local = NULL;
        cgul_convspec__set_flag_hash(&local, obj, flag_hash);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get whether the '0' flag is present.
     *
     * @return whether the '0' flag is present
     */
    virtual int
    get_flag_zero() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_convspec__get_flag_zero(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set whether the '0' flag is present.
     *
     * @param[in] flag_zero whether the '0' flag is present
     */
    virtual void
    set_flag_zero(int flag_zero)
    {
        cgul_exception_t local = NULL;
        cgul_convspec__set_flag_zero(&local, obj, flag_zero);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get whether the '-' flag is present.
     *
     * @return whether the '-' flag is present
     */
    virtual int
    get_flag_dash() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_convspec__get_flag_dash(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set whether the '-' flag is present.
     *
     * @param[in] flag_dash whether the '-' flag is present
     */
    virtual void
    set_flag_dash(int flag_dash)
    {
        cgul_exception_t local = NULL;
        cgul_convspec__set_flag_dash(&local, obj, flag_dash);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get whether the ' ' flag is present.
     *
     * @return whether the ' ' flag is present
     */
    virtual int
    get_flag_space() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_convspec__get_flag_space(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set whether the ' ' flag is present.
     *
     * @param[in] flag_space whether the ' ' flag is present
     */
    virtual void
    set_flag_space(int flag_space)
    {
        cgul_exception_t local = NULL;
        cgul_convspec__set_flag_space(&local, obj, flag_space);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get whether the '+' flag is present.
     *
     * @return whether the '+' flag is present
     */
    virtual int
    get_flag_plus() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_convspec__get_flag_plus(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set whether the '+' flag is present.
     *
     * @param[in] flag_plus whether the '+' flag is present
     */
    virtual void
    set_flag_plus(int flag_plus)
    {
        cgul_exception_t local = NULL;
        cgul_convspec__set_flag_plus(&local, obj, flag_plus);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Whether the field width is present.
     *
     * @return whether the field width is present
     */
    virtual int
    get_is_field_width_present() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_convspec__get_is_field_width_present(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set whether the field width is present.
     *
     * @param[in] is_field_width_present whether the field width is present
     */
    virtual void
    set_is_field_width_present(int is_field_width_present)
    {
        cgul_exception_t local = NULL;
        cgul_convspec__set_is_field_width_present(
            &local, obj, is_field_width_present);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Whether the field width is on the stack.  This occurs when the
     * conversion specifier uses '*' instead of an actual width.  It
     * makes no sense to call this function if
     * <code>get_is_field_width_present()</code> returns
     * <code>false</code>.
     *
     * @return whether the field width is on the stack
     */
    virtual int
    get_is_field_width_on_stack() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_convspec__get_is_field_width_on_stack(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set whether the field width is on the stack.  This occurs when
     * the conversion specifier uses '*' instead of an actual width.
     * Calling this method with <code>is_field_width_on_stack</code>
     * set to <code>true</code> does not implicitly mark the field
     * width as present.  You will still need to call
     * <code>set_is_field_width_present()</code> if that is the
     * intent.
     *
     * @param[in] is_field_width_on_stack whether the field width is on stack
     */
    virtual void
    set_is_field_width_on_stack(int is_field_width_on_stack)
    {
        cgul_exception_t local = NULL;
        cgul_convspec__set_is_field_width_on_stack(
            &local, obj, is_field_width_on_stack);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the field width.  It makes no sense to call this function
     * if <code>get_is_field_width_present</code> returns
     * <code>false</code> or
     * <code>get_is_field_width_on_stack()</code> returns
     * <code>true</code>.
     *
     * @return field width
     */
    virtual int
    get_field_width() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_convspec__get_field_width(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the field width.  Calling this method does not implicitly
     * mark the field width as present.  You will still need to call
     * <code>set_is_field_width_present()</code> if that is the
     * intent.  It also does not implicitly mark the field width as
     * not being on the stack.  You will still need to call
     * <code>set_is_field_width_on_stack()</code> if that is the
     * intent.
     *
     * @param[in] field_width field width
     */
    virtual void
    set_field_width(int field_width)
    {
        cgul_exception_t local = NULL;
        cgul_convspec__set_field_width(&local, obj, field_width);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Whether the precision is present.
     *
     * @return whether the precision is present
     */
    virtual int
    get_is_precision_present() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_convspec__get_is_precision_present(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set whether the precision is present.
     *
     * @param[in] is_precision_present whether the precision is present
     */
    virtual void
    set_is_precision_present(int is_precision_present)
    {
        cgul_exception_t local = NULL;
        cgul_convspec__set_is_precision_present(
            &local, obj, is_precision_present);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Whether the precision is on the stack.  This occurs when the
     * conversion specifier uses '*' instead of an actual precision.
     * It makes no sense to call this function if
     * <code>get_is_precision_present()</code> returns
     * <code>false</code>.
     *
     * @return whether the precision is on the stack
     */
    virtual int
    get_is_precision_on_stack() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_convspec__get_is_precision_on_stack(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set whether the precision is on the stack.  This occurs when
     * the conversion specifier uses '*' instead of an actual
     * precision.  Calling this method with
     * <code>is_precision_on_stack</code> set to <code>true</code>
     * does not implicitly mark the precision as present.  You will
     * still need to call <code>set_is_precision_present()</code> if
     * that is the intent.
     *
     * @param[in] is_precision_on_stack whether the precision is on stack
     */
    virtual void
    set_is_precision_on_stack(int is_precision_on_stack)
    {
        cgul_exception_t local = NULL;
        cgul_convspec__set_is_precision_on_stack(
            &local, obj, is_precision_on_stack);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the precision.  It makes no sense to call this function if
     * <code>get_is_precision_present</code> returns
     * <code>false</code> or <code>get_is_precision_on_stack()</code>
     * returns <code>true</code>.
     *
     * @return precision
     */
    virtual int
    get_precision() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_convspec__get_precision(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the precision.  Calling this method does not implicitly
     * mark the precision as present.  You will still need to call
     * <code>set_is_precision_present()</code> if that is the intent.
     * It also does not implicitly mark the precision as not being on
     * the stack.  You will still need to call
     * <code>set_is_precision_on_stack()</code> if that is the intent.
     *
     * @param[in] precision precision
     */
    virtual void
    set_precision(int precision)
    {
        cgul_exception_t local = NULL;
        cgul_convspec__set_precision(&local, obj, precision);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Whether the length modifier is present.
     *
     * @return whether the length modifier is present
     */
    virtual int
    get_is_length_modifier_present() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_convspec__get_is_length_modifier_present(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set whether the length modifier is present.
     *
     * @param[in] is_length_modifier_present whether the length modifier
     * is present
     */
    virtual void
    set_is_length_modifier_present(int is_length_modifier_present)
    {
        cgul_exception_t local = NULL;
        cgul_convspec__set_is_length_modifier_present(
            &local, obj, is_length_modifier_present);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the length modifier.  It makes no sense to call this function
     * if <code>get_is_length_modifier_present()</code>
     * returns <code>false</code>.
     *
     * @return length modifier
     */
    virtual const char*
    get_length_modifier() const
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_convspec__get_length_modifier(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the length modifier.  Calling this method does not
     * implicitly mark the length modifier as present.  You will still
     * need to call <code>set_is_length_modifier_present()</code> if
     * that is the intent.  If an error occurs, an exception is
     * thrown.
     *
     * @param[in] length_modifier length modifier
     */
    virtual void
    set_length_modifier(const char* length_modifier)
    {
        cgul_exception_t local = NULL;
        cgul_convspec__set_length_modifier(&local, obj, length_modifier);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the conversion specifier.  For example, this is the 'd'
     * character in the "%d" conversion specification.
     *
     * @return conversion specifier
     */
    virtual char
    get_cspec() const
    {
        char rv = '\0';
        cgul_exception_t local = NULL;
        rv = cgul_convspec__get_cspec(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the conversion specifier.  For example, this is the 'd'
     * character in the "%d" conversion specification.
     *
     * @param[in] cspec conversion specifier
     */
    virtual void
    set_cspec(char cspec)
    {
        cgul_exception_t local = NULL;
        cgul_convspec__set_cspec(&local, obj, cspec);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the underlying <code>cgul_convspec</code> object.
     *
     * @return underlying object
     */
    virtual cgul_convspec_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_convspec</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_convspec__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_convspec_t
    take_obj()
    {
        cgul_convspec_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_convspec_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_convspec__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /** Underlying <code>cgul_convspec</code> object. */
    cgul_convspec_t obj;

};

#endif  /*  CGUL_CONVSPEC_CXX_H  */
