#ifndef CGUL_CRC32_H
#define CGUL_CRC32_H

/**
 * The <code>cgul_crc32</code> class is designed to handle any of the
 * different common permutations on the CRC-32 algorithm.  This class
 * is parameterized to allow different generator polynomials, initial
 * values of the shift register, final XOR values, whether to append
 * the size, whether to XOR the size before appending, and whether the
 * data stream is reflected.
 *
 * This class was first written in java, and the naming convention for
 * methods in this class follows what is established by
 * java.security.MessageDigest instead what is established by
 * java.util.zip.CRC32.
 *
 * A simple example of how to use this class follows:
 *
 * <pre>
 *
 *     cgul_exception_t* cex = NULL;
 *     cgul_crc32_t crc32 = cgul_crc32__new(cex,
 *                                          0xedb88320,
 *                                          0xffffffff,
 *                                          0xffffffff,
 *                                          1);
 *
 *     unsigned long result = cgul_crc32__crc_block(cex,
 *                                                  crc,
 *                                                  "123456789",
 *                                                  9);
 *
 *     cgul_libc_printf(cex, "0x%lx\n", result);
 *
 * </pre>
 *
 * The above code will calculate the ZIP CRC for the 9 bytes that make
 * up the ASCII string "123456789".  The CRC that is printed to the
 * screen will be 0xcbf43926.
 *
 * You might also want to calculate the cksum/posix CRC.  You can
 * construct the right cgul_crc32 object as follows:
 *
 * <pre>
 *
 *     crc32 = cgul_crc32__new_appends_size(cex,
 *                                          0x04c11db7,
 *                                          0,
 *                                          0xffffffff,
 *                                          0,
 *                                          0);
 *
 * </pre>
 *
 * NOTE: If you use this class from multiple threads, you MUST provide
 * external synchronization to all methods so that you do not corrupt
 * internal data structures.
 *
 * @file
 * @author Paul Serice
 * @brief CRC-32
 * @see cgul_crc32_incr
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/** Opaque pointer to <code>cgul_crc32</code> instance. */
typedef struct cgul_crc32* cgul_crc32_t;

/**
 * Create a new <code>cgul_crc32</code> object.  Objects created with
 * this constructor do not append the size to the data stream that is
 * CRCed.  The caller is responsible for freeing the object by calling
 * <code>cgul_crc32__delete()</code>.  If memory cannot be allocated,
 * <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] g generator polynomial
 * @param[in] init initial value loaded into the accumulator
 * @param[in] final_xor value to xor with the calculated crc
 * @param[in] reflected whether to use the reflected algorithm
 * @return new <code>cgul_crc32</code> instance
 */
extern CGUL_EXPORT cgul_crc32_t
cgul_crc32__new(cgul_exception_t* cex,
                unsigned long g,
                unsigned long init,
                unsigned long final_xor,
                int reflected);

/**
 * Create a new <code>cgul_crc32</code> object.  Objects created with
 * this constructor append the size to the data stream that is CRCed.
 * The caller is responsible for freeing the object by calling
 * cgul_crc32__delete().  If memory cannot be allocated,
 * <code>NULL</code> is returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] g polynomial
 * @param[in] init initial value loaded into the accumulator
 * @param[in] final_xor value to xor with the calculated crc
 * @param[in] reflected whether to use the reflected algorithm
 * @param[in] size_xor value to xor with the size before appending
 * @return new <code>cgul_crc32</code> instance
 */
extern CGUL_EXPORT cgul_crc32_t
cgul_crc32__new_appends_size(cgul_exception_t* cex,
                             unsigned long g,
                             unsigned long init,
                             unsigned long final_xor,
                             int reflected,
                             size_t size_xor);

/**
 * This method frees all internally allocated memory.  Do not try to
 * access <code>crc</code> after calling this method.
 *
 * @param[in] crc <code>cgul_crc32</code> instance
 */
extern CGUL_EXPORT void
cgul_crc32__delete(cgul_crc32_t crc);

/**
 * Assign the current accumulator and size of <code>rhs</code> to
 * <code>lhs</code>.  This operation probably only makes sense if
 * <code>lhs</code> and <code>rhs</code> have been constructed
 * identically.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] lhs left-hand side
 * @param[in] rhs right-hand side
 */
extern CGUL_EXPORT void
cgul_crc32__assign(cgul_exception_t* cex,
                   cgul_crc32_t lhs,
                   cgul_crc32_t rhs);

/**
 * This method completes the CRC calculation by performing the size
 * XOR and the final XOR.  You will generally call one of the update()
 * methods before calling this method.  When this method returns, the
 * internal state of the object is reset so that it is ready to
 * calculate the next CRC.  The return value is the final value for
 * the CRC.
 *
 * @param[in] cex c-style exception
 * @param[in] crc <code>cgul_crc32</code> instance
 * @return final crc value
 */
extern CGUL_EXPORT unsigned long
cgul_crc32__crc(cgul_exception_t* cex,
                cgul_crc32_t crc);

/**
 * This is the method to call when you only have one block for which
 * you want to calculate a CRC.  You can also call this if you have
 * called one of the update() methods before, and you know that
 * <code>block</code> is your last block.  When this method returns,
 * the internal state of the object is reset so that it is ready to
 * calculate the next CRC.  <code>length</code> is the number of bytes
 * inside of <code>block</code> that you want to add to the CRC
 * calculation.  The return value is the final value for the CRC.
 *
 * @param[in] cex c-style exception
 * @param[in] crc <code>cgul_crc32</code> instance
 * @param[in] block block of data to add to the crc
 * @param[in] length length of <code>block</code>
 * @return final crc value
 */
extern CGUL_EXPORT unsigned long
cgul_crc32__crc_block(cgul_exception_t* cex,
                      cgul_crc32_t crc,
                      const void* block,
                      size_t length);

/**
 * This is the method to call when you want to calculate the CRC from
 * a range of bytes from one block.  You can also call this if you
 * have called one of the update() methods before, and you know that
 * <code>block</code> is your last block.  When this method returns,
 * the internal state of the object is reset so that it is ready to
 * calculate the next CRC.  <code>offset</code> is the index into
 * <code>block</code> where you want to add <code>length</code> bytes
 * to the CRC calculation.  The return value is the final value for
 * the CRC.
 *
 * @param[in] cex c-style exception
 * @param[in] crc <code>cgul_crc32</code> instance
 * @param[in] block block of data to add to the crc
 * @param[in] offset offset into <code>block</code>
 * @param[in] length length of <code>block</code>
 * @return final crc value
 */
extern CGUL_EXPORT unsigned long
cgul_crc32__crc_range_of_block(cgul_exception_t* cex,
                               cgul_crc32_t crc,
                               const void* block,
                               size_t offset,
                               size_t length);

/**
 * This is the method to call when you only have one file for which
 * you want to calculate a CRC.  You can also call this if you have
 * called one of the update() methods before (like
 * <code>cgul_crc32__update_file()</code>), and you know that
 * <code>f</code> is your last file.  When this method returns, the
 * internal state of the object is reset so that it is ready to
 * calculate the next CRC.  The return value is the final value for
 * the CRC.
 *
 * @param[in,out] cex c-style exception
 * @param[in] crc <code>cgul_crc32</code> instance
 * @param[in] f input file
 * @return final crc value
 */
extern CGUL_EXPORT unsigned long
cgul_crc32__crc_file(cgul_exception_t* cex,
                     cgul_crc32_t crc,
                     FILE* f);

/**
 * This is the method to call when you only have one file for which
 * you want to calculate a CRC.  You can also call this if you have
 * called one of the update() methods before (like
 * <code>cgul_crc32__update_fname()</code>), and you know that
 * <code>fname</code> is your last file.  When this method returns,
 * the internal state of the object is reset so that it is ready to
 * calculate the next CRC.  The return value is the final value for
 * the CRC.
 *
 * @param[in,out] cex c-style exception
 * @param[in] crc <code>cgul_crc32</code> instance
 * @param[in] fname input file name
 * @return final crc value
 */
extern CGUL_EXPORT unsigned long
cgul_crc32__crc_fname(cgul_exception_t* cex,
                      cgul_crc32_t crc,
                      const char* fname);

/**
 * This method updates the CRC by adding the byte <code>c</code> to
 * the CRC calculation.
 *
 * @param[in] cex c-style exception
 * @param[in] crc <code>cgul_crc32</code> instance
 * @param[in] c byte to add to crc
 */
extern CGUL_EXPORT void
cgul_crc32__update_byte(cgul_exception_t* cex,
                        cgul_crc32_t crc,
                        unsigned char c);

/**
 * This method updates the CRC by adding all the bytes in
 * <code>block</code> to the calculation.
 *
 * @param[in] cex c-style exception
 * @param[in] crc <code>cgul_crc32</code> instance
 * @param[in] block block of data to add to the crc
 * @param[in] length length of <code>block</code>
 */
extern CGUL_EXPORT void
cgul_crc32__update_block(cgul_exception_t* cex,
                         cgul_crc32_t crc,
                         const void* block,
                         size_t length);

/**
 * This method updates the CRC by adding a range of bytes in
 * <code>block</code> to the calculation.  The range starts at
 * <code>offset</code> and extends <code>length</code> bytes.
 *
 * @param[in] cex c-style exception
 * @param[in] crc <code>cgul_crc32</code> instance
 * @param[in] block block of data to add to the crc
 * @param[in] offset offset into <code>block</code>
 * @param[in] length length of <code>block</code>
 */
extern CGUL_EXPORT void
cgul_crc32__update_range_of_block(cgul_exception_t* cex,
                                  cgul_crc32_t crc,
                                  const void* block,
                                  size_t offset,
                                  size_t length);

/**
 * This method updates the CRC by adding all the bytes (starting at
 * the current offset) in the file <code>f</code> to the calculation.
 * If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] crc <code>cgul_crc32</code> instance
 * @param[in] f input file
 */
extern CGUL_EXPORT void
cgul_crc32__update_file(cgul_exception_t* cex,
                        cgul_crc32_t crc,
                        FILE* f);

/**
 * This method updates the CRC by adding all the bytes in the file
 * with name <code>fname</code> to the calculation.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] crc <code>cgul_crc32</code> instance
 * @param[in] fname input file name
 */
extern CGUL_EXPORT void
cgul_crc32__update_fname(cgul_exception_t* cex,
                         cgul_crc32_t crc,
                         const char* fname);

/**
 * Reset the <code>cgul_crc32</code> object for further use.
 *
 * @param[in] cex c-style exception
 * @param[in] crc <code>cgul_crc32</code> instance
 */
extern CGUL_EXPORT void
cgul_crc32__reset(cgul_exception_t* cex,
                  cgul_crc32_t crc);

/**
 * You can use this method to see the CRC lookup table that is being
 * used.  The returned value is the CRC lookup table for this object
 * as a string.  The caller is responsible for calling free() on the
 * string that is returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] crc <code>cgul_crc32</code> instance
 * @return crc lookup table as a string
 */
extern CGUL_EXPORT char*
cgul_crc32__get_table_as_string(cgul_exception_t* cex,
                                cgul_crc32_t crc);


CGUL_END_C

#endif  /*  CGUL_CRC32_H  */
