#ifndef CGUL_CRLF_FILE_H
#define CGUL_CRLF_FILE_H

/**
 * This class reads lines of text from DOS, Mac, or Unix text files
 * optionally stripping a leading UTF-8 byte order mark if present.
 *
 * @file
 * @author Paul Serice
 * @brief read dos, mac, and unix text files
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_crlf_file</code> instance. */
typedef struct cgul_crlf_file* cgul_crlf_file_t;

/**
 * This method creates a new <code>cgul_crlf_file</code> instance.
 * After this method returns,
 * <code>cgul_crlf_file__open_fname()</code>,
 * <code>cgul_crlf_file__open_file()</code>, or
 * <code>cgul_crlf_file__open_memory()</code> should be called before
 * calling any other method.  The client is responsible for calling
 * <code>cgul_crlf_file__delete()</code> on the object returned.  If
 * an error occurs, <code>NULL</code> is returned, and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_crlf_file</code> instance
 */
extern CGUL_EXPORT cgul_crlf_file_t
cgul_crlf_file__new(cgul_exception_t* cex);

/**
 * This method creates a new <code>cgul_crlf_file</code> instance and
 * calls <code>cgul_crlf_file__open_fname()</code> passing it
 * <code>fname</code>.  The file will be closed when this instance is
 * deleted.  The client is responsible for calling
 * <code>cgul_crlf_file__delete()</code> on the object returned.  If
 * an error occurs, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * If <code>is_block_buffered</code> is true, this method results in a
 * very fast <code>cgul_crlf_file</code> instance, but the underlying
 * file should be a regular file or any other file that is capable of
 * block buffering.  If the file is not capable of block buffering,
 * <code>is_block_buffered</code> should be false which results in a
 * much slower <code>cgul_crlf_file</code> instance, but one that
 * should be able to read from any type of file.  As a general rule,
 * if you are reading from a regular file or are working in batch
 * mode, <code>is_block_buffered</code> should be true to get the best
 * performance.  If you are reading from a line-oriented or
 * character-oriented device like a terminal or from a named pipe
 * across which small messages are being passed,
 * <code>is_block_buffered</code> should be false. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] fname file name
 * @param[in] is_block_buffered whether the file is block buffered
 * @return new <code>cgul_crlf_file</code> instance
 */
extern CGUL_EXPORT cgul_crlf_file_t
cgul_crlf_file__new_from_fname(cgul_exception_t* cex,
                               const char* fname,
                               int is_block_buffered);

/**
 * This method creates a new <code>cgul_crlf_file</code> instance and
 * calls <code>cgul_crlf_file__open_file()</code> passing it
 * <code>f</code>.  The class does not take ownership of
 * <code>f</code>.  Thus, the client is still responsible for calling
 * <code>fclose()</code> on it.  The client is also responsible for
 * calling <code>cgul_crlf_file__delete()</code> on the object
 * returned.  If an error occurs, <code>NULL</code> is returned, and
 * an exception is thrown.
 *
 * If <code>is_block_buffered</code> is true, this method results in a
 * very fast <code>cgul_crlf_file</code> instance, but the underlying
 * file should be a regular file or any other file that is capable of
 * block buffering.  If the file is not capable of block buffering,
 * <code>is_block_buffered</code> should be false which results in a
 * much slower <code>cgul_crlf_file</code> instance, but one that
 * should be able to read from any type of file.  As a general rule,
 * if you are reading from a regular file or are working in batch
 * mode, <code>is_block_buffered</code> should be true to get the best
 * performance.  If you are reading from a line-oriented or
 * character-oriented device like a terminal or from a named pipe
 * across which small messages are being passed,
 * <code>is_block_buffered</code> should be false. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] f file
 * @param[in] is_block_buffered whether the file is block buffered
 * @return new <code>cgul_crlf_file</code> instance
 */
extern CGUL_EXPORT cgul_crlf_file_t
cgul_crlf_file__new_from_file(cgul_exception_t* cex,
                              FILE* f,
                              int is_block_buffered);

/**
 * This method creates a new <code>cgul_crlf_file</code> instance and
 * calls <code>cgul_crlf_file__open_memory()</code> passing it
 * <code>buf</code> and <code>buf_size</code>.  This class does not
 * take ownership of <code>buf</code> so the client is still
 * responsible for freeing <code>buf</code> if necessary.  The client
 * is responsible for calling <code>cgul_crlf_file__delete()</code> on
 * the object returned.  If an error occurs, <code>NULL</code> is
 * returned, and an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] buf memory buffer
 * @param[in] buf_size size of <code>buf</code> in bytes
 * @return new <code>cgul_crlf_file</code> instance
 */
extern CGUL_EXPORT cgul_crlf_file_t
cgul_crlf_file__new_from_memory(cgul_exception_t* cex,
                                const char* buf,
                                size_t buf_size);

/**
 * This method deletes the <code>crlf_file</code> instance freeing all
 * internally allocated resources.  This does not include closing the
 * underlying file if it was obtained from the client by a call to
 * <code>cgul_crlf_file__new_from_file()</code> or
 * <code>cgul_crlf_file__open_file()</code>.  The client must not use
 * <code>crlf_file</code> after calling this method.
 *
 * @param[in] crlf_file crlf file
 */
extern CGUL_EXPORT void
cgul_crlf_file__delete(cgul_crlf_file_t crlf_file);

/**
 * Open the file with name <code>fname</code> and use it when getting
 * lines.  If a data source is already open, it will be closed before
 * this method attempts to open the new data source.  The new file
 * will be closed when this instance is deleted.  If an error occurs,
 * an exception is thrown. <p>
 *
 * If <code>is_block_buffered</code> is true, this method results in a
 * very fast <code>cgul_crlf_file</code> instance, but the underlying
 * file should be a regular file or any other file that is capable of
 * block buffering.  If the file is not capable of block buffering,
 * <code>is_block_buffered</code> should be false which results in a
 * much slower <code>cgul_crlf_file</code> instance, but one that
 * should be able to read from any type of file.  As a general rule,
 * if you are reading from a regular file or are working in batch
 * mode, <code>is_block_buffered</code> should be true to get the best
 * performance.  If you are reading from a line-oriented or
 * character-oriented device like a terminal or from a named pipe
 * across which small messages are being passed,
 * <code>is_block_buffered</code> should be false. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] crlf_file <code>cgul_crlf_file</code> instance
 * @param[in] fname file name
 * @param[in] is_block_buffered whether the file is block buffered
 */
extern CGUL_EXPORT void
cgul_crlf_file__open_fname(cgul_exception_t* cex,
                           cgul_crlf_file_t crlf_file,
                           const char* fname,
                           int is_block_buffered);

/**
 * Use the file <code>f</code> when getting lines.  If a data source
 * is already open, it is closed before attempting to open the new
 * data source.  This class does not take ownership of <code>f</code>.
 * Thus, the client is still responsible for calling
 * <code>fclose()</code> on <code>f</code>.  If an error occurs, an
 * exception is thrown. <p>
 *
 * If <code>is_block_buffered</code> is true, this method results in a
 * very fast <code>cgul_crlf_file</code> instance, but the underlying
 * file should be a regular file or any other file that is capable of
 * block buffering.  If the file is not capable of block buffering,
 * <code>is_block_buffered</code> should be false which results in a
 * much slower <code>cgul_crlf_file</code> instance, but one that
 * should be able to read from any type of file.  As a general rule,
 * if you are reading from a regular file or are working in batch
 * mode, <code>is_block_buffered</code> should be true to get the best
 * performance.  If you are reading from a line-oriented or
 * character-oriented device like a terminal or from a named pipe
 * across which small messages are being passed,
 * <code>is_block_buffered</code> should be false. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] crlf_file <code>cgul_crlf_file</code> instance
 * @param[in] f file
 * @param[in] is_block_buffered whether the file is block buffered
 */
extern CGUL_EXPORT void
cgul_crlf_file__open_file(cgul_exception_t* cex,
                          cgul_crlf_file_t crlf_file,
                          FILE* f,
                          int is_block_buffered);

/**
 * Use the memory buffer <code>buf</code> when getting lines.  If a
 * data source is already open, it will be closed before this method
 * attempts to open the new data source.  This class does not take
 * ownership of <code>buf</code>.  Thus, the client is still
 * responsible for freeing <code>buf</code> if necessary.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] crlf_file <code>cgul_crlf_file</code> instance
 * @param[in] buf memory buffer
 * @param[in] buf_size size of <code>buf</code> in bytes
 */
extern CGUL_EXPORT void
cgul_crlf_file__open_memory(cgul_exception_t* cex,
                            cgul_crlf_file_t crlf_file,
                            const char* buf,
                            size_t buf_size);

/**
 * Close the file.  After this method returns,
 * <code>cgul_crlf_file__open_fname()</code>,
 * <code>cgul_crlf_file__open_file()</code>, or
 * <code>cgul_crlf_file__open_memory()</code> should be called before
 * calling any other method.
 *
 * @param[in] cex c-style exception
 * @param[in] crlf_file <code>cgul_crlf_file</code> instance
 */
extern CGUL_EXPORT void
cgul_crlf_file__close(cgul_exception_t* cex,
                      cgul_crlf_file_t crlf_file);

/**
 * Return the name of the currently opened file.  If a
 * <code>FILE*</code> was opened instead of a file name,
 * <code>"FILE"</code> will be used.  If memory was opened instead of
 * a file name, <code>"MEMORY"</code> will be used.  The client must
 * not attempt to free the pointer returned.  This method throws an
 * exception only if a file is not currently open.
 *
 * @param[in] cex c-style exception
 * @param[in] crlf_file <code>cgul_crlf_file</code> instance
 * @return name of the currently opened file
 */
extern CGUL_EXPORT const char*
cgul_crlf_file__get_fname(cgul_exception_t* cex,
                          cgul_crlf_file_t crlf_file);

/**
 * Return the currently opened file.  If a file name or
 * <code>FILE*</code> was opened, a <code>FILE*</code> is returned.
 * If memory was opened, a <code>char*</code> is returned.  The client
 * must not do anything that would invalidate the pointer or corrupt
 * the associated data stream.  This method throws an exception only
 * if a file is not currently open.
 *
 * @param[in] cex c-style exception
 * @param[in] crlf_file <code>cgul_crlf_file</code> instance
 * @return currently opened file or memory
 */
extern CGUL_EXPORT const void*
cgul_crlf_file__get_file(cgul_exception_t* cex,
                         cgul_crlf_file_t crlf_file);

/**
 * This method returns whether the leading UTF-8 byte-order mark (BOM)
 * should be removed from the first line if it is present.
 *
 * @param[in] cex c-style exception
 * @param[in] crlf_file <code>cgul_crlf_file</code> instance
 * @return whether to strip a leading UTF-8 byte-order mark
 */
extern CGUL_EXPORT int
cgul_crlf_file__get_strip_utf8_bom(cgul_exception_t* cex,
                                   cgul_crlf_file_t crlf_file);

/**
 * By default, this class detects the leading UTF-8 byte-order mark
 * (BOM) and strips it from the first line returned by
 * <code>cgul_crlf_file__get_line()</code> if it is present.  It then
 * clears its internal flag so that BOMs internal to the text file
 * will be returned.  This is generally what you want because the
 * leading BOM is not significant but the internal BOMs are. <p>
 *
 * You can alter the way this class handles the leading BOM by calling
 * this method with <code>strip_utf8_bom</code> set to <code>0</code>.
 * This will cause the leading BOM to be returned as part of the first
 * line.  This can be useful, for example, if you just want to convert
 * the text file and are not interested in its contents. <p>
 *
 * It should be noted that most operating systems do not save UTF-8
 * text files with a leading BOM because UTF-8 is a character stream
 * and, as such, does not have byte-order problems; however, Microsoft
 * Windows adds the BOM to its UTF-8 text files presumably to help
 * distinguish UTF-8 text files from text files with different
 * encodings. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] crlf_file <code>cgul_crlf_file</code> instance
 * @param[in] strip_utf8_bom whether to strip leading UTF-8 byte-order mark
 */
extern CGUL_EXPORT void
cgul_crlf_file__set_strip_utf8_bom(cgul_exception_t* cex,
                                   cgul_crlf_file_t crlf_file,
                                   int strip_utf8_bom);

/**
 * Return the array of lines from the text file <code>fname</code> in
 * <code>*lines</code> and the number of lines in
 * <code>*line_count</code>.  Each line is stored in the array as a
 * C-style string.  The client is responsible for calling
 * <code>free()</code> on each line.  If an error occurs, an exception
 * is thrown. <p>
 *
 * To use this function, do something like the following: <p>
 *
 * <pre>
 *
 *     char** lines = NULL;
 *     unsigned long int line_count = 0;
 *     ...
 *     cgul_crlf_file__get_lines(cex, crlf, &lines, &line_count);
 *     if (*cex) {
 *         goto out;
 *     }
 *
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] crlf_file <code>cgul_crlf_file</code> instance
 * @param[out] lines ordered list of lines from the file
 * @param[out] line_count number of lines
 * @return list of lines from the file
 */
void
cgul_crlf_file__get_lines(cgul_exception_t* cex,
                          cgul_crlf_file_t crlf_file,
                          char*** lines,
                          unsigned long int* line_count);

/**
 * This method returns the next line of text from the underlying DOS,
 * Mac, or Unix text file.  The caller must not call
 * <code>free()</code> on the pointer returned as it points into a
 * larger block owned by this class.  If EOF is reached,
 * <code>NULL</code> is returned.  If an error occurs,
 * <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] crlf_file <code>cgul_crlf_file</code> instance
 * @return next line of text
 */
extern CGUL_EXPORT const char*
cgul_crlf_file__get_line(cgul_exception_t* cex,
                         cgul_crlf_file_t crlf_file);

/**
 * Get the line count for the last line returned by
 * <code>cgul_crlf_file__get_line</code>.  The line count is
 * one-based.  No attempt is made to prevent the return value from
 * overflowing.  So, the caller is responsible for verifying the
 * return value.
 *
 * @param[in,out] cex c-style exception
 * @param[in] crlf_file <code>cgul_crlf_file</code> instance
 * @return line count
 */
extern CGUL_EXPORT unsigned long
cgul_crlf_file__get_line_count(cgul_exception_t* cex,
                               cgul_crlf_file_t crlf_file);

/**
 * Get the line offset for the last line returned by
 * <code>cgul_crlf_file__get_line</code>.  The offset is zero-based.
 * If the underlying file is binary and random access, you can use the
 * return value to directly seek to the line as follows:
 *
 * <pre>
 *     cgul_crlf_file__fseek(cex, crlf_file, offset, SEEK_SET);
 * </pre>
 *
 * Because the prototype for <code>fseek()</code> requires a
 * <code>long</code> for the offset parameter, no attempt is made to
 * prevent the return value from overflowing.  So, the caller is
 * responsible for verifying the return value. <p>
 *
 * Note that the offset returned is the number of bytes from the start
 * of the file to the current line.  This is not necessarily the same
 * as the number of characters which depends on how the file is
 * encoded. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] crlf_file <code>cgul_crlf_file</code> instance
 * @return file offset
 */
extern CGUL_EXPORT unsigned long
cgul_crlf_file__get_line_offset(cgul_exception_t* cex,
                                cgul_crlf_file_t crlf_file);

/**
 * This method calls <code>cgul_libc__fseek()</code> on the underlying
 * <code>FILE*</code> object and resets the underlying
 * <code>cgul_crlf</code> object so that reading new lines can
 * continue at <code>offset</code> which is relative to
 * <code>whence</code>.  If an error occurs, an exception is
 * thrown. <p>
 *
 * You can call <code>cgul_crlf_file__get_line_offset()</code>
 * immediately after calling <code>cgul_crlf_file__get_line()</code>
 * to determine the offset for the beginning of the last line
 * returned. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] crlf_file <code>cgul_crlf_file</code> instance
 * @param[in] offset offset
 * @param[in] whence position to which offset is relative
 */
extern CGUL_EXPORT void
cgul_crlf_file__fseek(cgul_exception_t* cex,
                      cgul_crlf_file_t crlf_file,
                      unsigned long offset,
                      int whence);

/**
 * This method calls <code>cgul_libc__rewind()</code> on the
 * underlying <code>FILE*</code> object and resets the underlying
 * <code>cgul_crlf</code> object so that reading new lines can
 * continue at the beginning of the file.  This method throws an
 * exception if either <code>cgul_crlf_file__open_fname()</code> or
 * <code>cgul_crlf_file__open_file()</code> has not been called.  This
 * method also throws an exception if the underlying file is not
 * seekable.
 *
 * @param[in,out] cex c-style exception
 * @param[in] crlf_file <code>cgul_crlf_file</code> instance
 */
extern CGUL_EXPORT void
cgul_crlf_file__rewind(cgul_exception_t* cex,
                       cgul_crlf_file_t crlf_file);

/**
 * This method returns the size of the buffer used to read blocks out
 * of the underlying file.
 *
 * @param[in] cex c-style exception
 * @param[in] crlf_file <code>cgul_crlf_file</code> instance
 * @return buffer size
 */
extern CGUL_EXPORT unsigned long
cgul_crlf_file__get_buffer_size(cgul_exception_t* cex,
                                cgul_crlf_file_t crlf_file);

/**
 * This method sets the size of the buffer used to read blocks out of
 * the underlying file.  By default, the buffer size is 16K which is
 * efficient for processing large files en masse but may not be
 * efficient when randomly accessing small parts of the same file.  If
 * an error occurs allocating the new buffer, an exception is thrown,
 * and the original buffer will continue to be used.
 *
 * @param[in,out] cex c-style exception
 * @param[in] crlf_file <code>cgul_crlf_file</code> instance
 * @param[in] bsize new buffer size
 */
extern CGUL_EXPORT void
cgul_crlf_file__set_buffer_size(cgul_exception_t* cex,
                                cgul_crlf_file_t crlf_file,
                                unsigned long bsize);

CGUL_END_C

#endif  /*  CGUL_CRLF_FILE_H  */
