#ifndef CGUL_ENDIAN_CXX_H
#define CGUL_ENDIAN_CXX_H

/**
 * C++ wrapper for <code>cgul_endian</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_endian</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_endian.h"

/**
 * This class provides the C++ bindings for C <code>cgul_endian</code>
 * external functions.  The main purpose of this class is to convert
 * the C-style function calls and exception handling in
 * <code>cgul_endian</code> into C++-style function calls and
 * exception handling.
 *
 * @brief C++ bindings for <code>cgul_endian</code>
 */
class cgul_endian_cxx {

public:

    /**
     * Return whether the host is big endian.  This is a fast run-time
     * test, but calling it before each swap will adversely affect
     * performance.
     *
     * @return whether the host is big-endian
     */
    static int
    is_big_endian()
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_endian__is_big_endian(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return whether the host is little endian.  This is a fast run-time
     * test, but calling it before each swap will adversely affect
     * performance.
     *
     * @return whether the host is little endian
     */
    static int
    is_little_endian()
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_endian__is_little_endian(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Byte swap the first <code>value_size</code> bytes in the region of
     * memory pointed to by <code>value</code>.  If <code>value</code> is
     * in big-endian format, it will be converted to little-endian format
     * and vice versa. <p>
     *
     * The following example shows how to convert an unsigned, big-endian
     * 32-bit integer to the host format (assuming the host is either big
     * endian or little endian instead of one of the more exotic formats):
     *
     * <pre>
     *     cgul_uint32_t value;
     *     ...
     *
     *     if (cgul_endian_cxx::is_little_endian(cex)) {
     *         ...
     *         cgul_endian_cxx::swap(cex, (char*)&value, sizeof(value));
     *         ...
     *     }
     * </pre>
     *
     * @param[in,out] value value
     * @param[in] value_size size of <code>value</code>
     * @see cgul_int
     */
    static void
    swap(void* value,
         size_t value_size)
    {
        cgul_exception_t local = NULL;
        cgul_endian__swap(&local, value, value_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

};

#endif  /*  CGUL_ENDIAN_CXX_H  */
