#ifndef CGUL_EVENT_ROUTER_H
#define CGUL_EVENT_ROUTER_H

/**
 * Experience shows that peer-to-peer communication between decoupled
 * objects does not scale well.  All but the smallest object-oriented
 * programs must have some sort of centralized communications hub so
 * that decoupled classes can come together to form a coherent
 * program.  The two patterns that commonly do this are the
 * <i>mediator</i> and <i>router</i> patterns.  This class implements
 * the <i>router</i> pattern. <p>
 * 
 * This particular implementation of the <i>router</i> pattern is
 * specifically designed so that the router itself needs no knowledge
 * of the events that are being passed.  Thus, when you want to add
 * new events or remove existing events, you can do so without having
 * to edit or recompile this class.  This works by the event sender
 * and the event listener agreeing on a rendezvous string.  When
 * rendezvous strings match, an event sender will be able to broadcast
 * to all objects listening for a particular event.  Multiple
 * listeners and multiple senders are allowed for any event.  Events
 * are received by listeners in the order in which the listeners were
 * added. <p>
 * 
 * If you have ever tried to design a peer-to-peer communication
 * system for your objects, you will immediately recognize how much
 * easier this is than trying to make sure every peer has a pointer to
 * all the peers with which it communicates.  The primary benefit is
 * that your event listeners do not have to be passed a pointer to the
 * event sender in order to register using a method like Java's
 * <code>Observable.addObserver()</code> or .NET's
 * <code>Delegate</code>.  This means when you change your mind about
 * which objects should send or receive events, you do not have to
 * manually find every event sender or event listener involved in the
 * change and alter the code accordingly which quickly becomes a
 * maintenance nightmare. <p>
 *
 * On the other hand, with the approach provided by this class, your
 * event senders and event listeners just pass in a string to the
 * <code>cgul_event_router</code>.  If you want to add or remove new
 * event senders or listeners, all changes are local to the class to
 * which you are adding or removing; you do not have to go traipsing
 * through your code looking for peer objects. <p>
 * 
 * Another advantage of this approach is that all objects are capable
 * of sending and receiving @htmlonly <u>any</u> @endhtmlonly event,
 * and you get this flexibility without either having to clutter your
 * inheritance tree or build this functionality explicitly into each
 * object. <p>
 * 
 * The use of rendezvous strings makes maintenance much easier, but it
 * places a burden on the programmer to make sure the strings are
 * unique and that both the event sender and the event listener agree
 * on what those strings are.  If you want compiler support for
 * detecting these errors, I suggest you resort to the Motif trick
 * whereby you keep a central list of <code>const</code> strings that
 * you assign to variables that have the same name as the string.  If
 * you accidentally reuse a string, the compiler will complain that
 * its variable has already been defined.  If you accidentally
 * misspell the string at the point where it is used, the compiler
 * will complain that it cannot resolve the symbol.  This fixes most
 * of the known problems that arise when using arbitrary strings as
 * rendezvous points. <p>
 *
 * @note The short story for how to use this class with Managed C++
 * .NET classes is that you need to use the .NET <code>GCHandle</code>
 * class to convert your tracking handles to and from magic cookies
 * that map to <code>void*</code>.  For an example of this see the
 * following file which is part of the cgul source distribution: <p>
 *
 * <pre>
 *     tests/main_event_router_cxx_dot_net.cpp
 * </pre>
 *
 * @note This class @htmlonly <u>is not</u> @endhtmlonly thread-safe
 * by design.  If you want to use this class from multiple threads,
 * you must provide external synchronization. <p>
 * 
 * @note This class @htmlonly <u>is</u> @endhtmlonly reentrant by
 * design.  In practice, this means it is safe for a listener to
 * remove itself (or to add and remove other listeners) in the same
 * code that listens for events.  It also means that event listeners
 * can recursively send events. <p>
 * 
 * @file
 * @author Paul Serice
 * @brief Implements the <i>router</i> pattern.
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/**
 * Opaque pointer to a <code>cgul_event_router</code> instance.
 */
typedef struct cgul_event_router* cgul_event_router_t;

/**
 * Opaque pointer a <code>cgul_event_router__listener</code> instance.
 * An object of this type is passed back to the user by
 * <code>cgul_event_router__add_listener()</code>.  It can be used to
 * remove a listener or change the data passed to a listener.
 */
typedef struct cgul_event_router__listener* cgul_event_router__listener_t;

/**
 * In order to register to listen to events, you must implement a
 * function that is of type.  The function must return true after each
 * call in order for <code>event</code> to be sent to the remaining
 * listeners. <p>
 *
 * @param[in] rendezvous rendezvous string associated with this event
 * @param[in] event event your function has been sent
 * @param[in] data client data that was passed in when the listener was
 * added with <code>cgul_event_router__add_listener()</code>.
 * @return whether to continue sending <code>event</code> to other listeners
 */
typedef int (*cgul_event_router__callback_t)(const char* rendezvous,
                                             const void* event,
                                             const void* data);

/**
 * Return the single instance of this class.  Because
 * <code>cgul_event_router</code> is a singleton, the caller is
 * generally <b>not</b> responsible for cleaning up by calling
 * <code>cgul_event_router__delete()</code>.  If memory cannot be
 * allocated, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @return single instance of this class.
 */
extern CGUL_EXPORT cgul_event_router_t
cgul_event_router__get_instance(cgul_exception_t* cex);

/**
 * This method frees all internally allocated resources.  In a typical
 * program, it should only be called once (if at all) immediately
 * before the program exists. <p>
 *
 * @param[in] er event router
 */
extern CGUL_EXPORT void
cgul_event_router__delete(cgul_event_router_t er);

/**
 * This method adds <code>cb</code> to listen for the event specified
 * by the rendezvous string <code>rendezvous</code>.  A copy of
 * <code>rendezvous</code> is made so the client does not have to keep
 * <code>rendezvous</code> in scope. <p>
 *
 * Events will be sent to the listeners in the order in which the
 * listeners were added. <p>
 *
 * The "all" rendezvous string is special; if you use it, your
 * listener will receive all events without having to explicitly
 * register for each one.  Furthermore, the "all" event listeners will
 * receive the events before the normal event listeners.  This makes
 * the output of a logger that is attached to the "all" event sane
 * when your code recursively sends events. <p>
 *
 * The <code>cb</code> callback can be added more than once per
 * rendezvous string.  This makes it possible to use this class with
 * C++ code when using a single private "static" method to forward the
 * callback to a member method for more than one object of the C++
 * class.  When using this tactic, each C++ object has be to
 * registered individually, and this class makes it possible to do
 * so. <p>
 *
 * <code>cb</code> should return true if it wants the event to be sent
 * to all remaining listeners.  If <code>cb</code> returns false, the
 * event will not be sent to any function that was added after
 * <code>cb</code>. <p>
 *
 * This class is reentrant so it is safe to recursively add or remove
 * listeners or even send events from <code>cb</code>.  <p>
 *
 * If an error occurs, <code>NULL</code> will be returned, and an
 * exception will be thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] er event router
 * @param[in] rendezvous rendezvous string
 * @param[in] cb callback function
 * @param[in] data opaque client data to pass back to <code>cb</code>
 * @return listener
 */
extern CGUL_EXPORT cgul_event_router__listener_t
cgul_event_router__add_listener(cgul_exception_t* cex,
                                cgul_event_router_t er,
                                const char* rendezvous,
                                cgul_event_router__callback_t cb,
                                const void* data);

/**
 * Remove <code>listener</code>.  After calling this method,
 * <code>listener</code> is invalid and should not be used again by
 * the client. <p>
 *
 * This class is reentrant so it is safe to recursively add or remove
 * listeners or even send events from the callback associated with
 * <code>listener</code>.  <p>
 *
 * @param[in] cex c-style exception
 * @param[in] er event router
 * @param[in] listener listener
 */
extern CGUL_EXPORT void
cgul_event_router__remove_listener(cgul_exception_t* cex,
                                   cgul_event_router_t er,
                                   cgul_event_router__listener_t listener);

/**
 * Send <code>event</code> to all listeners registered for events
 * associated with the <code>rendezvous</code> string.  Events will be
 * sent to the listeners in the order in which the listeners were
 * added.  This class is reentrant so it is safe to call this method
 * from the callbacks associated with listeners thereby recursively
 * sending events.  If an error occurs, an exception will be thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] er event router
 * @param[in] rendezvous rendezvous string
 * @param[in] event event to send
 */
extern CGUL_EXPORT void
cgul_event_router__send_event(cgul_exception_t* cex,
                              cgul_event_router_t er,
                              const char* rendezvous,
                              const void* event);

/**
 * This method sets the client data that is passed back to
 * <code>listener</code> when it is sent an event.
 *
 * @param[in] cex c-style exception
 * @param[in] er event router
 * @param[in] listener listener
 * @param[in] data client data
 */
extern CGUL_EXPORT void
cgul_event_router__set_data(cgul_exception_t* cex,
                            cgul_event_router_t er,
                            cgul_event_router__listener_t listener,
                            const void* data);

/**
 * Get the count of listeners currently registered for the event
 * associated with the <code>rendezvous</code> string. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] er event router
 * @param[in] rendezvous rendezvous string
 * @return listener count
 */
extern CGUL_EXPORT size_t
cgul_event_router__get_listener_count(cgul_exception_t* cex,
                                      cgul_event_router_t er,
                                      const char* rendezvous);

CGUL_END_C

#endif  /*  CGUL_EVENT_ROUTER_H  */
