#ifndef CGUL_EXCEPTION_H
#define CGUL_EXCEPTION_H

/**
 * <i>Basics:</i><p>
 *
 * This class provides a generic object that you can use to unwind the
 * stack similar to throwing an exception in C++. <p>
 *
 * The basic idea is for the main program to initialize a
 * cgul_exception_t variable to <code>NULL</code>.  You then pass in
 * a pointer to the <code>cgul_exception_t</code> object to every
 * function you write.  If you get an error that you want to throw,
 * you allocate a cgul_exception_t object using cgul_exception__new()
 * and return to the caller.  Very simple; except ... for this to
 * work, every function you write <b>must</b> first check to see if
 * the <code>cgul_exception_t</code> object it is passed is not
 * <code>NULL</code> before it begins doing its work. <p>
 *
 * If the <code>cgul_exception_t</code> is not <code>NULL</code>, it
 * means an error has occurred, and the stack is unwinding.  Thus,
 * your function must return immediately without doing any work in
 * order to allow the stack to unwind.  Lastly, unlike with Java,
 * there is no garbage collection; so, the code that catches the
 * exception needs to free the memory associated with the exception by
 * calling <code>cgul_exception__delete()</code>. <p>
 *
 * It should also be noted that the same issues regarding throwing
 * exceptions in your destructors that pertain to C++ also pertain to
 * these exceptions.  Specifically, don't do it, but it is not the end
 * of the world if you do.  It only means that the exception you throw
 * from your destructor will mask the original exception.  Typically,
 * there is no reason to even pass in the thread's
 * <code>cgul_exception_t</code> pointer to your destructors.
 *
 * It should be noted that <code>cgul_exception__new()</code> is
 * unlike the other constructors in this library in that it does not
 * participate in the C-style exception mechanism.  You can clearly
 * see this because it does not accept a
 * <code>cgul_exception_t</code> pointer as its first argument.  The
 * reason for this is that unwinding the stack has to start somewhere,
 * and, in this scheme, it starts by assigning the value returned by
 * <code>cgul_exception__new()</code>.
 *
 * When you are reading the doxygen comments for other classes that
 * use this class, a function with <code>param[in] cex</code> only
 * tests the exception and returns if the stack is unwinding.  A
 * function with <code>param[in,out] cex</code> both tests the
 * exception and, if an error occurs, throws an exception by
 * allocating a new <code>cgul_exception</code> object. <p>
 *
 * <i>Usage Patterns:</i><p>
 *
 * A typical main() will look something like the following:
 *
 * <pre>
 *   
 *      int main(int argc, char* argv[])
 *      {
 *          int rv = 0;
 *
 *          // When an exception is thrown, the lower-level function will
 *          // create a new cgul_exception class and set "local" to point
 *          // to it.
 *          cgul_exception_t local = NULL;
 *
 *          // Because each function returns immediately if "local" is set,
 *          // we can write the functional part of the code clean of most
 *          // (if not all) in-line error checking.
 *          func1(&local, ...);
 *          func2(&local, ...);
 *          func3(&local, ...);
 *          func4(&local, ...);
 *
 *          // After the stack unwinds, we check for errors.
 *          if (local) {
 *              const char* errmsg = cgul_exception__get_message(local);
 *              fprintf(stderr, "*** %s: Error: %s\n", progname, errmsg);
 *              cgul_exception__delete(local);
 *              local = NULL;
 *              rv = 1;
 *          }
 *      
 *          return rv;
 *      }
 *
 * </pre>
 *
 * A typical constructor that throws only one or two exceptions will look
 * like the following:
 *
 * <pre>

 *     my_obj_t
 *     my_obj_new(cgul_exception_t* cex)
 *     {
 *         my_obj_t rv = NULL;
 *     
 *         assert(cex);
 *         if (*cex) {
 *             // An exception has already been thrown.  We need to allow the
 *             // stack to unwind by not doing anything substantive.
 *             goto out;
 *         }
 *     
 *         rv = calloc(1, sizeof(*rv));
 *         if (rv == NULL) {
 *             // Notice that these lower-level functions still have to check
 *             // for errors and throw exceptions.
 *             *cex = cgul_exception__get_out_of_memory_exception();
 *             goto out;
 *         }
 *
 *         // Initialize your object here.
 *     
 *      out:
 *     
 *         if (*cex && rv) {
 *             // Clean up a partially instantiated object.  Notice that your
 *             // destructors (here my_obj_delete()) are not passed the "cex"
 *             // pointer.  The reason for this is the same reason why C++
 *             // destructors should not throw exceptions.  Your destructors
 *             // always need to be able to clean up partially instantiated
 *             // objects which means they need to be callable from a context
 *             // in which an exception has already been raised and must not
 *             // themselves throw an exception.
 *             my_obj_delete(rv);
 *             rv = NULL;
 *         }
 *     
 *         return rv;
 *     }
 *
 * </pre>
 *
 * @file
 * @author Paul Serice
 * @brief cgul c-style exception
 */

#include "cgul_common.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_exception</code> instance. */
typedef struct cgul_exception* cgul_exception_t;

/**
 * A statically allocated <code>cgul_exception_t</code> object exists
 * to handle the special case of running out of memory because once
 * you run out of memory you will obviously have problems allocating a
 * <code>cgul_exception</code> object at that point.  If your code
 * runs out of memory, this is the exception that it should throw. <p>
 *
 * It is perfectly legal for the code that catches the exception to
 * call <code>cgul_exception__delete()</code> on the object returned
 * despite the object being statically allocated.  This works because
 * the <code>cgul_exception</code> class knows not to actually call
 * <code>free()</code> on this object.
 *
 * @return out of memory exception
 */
extern CGUL_EXPORT cgul_exception_t
cgul_exception__get_out_of_memory_exception(void);

/**
 * Create a new <code>cgul_exception</code> object using the
 * printf()-style format string <code>format</code>.  This method is a
 * very thin wrapper around
 * <code>cgul_exception__new_from_va_list()</code>. <p>
 *
 * In the special case where memory cannot be allocated for a new
 * <code>cgul_exception</code> instance,
 * <code>cgul_exception__get_out_of_memory_exception()</code> will be
 * returned in place of the exception that would otherwise be
 * returned.  <p>
 *
 * @param[in] format printf()-style format string
 * @param[in] ... variable number of arguments that satisfy the format string
 * @return new <code>cgul_exception</code> object
 */
extern CGUL_EXPORT cgul_exception_t
cgul_exception__new(const char* format, ...)
#if defined(__GNUC__) && !defined(__MINGW_H)
    __attribute__ ((format (printf, 1, 2)))
#endif
;

/**
 * Create a new <code>cgul_exception</code> object using the
 * printf()-style format string <code>format</code> and the variable
 * argument list <code>args</code>. <p>
 *
 * In the special case where memory cannot be allocated for a new
 * <code>cgul_exception</code> instance,
 * <code>cgul_exception__get_out_of_memory_exception()</code> will be
 * returned in place of the exception that would otherwise be
 * returned.  <p>
 *
 * @param[in] format printf()-style format string
 * @param args variable number of arguments that satisfy <code>format</code>
 * @return new <code>cgul_exception</code> object
 */
extern CGUL_EXPORT cgul_exception_t
cgul_exception__new_from_va_list(const char* format,
                                 va_list args);

/**
 * Free all internally allocated memory.  Do not try to access
 * <code>cex</code> after calling this method.
 *
 * @param[in] cex <code>cgul_exception</code> object
 */
extern CGUL_EXPORT void
cgul_exception__delete(cgul_exception_t cex);

/**
 * Get the message associated with the exception.
 *
 * @param[in] cex recently caught exception
 * @return message associated with the exception
 */
extern CGUL_EXPORT const char*
cgul_exception__get_message(cgul_exception_t cex);

CGUL_END_C

#endif  /*  CGUL_EXCEPTION_H  */
