#ifndef CGUL_HCHAR_H
#define CGUL_HCHAR_H

/**
 * This file defines <code>cgul_hchar_t</code> which is a 16-bit,
 * wide-character type.  It also defines the basic set of functions
 * that manipulate raw, 16-bit, wide-character strings.  It is called
 * "hchar" because it is half the size of <code>cgul_wchar_t</code>
 * (similar to how the <code>printf()</code> format string "%hu"
 * describes an <code>unsigned short</code>). <p>
 *
 * @file
 * @author Paul Serice
 * @brief cgul 16-bit wide-character support
 * @see cgul_hstring
 * @see cgul_unicode
 * @see cgul_wchar
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_int.h"

CGUL_BEGIN_C

/**
 * The <code>cgul_hchar_t</code> typedef always defines a 16-bit, wide
 * character which means it is always large enough to hold any Unicode
 * value in the Basic Multilingual Plane (BMP).  Surrogate pairs are
 * required for characters outside the BMP. <p>
 *
 * If you need to convert between <code>cgul_hchar_t</code> and
 * <code>wchar_t</code>, there are routines in
 * <code>cgul_unicode</code> that convert between UTF-16 and UTF-8.
 * You can then use an operating system dependent mechanism for
 * converting from UTF-8 to <code>wchar_t</code>.  Alternatively, it
 * may just be easier to write simple loop that copies each character
 * from a <code>cgul_hchar_t</code> string to a <code>wchar_t</code>
 * string or vice versa. <p>
 */
typedef cgul_uint16_t cgul_hchar_t;

/**
 * The NUL character in a <code>cgul_hchar_t</code> string.
 */
#define CGUL_HCHAR__NUL ((cgul_hchar_t)0)

/**
 * The Unicode Byte-Order Mark (BOM) character.
 */
#define CGUL_HCHAR__BOM ((cgul_hchar_t)0xfeff)

/**
 * The maximum number of bytes needed to hold any UTF-8 multi-byte
 * sequence that represents one Unicode character.  A surrogate pair
 * is required to represent Unicode values outside of the BMP (0x10000
 * - 0x10ffff).  A surrogate pair is two 16-bit characters or 4 bytes.
 */
#define CGUL_HCHAR__MB_LEN_MAX (4)

/**
 * The most annoying thing about using <code>cgul_hchar_t</code>
 * instead of <code>hchar_t</code> is that we lose support for the
 * compiler automatically generating wide-character strings by simply
 * prepending 'L' to the string literal.  Because we often need the
 * empty string, this class provides one as a convenience.
 */
extern CGUL_EXPORT const cgul_hchar_t CGUL_HCHAR__EMPTY_STRING[1];

/**
 * This function returns the number of 16-bit wide-characters in
 * <code>hs</code>.  The return value is determined by counting all
 * the wide characters that come before the trailing NUL wide
 * character.  The caller is responsible for insuring that
 * <code>hs</code> is NUL-terminated. <p>
 *
 * Note that this function does assumes the string is UCS-2, not
 * UTF-16, encoded.  Thus if the string is actually UTF-16 encoded,
 * surrogate pairs will count as two characters.  If you need an exact
 * count of unicode characters, use
 * <code>cgul_unicode__get_hchar_count()</code> instead. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] hs 16-bit wide-character string
 * @return number of 16-bit wide characters in <code>hs</code>
 */
extern CGUL_EXPORT size_t
cgul_hchar__hcslen(cgul_exception_t* cex,
                   const cgul_hchar_t* hs);

/**
 * This function copies the 16-bit wide characters in <code>src</code>
 * to <code>dst</code>.  The caller is responsible for making sure
 * that <code>src</code> is NUL-terminated and that <code>dst</code>
 * is large enough to hold all the characters in <code>src</code>
 * including the trailing NUL character.
 *
 * @param[in] cex c-style exception
 * @param[in] dst destination 16-bit wide-character string
 * @param[in] src source 16-bit wide-character string
 */
extern CGUL_EXPORT void
cgul_hchar__hcscpy(cgul_exception_t* cex,
                   cgul_hchar_t* dst,
                   const cgul_hchar_t* src);

/**
 * Perform a case-significant, 16-bit, wide-character string
 * comparison of <code>hs1</code> and <code>hs2</code>.  Return -1, 0,
 * or 1 if <code>hs1</code> is less than, equal to, or greater than
 * <code>hs2</code> respectively. <p>
 *
 * For simplicity, this function compares strings stricly by the
 * ordinal value of each character which should be sufficient, for
 * example, when comparing strings when inserting them as keys into a
 * <code>cgul_rbtree</code>; however, the comparison results are not
 * likely to be the same as what <code>strcoll()</code> would return
 * if used on the UTF-8 version of the strings. <p>
 *
 * @note This function does not take a <code>cgul_exception</code>
 * object as its first parameter in order to make it easier to use the
 * function as the comparison function of <code>cgul_rbtree</code>
 * objects. <p>
 *
 * @param[in] hs1 left-hand side
 * @param[in] hs2 right-hand side
 * @return -1, 0, or 1 if <code>hs1</code> is less than, equal to, or
 * greater than <code>hs2</code> respectively
 * @see cgul_hstring__compare()
 */
extern CGUL_EXPORT int
cgul_hchar__hcscmp(const cgul_hchar_t* hs1,
                   const cgul_hchar_t* hs2);

/**
 * Return whether the 16-bit, wide character <code>hc</code> is
 * considered to be white-space.  To avoid locale dependencies,
 * white-space is defined as any character from the following list:
 * ' ', '\\t', '\\n', '\\r', '\\f', '\\v'
 *
 * @param[in] cex c-style exception
 * @param[in] hc 16-bit, wide character
 * @return whether the character is white-space
 */
extern CGUL_EXPORT int
cgul_hchar__isspace(cgul_exception_t* cex,
                    cgul_hchar_t hc);

CGUL_END_C

#endif  /*  CGUL_HCHAR_H  */
