#ifndef CGUL_HCHAR_CXX_H
#define CGUL_HCHAR_CXX_H

/**
 * C++ wrapper for <code>cgul_hchar</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_hchar</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_hchar.h"

/**
 * This class provides the C++ bindings for C <code>cgul_hchar</code>
 * external functions.  The main purpose of this class is to convert
 * the C-style function calls and exception handling in
 * <code>cgul_hchar</code> into C++-style function calls and exception
 * handling.
 *
 * @brief C++ bindings for <code>cgul_hchar</code>
 */
class cgul_hchar_cxx {

public:

    /**
     * This function returns the number of 16-bit wide-characters in
     * <code>hs</code>.  The return value is determined by counting all
     * the wide characters that come before the trailing NUL wide
     * character.  The caller is responsible for insuring that
     * <code>hs</code> is NUL-terminated. <p>
     *
     * Note that this function does assumes the string is UCS-2, not
     * UTF-16, encoded.  Thus if the string is actually UTF-16 encoded,
     * surrogate pairs will count as two characters.  If you need an exact
     * count of unicode characters, use
     * <code>cgul_unicode_cxx::get_hchar_count()</code> instead. <p>
     *
     * @param[in] hs 16-bit wide-character string
     * @return number of 16-bit wide characters in <code>hs</code>
     */
    static size_t
    hcslen(const cgul_hchar_t* hs)
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_hchar__hcslen(&local, hs);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This function copies the 16-bit wide characters in <code>src</code>
     * to <code>dst</code>.  The caller is responsible for making sure
     * that <code>src</code> is NUL-terminated and that <code>dst</code>
     * is large enough to hold all the characters in <code>src</code>
     * including the trailing NUL character.
     *
     * @param[in] dst destination 16-bit wide-character string
     * @param[in] src source 16-bit wide-character string
     */
    static void
    hcscpy(cgul_hchar_t* dst,
           const cgul_hchar_t* src)
    {
        cgul_exception_t local = NULL;
        cgul_hchar__hcscpy(&local, dst, src);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform a case-significant, 16-bit, wide-character string
     * comparison of <code>hs1</code> and <code>hs2</code>.  Return -1, 0,
     * or 1 if <code>hs1</code> is less than, equal to, or greater than
     * <code>hs2</code> respectively. <p>
     *
     * For simplicity, this function compares strings stricly by the
     * ordinal value of each character which should be sufficient, for
     * example, when comparing strings when inserting them as keys into a
     * <code>cgul_rbtree</code>; however, the comparison results are not
     * likely to be the same as what <code>strcoll()</code> would return
     * if used on the UTF-8 version of the strings. <p>
     *
     * @note This function does not take a <code>cgul_exception</code>
     * object as its first parameter in order to make it easier to use the
     * function as the comparison function of <code>cgul_rbtree</code>
     * objects. <p>
     *
     * @param[in] hs1 left-hand side
     * @param[in] hs2 right-hand side
     * @return -1, 0, or 1 if <code>hs1</code> is less than, equal to, or
     * greater than <code>hs2</code> respectively
     * @see cgul_hstring_cxx::compare()
     */
    static int
    hcscmp(const cgul_hchar_t* hs1,
           const cgul_hchar_t* hs2)
    {
        return cgul_hchar__hcscmp(hs1, hs2);
    }

    /**
     * Return whether the 16-bit, wide character <code>hc</code> is
     * considered to be white-space.  To avoid locale dependencies,
     * white-space is defined as any character from the following list:
     * ' ', '\\t', '\\n', '\\r', '\\f', '\\v'
     *
     * @param[in] hc 16-bit, wide character
     * @return whether the character is white-space
     */
    static int
    isspace(cgul_hchar_t hc)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_hchar__isspace(&local, hc);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

};

#endif  /*  CGUL_HCHAR_CXX_H  */
