#ifndef CGUL_HEAP_H
#define CGUL_HEAP_H

/**
 * Binary heap implementation.  This class can be used to trivially
 * implement a priority queue.  By default, this class implements a
 * max-heap but can be used as a min-heap simply by negating the
 * output of the comparison function. <p>

 * @file
 * @author Paul Serice
 * @brief binary heap
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_heap_primitives.h"
#include "cgul_vector.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_heap</code> instance. */
typedef struct cgul_heap* cgul_heap_t;

/**
 * Create a new <code>cgul_heap</code> object that maintains heap
 * order using the comparison function <code>cf</code>.  The caller is
 * responsible for freeing the object by calling
 * <code>cgul_heap__delete()</code>.  If memory cannot be allocated,
 * <code>NULL</code> is returned, and an exception is thrown. <p>
 *
 * @note <code>cgul_heap</code> is by default a max-heap, but it can
 * be used as a min-heap simply by negating the output of the
 * comparison function. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] cf comparison function
 * @return new <code>cgul_heap</code> instance
 * @see cgul_heap__new_from_vector()
 */
extern CGUL_EXPORT cgul_heap_t
cgul_heap__new(cgul_exception_t* cex,
               cgul_heap_primitives__compare_t cf);

/**
 * Create a new <code>cgul_heap</code> object from the vector
 * <code>v</code>.  The new heap maintains heap order using the
 * comparison function <code>cf</code>.  The caller is responsible for
 * freeing the object by calling <code>cgul_heap__delete()</code>.
 * This class takes ownership of the vector so that calling
 * <code>cgul_heap__delete()</code> also deletes the vector.  If
 * memory cannot be allocated, <code>NULL</code> is returned, and an
 * exception is thrown.
 *
 * @note <code>cgul_heap</code> is by default a max-heap, but it can
 * be used as a min-heap simply by negating the output of the
 * comparison function. <p>
 *
 * @note The asymptotic complexity of this method is O(n) which is
 * more efficient than creating an empty heap and inserting n elements
 * which would be O(n log n).
 *
 * @param[in,out] cex c-style exception
 * @param[in] v vector
 * @param[in] cf comparison function
 * @return new <code>cgul_heap</code> instance
 */
extern CGUL_EXPORT cgul_heap_t
cgul_heap__new_from_vector(cgul_exception_t* cex,
                           cgul_vector_t v,
                           cgul_heap_primitives__compare_t cf);

/**
 * This method frees all internally allocated memory.  This method
 * does not free the elements stored in the heap.  The caller
 * allocated those elements so the caller is responsible for freeing
 * them.  Do not try to access <code>heap</code> after calling this method.
 *
 * @param[in] heap <code>cgul_heap</code> instance
 * @see cgul_heap__free_values()
 */
extern CGUL_EXPORT void
cgul_heap__delete(cgul_heap_t heap);

/**
 * This method calls <code>free()</code> on all the values in the heap
 * <code>heap</code>.  Because this is such a common operation, it is
 * an exception to the rule that cgul containers never free values.
 * This method should only ever be called immediately before calling
 * <code>cgul_heap__delete()</code> because it otherwise invalidates
 * the heap.  Because this function is basically an extension of
 * <code>cgul_heap__delete()</code>, it does not accept a
 * <code>cgul_exception</code> object as an input parameter.
 *
 * @param[in] heap <code>cgul_heap</code> instance
 */
extern CGUL_EXPORT void
cgul_heap__free_values(cgul_heap_t heap);

/**
 * Return true if the heap <code>heap</code> is empty.
 *
 * @param[in] cex c-style exception
 * @param[in] heap <code>cgul_heap</code> instance
 * @return whether <code>heap</code> is empty
 */
extern CGUL_EXPORT int
cgul_heap__is_empty(cgul_exception_t* cex,
                    cgul_heap_t heap);

/**
 * Push the value <code>value</code> onto the heap <code>heap</code>
 * maintaining heap order using the some comparison function that was
 * passed into the heap constructor.  If an error occurs, an exception
 * is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] heap <code>cgul_heap</code> instance
 * @param[in] value to push onto the heap
 */
extern CGUL_EXPORT void
cgul_heap__push(cgul_exception_t* cex,
                cgul_heap_t heap,
                void* value);

/**
 * Pop the top value from the heap <code>heap</code> maintaining heap
 * order using the some comparison function that was passed into the
 * heap constructor.  If the heap is empty, <code>NULL</code> is
 * returned.
 *
 * @param[in] cex c-style exception
 * @param[in] heap <code>cgul_heap</code> instance
 * @return top value
 */
extern CGUL_EXPORT void*
cgul_heap__pop(cgul_exception_t* cex,
               cgul_heap_t heap);

/**
 * Return the top value from the heap <code>heap</code> without
 * removing it from the heap.  If the heap is empty, <code>NULL</code>
 * is returned.
 *
 * @param[in] cex c-style exception
 * @param[in] heap <code>cgul_heap</code> instance
 * @return top value
 */
extern CGUL_EXPORT void*
cgul_heap__top(cgul_exception_t* cex,
               cgul_heap_t heap);

/**
 * Return the size of the heap <code>heap</code>.  This is a count of
 * the total number of elements stored in the heap.
 *
 * @param[in] cex c-style exception
 * @param[in] heap <code>cgul_heap</code> instance
 * @return size of the heap
 */
extern CGUL_EXPORT unsigned long int
cgul_heap__get_size(cgul_exception_t* cex,
                    cgul_heap_t heap);

/**
 * Swap the underlying data for <code>heap1</code> and <code>heap2</code>.
 * For large heaps, this should be much faster than trying to do the
 * same thing using removes and inserts.
 *
 * @param[in] cex c-style exception
 * @param[in] heap1 first <code>cgul_heap</code> instance
 * @param[in] heap2 second <code>cgul_heap</code> instance
 */
extern CGUL_EXPORT void
cgul_heap__swap(cgul_exception_t* cex,
                cgul_heap_t heap1,
                cgul_heap_t heap2);

CGUL_END_C

#endif  /*  CGUL_HEAP_H  */
