#ifndef CGUL_HEAP_CXX_H
#define CGUL_HEAP_CXX_H

/**
 * C++ wrapper for <code>cgul_heap</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_heap</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_heap_primitives_cxx.h"
#include "cgul_heap.h"
#include "cgul_vector_cxx.h"

/**
 * This class provides the C++ bindings for C <code>cgul_heap</code>
 * objects.  The main purpose of this class is to convert the C-style
 * function calls and exception handling in <code>cgul_heap</code>
 * into C++-style function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_heap</code>
 * @see cgul_heap_t
 */
class cgul_heap_cxx {

public:

    /**
     * Create a new <code>cgul_heap_cxx</code> object that maintains
     * heap order using the comparison function <code>cf</code>.  The
     * caller is responsible for freeing the object by calling
     * <code>delete</code> on the object.  If memory cannot be
     * allocated, an exception is thrown. <p>
     *
     * @note <code>cgul_heap_cxx</code> is by default a max-heap, but
     * it can be used as a min-heap simply by negating the output of
     * the comparison function. <p>
     *
     * @param[in] cf comparison function
     * @return new <code>cgul_heap_cxx</code> instance
     */
    cgul_heap_cxx(cgul_heap_primitives_cxx::compare_t cf)
    {
        cgul_vector_cxx v;
        initialize(v, cf);
    }

    /**
     * Create a new <code>cgul_heap_cxx</code> object from the vector
     * <code>v</code>.  The new heap maintains heap order using the
     * comparison function <code>cf</code>.  The caller is responsible
     * for freeing the object by calling <code>delete</code> on the
     * object.  This class takes ownership of the underlying
     * <code>cgul_vector</code> so that deleting the object also
     * deletes the underlying vector.  If memory cannot be allocated,
     * an exception is thrown.
     *
     * @note <code>cgul_heap_cxx</code> is by default a max-heap, but
     * it can be used as a min-heap simply by negating the output of
     * the comparison function. <p>
     *
     * @note The asymptotic complexity of this method is O(n) which is
     * more efficient than creating an empty heap and inserting n elements
     * which would be O(n log n).
     *
     * @param[in] v vector
     * @param[in] cf comparison function
     */
    cgul_heap_cxx(cgul_vector_cxx& v,
                  cgul_heap_primitives_cxx::compare_t cf)
    {
        initialize(v, cf);
    }

    /**
     * Create a new <code>cgul_heap_cxx</code> object by wrapping
     * an existing <code>cgul_heap</code> object.
     *
     * @param[in] rhs right-hand side
     */
    cgul_heap_cxx(cgul_heap_t rhs)
    {
        obj = rhs;
    }

    /**
     * This method frees all internally allocated memory.  This method
     * does not free the elements stored in the heap.  The caller
     * allocated those elements so the caller is responsible for freeing
     * them.  Do not try to access <code>heap</code> after calling this method.
     *
     * @see free_values()
     */
    virtual
    ~cgul_heap_cxx()
    {
        if (obj) {
            cgul_heap__delete(obj);
            obj = NULL;
        }
    }

    /**
     * This method calls <code>free()</code> on all the values in the
     * heap <code>heap</code>.  Because this is such a common
     * operation, it is an exception to the rule that cgul containers
     * never free values.  This method should only ever be called
     * immediately before calling <code>delete</code> on the object
     * because it otherwise invalidates the heap.
     */
    virtual void
    free_values()
    {
        cgul_heap__free_values(obj);
    }

    /**
     * Return true if the heap is empty.
     *
     * @return whether the heap is empty
     */
    virtual int
    is_empty() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_heap__is_empty(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Push the value <code>value</code> onto the heap maintaining
     * heap order using the some comparison function that was passed
     * into the heap constructor.  If an error occurs, an exception is
     * thrown.
     *
     * @param[in] value to push onto the heap
     */
    virtual void
    push(void* value)
    {
        cgul_exception_t local = NULL;
        cgul_heap__push(&local, obj, value);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Pop the top value from the heap maintaining heap order using
     * the some comparison function that was passed into the heap
     * constructor.  If the heap is empty, <code>NULL</code> is
     * returned.
     *
     * @return top value
     */
    virtual void*
    pop()
    {
        void* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_heap__pop(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the top value from the heap without removing it from the
     * heap.  If the heap is empty, <code>NULL</code> is returned.
     *
     * @return top value
     */
    virtual void*
    top()
    {
        void* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_heap__top(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the size of the heap.  This is a count of the total number
     * of elements stored in the heap.
     *
     * @return size of the heap
     */
    virtual unsigned long int
    get_size() const
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_heap__get_size(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Swap the underlying data for this object and <code>rhs</code>.
     * For large heaps, this should be much faster than trying to do the
     * same thing using removes and inserts.
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    swap(cgul_heap_cxx& rhs)
    {
        cgul_exception_t local = NULL;
        cgul_heap__swap(&local, this->obj, rhs.obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the underlying <code>cgul_heap</code> object.
     *
     * @return underlying object
     */
    virtual cgul_heap_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_heap</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_heap__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_heap_t
    take_obj()
    {
        cgul_heap_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_heap_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_heap__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_heap_cxx();
    cgul_heap_cxx(const cgul_heap_cxx&);
    cgul_heap_cxx& operator=(const cgul_heap_cxx&);

    /**
     * Initialize a new <code>cgul_heap_cxx</code> instance from the
     * vector <code>v</code> and the comparison function
     * <code>cf</code>.  This method takes ownership of the underlying
     * <code>cgul_vector</code>.
     *
     * @param[in] v vector
     * @param[in] cf comparison function
     */
    void
    initialize(cgul_vector_cxx& v,
               cgul_heap_primitives_cxx::compare_t cf)
    {
        cgul_exception_t local = NULL;
        obj = cgul_heap__new_from_vector(&local, v.take_obj(), cf);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /** Underlying <code>cgul_heap</code> object. */
    cgul_heap_t obj;

};

#endif  /*  CGUL_HEAP_CXX_H  */
