#ifndef CGUL_HEAP_PRIMITIVES_H
#define CGUL_HEAP_PRIMITIVES_H

/**
 * Primitive heap operations that can be performed on a Binary Heap
 * represented as an array.
 *
 * @file
 * @author Paul Serice
 * @brief primitive heap operations
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/**
 * This typedef is the interface the client must define in order for
 * the <code>cgul_heap_primitives</code> functions to maintain heap
 * order.  When defining the comparison function, it should compare
 * <code>e1</code> to <code>e2</code> and return less than zero, zero,
 * or greater than zero if <code>e1</code> is less than, equal to, or
 * greater than <code>e2</code>.
 *
 * @param[in] e1 first element
 * @param[in] e2 second element
 * @return result of comparison
 */
typedef int (*cgul_heap_primitives__compare_t)(const void* e1,
                                               const void* e2);

/**
 * This function converts the unordered array <code>v</code> into a
 * heap.  This is done using the bottom-up algorithm which is O(n).
 *
 * @param[in] cex c-style exception
 * @param[in] v array of pointers to elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] cf comparison function
 */
extern CGUL_EXPORT void
cgul_heap_primitives__heapify(cgul_exception_t* cex,
                              void** v,
                              size_t v_count,
                              cgul_heap_primitives__compare_t cf);

/**
 * Shift the element at <code>v[index]</code> down until it is greater
 * than or equal to all of its descendants as determined by the
 * comparison function <code>cf</code>.  (This code is for a
 * max-heap.) <p>
 *
 * If <code>floyd</code> is true, Floyd's Method is used as an
 * optimization to reduce the number of callbacks required to the
 * comparison function.  It works by initially assuming the value
 * being shifted down is smaller than both children causing the larger
 * child to be shifted up at each level until a leaf node is reached.
 * Then, <code>cgul_heap_primitives__shift_up()</code> is called to
 * find the correct location for the value.  This is an optimization
 * when forced to replace the root value with a value from the back of
 * the array which is probably smaller than all the values along the
 * path described.
 *
 * @note Do *not* use Floyd's Method until after the heap has been
 * fully constructed because <code>cgul_heap_primitives__shift_up()</code>
 * assumes the heap is fully constructed. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] v array of pointers to elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] index index of element to be shifted down
 * @param[in] use_floyds_method use Floyd's Method
 * @param[in] cf comparison function
 */
extern CGUL_EXPORT void
cgul_heap_primitives__shift_down(cgul_exception_t* cex,
                                 void** v,
                                 size_t v_count,
                                 size_t index,
                                 int use_floyds_method,
                                 cgul_heap_primitives__compare_t cf);

/**
 * Shift the element at <code>v[index]</code> up until it is the root
 * element or until it is greater than or equal to all of its
 * descendants but less than its parent as determined by the
 * comparison function <code>cf</code>.  (This code is for a
 * max-heap.)
 *
 * @param[in] cex c-style exception
 * @param[in] v array of pointers to elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] index index of element to be shifted down
 * @param[in] cf comparison function
 */
extern CGUL_EXPORT void
cgul_heap_primitives__shift_up(cgul_exception_t* cex,
                               void** v,
                               size_t v_count,
                               size_t index,
                               cgul_heap_primitives__compare_t cf);

CGUL_END_C

#endif  /*  CGUL_HEAP_PRIMITIVES_H  */
