#ifndef CGUL_LIBC_H
#define CGUL_LIBC_H

/**
 * Wrapper around common functions that converts from tradition
 * c-style error reporting to throwing a c-style exception using
 * <code>cgul_exception</code>.
 *
 * @file
 * @author Paul Serice
 * @brief wrapper for throwing c-style exceptions from common functions
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/**
 * This method lets you test out-of-memory exception handling by
 * forcing <code>cgul_libc__calloc()</code>,
 * <code>cgul_libc__malloc()</code>, and
 * <code>cgul_libc__realloc()</code> to skip memory allocation and
 * immediately return <code>NULL</code>.  You should set
 * <code>out_of_memory</code> immediately before the allocation call
 * you want to test and clear it immediately after.
 *
 * @param[in] cex c-style exception
 * @param[in] out_of_memory whether to simulate out-of-memory condition
 */
extern CGUL_EXPORT void
cgul_libc__set_out_of_memory(cgul_exception_t* cex,
                             int out_of_memory);

/**
 * Wrapper for <code>calloc()</code>.  On success, a pointer to a
 * block of memory that is large enough to hold an array of
 * <code>element_count</code> elements is returned.  The block of
 * memory is initialized with all zeros, and the client must call
 * <code>free()</code> on the value returned.  On failure, an
 * exception is thrown, and <code>NULL</code> is returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] element_count number of elements to allocate space for
 * @param[in] element_size size of one element
 * @return pointer to a block of memory
 */
extern CGUL_EXPORT void*
cgul_libc__calloc(cgul_exception_t* cex,
                  size_t element_count,
                  size_t element_size);

/**
 * Wrapper for <code>malloc()</code>.  On success, a pointer to a
 * block of memory that is <code>new_size</code> bytes long is
 * returned uninitialized, and the client must call
 * <code>free()</code> on the value returned.  On failure, an
 * exception is thrown, and <code>NULL</code> is returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] new_size new size for the block of memory to return
 * @return point to a block of memory
 */
extern CGUL_EXPORT void*
cgul_libc__malloc(cgul_exception_t* cex,
                  size_t new_size);

/**
 * Wrapper for <code>realloc()</code>.  On success, a pointer to
 * memory that holds <code>new_size</code> bytes is returned, and the
 * client must eventually call <code>free()</code> on the value
 * returned.  The client must not call <code>free()</code> on
 * <code>orig_pointer</code> if this function returns
 * successfully. <p>
 *
 * If an error occurs, an exception will be thrown, and
 * <code>orig_pointer</code> will be returned, and the client will
 * still need to call <code>free()</code> on
 * <code>orig_pointer</code>.  (Notice that this is different behavior
 * from <code>realloc()</code> which returns <code>NULL</code> if an
 * error occurs.)
 *
 * @param[in,out] cex c-style exception
 * @param[in] orig_pointer original pointer
 * @param[in] new_size new size for the pointer that is returned
 * @return pointer to a block of memory
 */
extern CGUL_EXPORT void*
cgul_libc__realloc(cgul_exception_t* cex,
                   void* orig_pointer,
                   size_t new_size);

/**
 * Wrapper for <code>memset()</code>.  Set a region of memory pointed
 * to by <code>buf</code> by repeatedly copying <code>c</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] buf buffer where memory will be set
 * @param[in] c byte repeatedly used to set the memory region
 * @param[in] n buffer size
 * @return <code>buf</code>
 */
extern CGUL_EXPORT void*
cgul_libc__memset(cgul_exception_t* cex,
                  void *buf,
                  int c,
                  size_t n);

/**
 * Wrapper for <code>memcpy()</code>.  Copy the first <code>n</code>
 * bytes from the <code>src</code> buffer to the <code>dst</code>
 * buffer.  It is important that the <code>src</code> and
 * <code>dst</code> buffers do not overlap.  If they do, you should
 * use <code>cgul_libc__memmove()</code> instead.
 *
 * @param[in] cex c-style exception
 * @param[in] dst destination buffer
 * @param[in] src source buffer
 * @param[in] n buffer size
 * @return <code>dst</code>
 */
extern CGUL_EXPORT void*
cgul_libc__memcpy(cgul_exception_t* cex,
                  void* dst,
                  const void* src,
                  size_t n);

/**
 * Wrapper for <code>memmove()</code>.  Copy the first <code>n</code>
 * bytes from the <code>src</code> buffer to the <code>dst</code>
 * buffer.  The <code>src</code> and <code>dst</code> buffers may
 * overlap.
 *
 * @param[in] cex c-style exception
 * @param[in] dst destination buffer
 * @param[in] src source buffer
 * @param[in] n buffer size
 * @return <code>dst</code>
 */
extern CGUL_EXPORT void*
cgul_libc__memmove(cgul_exception_t* cex,
                   void* dst,
                   const void* src,
                   size_t n);

/**
 * Wrapper for <code>fopen()</code>.  It opens the file with name
 * <code>fname</code> using the mode specified by <code>mode</code>.
 * If an error occurs, <code>NULL</code> is returned and an exception
 * is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] fname file name
 * @param[in] mode open file mode
 * @return opened file
 */
extern CGUL_EXPORT FILE*
cgul_libc__fopen(cgul_exception_t* cex,
                 const char* fname,
                 const char* mode);

/**
 * Wrapper for <code>freopen()</code>.  It reopens <code>f</code> so
 * that the underlying file descriptor points to <code>fname</code>
 * opened using <code>mode</code>.  If an error occurs,
 * <code>NULL</code> is returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] fname file name
 * @param[in] mode open file mode
 * @param[in] f file object to reopen
 * @return opened file
 */
extern CGUL_EXPORT FILE*
cgul_libc__freopen(cgul_exception_t* cex,
                   const char* fname,
                   const char* mode,
                   FILE* f);

/**
 * Wrapper for <code>fclose()</code>.  It closes <code>f</code>.  If
 * an error occurs, an exception is thrown.  You should be careful
 * using this function because you will almost always want to close
 * files if the main part of your code throws an exception.  Thus, if
 * the stack is unwinding, to avoid this function from overwriting the
 * current exception, you should probably call <code>fclose()</code>
 * directly instead of calling this function.
 *
 * @param[in,out] cex c-style exception
 * @param[in] f file object to close
 */
extern CGUL_EXPORT void
cgul_libc__fclose(cgul_exception_t* cex,
                  FILE *f);

/**
 * Wrapper for <code>fflush()</code>.  It flushes the file
 * <code>f</code>, or if <code>f</code> is <code>NULL</code>, it
 * flushes all open files.  If an error occurs, an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] f file object to close
 */
extern CGUL_EXPORT void
cgul_libc__fflush(cgul_exception_t* cex,
                  FILE* f);

/**
 * Wrapper for <code>fread()</code>.  It reads
 * <code>element_count</code> elements of size
 * <code>element_size</code> each into <code>buf</code> from
 * <code>f</code>.  On success, the number of elements read (which is
 * not necessarily the same as the number of bytes read) is returned.
 * If EOF is reached, a byte count of 0 is returned.  On failure,
 * <code>(size_t)-1</code> is returned and an exception is thrown.
 *
 * @param[in,out] cex cex c-style exception
 * @param[in] buf destination buffer
 * @param[in] element_size size of each element to read
 * @param[in] element_count number of elements to read
 * @param[in] f file from which to read
 * @return number of elements read
 * @see cgul_libc__fread_block()
 */
extern CGUL_EXPORT size_t
cgul_libc__fread(cgul_exception_t* cex,
                 void* buf,
                 size_t element_size,
                 size_t element_count,
                 FILE* f);

/**
 * Read the block <code>block</code> of size <code>block_size</code>
 * bytes from the file <code>fin</code>.  Unlike
 * <code>cgul_libc__fread()</code>, the read is restarted if
 * interrupted.  This function returns the total number of bytes read.
 * If an error occurs, the total number of bytes read is returned
 * (which could be a short count), and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] block block
 * @param[in] block_size size of block in bytes
 * @param[in] fin input file
 * @return total number of bytes read
 * @see cgul_libc__fread()
 */
extern CGUL_EXPORT size_t
cgul_libc__fread_block(cgul_exception_t* cex,
                       void* block,
                       size_t block_size,
                       FILE* fin);

/**
 * Wrapper for <code>fwrite()</code>.  It writes
 * <code>element_count</code> elements of size
 * <code>element_size</code> each from <code>buf</code> into
 * <code>f</code>.  On success, the number of elements written (which
 * is not necessarily the same as the number of bytes written) is
 * returned.  On failure, <code>(size_t)-1</code> is returned and an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] buf source buffer
 * @param[in] element_size size of each element to write
 * @param[in] element_count number of elements to write
 * @param[in] f file to which to write
 * @return number of elements written
 * @see cgul_libc__fwrite_block()
 */
extern CGUL_EXPORT size_t
cgul_libc__fwrite(cgul_exception_t* cex,
                  const void* buf,
                  size_t element_size,
                  size_t element_count,
                  FILE* f);

/**
 * Write the block <code>block</code> of size <code>block_size</code>
 * bytes to the output file <code>fout</code>.  Unlike
 * <code>cgul_libc__fwrite()</code>, the write is restarted if
 * interrupted.  This function returns the total number of bytes
 * written.  If an error occurs, the total number of bytes written is
 * returned (which could be a short count), and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] block block
 * @param[in] block_size size of block in bytes
 * @param[in] fout output file
 * @return total number of bytes written
 * @see cgul_libc__fwrite()
 */
extern CGUL_EXPORT size_t
cgul_libc__fwrite_block(cgul_exception_t* cex,
                        const void* block,
                        size_t block_size,
                        FILE* fout);

/**
 * Load the binary file <code>fin</code> into memory saving a copy of
 * the starting location of the file in memory to <code>*ptr</code>
 * and its length to <code>*length</code> if <code>length</code> is
 * not <code>NULL</code>.  The caller is responsible for calling
 * <code>free()</code> on <code>*ptr</code>.  If an error occurs, an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] fin input file
 * @param[out] ptr pointer to in-memory copy of file
 * @param[out] length length in bytes of the in-memory copy of the file
 * @see cgul_libc__save_file()
 */
extern CGUL_EXPORT void
cgul_libc__load_file(cgul_exception_t* cex,
                     FILE* fin,
                     char** ptr,
                     size_t* length);

/**
 * Load the binary file with name <code>fname</code> into memory
 * saving a copy of the starting location of the file in memory to
 * <code>*ptr</code> and its length to <code>*length</code> if
 * <code>length</code> is not <code>NULL</code>.  The caller is
 * responsible for calling <code>free()</code> on <code>*ptr</code>.
 * If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] fname name of input file
 * @param[out] ptr pointer to in-memory copy of file
 * @param[out] length length in bytes of the in-memory copy of the file
 * @see cgul_libc__save_fname()
 */
extern CGUL_EXPORT void
cgul_libc__load_fname(cgul_exception_t* cex,
                      const char* fname,
                      char** ptr,
                      size_t* length);

/**
 * Save the binary block of memory <code>block</code> extending for
 * <code>block_size</code> bytes to the output file <code>fout</code>.
 * If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] fout output file
 * @param[in] block block
 * @param[in] block_size size of block in bytes
 * @see cgul_libc__load_file()
 */
extern CGUL_EXPORT void
cgul_libc__save_file(cgul_exception_t* cex,
                     FILE* fout,
                     const char* block,
                     size_t block_size);

/**
 * Save the binary block of memory <code>block</code> extending for
 * <code>block_size</code> bytes to the output file with name
 * <code>fname</code>.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] fname name of the output file
 * @param[in] block block
 * @param[in] block_size size of block in bytes
 * @see cgul_libc__load_fname()
 */
extern CGUL_EXPORT void
cgul_libc__save_fname(cgul_exception_t* cex,
                      const char* fname,
                      const char* block,
                      size_t block_size);

/**
 * Copy everything from <code>fin</code> to <code>fout</code>.  If an
 * error occurs, an exception is thrown.  This is not a wrapper around
 * any libc function.
 *
 * @param[in,out] cex c-style exception
 * @param[in] fin input file
 * @param[in] fout output file
 */
extern CGUL_EXPORT void
cgul_libc__copy_file(cgul_exception_t* cex,
                     FILE* fin,
                     FILE* fout);

/**
 * Wrapper for <code>ftell()</code>.  On success, the file position of
 * <code>f</code> is returned.  On failure, -1 is returned, and an
 * exception is thrown.
 * 
 * @param[in,out] cex c-style exception
 * @param[in] f file
 * @return file position
 */
extern CGUL_EXPORT long
cgul_libc__ftell(cgul_exception_t* cex,
                 FILE* f);

/**
 * Wrapper for <code>fseek()</code>.  On success, the file is
 * positioned according to the values of <code>offset</code> and
 * <code>whence</code> where <code>offset</code> is relative to the
 * beginning of the file, the current position, or the end of the file
 * if <code>whence</code> is <code>SEEK_SET</code>,
 * <code>SEEK_CUR</code>, or <code>SEEK_END</code> respectively.  On
 * failure, an exception is thrown.
 * 
 * @param[in,out] cex c-style exception
 * @param[in] f file
 * @param[in] offset offset
 * @param[in] whence position to which offset is relative
 */
extern CGUL_EXPORT void
cgul_libc__fseek(cgul_exception_t* cex,
                 FILE* f,
                 long offset,
                 int whence);

/**
 * Wrapper for <code>rewind()</code> to rewind <code>f</code>.  An
 * exception is thrown if <code>f</code> is not seekable.
 *
 * @param[in,out] cex c-style exception
 * @param[in] f file
 */
extern CGUL_EXPORT void
cgul_libc__rewind(cgul_exception_t* cex,
                  FILE* f);

/**
 * Wrapper for <code>fgetpos()</code>.  On success, the current file
 * position for <code>f</code> is stored in pos <code>pos</code>.  On
 * failure, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] f file
 * @param[out] pos current file position
 */
extern CGUL_EXPORT void
cgul_libc__fgetpos(cgul_exception_t* cex,
                   FILE* f,
                   fpos_t* pos);

/**
 * Wrapper for <code>fsetpos()</code>.  On success, the current file
 * position for <code>f</code> is set to <code>pos</code>.  On
 * failure, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] f file
 * @param[out] pos current file position
 */
extern CGUL_EXPORT void
cgul_libc__fsetpos(cgul_exception_t* cex,
                   FILE* f,
                   fpos_t* pos);

/**
 * Wrapper for <code>tmpfile()</code>.  On success, a file opened for
 * reading and writting is returned.  On failure, <code>NULL</code> is
 * returned and an exception is thrown.  When opening a file using
 * glibc-2.3.4, <code>strace</code> returned the following which says
 * the file is opened without race conditions and is immediately
 * unlinked so that the file is removed even if your program exits
 * unexpectedly:
 *
 * <pre>
 *     open("/tmp/tmpfGQda9F", O_RDWR|O_CREAT|O_EXCL, 0600) = 3
 *     unlink("/tmp/tmpfGQda9F")               = 0
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @return temporary file
 */
extern CGUL_EXPORT FILE*
cgul_libc__tmpfile(cgul_exception_t* cex);

/**
 * Wrapper for <code>cgul_libc__vprintf()</code>.  On success, the
 * number of characters printed to <code>stdout</code> is returned.  On
 * failure, -1 is returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] format format string
 * @param[in] ... variable argument list that matches <code>format</code>
 * @return number of characters read
 */
extern CGUL_EXPORT int
cgul_libc__printf(cgul_exception_t* cex,
                  const char* format,
                  ...)
#if defined(__GNUC__) && !defined(__MINGW_H)
    __attribute__ ((format (printf, 2, 3)))
#endif
;

/**
 * Wrapper for <code>vprintf()</code>.  On success, the number of
 * characters printed to <code>stdout</code> is returned.  On failure,
 * -1 is returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] format format string
 * @param[in] args variable argument list that matches <code>format</code>
 * @return number of characters read
 */
extern CGUL_EXPORT int
cgul_libc__vprintf(cgul_exception_t* cex,
                   const char* format,
                   va_list args);

/**
 * Wrapper for <code>cgul_libc__vfprintf()</code>.  On success, the
 * number of characters printed to <code>f</code> is returned.  On
 * failure, -1 is returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] f output file
 * @param[in] format format string
 * @param[in] ... variable argument list that matches <code>format</code>
 * @return number of characters read
 */
extern CGUL_EXPORT int
cgul_libc__fprintf(cgul_exception_t* cex,
                   FILE* f,
                   const char* format,
                   ...)
#if defined(__GNUC__) && !defined(__MINGW_H)
    __attribute__ ((format (printf, 3, 4)))
#endif
;

/**
 * Wrapper for <code>vfprintf()</code>.  On success, the number of
 * characters printed to <code>f</code> is returned.  On failure, -1
 * is returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] f output file
 * @param[in] format format string
 * @param[in] args variable argument list that matches <code>format</code>
 * @return number of characters read
 */
extern CGUL_EXPORT int
cgul_libc__vfprintf(cgul_exception_t* cex,
                    FILE* f,
                    const char* format,
                    va_list args);

/**
 * Wrapper for <code>fgetc()</code>.  Return the next character from
 * <code>f</code>.  If the EOF is reached, <code>EOF</code> is
 * returned.  If an error occurs, <code>EOF</code> is returned and an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] f input file
 * @return character read from <code>f</code>
 */
extern CGUL_EXPORT int
cgul_libc__fgetc(cgul_exception_t* cex,
                 FILE* f);

/**
 * Wrapper for <code>fputc()</code>.  Cast <code>c</code> to an
 * <code>unsigned char</code> and write it to <code>f</code>.  If an
 * error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] c character to write
 * @param[in] f output file
 */
extern CGUL_EXPORT void
cgul_libc__fputc(cgul_exception_t* cex,
                 int c,
                 FILE* f);

/**
 * Wrapper for <code>fgets()</code>.  Read a string from
 * <code>f</code>.  The string will be returned in <code>buf</code>
 * and will be null terminated.  The string is limited to
 * <code>buf_size - 1</code> or the first '\\n' character.  On
 * success, <code>buf</code> is returned.  If EOF is reached,
 * <code>NULL</code> is returned.  If an error occurs,
 * <code>NULL</code> is returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] buf buffer where string is stored
 * @param[in] buf_size size of <code>buf</code>
 * @param[in] f input file
 * @return <code>buf</code> on success or <code>NULL</code> on EOF or error
 */
extern CGUL_EXPORT char*
cgul_libc__fgets(cgul_exception_t* cex,
                 char* buf,
                 int buf_size,
                 FILE* f);

/**
 * Wrapper for <code>fputs</code>.  Write <code>s</code> to
 * <code>f</code>.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s string
 * @param[in] f file to which <code>s</code> is written
 */
extern CGUL_EXPORT void
cgul_libc__fputs(cgul_exception_t* cex,
                 const char* s,
                 FILE* f);

/**
 * Wrapper for <code>puts</code>.  It writes the string <code>s</code>
 * to <code>stdout</code>.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s string to write
 */
extern CGUL_EXPORT void
cgul_libc__puts(cgul_exception_t* cex,
                const char* s);

/**
 * Wrapper for <code>strlen()</code> to determine the string length of
 * <code>s</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] s string
 * @return length of <code>s</code>
 */
extern CGUL_EXPORT size_t
cgul_libc__strlen(cgul_exception_t* cex,
                  const char* s);

/**
 * Wrapper for <code>strcpy()</code> to copy <code>src</code> to
 * <code>dst</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] dst destination string
 * @param[in] src source string
 * @return <code>dst</code>
 */
extern CGUL_EXPORT char*
cgul_libc__strcpy(cgul_exception_t* cex,
                  char* dst,
                  const char* src);

/**
 * This method implements <code>strncpy()</code> functionality except
 * it always <code>null</code> terminates the result.
 *
 * @param[in] cex c-style exception
 * @param[in] dst destination string
 * @param[in] dst_size size in bytes of <code>dst</code>
 * @param[in] src source string
 * @return <code>dst</code>
 */
extern CGUL_EXPORT char*
cgul_libc__strncpy(cgul_exception_t* cex,
                   char* dst,
                   const char* src,
                   size_t dst_size);

/**
 * Wrapper for <code>strcpy()</code> to concatenate <code>src</code>
 * with <code>dst</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] dst destination string
 * @param[in] src source string
 * @return <code>dst</code>
 */
extern CGUL_EXPORT char*
cgul_libc__strcat(cgul_exception_t* cex,
                  char* dst,
                  const char *src);

/**
 * Wrapper for <code>strcmp()</code>.  It compares <code>s1</code> with
 * <code>s2</code> and returns less than zero if <code>s1</code> is
 * less than <code>s2</code>.  It returns greater than zero if
 * <code>s1</code> is greater than <code>s2</code>, and it returns
 * zero if <code>s1</code> is equal to <code>s2</code>.
 *
 * @param[in] cex c-style exception
 * @param s1 first string to compare
 * @param s2 second string to compare
 * @return result of comparison
 */
extern CGUL_EXPORT int
cgul_libc__strcmp(cgul_exception_t* cex,
                  const char* s1,
                  const char* s2);

/**
 * Return a duplicate of <code>s</code>.  The client is responsible
 * for calling <code>free()</code> on the pointer returned.  If an
 * error occurs, <code>NULL</code> is returned and an exception is
 * thrown.  This is not a wrapper around <code>strdup()</code> because
 * <code>strdup()</code> is not portable.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s string to duplicate
 * @return duplicate string
 */
extern CGUL_EXPORT char*
cgul_libc__strdup(cgul_exception_t* cex,
                  const char* s);

/**
 * This function is a thin wrapper around
 * <code>cgul_libc__vsnprintf()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] dst destination string
 * @param[in] dst_size size of <code>dst</code>
 * @param[in] format format string
 * @param[in] ... variable length arguments that match <code>format</code>
 * @return length of the string copied to <code>dst</code>
 * @see cgul_libc__vsnprintf
 */
extern CGUL_EXPORT int
cgul_libc__snprintf(cgul_exception_t* cex,
                    char* dst,
                    long dst_size,
                    const char* format,
                    ...)
#if defined(__GNUC__) && !defined(__MINGW_H)
    __attribute__ ((format (printf, 4, 5)))
#endif
;

/**
 * This function is not a wrapper around <code>snprintf()</code>
 * because <code>snprintf()</code> is not generally portable.
 * Instead, this function uses the sprintf()-like implementation
 * provided by <code>cgul_string__append_vsprintf()</code>. <p>
 *
 * This function expands <code>format</code> and writes up to a
 * maximum of <code>dst_size - 1</code> bytes to <code>dst</code>
 * truncating the expanded string if necessary.  It always
 * null-terminates <code>dst</code>.  On success, the entire length of
 * the expanded string is returned even if <code>dst</code> only holds
 * a truncated version of that string.  On failure, -1 is returned and
 * an exception is thrown. <p>
 *
 * NOTE: If you pass in <code>0</code> for <code>dst_size</code> and
 * (optionally) <code>NULL</code> for <code>dst</code>, this function
 * will not attempt to alter <code>dst</code>, but it will return the
 * full length of the expanded format string.  You can use this return
 * value to allocate the right amount of space so that you can turn
 * around and call <code>cgul_libc__vsprintf()</code>.  However, you
 * are probably better off calling <code>cgul_libc__asprintf()</code>
 * which does the allocation and format string expansion in a single
 * call.
 *
 * @param[in] cex c-style exception
 * @param[in] dst destination string
 * @param[in] dst_size size of <code>dst</code>
 * @param[in] format format string
 * @param[in] args variable length arguments that match <code>format</code>
 * @return length of the string copied to <code>dst</code>
 * @see cgul_libc__asprintf
 */
extern CGUL_EXPORT int
cgul_libc__vsnprintf(cgul_exception_t* cex,
                     char* dst,
                     long dst_size,
                     const char* format,
                     va_list args);

/**
 * This function is a thin wrapper around
 * <code>cgul_libc__vasprintf()</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] format format string
 * @param[in] ... variable arguments that match <code>format</code>
 * @return newly allocated and formatted string
 */
extern CGUL_EXPORT char*
cgul_libc__asprintf(cgul_exception_t* cex,
                    const char* format,
                    ...)
#if defined(__GNUC__) && !defined(__MINGW_H)
    __attribute__ ((format (printf, 2, 3)))
#endif
;

/**
 * This function mimics the BSD vasprintf() function by allocating
 * enough memory before calling sprintf() to guarantee that no buffer
 * overflows occur.  The caller is responsible for calling free() on
 * the pointer returned.  If this method fails, an exception is thrown
 * and <code>NULL</code> is returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] format format string
 * @param[in] args variable arguments that match <code>format</code>
 * @return newly allocated and formatted string
 */
extern CGUL_EXPORT char*
cgul_libc__vasprintf(cgul_exception_t* cex,
                     const char* format,
                     va_list args);

/**
 * Get the base name of <code>name</code>.  It returns everything
 * after the last slash or the last backslash.  If no slash or
 * backslash can be found, it simply returns <code>name</code>.  The
 * client is responsible for calling <code>free()</code> on the
 * pointer returned.  If memory cannot be allocated, <code>NULL</code>
 * is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] name name of file
 * @return base name of <code>name</code>
 */
extern CGUL_EXPORT char*
cgul_libc__basename(cgul_exception_t* cex,
                    const char* name);

/**
 * Get the directory name of <code>name</code>.  It returns everything
 * before the last slash or the last backslash.  If no slash or
 * backslash can be found, it simply returns <code>"."</code>.  The
 * client is responsible for calling <code>free()</code> on the
 * pointer returned.  If memory cannot be allocated, <code>NULL</code>
 * is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] name name of file
 * @return directory name of <code>name</code>
 */
extern CGUL_EXPORT char*
cgul_libc__dirname(cgul_exception_t* cex,
                   const char* name);

/**
 * Get the file extension of the base name of <code>name</code>.  It
 * returns everything after and including the last dot in the base
 * name of <code>name</code>.  If no dot can be found, it simply
 * returns an empty string.  The client is responsible for calling
 * <code>free()</code> on the pointer returned.  If memory cannot be
 * allocated, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * For example, the base name of "/foo/bar.d/" is "bar.d" (despite the
 * trailing path separator).  Thus, the extension is ".d".  This
 * differs from how the Tcl "file extension" command works which
 * returns an empty string in this case (because of the path
 * separator). <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] name name of file
 * @return file extension of <code>name</code>
 * @see cgul_libc__rootname()
 */
extern CGUL_EXPORT char*
cgul_libc__extension(cgul_exception_t* cex,
                     const char* name);

/**
 * Get the root name of <code>name</code>.  This function joins the
 * directory name and base name of <code>name</code> after stripping
 * off the extension.  The client is responsible for calling
 * <code>free()</code> on the pointer returned.  If memory cannot be
 * allocated, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * For example, the base name of "/foo/bar.d/" is "bar.d" (despite the
 * trailing path separator).  Thus, the extension is ".d" and the root
 * name is "/foo/bar".  This differs from how the Tcl "file extension"
 * command works which returns "/foo/bar.d/" in this case (because of
 * the path separator). <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] name name of file
 * @return root name of <code>name</code>
 */
extern CGUL_EXPORT char*
cgul_libc__rootname(cgul_exception_t* cex,
                    const char* name);

/**
 * This method strips the UTF-8 Byte-Order Mark (BOM) from
 * <code>f</code>.  This method returns <code>1</code> if the BOM is
 * stripped; otherwise, it returns <code>0</code>.  If an error
 * occurs, <code>0</code> is returned, and an exception is thrown. <p>
 *
 * IMPORTANT: Because of the way this method is written, it can only
 * be called on files that allow random access.  Do not try to use it
 * on <code>stdin</code> for example.
 *
 * @param[in,out] cex c-style exception
 * @param[in] f file
 * @return whether the utf8 byte order mark was stripped
 */
extern CGUL_EXPORT int
cgul_libc__strip_utf8_byte_order_mark(cgul_exception_t* cex,
                                      FILE* f);

CGUL_END_C

#endif  /*  CGUL_LIBC_H  */
