#ifndef CGUL_LINE_WRAPPER_H
#define CGUL_LINE_WRAPPER_H

/**
 * The functions in this file wrap lines based on a certain number of
 * characters per line (which is suitable for formatting output for a
 * terminal) or based on the variable-length spacing between
 * characters in a BDF font loaded by the <code>cgul_bdf</code>
 * class.
 *
 * @file
 * @author Paul Serice
 * @brief line wrapper
 */

#include "cgul_common.h"
#include "cgul_bdf.h"
#include "cgul_exception.h"
#include "cgul_string.h"
#include "cgul_wstring.h"

CGUL_BEGIN_C

/**
 * This function wraps the UTF-8 string <code>lines</code> at
 * <code>column</code> and returns the result in <code>lines</code>.
 * This is used for formatting text destined for the console.  It
 * assumes all characters have the same, unspecified width.  So the
 * wrapping is based on the maximum number of characters per line. <p>
 *
 * It is permissible for <code>lines</code> to have embedded new-line
 * characters before calling this method.  When an embedded new-line
 * character is encountered, this method will honor it and force the
 * start of a new line.
 *
 * If an error occurs, an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in,out] lines lines to wrap
 * @param[in] column column at which to wrap
 */
extern CGUL_EXPORT void
cgul_line_wrapper__wrap_utf8_at_column(cgul_exception_t* cex,
                                       cgul_string_t lines,
                                       unsigned int column);

/**
 * This function wraps the UTF-32 string <code>wlines</code> at
 * <code>column</code> and returns the result in <code>wlines</code>.
 * This is used for formatting text destined for the console.  It
 * assumes all characters have the same, unspecified width.  So the
 * wrapping is based on the maximum number of wide characters per
 * line. <p>
 *
 * It is permissible for <code>wlines</code> to have embedded new-line
 * wide characters before calling this method.  When an embedded
 * new-line wide character is encountered, this method will honor it
 * and force the start of a new line.
 *
 * If an error occurs, an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in,out] wlines lines to wrap
 * @param[in] column column at which to wrap
 */
extern CGUL_EXPORT void
cgul_line_wrapper__wrap_utf32_at_column(cgul_exception_t* cex,
                                        cgul_wstring_t wlines,
                                        unsigned int column);

/**
 * This function wraps the UTF-8 string <code>lines</code> at
 * <code>width</code> and returns the result in <code>lines</code>.
 * This is used for formatting text destined to be drawn on a
 * graphical surface.  The amount of space each line consumes is
 * calculated using the BDF font <code>bdf</code> by calling
 * <code>cgul_bdf__get_utf8_extents()</code>.  Thus, in order for
 * this function to be accurate, your code must draw text as explained
 * in the comments to <code>cgul_bdf__get_utf8_extents()</code> and
 * the value you pass in here for <code>substitute</code> must
 * match. <p>
 *
 * It is permissible for <code>lines</code> to have embedded new-line
 * characters before calling this method.  When an embedded new-line
 * character is encountered, this method will honor it and force the
 * start of a new line. <p>
 *
 * If <code>substitute</code> is set, this method assumes the code
 * that draws the text will substitute missing glyphs with the default
 * glyph for the font, e.g., an empty box.  If <code>substitute</code>
 * is not set, this method will throw an exception if it cannot find
 * the glyph for one of the characters in <code>lines</code>. <p>
 *
 * If you just need to calculate the width of a string in pixels, you
 * can use <code>cgul_bdf__get_utf8_extents()</code> instead. <p>
 *
 * If an error occurs, an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in,out] lines lines to wrap
 * @param[in] bdf BDF font
 * @param[in] substitute whether to substitute missing glyphs with the default
 * @param[in] max_line_width maximum line width before wrapping occurs
 */
extern CGUL_EXPORT void
cgul_line_wrapper__wrap_utf8_at_width(cgul_exception_t* cex,
                                      cgul_string_t lines,
                                      cgul_bdf_t bdf,
                                      int substitute,
                                      unsigned int max_line_width);

/**
 * This function wraps the UTF-32 string <code>wlines</code> at
 * <code>width</code> and returns the result in <code>wlines</code>.
 * This is used for formatting text destined to be drawn on a
 * graphical surface.  The amount of space each line consumes is
 * calculated using the BDF font <code>bdf</code> by calling
 * <code>cgul_bdf__get_utf32_extents()</code>.  Thus, in order for
 * this function to be accurate, your code must draw text as explained
 * in the comments to <code>cgul_bdf__get_utf32_extents()</code> and
 * the value you pass in here for <code>substitute</code> must
 * match. <p>
 *
 * It is permissible for <code>wlines</code> to have embedded new-line
 * characters before calling this method.  When an embedded new-line
 * character is encountered, this method will honor it and force the
 * start of a new line. <p>
 *
 * If <code>substitute</code> is set, this method assumes the code
 * that draws the text will substitute missing glyphs with the default
 * glyph for the font, e.g., an empty box.  If <code>substitute</code>
 * is not set, this method will throw an exception if it cannot find
 * the glyph for one of the characters in <code>lines</code>. <p>
 *
 * If you just need to calculate the width of a string in pixels, you
 * can use <code>cgul_bdf__get_utf32_extents()</code> instead. <p>
 *
 * If an error occurs, an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in,out] wlines lines to wrap
 * @param[in] bdf BDF font
 * @param[in] substitute whether to substitute missing glyphs with the default
 * @param[in] max_line_width maximum line width before wrapping occurs
 */
extern CGUL_EXPORT void
cgul_line_wrapper__wrap_utf32_at_width(cgul_exception_t* cex,
                                       cgul_wstring_t wlines,
                                       cgul_bdf_t bdf,
                                       int substitute,
                                       unsigned int max_line_width);

CGUL_END_C

#endif  /*  CGUL_LINE_WRAPPER_H  */
