#ifndef CGUL_LIST_CXX_H
#define CGUL_LIST_CXX_H

/**
 * C++ wrapper for <code>cgul_list</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_list</code>
 */

#include "cgul_cache_cxx.h"
#include "cgul_exception_cxx.h"
#include "cgul_list.h"
#include "cgul_list_node_cxx.h"

/**
 * This class provides the C++ bindings for C <code>cgul_list</code>
 * objects.  The main purpose of this class is to convert the C-style
 * function calls and exception handling in <code>cgul_list</code>
 * into C++-style function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_list</code>
 */
class cgul_list_cxx {

public:

    /**
     * This typedef is the interface for the comparison function used
     * by <code>sort()</code> and <code>insert_sorted()</code>.
     */
    typedef cgul_list__compare_t compare_t;

    /**
     * This typedef is the interface for the combining function used
     * by the following methods: <p>
     *
     * <pre>
     *     <code>foldl()</code>
     *     <code>foldr()</code>
     * </pre>
     *
     * @param[in] value value
     * @param[in] data client data
     * @return whether to continue
     */
    typedef int (*fold_t)(void* value,
                          void* data);

    /**
     * This typedef is the interface for the callback function used by
     * the following methods: <p>
     *
     * <pre>
     *     <code>traverse()</code>
     *     <code>traverse_range()</code>
     * </pre>
     *
     * @param[in] l list
     * @param[in] n node
     * @param[in] data client data
     * @return whether to continue
     */
    typedef int (*traverse_t)(cgul_list_cxx* l,
                              cgul_list_node_cxx* n,
                              void* data);
    
    /**
     * Create a new <code>cgul_list_cxx</code> object.  If memory
     * cannot be allocated, an exception is thrown.
     *
     * @return new <code>cgul_list_cxx</code> instance
     */
    cgul_list_cxx()
    {
        cgul_exception_t local = NULL;
        obj = cgul_list__new(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_list_cxx</code> object by wrapping an
     * existing <code>cgul_list</code> object.
     *
     * @param[in] rhs right-hand side
     */
    cgul_list_cxx(cgul_list_t rhs)
    {
        obj = rhs;
    }

    /**
     * This method frees all internally allocated memory.  Do not try
     * to access this instance after calling this method.
     *
     * @see free_values()
     */
    virtual
    ~cgul_list_cxx()
    {
        if (obj) {
            cgul_list__delete(obj);
            obj = NULL;
        }
    }

    /**
     * This method calls <code>free()</code> on all the values in the
     * list.  Because this is such a common operation, it is an
     * exception to the rule that cgul containers never free values.
     * This method should only ever be called immediately before
     * calling <code>delete</code> because it otherwise invalidates
     * the list.
     */
    virtual void
    free_values()
    {
        cgul_list__free_values(obj);
    }

    /**
     * Get the size of the cache of nodes.
     *
     * @return size of the cache of nodes
     */
    virtual unsigned long int
    get_cache_size() const
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_list__get_cache_size(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the size of the cache of nodes.
     *
     * For efficiency, the <code>cgul_list_cxx</code> can keep a
     * cache of nodes that can be reused.  In many situations, this
     * can greatly reduce the number of calls to <code>malloc()</code>
     * which can greatly improve performance and reduce memory
     * fragmentation. <p>
     *
     * You can release all the cached nodes and disable caching of nodes
     * by setting the cache size to 0.  This is the default. <p>
     *
     * If an error occurs (while enlarging the cache), an exception is
     * thrown. <p>
     *
     * @param[in] size of the cache of nodes
     */
    virtual void
    set_cache_size(unsigned long int size)
    {
        cgul_exception_t local = NULL;
        cgul_list__set_cache_size(&local, obj, size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return true if the list is empty.
     *
     * @return whether the list is empty
     */
    virtual int
    is_empty() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_list__is_empty(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the first element on the list.  You can use
     * <code>cgul_list_node_cxx::get_next()</code> to iterate over
     * the list by starting with the value returned by this method.
     * If the list is empty, <code>NULL</code> is returned.
     *
     * @return first element
     */
    virtual cgul_list_node_cxx*
    get_front() const
    {
        cgul_list_node_cxx* rv = NULL;
        cgul_exception_t local = NULL;
        rv = (cgul_list_node_cxx*)cgul_list__get_front(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the last element on the list.  You can use
     * <code>cgul_list_node_cxx::get_prev()</code> to iterate over
     * the list by starting with the value returned by this method.
     * If the list is empty, <code>NULL</code> is returned.
     *
     * @return last element
     */
    virtual cgul_list_node_cxx*
    get_back() const
    {
        cgul_list_node_cxx* rv = NULL;
        cgul_exception_t local = NULL;
        rv = (cgul_list_node_cxx*)cgul_list__get_back(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Insert <code>value</code> at the front of the list.  If
     * successful, the inserted node is returned; otherwise, an
     * exception is thrown.
     *
     * @param[in] value value to add at the front of the list
     * @return node that holds <code>value</code>
     */
    virtual cgul_list_node_cxx*
    push_front(const void* value)
    {
        cgul_list_node_cxx* rv = NULL;
        cgul_exception_t local = NULL;
        rv = (cgul_list_node_cxx*)cgul_list__push_front(&local, obj, value);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Remove the first node from the list.  Return the value that was
     * stored in node.  If the list is empty, <code>NULL</code> is
     * returned.
     *
     * @return value that was stored in the front node
     */
    virtual void*
    pop_front()
    {
        void* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_list__pop_front(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Insert <code>value</code> at the back of the list.  If
     * successful, the return value is the inserted node; otherwise,
     * an exception is thrown.
     *
     * @param[in] value value to add at the back of the list
     * @return node that holds <code>value</code>
     */
    virtual cgul_list_node_cxx*
    push_back(const void* value)
    {
        cgul_list_node_cxx* rv = NULL;
        cgul_exception_t local = NULL;
        rv = (cgul_list_node_cxx*)cgul_list__push_back(&local, obj, value);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Remove the last value from the list.  Return the value that was
     * stored in node.  If the list is empty, <code>NULL</code> is
     * returned.
     *
     * @return value that was stored in the back node
     */
    virtual void*
    pop_back()
    {
        void* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_list__pop_back(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Move <code>n2</code> before <code>n1</code> in the list.  If
     * <code>n1 == NULL</code>, move <code>n2</code> to before the
     * <code>NULL</code> at the end of the list.  <code>n1</code> and
     * <code>n2</code> must both already exist on the same list. <p>
     *
     * For example, to rotate the list wrapping the first element around
     * to the back of the list, do the following: <p>
     *
     * <pre>
     *     lst.move_before_node(NULL, lst.get_front());
     * </pre>
     *
     * @param[in] n1 insert before this node
     * @param[in] n2 node to insert
     */
    virtual void
    move_before_node(cgul_list_node_cxx* n1,
                     cgul_list_node_cxx* n2)
    {
        cgul_exception_t local = NULL;
        cgul_list__move_before_node(&local,
                                    obj,
                                    (cgul_list_node_t)n1,
                                    (cgul_list_node_t)n2);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Move <code>n2</code> after <code>n1</code> in the list.  If
     * <code>n1 == NULL</code>, move <code>n2</code> to after the
     * <code>NULL</code> at the beginning of the list.
     * <code>n1</code> and <code>n2</code> must both already exist on
     * the same list. <p>
     *
     * For example, to rotate the list wrapping the last element around
     * to the front of the list, do the following: <p>
     *
     * <pre>
     *     lst.move_after_node(NULL, lst.get_back());
     * </pre>
     *
     * @param[in] n1 insert after this node
     * @param[in] n2 node to insert
     */
    virtual void
    move_after_node(cgul_list_node_cxx* n1,
                    cgul_list_node_cxx* n2)
    {
        cgul_exception_t local = NULL;
        cgul_list__move_after_node(&local,
                                   obj,
                                   (cgul_list_node_t)n1,
                                   (cgul_list_node_t)n2);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Insert <code>value</code> before node <code>n</code> in the
     * list.  If <code>n == NULL</code>, then insert before the
     * <code>NULL</code> at the end of the list.  If successful, the
     * return value is the inserted node; otherwise, an exception is
     * thrown.
     *
     * @param[in] n inserting before this node
     * @param[in] value value to insert
     * @return node that holds <code>value</code>
     */
    virtual cgul_list_node_cxx*
    insert_before_node(cgul_list_node_cxx* n,
                       const void* value)
    {
        cgul_list_node_cxx* rv = NULL;
        cgul_exception_t local = NULL;
        rv = (cgul_list_node_cxx*)
            cgul_list__insert_before_node(&local,
                                          obj,
                                          (cgul_list_node_t)n,
                                          value);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Insert <code>value</code> after node <code>n</code> in the
     * list.  If <code>n == NULL</code>, then insert after the
     * <code>NULL</code> at the beginning of the list.  If successful,
     * the return value is the inserted node; otherwise, an exception
     * is thrown.
     *
     * @param[in] n inserting after this node
     * @param[in] value value to insert
     * @return node that holds <code>value</code>
     */
    virtual cgul_list_node_cxx*
    insert_after_node(cgul_list_node_cxx* n,
                      const void* value)
    {
        cgul_list_node_cxx* rv = NULL;
        cgul_exception_t local = NULL;
        rv = (cgul_list_node_cxx*)
            cgul_list__insert_after_node(&local,
                                         obj,
                                         (cgul_list_node_t)n,
                                         value);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Insert <code>value</code> into the list in sort order as
     * determined by the comparison function <code>f</code>.  If
     * successful, the return value is the inserted node; otherwise,
     * an exception is thrown. <p>
     *
     * This method assumes the list is already sorted.  If not, call
     * <code>sort()</code> before calling this method. <p>
     *
     * This method is O(N).  Thus, this method is more efficient for a
     * single insert on an already sorted list than
     * <code>sort()</code> which is O(N log N), but calling
     * this method to perform N inserts is O(N^2) which is not as
     * efficient as N calls to <code>push_back()</code>
     * followed by a single call to <code>sort()</code>. <p>
     *
     * If you need to constantly maintain a sorted list,
     * <code>cgul_rbtree_cxx</code> or <code>cgul_multimap_cxx</code>
     * are usually better solutions. <p>
     *
     * @param[in] value value to insert
     * @param[in] f comparison function
     * @return node that holds <code>value</code>
     */
    virtual cgul_list_node_cxx*
    insert_sorted(const void* value,
                  compare_t f)
    {
        cgul_list_node_cxx* rv = NULL;
        cgul_exception_t local = NULL;
        rv = (cgul_list_node_cxx*)
                 cgul_list__insert_sorted(&local, obj, value, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Sort the listusing the comparison function <code>f</code>. <p>
     *
     * This method uses merge sort.  The main advantage of this method
     * over <code>cgul_merge_sort()</code> is that it can sort the list
     * in-line whereas <code>cgul_merge_sort</code> needs an extra copy of
     * the list. <p>
     *
     * This method is O(N log N).  Thus, this method is more efficient
     * when sorting a large number of items, but
     * <code>insert_sorted()</code> is O(N) so it is faster when
     * inserting just one item into an already sorted list. <p>
     *
     * If you need to constantly maintain a sorted list,
     * <code>cgul_rbtree_cxx</code> or <code>cgul_multimap_cxx</code>
     * are usually better solutions. <p>
     *
     * If you need to merge two sorted lists, <code>merge()</code> is
     * usually a better solution. <p>
     *
     * @param[in] f comparison function
     * @see cgul_merge_sort
     * @see cgul_list_cxx::merge
     */
    virtual void
    sort(compare_t f)
    {
        cgul_exception_t local = NULL;
        cgul_list__sort(&local, obj, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Merge the sorted list <code>l</code> with <code>this</code>
     * list.  The sort order is determined by the comparison function
     * <code>f</code>.  When this method returns, all the nodes will
     * be in <code>this</code> list and <code>l</code> will be
     * empty. <p>
     *
     * This method is extremely efficient doing its job in a single
     * pass, i.e., O(N).  This method should be the first thing that
     * comes to mind when you need to add a small, unsorted list to a
     * large, sorted list.  In this case, sort the small list using
     * <code>cgul_list__sort()</code> before calling this method to
     * merge the small list with the large list. <p>
     *
     * @param[in,out] l list to merge into this list
     * @param[in] f comparison function
     */
    virtual void
    merge(cgul_list_cxx& l,
          compare_t f)
    {
        cgul_exception_t local = NULL;
        cgul_list__merge(&local, obj, l.get_obj(), f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Remove all elements from the list. <p>
     *
     * Strictly as a convenience, this method is an exception to the rule
     * that cgul containers never free values.  If you pass in a
     * <code>values_cache_cxx</code> instance that is not <code>NULL</code>,
     * the values will be put back on their cache. <p>
     *
     * @param[in] values_cache values cache
     * @see cgul_cache_cxx::get_freer()
     */
    virtual void
    clear(cgul_cache_cxx* values_cache)
    {
        cgul_exception_t local = NULL;
        cgul_list__clear(&local,
                         obj,
                         (values_cache ? values_cache->get_obj() : NULL));
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Remove the node <code>n</code> from the list.  If
     * <code>value_out</code> is not <code>NULL</code>, the value held
     * by <code>n</code> is returned in <code>*value_out</code>. <p>
     *
     * It is almost always a mistake to naively call this method while
     * iterating over the tree.  The problem is that calling this
     * method invalidates the node making it impossible to call
     * <code>node->get_next()</code> afterward.  The solution is
     * simple.  Just call <code>node->get_next()</code> before calling
     * this method:
     *
     * <pre>
     *
     *     curr = l->get_front();
     *     for ( ; curr ; curr = next) {
     *         next = curr->get_next();
     *         l->remove_node(curr, &value_out);
     *         free(value_out);
     *     }
     *
     * </pre>
     *
     * Alternatively, you can use <code>remove_range()</code>,
     * <code>traverse()</code>, or <code>traverse_range()</code>. <p>
     *
     * If you use one of the traversal methods, it is safe to call
     * this method naively.  For example, if your tree stores
     * <code>cgul_string_cxx</code> objects for its values, another
     * way to remove nodes is as follows: <p>
     *
     * <pre>
     *
     *     static int
     *     remove_node_cb(cgul_list_cxx* l,
     *                    cgul_list_node_cxx* n,
     *                    void* data)
     *     {
     *         cgul_string_cxx* value_out = NULL;
     *         l->remove_node(n, (void**)&value_out);
     *         delete value_out;
     *         return 1;
     *     }
     *
     *     l->traverse(&remove_node_cb, NULL);
     *
     * </pre>
     *
     * @param[in] n node to remove
     * @param[out] value_out value
     */
    virtual void
    remove_node(cgul_list_node_cxx* n,
                void** value_out)
    {
        cgul_exception_t local = NULL;
        cgul_list__remove_node(&local, obj, (cgul_list_node_t)n, value_out);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Remove the nodes between <code>n_begin</code> (inclusive) and
     * <code>n_end</code> (inclusive).  If <code>n_begin</code> is NULL,
     * start at the front of the list.  If <code>n_end</code> is NULL or
     * is not found, continue removing until the end of the list. <p>
     *
     * Strictly as a convenience, this method is an exception to the rule
     * that cgul containers never free values.  If you pass in a
     * <code>values_cache_cxx</code> instance that is not <code>NULL</code>,
     * the values will be put back on their cache. <p>
     *
     * @param[in] n_begin begin removing with and including this node
     * @param[in] n_end end removing with and including this node
     * @param[in] values_cache values cache
     * @see cgul_cache_cxx::get_freer()
     */
    virtual void
    remove_range(cgul_list_node_cxx* n_begin,
                 cgul_list_node_cxx* n_end,
                 cgul_cache_cxx* values_cache)
    {
        cgul_exception_t local = NULL;
        cgul_list__remove_range(
            &local,
            obj,
            (cgul_list_node_t)n_begin,
            (cgul_list_node_t)n_end,
            (values_cache ? values_cache->get_obj() : NULL));
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Remove all the nodes that have <code>value</code> for their value.
     * If at least one node was removed, this method returns
     * <code>1</code>; otherwise, it returns <code>0</code>. <p>
     *
     * Strictly as a convenience, this method is an exception to the rule
     * that cgul containers never free values.  If you pass in a
     * <code>values_cache_cxx</code> instance that is not <code>NULL</code>,
     * the values will be put back on their cache. <p>
     *
     * @param[in] value values to remove from the list
     * @param[in] values_cache values cache
     * @see cgul_cache_cxx::get_freer()
     */
    virtual int
    remove_value(const void* value,
                 cgul_cache_cxx* values_cache)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_list__remove_value(
                 &local,
                 obj,
                 value,
                 (values_cache ? values_cache->get_obj() : NULL));
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the size of the list, i.e., the number of elements in the
     * list.
     *
     * @return size of the list
     */
    virtual unsigned long int
    get_size() const
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_list__get_size(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Swap the underlying data for this object and <code>rhs</code>.
     * For large lists, this should be much faster than trying to do the
     * same thing using removes and inserts.
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    swap(cgul_list_cxx& rhs)
    {
        cgul_exception_t local = NULL;
        cgul_list__swap(&local, this->obj, rhs.obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method performs a left fold of the list with the combining
     * function <code>f</code>.  <code>f</code> is called once for
     * each value in the list starting at the front of the list and
     * iterating forward to the end of the list. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * value.  The second parameter passed into <code>f</code> is the
     * client data which is where the result of the fold should be
     * accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order
     * for iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldl(fold_t f,
          void* data)
    {
        void* value = NULL;
        cgul_list_node_t curr = NULL;
        cgul_list_node_t next = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_list__foldl()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over each node. */
        curr = cgul_list__get_front(&local, obj);
        for ( ; curr ; curr = next) {
            /* Get the next node now in case f deletes curr. */
            next = cgul_list_node__get_next(&local, curr);
            /* Call back. */
            value = (void*)cgul_list_node__get_value(&local, curr);
            if (!(*f)(value, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a right fold of the list with the
     * combining function <code>f</code>.  <code>f</code> is called
     * once for each value in the list starting at the back of the
     * list and iterating backward to the front of the list. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * value.  The second parameter passed into <code>f</code> is the
     * client data which is where the result of the fold should be
     * accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldr(fold_t f,
          void* data)
    {
        void* value = NULL;
        cgul_list_node_t curr = NULL;
        cgul_list_node_t prev = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_list__foldr()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over each node. */
        curr = cgul_list__get_back(&local, obj);
        for ( ; curr ; curr = prev) {
            /* Get the previous node now in case f deletes curr. */
            prev = cgul_list_node__get_prev(&local, curr);
            /* Call back. */
            value = (void*)cgul_list_node__get_value(&local, curr);
            if (!(*f)(value, data)) {
                break;
            }
        }
    }

    /**
     * Traverse all nodes passing each node to the function
     * <code>f</code>. <p>
     *
     * The first parameter passed into <code>f</code> is the list
     * <code>this</code>.  The second paramenter passed into
     * <code>f</code> is the node for this iteration.  The third
     * parameter passed into <code>f</code> is the client data
     * <code>data</code>. <p>
     *
     * <code>f</code> is provided with a safe context in which it can
     * remove the node that is passed into <code>f</code> by calling
     * <code>remove_node()</code>. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * the traversal to continue. <p>
     *
     * NOTE: It is not strictly necessary that you use
     * <code>traverse()</code> or <code>traverse_range()</code> in
     * order to iterate over the list elements.  In fact, I would
     * recommend that you use
     * <code>cgul_list_node_cxx::get_next()</code> for most of your
     * iteration needs.  If you need to remove nodes though, you
     * should probably use this method.  If you look at this method's
     * source however, you will see that the
     * <code>cgul_list_node_cxx</code> class provides all the public
     * methods you need to safely remove nodes while you are iterating
     * over the list, but you do have to be careful.
     *
     * @param[in] f traversal callback function
     * @param[in] data client data passed to <code>f</code>
     * @see traverse_range()
     */
    virtual void
    traverse(traverse_t f,
             void* data)
    {
        traverse_range(NULL, NULL, f, data);
    }

    /**
     * Traverse all nodes starting with <code>first</code> (inclusive) and
     * ending with <code>last</code> (inclusive) passing each node to the
     * function <code>f</code>.  If you know the first node, but are not
     * sure of the last node, just use <code>NULL</code> as the last node.
     * This will cause this method to iterate until it reaches the end of
     * the list.  You can than have <code>f</code> return 0 when it
     * determines that the last node has been reached. <p>
     *
     * The first parameter passed into <code>f</code> is the list
     * <code>this</code>.  The second paramenter passed into
     * <code>f</code> is the node for this iteration.  The third
     * parameter passed into <code>f</code> is the client data
     * <code>data</code>. <p>
     *
     * <code>f</code> is provided with a safe context in which it can
     * remove the node that is passed into <code>f</code> by calling
     * <code>remove_node()</code>. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * the traversal to continue. <p>
     *
     * If <code>first</code> is <code>NULL</code>, iteration starts at the
     * beginning of the list.  If <code>last</code> is <code>NULL</code>,
     * iteration stops at the end of the list. <p>
     *
     * NOTE: It is not strictly necessary that you use
     * <code>traverse()</code> or <code>traverse_range()</code> in
     * order to iterate over the list elements.  In fact, I would
     * recommend that you use
     * <code>cgul_list_node_cxx::get_next()</code> for most of your
     * iteration needs.  If you need to remove nodes though, you
     * should probably use this method.  If you look at this method's
     * source however, you will see that the
     * <code>cgul_list_node_cxx</code> class provides all the public
     * methods you need to safely remove nodes while you are iterating
     * over the list, but you do have to be careful.
     *
     * @param[in] first first node in range
     * @param[in] last last node in range
     * @param[in] f traversal callback function
     * @param[in] data client data passed to <code>f</code>
     * @see traverse()
     */
    virtual void
    traverse_range(cgul_list_node_cxx* first,
                   cgul_list_node_cxx* last,
                   traverse_t f,
                   void* data)
    {
        cgul_list_node_cxx* curr = NULL;
        cgul_list_node_cxx* next = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_list__traverse_range()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */
        
        if (first == NULL) {
            first = (cgul_list_node_cxx*)cgul_list__get_front(&local, obj);
        }
        if (last == NULL) {
            last = (cgul_list_node_cxx*)cgul_list__get_back(&local, obj);
        }

        /* Iterate over the nodes. */
        for (curr = first ; curr ; curr = next) {
            /* Get next now in case f deletes curr. */
            next = (cgul_list_node_cxx*)
                       cgul_list_node__get_next(&local, (cgul_list_node_t)curr);
            /* Call back. */
            if (!(*f)(this, curr, data)) {
                break;
            }
            /* We are done if we just processed the last node. */
            if (curr == last) {
                break;
            }
        }
    }

    /**
     * Get the underlying <code>cgul_list</code> object.
     *
     * @return underlying object
     */
    virtual cgul_list_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_list</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_list__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_list_t
    take_obj()
    {
        cgul_list_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_list_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_list__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_list_cxx(const cgul_list_cxx&);
    cgul_list_cxx& operator=(const cgul_list_cxx&);

    /** Underlying <code>cgul_list</code> object. */
    cgul_list_t obj;

};

#endif  /*  CGUL_LIST_CXX_H  */
