#ifndef CGUL_MICROXML_CXX_H
#define CGUL_MICROXML_CXX_H

/**
 * C++ wrapper for <code>cgul_microxml</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_microxml</code>
 */

#include "cgul_microxml.h"
#include "cgul_exception_cxx.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_microxml</code> objects.  The main purpose of this class
 * is to convert the C-style function calls and exception handling in
 * <code>cgul_microxml</code> into C++-style function calls and
 * exception handling.
 *
 * @brief C++ bindings for <code>cgul_microxml</code>
 */
class cgul_microxml_cxx {

public:

    /**
     * Static method used to escape the special MicroXML characters '&',
     * '<', '>', '"', and "'" in the string <code>s</code>.  The client is
     * repsonsible for calling <code>free()</code> on the pointer
     * returned.  If an error occurs, an exception is thrown.
     *
     * @param[in] s UTF-8 input string
     * @return UTF-8 escaped output string
     */
    static char*
    escape(const char* s)
    {
        char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__escape(&local, s);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Default constructor for <code>cgul_microxml_cxx</code>.
     * <code>cgul_microxml__set_stream_reader()</code> should be
     * called before calling <code>read_next_token()</code>.  If an
     * error occurs, an exception is thrown.
     *
     * @see set_maximum_text_length()
     * @see set_skip_comments()
     * @see set_skip_white_space()
     */
    cgul_microxml_cxx()
    {
        cgul_exception_t local = NULL;
        obj = cgul_microxml__new(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Construct a new <code>cgul_microxml_cxx</code> instance and
     * call <code>open_fname()</code> passing it <code>fname</code>.
     * The file will be closed when this instance is deleted.  If an
     * error occurs, an exception is thrown. <p>
     *
     * @param[in] fname file name
     * @see set_text_size()
     * @see set_comment_size()
     * @see set_skip_comments()
     * @see set_skip_white_space()
     */
    cgul_microxml_cxx(const char* fname)
    {
        cgul_exception_t local = NULL;
        obj = cgul_microxml__new_from_fname(&local, fname);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Construct a new <code>cgul_microxml_cxx</code> instance and
     * call <code>open_file()</code> passing it <code>f</code>.  The
     * class does not take ownership of <code>f</code>.  Thus, the
     * client is still responsible for calling <code>fclose()</code>
     * on it.  If an error occurs, an exception is thrown.
     *
     * @param[in] f file
     * @see set_text_size()
     * @see set_comment_size()
     * @see set_skip_comments()
     * @see set_skip_white_space()
     */
    cgul_microxml_cxx(FILE* f)
    {
        cgul_exception_t local = NULL;
        obj = cgul_microxml__new_from_file(&local, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Construct a new <code>cgul_microxml_cxx</code> instance and
     * call <code>open_memory()</code> passing it <code>buffer</code>
     * and <code>buffer_size</code>.  This class does not take
     * ownership of <code>buffer</code> so the client is still
     * responsible for freeing <code>buffer</code> if necessary.  If
     * an error occurs, an exception is thrown. <p>
     *
     * @param[in] buffer memory buffer
     * @param[in] buffer_size size of <code>buffer</code> in bytes
     * @see set_text_size()
     * @see set_comment_size()
     * @see set_skip_comments()
     * @see set_skip_white_space()
     */
    cgul_microxml_cxx(const char* buffer,
                      size_t buffer_size)
    {
        cgul_exception_t local = NULL;
        obj = cgul_microxml__new_from_memory(&local, buffer, buffer_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Construct a new <code>cgul_microxml_cxx</code> instance and
     * call <code>set_stream_reader()</code> passing it
     * <code>sr</code>.  If an error occurs, an exception is
     * thrown. <p>
     *
     * @param[in] sr stream reader
     * @see set_maximum_text_length()
     * @see set_skip_comments()
     * @see set_skip_white_space()
     */
    cgul_microxml_cxx(cgul_stream_reader_t sr)
    {
        cgul_exception_t local = NULL;
        obj = cgul_microxml__new_from_stream_reader(&local, sr);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Destructor.
     */
    virtual
    ~cgul_microxml_cxx()
    {
        if (obj) {
            cgul_microxml__delete(obj);
            obj = NULL;
        }
    }

    /**
     * Open the MicroXML file with file name <code>fname</code>.  If a
     * file or memory buffer is already open, it is closed before
     * attempting to open the new file.  The new file will be closed when
     * this instance is deleted.  If an error occurs, an exception is
     * thrown.
     *
     * @param[in] fname file name
      */
    virtual void
    open_fname(const char* fname)
    {
        cgul_exception_t local = NULL;
        cgul_microxml__open_fname(&local, obj, fname);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Open the MicroXML file <code>f</code>.  If a file or memory buffer
     * is already open, it is closed before attempting to open the new
     * file.  This class does not take ownership of <code>f</code>.  Thus,
     * the client is still responsible for calling <code>fclose()</code>
     * on <code>f</code>.  If an error occurs, an exception is thrown.
     *
     * @param[in] f input file name
      */
    virtual void
    open_file(FILE* f)
    {
        cgul_exception_t local = NULL;
        cgul_microxml__open_file(&local, obj, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Open the MicroXML file contained in the memory buffer
     * <code>buffer</code> holding size <code>buffer_size</code> bytes
     * (not Unicode characters).  If a file or memory buffer is already
     * open, it is closed before attempting to open the new memory buffer.
     * This class does not take ownership of <code>buffer</code>.  Thus,
     * the client is still responsible for freeing the buffer if
     * necessary.  This method does not alter <code>buffer</code>, but
     * <code>buffer</code> should not be changed while it is being used by
     * this class.  If an error occurs, an exception is thrown.
     *
     * @param[in] buffer memory buffer
     * @param[in] buffer_size buffer size in bytes (not Unicode characters)
     */
    virtual void
    open_memory(const char* buffer,
                size_t buffer_size)
    {
        cgul_exception_t local = NULL;
        cgul_microxml__open_memory(&local, obj, buffer, buffer_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Close the open MicroXML file (if any).  If an error occurs, an
     * exception is thrown. <p>
     */
    virtual void
    close()
    {
        cgul_exception_t local = NULL;
        cgul_microxml__close(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the stream reader.  If no stream reader is currently
     * selected, <code>NULL</code> is returned.
     *
     * @return stream reader
     */
    virtual cgul_stream_reader_t
    get_stream_reader()
    {
        cgul_stream_reader_t rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__get_stream_reader(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the stream reader <code>sr</code>.  This class does not take
     * ownership of <code>sr</code>.  Thus, the client is still
     * responsible for deleting it.  If <code>sr</code> is
     * <code>NULL</code>, the old stream reader will no longer be used
     * making it safe to delete the old stream reader without having to
     * set a new one.  If an error occurs, an exception is thrown. <p>
     *
     * The only constraint on <code>sr</code> is that it must yield
     * Unicode characters.  If the stream is UTF-8 or starts with a
     * Unicode Byte-Order Mark (BOM), <code>cgul_stream_reader_cxx</code>
     * should automatically detect the correct decoder to use; otherwise,
     * manually set <code>sr</code> to use one of the following decoders: <p>
     *
     * <code>
     *     cgul_stream_reader_cxx::decode_utf8
     *     cgul_stream_reader_cxx::decode_utf16be
     *     cgul_stream_reader_cxx::decode_utf16le
     *     cgul_stream_reader_cxx::decode_utf32be
     *     cgul_stream_reader_cxx::decode_utf32le
     * </code>
     *
     * If you need to parse MicroXML fragments, multiple block readers
     * can be added to the stream reader at any time using
     * <code>cgul_stream_reader_cxx::add_block_reader()</code>. <p>
     *
     * @param[in] sr stream reader
     * @see add_block_reader()
     */
    virtual void
    set_stream_reader(cgul_stream_reader_t sr)
    {
        cgul_exception_t local = NULL;
        cgul_microxml__set_stream_reader(&local, obj, sr);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the presentation name.  If an input source is not currently
     * set, an exception is thrown.
     *
     * @return presentation name
     */
    virtual const char*
    get_presentation_name()
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__get_presentation_name(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the presentation name.  The presentation name is primarily used
     * for error reporting and can be set to anything, but it does need to
     * be set again if the input file changes.  If the presentation name
     * is not set, it defaults to the file name that was opened or to
     * <code>"FILE"</code> or <code>"MEMORY"</code> if using a
     * <code>FILE*</code> or memory buffer respectively.  An exception is
     * thrown if an attempt is made to set the presentation name without
     * first setting the input file.
     *
     * @param[in] presentation_name presentation name
     */
    virtual void
    set_presentation_name(const char* presentation_name)
    {
        cgul_exception_t local = NULL;
        cgul_microxml__set_presentation_name(&local, obj, presentation_name);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the maximum entity length.  If set to <code>0</code>, the
     * length is limited only by available memory.  The default is 16K.
     *
     * @return maximum entity length
     */
    virtual size_t
    get_maximum_entity_length()
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__get_maximum_entity_length(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the maximum entity length.  If set to <code>0</code>, the
     * length is limited only by available memory.  The default is 16K.
     *
     * @param[in] entity_length_max maximum entity length
     */
    virtual void
    set_maximum_entity_length(size_t entity_length_max)
    {
        cgul_exception_t local = NULL;
        cgul_microxml__set_maximum_entity_length(
            &local, obj, entity_length_max);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the starting line number for the current token.  If no input
     * source is set, <code>0</code> is returned, and an exception is
     * thrown.
     *
     * @return current line number
     */
    virtual size_t
    get_line_number()
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__get_line_number(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the starting column number for the current token.  If no
     * input source is set, <code>0</code> is returned, and an exception
     * is thrown.
     *
     * @return current column number
     */
    virtual size_t
    get_column_number()
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__get_column_number(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Read the next token from the MicroXML input source.  This method
     * can be used to implement a pull parser.  If an error occurs, an
     * exception is thrown. <p>
     *
     * @see is_document_start()
     * @see is_document_end()
     * @see is_element_start()
     * @see is_element_end()
     * @see is_text()
     * @see is_comment()
     */
    virtual void
    read_next_token()
    {
        cgul_exception_t local = NULL;
        cgul_microxml__read_next_token(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return whether the current token indicates the start of the
     * document.  This is always the first token.
     *
     * @return whether the current token indicates the start of the document
     */
    virtual bool
    is_document_start()
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__is_document_start(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (rv != 0);
    }

    /**
     * Return whether the current token indicates the end of the document.
     * This is always the last token if no errors occurr.  Attempting to
     * get the next token after this one results in an error.
     *
     * @return whether the current token indicates the end of the document
     */
    virtual bool
    is_document_end()
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__is_document_end(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (rv != 0);
    }

    /**
     * Return whether the start tag for an element has been read.  Use
     * <code>get_element_name()</code> and
     * <code>get_attributes()</code> to get the name and attributes
     * for the element.  Use <code>is_empty_element()</code> to
     * determine if the element is an empty element (e.g.,
     * <code>&lt;foo/&gt;</code>).
     *
     * @return whether the start tag for an element has been read
     */
    virtual bool
    is_element_start()
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__is_element_start(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (rv != 0);
    }

    /**
     * Return whether the end tag for an element has been read or, if an
     * empty element was parsed earlier, whether an end tag is being
     * simulated on behalf of the earlier empty element.  This way the
     * following two MicroXML samples generate the same token stream: <p>
     *
     * <pre>
     *     &lt;foo&gt;&lt;/foo&gt;
     *     &lt;foo/&gt;
     * </pre>
     *
     * Use <code>get_element_name()</code> to get the name of the
     * element.  Use <code>is_empty_element()</code> to determine if
     * the end tag is being simulated because the matching start tag
     * is an empty element. <p>
     *
     * @return whether the end tag for an element has been read
     */
    virtual bool
    is_element_end()
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__is_element_end(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (rv != 0);
    }

    /**
     * Return whether text has been read.  Use <code>get_text()</code>
     * to get the text.
     *
     * @return whether text has been read
     */
    virtual bool
    is_text()
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__is_text(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (rv != 0);
    }

    /**
     * Return whether a comment has been read.  Use
     * <code>get_comment()</code> to get the comment.
     *
     * @return whether comment has been read
     */
    virtual bool
    is_comment()
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__is_comment(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (rv != 0);
    }

    /**
     * Return the name of the current element.  The string is owned by
     * this class so the client must not attempt to free it.  This
     * method is valid when processing the start or end tag of an
     * element; otherwise, an exception is thrown. <p>
     *
     * Element names are interned by the <code>cgul_microxml_cxx</code>
     * instance making them valid until the <code>cgul_microxml_cxx</code>
     * instance is deleted or reset.  It is common for elements to be
     * repeated.  In this case, this method returns the same pointer for
     * the repeated elements which allows DOMs to be created in a memory
     * efficient manner.  Because the element names are shared, it is
     * important that the client treat them as immutable which should be
     * natural in most contexts. <p>
     *
     * @return name associated with token
     */
    virtual const char*
    get_element_name()
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__get_element_name(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the value for the attribute name
     * <code>attribute_name</code>.  If no attribute has that name,
     * <code>NULL</code> is returned.  Ownership of the attribute value
     * remains with the <code>cgul_microxml_cxx</code> instance so the client
     * must not attempt to delete it.  This method is valid when
     * processing the start or end tag of an element; otherwise,
     * an exception is thrown. <p>
     *
     * Unlike element names and attribute names, attribute values are only
     * valid until the next token is read unless
     * <code>cgul_string__take_value()</code> is called on the returned
     * attribute value in which case the client is responsible for calling
     * <code>free()</code> on the value taken. <p>
     *
     * The attribute value can be converted to numeric types using methods
     * from <code>cgul_string</code> like
     * <code>cgul_string__to_int()</code>. <p>
     *
     * @param[in] attribute_name attribute name
     * @return value for the given attribute name or <code>NULL</code>
     * @see get_attributes()
     * @see take_attributes()
     */
    virtual cgul_string_t
    get_attribute_value(const char* attribute_name)
    {
        cgul_string_t rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__get_attribute_value(&local, obj, attribute_name);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the entire attribute map associated with the current
     * element.  The keys and values stored in the attributes map are
     * owned by this class so the client must not attempt to free or
     * delete them.  This method is valid when processing the start
     * tag of an element; otherwise, an exception is thrown. <p>
     *
     * The keys in the attributes map correspond to the names of the
     * attributes and are C-style strings.  The keys are valid until this
     * <code>cgul_microxml_cxx</code> instance is deleted or reset. <p>
     *
     * The values in the attributes map correspond to the attribute values
     * and are <code>cgul_string</code> instances (not C-style strings).
     * Unlike the keys, the values are only valid until the next token is
     * read unless <code>cgul_string__take_value()</code> is called on the
     * the attribute values in which case the client is responsible for
     * calling <code>free()</code> on each value taken. <p>
     *
     * The attribute values can be converted to numeric types using
     * methods from <code>cgul_string</code> like
     * <code>cgul_string__to_int()</code>. <p>
     *
     * You can iterate over the attribute names and values as follows: <p>
     *
     * <code>
     *     cgul_hash_node_t attr = NULL;
     *     attr = cgul_hash__get_front(cex, attrs);
     *     for ( ; attr ; attr = cgul_hash_node__get_next(cex, attr)) {
     *      // Get the key/value pair for this attribute.
     *      const char* key = (const char*)cgul_hash_node__get_key(cex, attr);
     *      cgul_string_t tmp = (cgul_string_t)cgul_hash_node__get_value(cex, attr);
     *      const char* value = cgul_string__get_value(cex, tmp);
     *      ...
     *     }
     * </code>
     *
     * @return map of attributes associated with token
     * @see take_attributes()
     * @see get_attribute_value()
     */
    virtual cgul_hash_t
    get_attributes()
    {
        cgul_hash_t rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__get_attributes(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the attributes map associated with the current element.
     * This method transfers ownership of the attributes map requiring the
     * client to delete the map by calling
     * <code>cgul_microxml__delete_attributes()</code>.  The element
     * attributes can only be taken once per element.  This method is
     * valid when processing the start tag of an element; otherwise,
     * an exception is thrown. <p>
     *
     * The keys in the attributes map correspond to the names of the
     * attributes and are C-style strings.  The keys are only valid until
     * <code>cgul_microxml__delete_attributes()</code> is called or until
     * this <code>cgul_microxml_cxx</code> instance is deleted or reset.  So
     * do *not* try use the keys after deleting <code>uxml</code>!  (The
     * reason for this is that attribute names are interned in a symbol
     * table owned by this class so that multiple occurrences of
     * the same attribute name can share the same symbol without requiring
     * extra memory for a duplicate string.) <p>
     *
     * The values in the attributes map correspond to the attribute values
     * and are <code>cgul_string</code> instances (not C-style strings).
     * Unlike the keys, the values are valid until
     * <code>cgul_microxml__delete_attributes()</code> is called even if
     * <code>uxml</code> has been deleted or reset. <p>
     *
     * The attribute values can be converted to numeric types using
     * methods from <code>cgul_string</code> like
     * <code>cgul_string__to_int()</code>. <p>
     *
     * You can iterate over the attribute names and values as follows: <p>
     *
     * <code>
     *     cgul_hash_node_t attr = NULL;
     *     attr = cgul_hash__get_front(cex, attrs);
     *     for ( ; attr ; attr = cgul_hash_node__get_next(cex, attr)) {
     *      // Get the key/value pair for this attribute.
     *      const char* key = (const char*)cgul_hash_node__get_key(cex, attr);
     *      cgul_string_t tmp = (cgul_string_t)cgul_hash_node__get_value(cex, attr);
     *      const char* value = cgul_string__get_value(cex, tmp);
     *      ...
     *     }
     * </code>
     *
     * @return map of attributes associated with token
     * @see get_attributes()
     * @see get_attribute_value()
     */
    virtual cgul_hash_t
    take_attributes()
    {
        cgul_hash_t rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__take_attributes(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Convenience method used to delete the attributes map
     * <code>attributes</code> returned by the method
     * <code>take_attributes()</code>.  The client must not
     * use <code>attributes</code> after this method returns. <p>
     *
     * @note Do not use this function for attribute maps returned by
     * <code>get_attributes()</code>.  Only use it on attributes maps
     * returned by <code>take_attributes()</code>. <p>
     *
     * @param[in] attributes attributes map
     */
    virtual void
    delete_attributes(cgul_hash_t attributes)
    {
        cgul_exception_t local = NULL;
        cgul_microxml__delete_attributes(&local, obj, attributes);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return whether the element is an empty element (e.g.,
     * <code>&lt;foo/&gt;</code>).  This method is valid when processing
     * the start or end tag of an element; otherwise, an exception is
     * thrown.
     *
     * @return whether the empty-element syntax was used to close the start tag
     */
    virtual bool
    is_empty_element()
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__is_empty_element(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (rv != 0);
    }

    /**
     * Return the current text.  The returned value is owned by this
     * class so the client must not attempt to delete it.  This method
     * is valid when processing text; otherwise, an exception is
     * thrown. <p>
     *
     * The text returned is only valid until the next token is read unless
     * <code>cgul_string__take_value()</code> is called on the returned
     * value in which case the client is responsible for calling
     * <code>free()</code> on the text taken. <p>
     *
     * The text can be converted to numeric types using methods from
     * <code>cgul_string</code> like
     * <code>cgul_string__to_int()</code>. <p>
     *
     * @return text associated with token
     * @see set_maximum_text_length()
     * @see set_skip_white_space()
     */
    virtual cgul_string_t
    get_text()
    {
        cgul_string_t rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__get_text(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return whether the current text is comprised of only white
     * space characters.  The MicroXML specification defines white
     * space as tab, new line, or space characters.  This method is
     * valid when processing text; an exception is thrown.
     *
     * @return whether text is pure white space
     */
    virtual bool
    is_white_space()
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__is_white_space(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (rv != 0);
    }

    /**
     * Return whether to skip text under the root element that is pure
     * white space.  By default, this white space is not skipped.  For
     * white space in the prolog and epilog, see
     * <code>get_skip_prolog_or_epilog_tokens()</code>.
     *
     * @return whether to skip text that is pure white space
     * @see get_skip_prolog_or_epilog_tokens()
     */
    virtual bool
    get_skip_white_space()
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__get_skip_white_space(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (rv != 0);
    }

    /**
     * Set whether to skip text under the root element that is pure white
     * space.  This is convenient when the document being parsed is
     * formatted with line breaks and indentation that are not
     * significant.  By default, this white space is not skipped.  For
     * white space in the prolog and epilog, see
     * <code>set_skip_prolog_or_epilog_tokens()</code>. <p>
     *
     * If enabled, <code>set_maximum_text_length()</code>
     * should be used to increase the maximum size of each block of text
     * (if necessary) so that at least one character that is not white
     * space is always included in each block of significant white
     * space. <p>
     *
     * @param[in] skip whether to skip text that is pure white space
     * @see set_maximum_text_length()
     * @see set_skip_prolog_or_epilog_tokens()
     */
    virtual void
    set_skip_white_space(int skip)
    {
        cgul_exception_t local = NULL;
        cgul_microxml__set_skip_white_space(&local, obj, skip);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the maximum size in Unicode characters of text returned
     * by <code>get_text()</code>.  If set to <code>0</code>, the
     * length is limited only by available memory.  The default is
     * 16K.
     *
     * @return maximum text length in Unicode characters
     */
    virtual size_t
    get_maximum_text_length()
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__get_maximum_text_length(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the maximum size in Unicode characters of text returned by
     * <code>get_text()</code>.  If set to <code>0</code>, the length
     * is limited only by available memory.  The default is 16K.
     *
     * @param[in] text_length_max maximum text length in Unicode characters
     */
    virtual void
    set_maximum_text_length(size_t text_length_max)
    {
        cgul_exception_t local = NULL;
        cgul_microxml__set_maximum_text_length(&local, obj, text_length_max);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the current comment.  The returned value is owned by this
     * class so the client must not attempt to delete it.  This method is
     * valid when processing comments; otherwise, an exception is thrown. <p>
     *
     * The comment returned is only valid until the next token is read
     * unless <code>cgul_string__take_value()</code> is called on the
     * returned value in which case the client is responsible for calling
     * <code>free()</code> on the comment taken. <p>
     *
     * @return comment associated with token
     * @see set_skip_comments()
     */
    virtual cgul_string_t
    get_comment()
    {
        cgul_string_t rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__get_comment(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return whether to skip comments under the root element.  By
     * default, these comments are skipped.  For comments in the
     * prolog and epilog, see
     * <code>get_skip_prolog_or_epilog_tokens()</code>.
     *
     * @return whether to skip text that is pure white space
     * @see get_skip_prolog_or_epilog_tokens()
     */
    virtual bool
    get_skip_comments()
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__get_skip_comments(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (rv != 0);
    }

    /**
     * Set whether to skip comments under the root element.  By default,
     * these comments are skipped.  For comments in the prolog and epilog,
     * see <code>set_skip_prolog_or_epilog_tokens()</code>.
     *
     * @param[in] skip whether to skip text that is pure white space
     * @see set_skip_prolog_or_epilog_tokens()
     */
    virtual void
    set_skip_comments(int skip)
    {
        cgul_exception_t local = NULL;
        cgul_microxml__set_skip_comments(&local, obj, skip);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the maximum comment length.  If set to <code>0</code>, the
     * length is limited only by available memory.  The default is 16K.
     *
     * @return maximum comment length
     */
    virtual size_t
    get_maximum_comment_length()
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__get_maximum_comment_length(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the maximum comment length.  If set to <code>0</code>, the
     * length is limited only by available memory.  The default is 16K.
     *
     * @param[in] comment_length_max maximum comment length
     */
    virtual void
    set_maximum_comment_length(size_t comment_length_max)
    {
        cgul_exception_t local = NULL;
        cgul_microxml__set_maximum_comment_length(
            &local, obj, comment_length_max);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return whether to skip tokens in the prolog or epilog.  These
     * are the white space and comment tokens that surround the root
     * element.  By default, these tokens are skipped.  For white
     * space and comments under the root element, see
     * <code>get_skip_white_space()</code> and
     * <code>get_skip_comments()</code>.
     *
     * @return whether tokens in the prolog or epilog
     * @see get_skip_white_space()
     * @see get_skip_comments()
     */
    virtual bool
    get_skip_prolog_or_epilog_tokens()
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_microxml__get_skip_prolog_or_epilog_tokens(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (rv != 0);
    }

    /**
     * Set whether to skip tokens in the prolog or epilog.  These are
     * the white space and comment tokens that surround the root
     * element.  By default, these tokens are skipped.  For white
     * space and comments under the root element, see
     * <code>set_skip_white_space()</code> and
     * <code>set_skip_comments()</code>.
     *
     * @param[in] skip whether tokens in the prolog or epilog
     * @see set_skip_white_space()
     * @see set_skip_comments()
     */
    virtual void
    set_skip_prolog_or_epilog_tokens(int skip)
    {
        cgul_exception_t local = NULL;
        cgul_microxml__set_skip_prolog_or_epilog_tokens(&local, obj, skip);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }


    /**
     * Get the underlying <code>cgul_microxml</code> object.
     *
     * @return underlying object
     */
    virtual cgul_microxml_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_microxml</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_microxml__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_microxml_t
    take_obj()
    {
        cgul_microxml_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_microxml_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_microxml__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /** Underlying <code>cgul_microxml</code> object. */
    cgul_microxml_t obj;

};



#endif  /*  CGUL_MICROXML_CXX_H  */
