#ifndef CGUL_MULTIMAP_H
#define CGUL_MULTIMAP_H

/**
 * Multimap implementation. <p>
 *
 * @file
 * @author Paul Serice
 * @brief multimap implementation
 */

#include "cgul_common.h"
#include "cgul_cache.h"
#include "cgul_exception.h"
#include "cgul_multimap_node.h"
#include "cgul_rbtree.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_multimap</code> instance. */
typedef struct cgul_multimap* cgul_multimap_t;

/**
 * This typedef is the interface the client must define in order for
 * <code>cgul_multimap</code> to sort your keys as they are inserted.
 * When you define your comparison function, you should compare
 * <code>key1</code> to <code>key2</code> and return less than zero,
 * zero, or greater than zero if <code>key1</code> is less than, equal
 * to, or greater than <code>key2</code>.
 *
 * @param[in] key1 first key
 * @param[in] key2 second key
 * @return result of comparison
 */
typedef int (*cgul_multimap__compare_t)(const void* key1,
                                        const void* key2);

/**
 * This typedef is the interface for the combining function used by
 * the following methods: <p>
 *
 * <pre>
 *     <code>cgul_multimap__foldl_keys()</code>
 *     <code>cgul_multimap__foldr_keys()</code>
 * </pre>
 *
 * The client must not alter <code>key</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] key key
 * @param[in] data client data
 * @return whether to continue
 */
typedef int (*cgul_multimap__fold_key_t)(cgul_exception_t* cex,
                                         const void* key,
                                         void* data);

/**
 * This typedef is the interface for the combining function used by
 * the following methods: <p>
 *
 * <pre>
 *     <code>cgul_multimap__foldl_values()</code>
 *     <code>cgul_multimap__foldr_values()</code>
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @param[in] value value
 * @param[in] data client data
 * @return whether to continue
 */
typedef int (*cgul_multimap__fold_value_t)(cgul_exception_t* cex,
                                           void* value,
                                           void* data);

/**
 * This typedef is the interface for the combining function used by
 * the following methods: <p>
 *
 * <pre>
 *     <code>cgul_multimap__foldl_pairs()</code>
 *     <code>cgul_multimap__foldr_pairs()</code>
 * </pre>
 *
 * The client must not alter <code>key</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] key key
 * @param[in] value value
 * @param[in] data client data
 * @return whether to continue
 */
typedef int (*cgul_multimap__fold_pair_t)(cgul_exception_t* cex,
                                          const void* key,
                                          void* value,
                                          void* data);

/**
 * This typedef is the interface for the callback function used by the
 * following methods: <p>
 *
 * <pre>
 *     <code>cgul_multimap__traverse()</code>
 *     <code>cgul_multimap__traverse_range()</code>
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] n node
 * @param[in] data client data
 * @return whether to continue
 */
typedef int (*cgul_multimap__traverse_t)(cgul_exception_t* cex,
                                         cgul_multimap_t mm,
                                         cgul_multimap_node_t n,
                                         void* data);

/**
 * Create a new <code>cgul_multimap</code> object.
 * <code>compare</code> is the comparison function that allows the
 * container to maintain its sorted order.  If memory cannot be
 * allocated, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * The client is responsible for freeing the object by calling
 * <code>cgul_multimap__delete()</code>.  Also, the client is
 * responsible for freeing the key/value pairs only after their nodes
 * have been permanently removed from the <code>cgul_multimap</code>.
 * The <code>cgul_multimap__traverse()</code> method can be used to
 * remove each node safely before calling
 * <code>cgul_multimap__delete()</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] compare comparison function
 * @return new <code>cgul_multimap</code> instance
 * @see cgul_multimap__compare_t
 */
extern CGUL_EXPORT cgul_multimap_t
cgul_multimap__new(cgul_exception_t* cex,
                   cgul_multimap__compare_t compare);

/**
 * This method frees all internally allocated memory.  This does not
 * include the key/value pairs stored in the multimap.  The client is
 * responsible for freeing the key/value pairs when the client thinks
 * it is convenient and safe to do so; however, the client must
 * understand that freeing a key before removing its node invalidates
 * the data structure. <p>
 *
 * As a convenience, you may want to call
 * <code>cgul_multimap__clear()</code> before calling this method in
 * order to properly put the keys and values back on their respective
 * <code>cgul_cache</code> objects (if <code>cgul_cache</code>
 * objects are being used) before you delete the multimap.  If
 * <code>cgul_cache</code> objects are not being used, the client
 * needs to arrange some other mechanism to free the keys or
 * values. <p>
 *
 * Do not try to access <code>mm</code> after calling this method.
 *
 * @param[in] mm <code>cgul_multimap</code> instance
 * @see cgul_multimap__free_keys()
 * @see cgul_multimap__free_values()
 */
extern CGUL_EXPORT void
cgul_multimap__delete(cgul_multimap_t mm);

/**
 * This method calls <code>free()</code> on all the keys in the
 * multimap <code>mm</code>.  Because this is such a common operation,
 * it is an exception to the rule that cgul containers never free
 * keys.  This method should only ever be called immediately before
 * calling <code>cgul_multimap__delete()</code> because it otherwise
 * invalidates the multimap.  Because this function is basically an
 * extension of <code>cgul_multimap__delete()</code>, it does not
 * accept a <code>cgul_exception</code> object as an input parameter.
 *
 * @param[in] mm <code>cgul_multimap</code> instance
 */
extern CGUL_EXPORT void
cgul_multimap__free_keys(cgul_multimap_t mm);

/**
 * This method calls <code>free()</code> on all the values in the
 * multimap <code>mm</code>.  Because this is such a common operation,
 * it is an exception to the rule that cgul containers never free
 * values.  This method should only ever be called immediately before
 * calling <code>cgul_multimap__delete()</code> because it otherwise
 * invalidates the multimap.  Because this function is basically an
 * extension of <code>cgul_multimap__delete()</code>, it does not
 * accept a <code>cgul_exception</code> object as an input parameter.
 *
 * @param[in] mm <code>cgul_multimap</code> instance
 */
extern CGUL_EXPORT void
cgul_multimap__free_values(cgul_multimap_t mm);

/**
 * Whether the multimap <code>mm</code> is empty.
 *
 * @param[in] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @return whether the multimap <code>mm</code> is empty
 */
extern CGUL_EXPORT int
cgul_multimap__is_empty(cgul_exception_t* cex,
                        cgul_multimap_t mm);

/**
 * Insert the <code>(key, value)</code> pair into the multimap
 * <code>mm</code>.  If <code>node</code> is not <code>NULL</code>, it
 * will hold the newly inserted node upon completion.  The return
 * value is the count of how many values in the multimap are now
 * associated with <code>key</code>.  If an error occurs,
 * <code>0</code> is returned, <code>*node</code> is set to
 * <code>NULL</code> (if <code>node</code> is not <code>NULL</code>),
 * and an exception is thrown. <p>
 *
 * It is important to understand that the <code>cgul_multimap</code>
 * class (like all containers in the cgul library) does not take
 * ownership of <code>key</code> or <code>value</code>.  It also does
 * not attempt to make a copy of the thing pointed to by
 * <code>key</code> or <code>value</code>.  This gives the user
 * complete control over the lifetime of these objects and gives a
 * real performance boost in many cases; however, this means that
 * <code>key</code> must not be invalidated while it is still being
 * used by this class. <p>
 *
 * Most containers in the cgul library, map one key to one value.
 * This class is different.  It maps one key to many values.  Thus,
 * this class really only needs one instance of the key even though
 * the insert method is called many times, and in fact, this class
 * only keeps the first instance of <code>key</code> that is inserted.
 * Thus, it is important to check the value returned by this method to
 * determine when to free your keys. <p>
 *
 * For example, if you use <code>malloc()</code> to dynamically
 * allocate a new <code>key</code> every time this method is called,
 * the key should probably be freed for all but the first insert:
 *
 * <pre>
 *
 *     count = cgul_multimap__insert(cex, mm, key, value, NULL);
 *     if (*cex) {
 *         goto out;
 *     }
 *     if (count > 1) {
 *         free(key);
 *     }
 *
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] key key
 * @param[in] value value
 * @param[out] node newly inserted node
 * @return count of values now associated with <code>key</code>
 */
extern CGUL_EXPORT unsigned long int
cgul_multimap__insert(cgul_exception_t* cex,
                      cgul_multimap_t mm,
                      void* key,
                      void* value,
                      cgul_multimap_node_t* node);

/**
 * This method works the same as <code>cgul_multimap__insert()</code>
 * except if <code>hint</code> is not <code>NULL</code> it calls the
 * comparison function (passed into <code>cgul_multimap__new()</code>)
 * to determine if <code>key</code> is equal to the key for
 * <code>hint</code>.  If so, the insertion can occur in constant time
 * instead of logarithmic time because <code>hint</code> already holds
 * a back reference to the underlying node for <code>key</code>. <p>
 *
 * If <code>hint</code> is <code>NULL</code> or <code>key</code> does
 * not equal the key for <code>hint</code>, this method simply
 * forwards the request to <code>cgul_multimap__insert()</code>, and
 * the insertion occurs in logarithmic time. <p>
 * 
 * @param[in,out] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] key key
 * @param[in] value value
 * @param[in] hint hint
 * @param[out] node newly inserted node
 * @return count of values now associated with <code>key</code>
 */
extern CGUL_EXPORT unsigned long int
cgul_multimap__insert_with_hint(cgul_exception_t* cex,
                                cgul_multimap_t mm,
                                void* key,
                                void* value,
                                cgul_multimap_node_t hint,
                                cgul_multimap_node_t* node);

/**
 * This method finds the smallest node associated with
 * <code>key</code> in the multimap <code>mm</code> and returns it in
 * <code>node</code>.  The return value of this method is the total
 * number of values associated with <code>key</code>.  If
 * <code>key</code> cannot be found, <code>node</code> will be set to
 * <code>NULL</code>, and <code>0</code> will be returned. <p>
 *
 * The caller can use the return value to iterate over all the values
 * associated with <code>key</code>.  For example, the following shows
 * how to iterate over all the values associated with <code>key</code>
 * in the order in which they were inserted: <p>
 *
 * <pre>
 *
 *     count = cgul_multimap__find(cex, mm, key, 0, &node);
 *     for (i = 0 ; i < count ; ++i) {
 *         ...
 *         node = cgul_multimap_node__get_next(cex, node, 0);
 *     }
 *
 * </pre>
 *
 * The primary sort criterion is the user's <code>compare</code>
 * function, of course, but the <code>older_is_greater</code> boolean
 * is used as a secondary sort criterion.  When a key maps to many
 * values, the <code>older_is_greater</code> boolean determines how to
 * sort the older nodes relative to the newer nodes. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] key key
 * @param[in] older_is_greater whether older nodes are greater than newer nodes
 * @param[out] node smallest node associated with <code>key</code>
 * @return count of values associated with <code>key</code>
 */
extern CGUL_EXPORT unsigned long int
cgul_multimap__find(cgul_exception_t* cex,
                    cgul_multimap_t mm,
                    void* key,
                    int older_is_greater,
                    cgul_multimap_node_t* node);

/**
 * Find the node having the largest key less than or equal to the
 * search key <code>search_key</code> and return it in
 * <code>*node</code>.  If such a key does not exist,
 * <code>NULL</code> is returned instead.  This method returns the
 * number of values that directly map to the floor. <p>
 *
 * The iterator returned in <code>*node</code> is for the floor key's
 * smallest value as determined by <code>older_is_greater</code> such
 * that forward iteration includes all the values for the floor
 * node. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] search_key search key
 * @param[in] older_is_greater whether older nodes are greater than newer nodes
 * @param[out] node floor or <code>NULL</code>
 * @return count of values associated with the floor
 */
extern CGUL_EXPORT unsigned long int
cgul_multimap__find_floor(cgul_exception_t* cex,
                          cgul_multimap_t mm,
                          void* search_key,
                          int older_is_greater,
                          cgul_multimap_node_t* node);

/**
 * Find the node having the smallest key greater than or equal to the
 * search key <code>search_key</code> and return it in
 * <code>*node</code>.  If such a key does not exist,
 * <code>NULL</code> is returned instead.  This method returns the
 * number of values that directly map to the ceiling. <p>
 *
 * The iterator returned in <code>*node</code> is for the ceiling
 * key's largest value as determined by <code>older_is_greater</code>
 * such that reverse iteration includes all the values for the ceiling
 * node. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] search_key search key
 * @param[in] older_is_greater whether older nodes are greater than newer nodes
 * @param[out] node ceiling or <code>NULL</code>
 * @return count of values associated with the ceiling
 */
extern CGUL_EXPORT unsigned long int
cgul_multimap__find_ceiling(cgul_exception_t* cex,
                            cgul_multimap_t mm,
                            void* search_key,
                            int older_is_greater,
                            cgul_multimap_node_t* node);

/**
 * This method removes <code>key</code> and all the values associated
 * with <code>key</code> from the multimap <code>mm</code> and returns
 * <code>1</code> on success.  On failure, this method returns
 * <code>0</code> which indicates that <code>key</code> could not be
 * found in the multimap. <p>
 *
 * Strictly as a convenience, this method is an exception to the rule
 * that cgul containers never free keys or values.  If you pass in
 * <code>keys_cache</code> or <code>values_cache</code> instances that
 * are not <code>NULL</code>, the keys or values will be put back on
 * their respective caches. <p>
 *
 * If you need to remove just one value associated with
 * <code>key</code> instead of all the values associated with
 * <code>key</code>, use <code>cgul_multimap__remove_node()</code>
 * instead. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] key key
 * @param[in] keys_cache keys cache
 * @param[in] values_cache values cache
 * @return whether the nodes associated with <code>key</code> were removed
 * @see cgul_cache__get_freer()
 */
extern CGUL_EXPORT int
cgul_multimap__remove(cgul_exception_t* cex,
                      cgul_multimap_t mm,
                      void* key,
                      cgul_cache_t keys_cache,
                      cgul_cache_t values_cache);

/**
 * Remove <code>node</code> from the multimap <code>mm</code>.  The
 * key and value pointers stored in the node that is to be removed
 * will be returned in <code>key_out</code> and <code>value_out</code>
 * if you pass in pointers that are not <code>NULL</code>, but the
 * value for the key that is returned will be <code>NULL</code> if the
 * multimap is still using the key.  This method returns the number of
 * values still associated with the same key. <p>
 *
 * Because the key in <code>node</code> might map to more than one
 * value, the caller should never invalidate the key until after this
 * method returns a count of zero or, equivalently, a value for
 * <code>key</code> that is not <code>NULL</code>: <p>
 *
 * <pre>
 *
 *     cgul_multimap__remove_node(cex, mm, node, &key, &value);
 *     if (key) {
 *         free(key);
 *     }
 *     free(value);
 *
 * </pre>
 *
 * It is almost always a mistake to naively call this method while
 * iterating over the multimap.  The problem is that calling this
 * method invalidates the node making it impossible to call
 * <code>cgul_multimap_node__get_next()</code> afterward.  The
 * solution is simple.  Just call
 * <code>cgul_multimap_node__get_next()</code> before calling this
 * method:
 *
 * <pre>
 *
 *     count = cgul_multimap__find(cex, mm, key, 1, &curr);
 *     if (curr) {
 *         for (i = 0 ; i < count ; ++i, curr = next) {
 *             next = cgul_multimap_node__get_next(cex, curr, 1);
 *             cgul_multimap__remove_node(cex,
 *                                        mm,
 *                                        curr,
 *                                        &key_out,
 *                                        &value_out);
 *             if (key_out) {
 *                 free(key_out);
 *             }
 *             free(value_out);
 *         }
 *     }
 *
 * </pre>
 *
 * Alternatively, you can use <code>cgul_multimap__remove()</code>,
 * <code>cgul_multimap__remove_range()</code>,
 * <code>cgul_multimap__traverse()</code>, or
 * <code>cgul_multimap__traverse_range()</code>. <p>
 *
 * If you use one of the traversal methods, it is safe to call this
 * method naively.  For example, if your multimap stores
 * <code>cgul_string</code> objects for both its keys and values,
 * another way to remove all the values associated with a key as
 * follows: <p>
 *
 * <pre>
 *
 *     static int
 *     remove_node_cb(cgul_multimap_t mm,
 *                    cgul_multimap_node_t node,
 *                    void* data)
 *     {
 *         cgul_string_t key = NULL;
 *         cgul_string_t value = NULL;
 *         cgul_exception_t local = NULL;
 *         cgul_multimap__remove_node(&local,
 *                                    mm,
 *                                    node,
 *                                    (void**)&key,
 *                                    (void**)&value);
 *         if (key) {
 *             cgul_string__delete(key);
 *         }
 *         cgul_string__delete(value);
 *         return 1;
 *     }
 *
 *     cgul_multimap_node_t first = NULL;
 *     cgul_multimap_node_t last = NULL;
 *     cgul_multimap__find(cex, mm, key, 1, &first);
 *     cgul_multimap__find(cex, mm, key, 0, &last);
 *     if (first && last) {
 *         cgul_multimap__traverse_range(cex,
 *                                       mm,
 *                                       first,
 *                                       last,
 *                                       1,
 *                                       &remove_node_cb,
 *                                       NULL);
 *     }
 *
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] node node
 * @param[out] key_out pointer for the key stored in the node
 * @param[out] value_out pointer for the value stored in the node
 * @return count of values now associated with <code>key</code>
 */
extern CGUL_EXPORT unsigned long int
cgul_multimap__remove_node(cgul_exception_t* cex,
                           cgul_multimap_t mm,
                           cgul_multimap_node_t node,
                           void** key_out,
                           void** value_out);

/**
 * This method removes nodes in the range <code>first</code>
 * (inclusive) to <code>last</code> (inclusive).  Strictly as a
 * convenience, this method is an exception to the rule that cgul
 * containers never free keys or values.  If you pass in
 * <code>keys_cache</code> or <code>values_cache</code> instances that
 * are not <code>NULL</code>, the keys or values will be put back on
 * their respective caches. <p>
 *
 * The primary sort criterion is the user's <code>compare</code>
 * function, of course, but the <code>older_is_greater</code> boolean
 * is used as a secondary sort criterion.  When a key maps to many
 * values, the <code>older_is_greater</code> boolean determines how to
 * sort the older nodes relative to the newer nodes. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] first first node in range
 * @param[in] last last node in range
 * @param[in] older_is_greater whether older nodes are greater than newer nodes
 * @param[in] keys_cache keys cache
 * @param[in] values_cache values cache
 * @return whether any nodes were removed.
 * @see cgul_cache__get_freer()
 */
extern CGUL_EXPORT int
cgul_multimap__remove_range(cgul_exception_t* cex,
                            cgul_multimap_t mm,
                            cgul_multimap_node_t first,
                            cgul_multimap_node_t last,
                            int older_is_greater,
                            cgul_cache_t keys_cache,
                            cgul_cache_t values_cache);

/**
 * This method removes the front node.  The key and value pointers
 * stored in the node that is to be removed will be returned in
 * <code>key_out</code> and <code>value_out</code> if you pass in
 * pointers that are not <code>NULL</code>.  Note however that because
 * keys map to multiple values, the key for this node may still be in
 * use by another node.  If this is the case, <code>NULL</code> is
 * returned in <code>key_out</code>.  An example of how to use this
 * method follows: <p>
 *
 * <pre>
 *
 *     cgul_multimap__remove_front(cex, mm, 0, &key_out, &value_out);
 *     if (key_out) {
 *         free(key_out);
 *     }
 *     free(value_out);
 *
 * </pre>
 *
 * The primary sort criterion is the user's <code>compare</code>
 * function, of course, but the <code>older_is_greater</code> boolean
 * is used as a secondary sort criterion.  When a key maps to many
 * values, the <code>older_is_greater</code> boolean determines how to
 * sort the older nodes relative to the newer nodes.  See
 * <code>cgul_multimap__get_front()</code> for more details. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] older_is_greater whether older nodes are greater than newer nodes
 * @param[out] key_out pointer for the key stored in the node
 * @param[out] value_out pointer for the value stored in the node
 * @return whether the node was removed
 */
extern CGUL_EXPORT int
cgul_multimap__remove_front(cgul_exception_t* cex,
                            cgul_multimap_t mm,
                            int older_is_greater,
                            void** key_out,
                            void** value_out);

/**
 * This method removes the back node.  The key and value pointers
 * stored in the node that is to be removed will be returned in
 * <code>key_out</code> and <code>value_out</code> if you pass in
 * pointers that are not <code>NULL</code>.  Note however that because
 * keys map to multiple values, the key for this node may still be in
 * use by another node.  If this is the case, <code>NULL</code> is
 * returned in <code>key_out</code>.  An example of how to use this
 * method follows: <p>
 *
 * <pre>
 *
 *     cgul_multimap__remove_back(cex, mm, 1, &key_out, &value_out);
 *     if (key_out) {
 *         free(key_out);
 *     }
 *     free(value_out);
 *
 * </pre>
 *
 * The primary sort criterion is the user's <code>compare</code>
 * function, of course, but the <code>older_is_greater</code> boolean
 * is used as a secondary sort criterion.  When a key maps to many
 * values, the <code>older_is_greater</code> boolean determines how to
 * sort the older nodes relative to the newer nodes.  See
 * <code>cgul_multimap__get_back()</code> for more details. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] older_is_greater whether older nodes are greater than newer nodes
 * @param[out] key_out pointer for the key stored in the node
 * @param[out] value_out pointer for the value stored in the node
 * @return whether the node was removed
 */
extern CGUL_EXPORT int
cgul_multimap__remove_back(cgul_exception_t* cex,
                           cgul_multimap_t mm,
                           int older_is_greater,
                           void** key_out,
                           void** value_out);

/**
 * Return the front node according to sort order.  This operation is
 * O(1).  If the multimap is empty, <code>NULL</code> is returned.
 * This class does not have to search the multimap in order to find
 * the front node because it indirectly keeps a pointer to the front
 * node up to date. <p>
 *
 * The primary sort criterion is the user's <code>compare</code>
 * function, of course, but the <code>older_is_greater</code> boolean
 * is used as a secondary sort criterion.  When a key maps to many
 * values, the <code>older_is_greater</code> boolean determines how to
 * sort the older nodes relative to the newer nodes. <p>
 *
 * The way the <code>older_is_greater</code> boolean works in this
 * implementation is that each key maps to a linked list of values.
 * The <code>compare</code> function only determines the relative
 * order of the keys.  To order the values, this implementation always
 * pushes newer values onto the front of the list.  Thus, the newest
 * value is always at the front of the list, and the oldest value is
 * always at the back of the list.  If <code>older_is_greater</code>
 * is true, the sort order is from front to back.  If
 * <code>older_is_greater</code> is false, the sort order is from back
 * to front. <p>
 *
 * For example, if you use this class to implement a priority queue
 * and want to get the node with the lowest priority that has been
 * waiting the longest, you would call this method with
 * <code>older_is_greater</code> set to <code>0</code>. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] older_is_greater whether older nodes are greater than newer nodes
 * @return front node
 * @see cgul_multimap_node__get_next()
 * @see cgul_multimap_node__get_prev()
 */
extern CGUL_EXPORT cgul_multimap_node_t
cgul_multimap__get_front(cgul_exception_t* cex,
                         cgul_multimap_t mm,
                         int older_is_greater);

/**
 * Return the back node according to sort order.  This operation is
 * O(1).  If the multimap is empty, <code>NULL</code> is returned.
 * This class does not have to search the multimap in order to find
 * the back node because it indirectly keeps a pointer to the back
 * node up to date.
 *
 * The primary sort criterion is the user's <code>compare</code>
 * function, of course, but the <code>older_is_greater</code> boolean
 * is used as a secondary sort criterion.  When a key maps to many
 * values, the <code>older_is_greater</code> boolean determines how to
 * sort the older nodes relative to the newer nodes. <p>
 *
 * The way the <code>older_is_greater</code> boolean works in this
 * implementation is that each key maps to a linked list of values.
 * The <code>compare</code> function only determines the relative
 * order of the keys.  To order the values, this implementation always
 * pushes newer values onto the front of the list.  Thus, the newest
 * value is always at the front of the list, and the oldest value is
 * always at the back of the list.  If <code>older_is_greater</code>
 * is true, the sort order is from front to back.  If
 * <code>older_is_greater</code> is false, the sort order is from back
 * to front. <p>
 *
 * For example, if you use this class to implement a priority queue
 * and want to get the node with the highest priority that has been
 * waiting the longest, you would call this method with
 * <code>older_is_greater</code> set to <code>1</code>. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] older_is_greater whether older nodes are greater than newer nodes
 * @return back node
 * @see cgul_multimap_node__get_next()
 * @see cgul_multimap_node__get_prev()
 */
extern CGUL_EXPORT cgul_multimap_node_t
cgul_multimap__get_back(cgul_exception_t* cex,
                        cgul_multimap_t mm,
                        int older_is_greater);

/**
 * Return the oldest node according to chronological order (i.e., the
 * order in which the nodes are inserted).  This operation is O(1).
 * If the multimap is empty, <code>NULL</code> is returned.  This
 * class does not have to search the multimap in order to find the
 * oldest node because it efficiently keeps a direct pointer to the
 * oldest node up to date. <p>
 *
 * The following example shows how to iterate over the entire multimap
 * in chronological order:
 *
 * <pre>
 *     cgul_multimap_node_t n = cgul_multimap__get_oldest(cex, mm);
 *     for ( ; n ; n = cgul_multimap_node__get_younger(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @return oldest node
 * @see cgul_multimap_node__get_younger()
 */
extern CGUL_EXPORT cgul_multimap_node_t
cgul_multimap__get_oldest(cgul_exception_t* cex,
                          cgul_multimap_t mm);

/**
 * Set the oldest node in the multimap to be <code>n</code>.  Calling
 * this method has the potential to confuse iterators and should be
 * handled with roughly the same level of caution as calling
 * <code>cgul_multimap__remove_node()</code>. <p>
 *
 * This method could be used, for example, if your code expires the
 * oldest node in the multimap, and you want to force <code>n</code> to be
 * the next node to expire. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] n node
 * @see cgul_multimap__set_youngest()
 */
extern CGUL_EXPORT void
cgul_multimap__set_oldest(cgul_exception_t* cex,
                          cgul_multimap_t mm,
                          cgul_multimap_node_t n);

/**
 * Return the youngest node according to chronological order (i.e.,
 * the order in which the nodes are inserted).  This operation is
 * O(1).  If the multimap is empty, <code>NULL</code> is returned.
 * This class does not have to search the multimap in order to find
 * the youngest node because it efficiently keeps a direct pointer to
 * the youngest node up to date. <p>
 *
 * The following example shows how to iterate over the entire multimap
 * in reverse chronological order:
 *
 * <pre>
 *     cgul_multimap_node_t n = cgul_multimap__get_youngest(cex, mm);
 *     for ( ; n ; n = cgul_multimap_node__get_older(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @return youngest node
 * @see cgul_multimap_node__get_older()
 */
extern CGUL_EXPORT cgul_multimap_node_t
cgul_multimap__get_youngest(cgul_exception_t* cex,
                            cgul_multimap_t mm);

/**
 * Set the youngest node in the multimap to be <code>n</code>.
 * Calling this method has the potential to confuse iterators and
 * should be handled with roughly the same level of caution as calling
 * <code>cgul_multimap__remove_node()</code>. <p>
 *
 * This method could be used, for example, if your code expires the
 * least-recently used (LRU) node.  By calling
 * <code>cgul_multimap__set_youngest()</code> each time a node is used,
 * the least-recently used node will be the oldest node in the
 * multimap. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] n node
 * @see cgul_multimap__set_oldest()
 */
extern CGUL_EXPORT void
cgul_multimap__set_youngest(cgul_exception_t* cex,
                            cgul_multimap_t mm,
                            cgul_multimap_node_t n);

/**
 * This method clears the multimap by removing each node individually.
 * Strictly as a convenience, this method is an exception to the rule
 * that cgul containers never free keys or values.  If you pass in
 * <code>keys_cache</code> or <code>values_cache</code> instances that
 * are not <code>NULL</code>, the keys or values will be put back on
 * their respective caches. <p>
 *
 * Another easy way to clear your multimap is to just keep removing
 * nodes from the front of the multimap freeing the key/value pairs as
 * you go:
 *
 * <pre>
 *     void* key = NULL;
 *     void* value = NULL;
 *     while (cgul_multimap__remove_front(cex, mm, 0, &key, &value)) {
 *         if (key) {
 *             free(key);
 *         }
 *         free(value);
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] keys_cache keys cache
 * @param[in] values_cache values cache
 * @see cgul_multimap__remove_range
 * @see cgul_cache__get_freer()
 */
extern CGUL_EXPORT void
cgul_multimap__clear(cgul_exception_t* cex,
                     cgul_multimap_t mm,
                     cgul_cache_t keys_cache,
                     cgul_cache_t values_cache);

/**
 * Return the total number of values stored in <code>mm</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @return size of the multimap
 */
extern CGUL_EXPORT unsigned long int
cgul_multimap__get_size(cgul_exception_t* cex,
                        cgul_multimap_t mm);

/**
 * Get the underlying tree of keys.  This is needed by
 * <code>cgul_multimap_cxx::foldl_keys()</code> and
 * <code>cgul_multimap_cxx::foldr_keys()</code> in order to
 * efficiently find the keys.
 */
extern CGUL_EXPORT cgul_rbtree_t
cgul_multimap__get_tree(cgul_exception_t* cex,
                        cgul_multimap_t mm);

/**
 * Swap the underlying data for <code>mm1</code> and <code>mm2</code>.
 * For large multimaps, this should be much faster than trying to do
 * the same thing using removes and inserts.
 *
 * @param[in] cex c-style exception
 * @param[in] mm1 first <code>cgul_multimap</code> instance
 * @param[in] mm2 second <code>cgul_multimap</code> instance
 */
extern CGUL_EXPORT void
cgul_multimap__swap(cgul_exception_t* cex,
                    cgul_multimap_t mm1,
                    cgul_multimap_t mm2);

/**
 * This method performs a left fold of the multimap <code>mm</code>
 * with the combining function <code>f</code>.  <code>f</code> is
 * called once for each key in the multimap starting at the front of
 * the multimap and iterating forward to the end of the multimap. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current key.  The third parameter passed into
 * <code>f</code> is the client data which is where the result of the
 * fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_multimap__foldl_keys(cgul_exception_t* cex,
                          cgul_multimap_t mm,
                          cgul_multimap__fold_key_t f,
                          void* data);

/**
 * This method performs a right fold of the multimap <code>mm</code>
 * with the combining function <code>f</code>.  <code>f</code> is
 * called once for each key in the multimap starting at the back of
 * the multimap and iterating backward to the front of the
 * multimap. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current key.  The third parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_multimap__foldr_keys(cgul_exception_t* cex,
                          cgul_multimap_t mm,
                          cgul_multimap__fold_key_t f,
                          void* data);

/**
 * This method performs a left fold of the multimap <code>mm</code>
 * with the combining function <code>f</code>.  <code>f</code> is
 * called once for each value in the multimap starting at the front of
 * the multimap and iterating forward to the end of the multimap. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current value.  The third parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] older_is_greater whether older nodes are greater than newer nodes
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_multimap__foldl_values(cgul_exception_t* cex,
                            cgul_multimap_t mm,
                            int older_is_greater,
                            cgul_multimap__fold_value_t f,
                            void* data);

/**
 * This method performs a right fold of the multimap <code>mm</code>
 * with the combining function <code>f</code>.  <code>f</code> is
 * called once for each value in the multimap starting at the back of
 * the multimap and iterating backward to the front of the
 * multimap. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current value.  The third parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] older_is_greater whether older nodes are greater than newer nodes
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_multimap__foldr_values(cgul_exception_t* cex,
                            cgul_multimap_t mm,
                            int older_is_greater,
                            cgul_multimap__fold_value_t f,
                            void* data);

/**
 * This method performs a left fold of the multimap <code>mm</code>
 * with the combining function <code>f</code>. <code>f</code> is
 * called once for each key/value pair in the multimap starting at the
 * front of the multimap and iterating forward to the end of the
 * multimap. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current key.  The third parameter passed into
 * <code>f</code> is the current value.  The fourth parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * The primary sort criterion is the user's <code>compare</code>
 * function, of course, but the <code>older_is_greater</code> boolean
 * is used as a secondary sort criterion.  When a key maps to many
 * values, the <code>older_is_greater</code> boolean determines how to
 * sort the older nodes relative to the newer nodes. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] older_is_greater whether older nodes are greater than newer nodes
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_multimap__foldl_pairs(cgul_exception_t* cex,
                           cgul_multimap_t mm,
                           int older_is_greater,
                           cgul_multimap__fold_pair_t f,
                           void* data);

/**
 * This method performs a right fold of the multimap <code>mm</code>
 * with the combining function <code>f</code>.  <code>f</code> is
 * called once for each key/value pair in the multimap starting at the
 * back of the multimap and iterating backward to the front of the
 * multimap. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current key.  The third parameter passed into
 * <code>f</code> is the current value.  The fourth parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * The primary sort criterion is the user's <code>compare</code>
 * function, of course, but the <code>older_is_greater</code> boolean
 * is used as a secondary sort criterion.  When a key maps to many
 * values, the <code>older_is_greater</code> boolean determines how to
 * sort the older nodes relative to the newer nodes. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] older_is_greater whether older nodes are greater than newer nodes
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_multimap__foldr_pairs(cgul_exception_t* cex,
                           cgul_multimap_t mm,
                           int older_is_greater,
                           cgul_multimap__fold_pair_t f,
                           void* data);

/**
 * Traverse all nodes passing each node to the function
 * <code>f</code>. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the multimap <code>mm</code>.  The third paramenter
 * passed into <code>f</code> is the node for this iteration.  The
 * fourth parameter passed into <code>f</code> is the client data
 * <code>data</code>. <p>
 *
 * <code>f</code> is provided with a safe context in which it can
 * remove the node that is passed into <code>f</code> by calling
 * <code>cgul_multimap__remove_node()</code>. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * the traversal to continue. <p>
 *
 * The primary sort criterion is the user's <code>compare</code>
 * function, of course, but the <code>older_is_greater</code> boolean
 * is used as a secondary sort criterion.  When a key maps to many
 * values, the <code>older_is_greater</code> boolean determines how to
 * sort the older nodes relative to the newer nodes. <p>
 *
 * NOTE: It is not strictly necessary that you use
 * <code>cgul_multimap__traverse()</code> or
 * <code>cgul_multimap__traverse_range()</code> in order to iterate over
 * the multimap elements.  In fact, I would recommend that you use
 * <code>cgul_multimap_node__get_next()</code> for most of your iteration
 * needs.  If you need to remove nodes though, you should probably use
 * this method.  If you look at this method's source however, you will
 * see that the <code>cgul_multimap_node</code> class provides all the
 * public methods you need to safely remove nodes while you are
 * iterating over the multimap, but you do have to be careful.
 *
 * @param[in,out] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] older_is_greater whether older nodes are greater than newer nodes
 * @param[in] f traversal callback function
 * @param[in] data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_multimap__traverse(cgul_exception_t* cex,
                        cgul_multimap_t mm,
                        int older_is_greater,
                        cgul_multimap__traverse_t f,
                        void* data);

/**
 * Traverse all nodes starting with <code>first</code> (inclusive) and
 * ending with <code>last</code> (inclusive) passing each node to the
 * function <code>f</code>.  If you know the first node, but are not
 * sure of the last node, just use <code>NULL</code> as the last node.
 * This will cause this method to iterate until it reaches the end of
 * the multimap.  You can than have <code>f</code> return 0 when it
 * determines that the last node has been reached. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the multimap <code>mm</code>.  The third paramenter
 * passed into <code>f</code> is the node for this iteration.  The
 * fourth parameter passed into <code>f</code> is the client data
 * <code>data</code>. <p>
 *
 * <code>f</code> is provided with a safe context in which it can
 * remove the node that is passed into <code>f</code> by calling
 * <code>cgul_multimap__remove_node()</code>. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * the traversal to continue. <p>
 *
 * If <code>first</code> is <code>NULL</code>, iteration starts at the
 * beginning of the multimap.  If <code>last</code> is <code>NULL</code>,
 * iteration stops at the end of the multimap. <p>
 *
 * The primary sort criterion is the user's <code>compare</code>
 * function, of course, but the <code>older_is_greater</code> boolean
 * is used as a secondary sort criterion.  When a key maps to many
 * values, the <code>older_is_greater</code> boolean determines how to
 * sort the older nodes relative to the newer nodes. <p>
 *
 * NOTE: It is not strictly necessary that you use
 * <code>cgul_multimap__traverse()</code> or
 * <code>cgul_multimap__traverse_range()</code> in order to iterate over
 * the multimap elements.  In fact, I would recommend that you use
 * <code>cgul_multimap_node__get_next()</code> for most of your
 * iteration needs.  If you need to remove nodes though, you should
 * probably use this method.  If you look at this method's source
 * however, you will see that the <code>cgul_multimap_node</code> class
 * provides all the public methods you need to safely remove nodes
 * while you are iterating over the multimap, but you do have to be
 * careful.
 *
 * @param[in,out] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] first first node in range
 * @param[in] last last node in range
 * @param[in] older_is_greater whether older nodes are greater than newer nodes
 * @param[in] f traversal callback function
 * @param[in] data client data passed to <code>f</code>
 * @see cgul_multimap__traverse
 */
extern CGUL_EXPORT void
cgul_multimap__traverse_range(cgul_exception_t* cex,
                              cgul_multimap_t mm,
                              cgul_multimap_node_t first,
                              cgul_multimap_node_t last,
                              int older_is_greater,
                              cgul_multimap__traverse_t f,
                              void* data);

CGUL_END_C

#endif  /*  CGUL_MULTIMAP_H  */
