#ifndef CGUL_MULTIMAP_CXX_H
#define CGUL_MULTIMAP_CXX_H

/**
 * C++ wrapper for <code>cgul_multimap</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_multimap</code>
 */

#include "cgul_cache_cxx.h"
#include "cgul_exception_cxx.h"
#include "cgul_multimap.h"
#include "cgul_multimap_node_cxx.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_multimap</code> objects.  The main purpose of this class
 * is to convert the C-style function calls and exception handling in
 * <code>cgul_multimap</code> into C++-style function calls and
 * exception handling.
 *
 * @brief C++ bindings for <code>cgul_multimap</code>
 * @see cgul_multimap_t
 */
class cgul_multimap_cxx {

public:

    /**
     * This typedef is the interface the client must define in order for
     * <code>cgul_multimap_cxx</code> to sort your keys as they are inserted.
     *
     * Arguably, this typedef should be an abstract base class, a pure
     * virtual method, or something that is the functional equivalent
     * of Java's "interface" keyword, but it doesn't really buy
     * anything except added complexity.
     */
    typedef cgul_multimap__compare_t compare_t;

    /**
     * This typedef is the interface for the combining function used
     * by the following methods: <p>
     *
     * <pre>
     *     <code>foldl_keys()</code>
     *     <code>foldr_keys()</code>
     * </pre>
     *
     * @param[in] key key
     * @param[in] data client data
     * @return whether to continue
     */
    typedef int (*fold_key_t)(const void* key,
                              void* data);

    /**
     * This typedef is the interface for the combining function used
     * by the following methods: <p>
     *
     * <pre>
     *     <code>foldl_values()</code>
     *     <code>foldr_values()</code>
     * </pre>
     *
     * @param[in] value value
     * @param[in] data client data
     * @return whether to continue
     */
    typedef int (*fold_value_t)(void* value,
                                void* data);

    /**
     * This typedef is the interface for the combining function used
     * by the following methods: <p>
     *
     * <pre>
     *     <code>foldl_pairs()</code>
     *     <code>foldr_pairs()</code>
     * </pre>
     *
     * @param[in] key key
     * @param[in] value value
     * @param[in] data client data
     * @return whether to continue
     */
    typedef int (*fold_pair_t)(const void* key,
                               void* value,
                               void* data);

    /**
     * This typedef is the interface for the callback function used by
     * the following methods: <p>
     *
     * <pre>
     *     <code>traverse()</code>
     *     <code>traverse_range()</code>
     * </pre>
     *
     * @param[in] mm multimap
     * @param[in] n node
     * @param[in] data client data
     * @return whether to continue
     */
    typedef int (*traverse_t)(cgul_multimap_cxx* mm,
                              cgul_multimap_node_cxx* n,
                              void* data);

    /**
     * Create a new <code>cgul_multimap</code> object.
     * <code>compare</code> is the comparison function that allows the
     * container to maintain its sorted order.  If memory cannot be
     * allocated, <code>NULL</code> is returned, and an exception is
     * thrown. <p>
     *
     * The client is responsible for deleting the object.  Also, the
     * client is responsible for freeing the key/value pairs only
     * after their nodes have been permanently removed from the
     * <code>cgul_multimap</code>.  The <code>traverse()</code> method
     * can be used to remove each node safely before deleting the
     * object.
     *
     * @param[in] compare comparison function
     * @see compare_t
     */
    cgul_multimap_cxx(compare_t compare)
    {
        cgul_exception_t local = NULL;
        obj = cgul_multimap__new(&local, compare);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_multimap_cxx</code> object by wrapping
     * an existing <code>cgul_multimap</code> object.
     *
     * @param[in] rhs right-hand side
     */
    cgul_multimap_cxx(cgul_multimap_t rhs)
    {
        obj = rhs;
    }

    /**
     * This method frees all internally allocated memory.  This does not
     * include the key/value pairs stored in the multimap.  The client is
     * responsible for freeing the key/value pairs when the client thinks
     * it is convenient and safe to do so; however, the client must
     * understand that freeing a key before removing its node invalidates
     * the data structure. <p>
     *
     * As a convenience, you may want to call <code>clear()</code>
     * before calling this method in order to properly put the keys
     * and values back on their respective <code>cgul_cache_cxx</code>
     * objects (if <code>cgul_cache_cxx</code> objects are being used)
     * before you delete the multimap.  If <code>cgul_cache_cxx</code> objects
     * are not being used, the client needs to arrange some other
     * mechanism to free the keys or values. <p>
     *
     * @see free_keys()
     * @see free_values()
     */
    virtual
    ~cgul_multimap_cxx()
    {
        if (obj) {
            cgul_multimap__delete(obj);
            obj = NULL;
        }
    }

    /**
     * This method calls <code>free()</code> on all the keys in the
     * multimap.  Because this is such a common operation, it is an
     * exception to the rule that cgul containers never free keys.
     * This method should only ever be called immediately before
     * calling <code>delete</code> because it otherwise invalidates
     * the multimap.
     */
    virtual void
    free_keys()
    {
        cgul_multimap__free_keys(obj);
    }

    /**
     * This method calls <code>free()</code> on all the values in the
     * multimap.  Because this is such a common operation, it is an
     * exception to the rule that cgul containers never free values.
     * This method should only ever be called immediately before
     * calling <code>delete</code> because it otherwise invalidates
     * the multimap.
     */
    virtual void
    free_values()
    {
        cgul_multimap__free_values(obj);
    }

    /**
     * Whether the multimap is empty.
     *
     * @return whether the multimap <code>mm</code> is empty
     */
    virtual int
    is_empty() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_multimap__is_empty(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Insert the <code>(key, value)</code> pair into the multimap.
     * If <code>node</code> is not <code>NULL</code>, it will hold the
     * newly inserted node upon completion.  The return value is the
     * count of how many values in the multimap are now associated
     * with <code>key</code>.  If an error occurs, an exception is
     * thrown. <p>
     *
     * It is important to understand that this class (like all
     * containers in the cgul library) do not take ownership of
     * <code>key</code> or <code>value</code>.  It also does not
     * attempt to make a copy of the thing pointed to by
     * <code>key</code> or <code>value</code>.  This gives the user
     * complete control over the lifetime of these objects and gives a
     * real performance boost in many cases; however, this means that
     * <code>key</code> must not be invalidated while it is still
     * being used by this class. <p>
     *
     * Most containers in the cgul library, map one key to one value.
     * This class is different.  It maps one key to many values.
     * Thus, this class really only needs one instance of the key even
     * though the insert method is called many times, and in fact,
     * this class only keeps the first instance of <code>key</code>
     * that is inserted.  Thus, it is important to check the value
     * returned by this method to determine when to free your
     * keys. <p>
     *
     * For example, if you use <code>malloc()</code> to dynamically
     * allocate a new <code>key</code> every time this method is called,
     * the key should probably be freed for all but the first insert:
     *
     * <pre>
     *
     *     if (mm->insert(key, value, &node) > 1) {
     *         free(key);
     *     }
     *
     * </pre>
     * 
     * @param[in] key key
     * @param[in] value value
     * @param[out] node newly inserted node
     * @return count of values now associated with <code>key</code>
     */
    virtual unsigned long int
    insert(void* key,
           void* value,
           cgul_multimap_node_cxx** node = NULL)
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_multimap__insert(&local,
                                   obj,
                                   key,
                                   value,
                                   (cgul_multimap_node_t*)node);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method works the same as <code>insert()</code>
     * except if <code>hint</code> is not <code>NULL</code> it calls the
     * comparison function (passed into the constructor)
     * to determine if <code>key</code> is equal to the key for
     * <code>hint</code>.  If so, the insertion can occur in constant time
     * instead of logarithmic time because <code>hint</code> already holds
     * a back reference to the underlying node for <code>key</code>. <p>
     *
     * If <code>hint</code> is <code>NULL</code> or <code>key</code> does
     * not equal the key for <code>hint</code>, this method simply
     * forwards the request to <code>insert()</code>, and
     * the insertion occurs in logarithmic time. <p>
     * 
     * @param[in] key key
     * @param[in] value value
     * @param[in] hint hint
     * @param[out] node newly inserted node
     * @return count of values now associated with <code>key</code>
     */
    virtual unsigned long int
    insert_with_hint(void* key,
                     void* value,
                     cgul_multimap_node_cxx* hint,
                     cgul_multimap_node_cxx** node = NULL)
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_multimap__insert_with_hint(&local,
                                             obj,
                                             key,
                                             value,
                                             (cgul_multimap_node_t)hint,
                                             (cgul_multimap_node_t*)node);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method finds the smallest node associated with
     * <code>key</code> in the multimap and returns it in
     * <code>node</code>.  The return value of this method is the total
     * number of values associated with <code>key</code>.  If
     * <code>key</code> cannot be found, <code>node</code> will be set to
     * <code>NULL</code>, and <code>0</code> will be returned. <p>
     *
     * The caller can use the return value to iterate over all the values
     * associated with <code>key</code>.  For example, the following shows
     * how to iterate over all the values associated with <code>key</code>
     * in the order in which they were inserted: <p>
     *
     * <pre>
     *
     *     count = mm->find(key, 0, &node);
     *     for (i = 0 ; i < count ; ++i) {
     *         ...
     *         node = node->get_next(0);
     *     }
     *
     * </pre>
     *
     * The primary sort criterion is the user's <code>compare</code>
     * function, of course, but the <code>older_is_greater</code> boolean
     * is used as a secondary sort criterion.  When a key maps to many
     * values, the <code>older_is_greater</code> boolean determines how to
     * sort the older nodes relative to the newer nodes. <p>
     *
     * @param[in] key key
     * @param[in] older_is_greater whether older nodes are greater
     * than newer nodes
     * @param[out] node smallest node associated with <code>key</code>
     * @return count of values associated with <code>key</code>
     */
    virtual unsigned long int
    find(void* key,
         int older_is_greater,
         cgul_multimap_node_cxx** node)
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_multimap__find(&local,
                                 obj,
                                 key,
                                 older_is_greater,
                                 (cgul_multimap_node_t*)node);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Find the node having the largest key less than or equal to the
     * search key <code>search_key</code> and return it in
     * <code>*node</code>.  If such a key does not exist,
     * <code>NULL</code> is returned instead.  This method returns the
     * number of values that directly map to the floor. <p>
     *
     * The iterator returned in <code>*node</code> is for the floor key's
     * smallest value as determined by <code>older_is_greater</code> such
     * that forward iteration includes all the values for the floor
     * node. <p>
     *
     * @param[in] search_key search key
     * @param[in] older_is_greater whether older nodes are greater
     * than newer nodes
     * @param[out] node floor or <code>NULL</code>
     * @return count of values associated with the floor
     */
    virtual unsigned long int
    find_floor(void* search_key,
               int older_is_greater,
               cgul_multimap_node_cxx** node)
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_multimap__find_floor(&local,
                                       obj,
                                       search_key,
                                       older_is_greater,
                                       (cgul_multimap_node_t*)node);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Find the node having the smallest key greater than or equal to
     * the search key <code>search_key</code> and return it in
     * <code>*node</code>.  If such a key does not exist,
     * <code>NULL</code> is returned instead.  This method returns the
     * number of values that directly map to the ceiling.
     *
     * The iterator returned in <code>*node</code> is for the ceiling
     * key's largest value as determined by <code>older_is_greater</code>
     * such that reverse iteration includes all the values for the ceiling
     * node. <p>
     *
     * @param[in] search_key search key
     * @param[in] older_is_greater whether older nodes are greater
     * than newer nodes
     * @param[out] node ceiling or <code>NULL</code>
     * @return count of values associated with the ceiling
     */
    virtual unsigned long int
    find_ceiling(void* search_key,
                 int older_is_greater,
                 cgul_multimap_node_cxx** node)
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_multimap__find_ceiling(&local,
                                         obj,
                                         search_key,
                                         older_is_greater,
                                         (cgul_multimap_node_t*)node);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method removes <code>key</code> and all the values associated
     * with <code>key</code> from the multimap and returns
     * <code>1</code> on success.  On failure, this method returns
     * <code>0</code> which indicates that <code>key</code> could not be
     * found in the multimap. <p>
     *
     * Strictly as a convenience, this method is an exception to the rule
     * that cgul containers never free keys or values.  If you pass in
     * <code>keys_cache</code> or <code>values_cache</code> instances that
     * are not <code>NULL</code>, the keys or values will be put back on
     * their respective caches. <p>
     *
     * If you need to remove just one value associated with
     * <code>key</code> instead of all the values associated with
     * <code>key</code>, use <code>remove_node()</code>
     * instead. <p>
     *
     * @param[in] key key
     * @param[in] keys_cache keys cache
     * @param[in] values_cache values cache
     * @return whether the nodes associated with <code>key</code> were removed
     * @see cgul_cache_cxx::get_freer()
     */
    virtual int
    remove(void* key,
           cgul_cache_cxx* keys_cache = NULL,
           cgul_cache_cxx* values_cache = NULL)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_multimap__remove(
                 &local,
                 obj,
                 key,
                 (keys_cache ? keys_cache->get_obj() : NULL),
                 (values_cache ? values_cache->get_obj() : NULL));
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Remove <code>node</code> from the multimap.  The
     * key and value pointers stored in the node that is to be removed
     * will be returned in <code>key_out</code> and <code>value_out</code>
     * if you pass in pointers that are not <code>NULL</code>, but the
     * value for the key that is returned will be <code>NULL</code> if the
     * multimap is still using the key.  This method returns the number of
     * values still associated with the same key. <p>
     *
     * Because the key in <code>node</code> might map to more than one
     * value, the caller should never invalidate the key until after this
     * method returns a count of zero or, equivalently, a value for
     * <code>key</code> that is not <code>NULL</code>: <p>
     *
     * <pre>
     *
     *     mm->remove_node(node, &key, &value);
     *     if (key) {
     *         free(key);
     *     }
     *     free(value);
     *
     * </pre>
     *
     * It is almost always a mistake to naively call this method while
     * iterating over the multimap.  The problem is that calling this
     * method invalidates the node making it impossible to call
     * <code>node->get_next()</code> afterward.  The solution is
     * simple.  Just call <code>node->get_next()</code> before calling
     * this method:
     *
     * <pre>
     *
     *     count = mm->find(key, 1, &curr);
     *     if (curr) {
     *         for (i = 0 ; i < count ; ++i, curr = next) {
     *             next = curr->get_next(1);
     *             mm->remove_node(curr, &key_out, &value_out);
     *             if (key_out) {
     *                 free(key_out);
     *             }
     *             free(value_out);
     *         }
     *     }
     *
     * </pre>
     *
     * Alternatively, you can use <code>remove()</code>,
     * <code>remove_range()</code>, <code>traverse()</code>, or
     * <code>traverse_range()</code>. <p>
     *
     * If you use one of the traversal methods, it is safe to call this
     * method naively.  For example, if your multimap stores
     * <code>cgul_string_cxx*</code> objects for both its keys and values,
     * another way to remove all the values associated with a key as
     * follows: <p>
     *
     * <pre>
     *
     *     static int
     *     remove_node_cb(cgul_multimap_t mm,
     *                    cgul_multimap_node_t node,
     *                    void* data)
     *     {
     *         cgul_string_cxx* key = NULL;
     *         cgul_string_cxx* value = NULL;
     *         mm->remove_node(node, (void**)&key, (void**)&value);
     *         if (key) {
     *             delete key;
     *         }
     *         delete value;
     *         return 1;
     *     }
     *
     *     cgul_multimap_node_cxx* first = NULL;
     *     cgul_multimap_node_cxx* last = NULL;
     *     mm->find(key, 1, &first);
     *     mm->find(key, 0, &last);
     *     if (first && last) {
     *         mm->traverse_range(first, last, 1, &remove_node_cb, NULL);
     *     }
     *
     * </pre>
     *
     * @param[in] node node
     * @param[out] key_out pointer for the key stored in the node
     * @param[out] value_out pointer for the value stored in the node
     * @return count of values now associated with <code>key</code>
     */
    virtual unsigned long int
    remove_node(cgul_multimap_node_cxx* node,
                void** key_out = NULL,
                void** value_out = NULL)
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_multimap__remove_node(&local,
                                        obj,
                                        (cgul_multimap_node_t)node,
                                        key_out,
                                        value_out);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method removes nodes in the range <code>first</code>
     * (inclusive) to <code>last</code> (inclusive).  Strictly as a
     * convenience, this method is an exception to the rule that cgul
     * containers never free keys or values.  If you pass in
     * <code>keys_cache</code> or <code>values_cache</code> instances that
     * are not <code>NULL</code>, the keys or values will be put back on
     * their respective caches. <p>
     *
     * The primary sort criterion is the user's <code>compare</code>
     * function, of course, but the <code>older_is_greater</code> boolean
     * is used as a secondary sort criterion.  When a key maps to many
     * values, the <code>older_is_greater</code> boolean determines how to
     * sort the older nodes relative to the newer nodes. <p>
     *
     * @param[in] first first node in range
     * @param[in] last last node in range
     * @param[in] older_is_greater whether older nodes are greater
     * than newer nodes
     * @param[in] keys_cache keys cache
     * @param[in] values_cache values cache
     * @return whether any nodes were removed.
     * @see cgul_cache_cxx::get_freer()
     */
    virtual int
    remove_range(cgul_multimap_node_cxx* first,
                 cgul_multimap_node_cxx* last,
                 int older_is_greater,
                 cgul_cache_cxx* keys_cache = NULL,
                 cgul_cache_cxx* values_cache = NULL)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_multimap__remove_range(
                 &local,
                 obj,
                 (cgul_multimap_node_t)first,
                 (cgul_multimap_node_t)last,
                 older_is_greater,
                 (keys_cache ? keys_cache->get_obj() : NULL),
                 (values_cache ? values_cache->get_obj() : NULL));
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method removes the front node.  The key and value pointers
     * stored in the node that is to be removed will be returned in
     * <code>key_out</code> and <code>value_out</code> if you pass in
     * pointers that are not <code>NULL</code>.  Note however that because
     * keys map to multiple values, the key for this node may still be in
     * use by another node.  If this is the case, <code>NULL</code> is
     * returned in <code>key_out</code>.  An example of how to use this
     * method follows: <p>
     *
     * <pre>
     *
     *     mm->remove_front(0, &key_out, &value_out);
     *     if (key_out) {
     *         free(key_out);
     *     }
     *     free(value_out);
     *
     * </pre>
     *
     * The primary sort criterion is the user's <code>compare</code>
     * function, of course, but the <code>older_is_greater</code> boolean
     * is used as a secondary sort criterion.  When a key maps to many
     * values, the <code>older_is_greater</code> boolean determines how to
     * sort the older nodes relative to the newer nodes.  See
     * <code>get_front()</code> for more details. <p>
     *
     * @param[in] older_is_greater whether older nodes are greater
     * than newer nodes
     * @param[out] key_out pointer for the key stored in the node
     * @param[out] value_out pointer for the value stored in the node
     * @return whether the node was removed
     */
    virtual int
    remove_front(int older_is_greater,
                 void** key_out,
                 void** value_out)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_multimap__remove_front(&local,
                                         obj,
                                         older_is_greater,
                                         key_out,
                                         value_out);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method removes the back node.  The key and value pointers
     * stored in the node that is to be removed will be returned in
     * <code>key_out</code> and <code>value_out</code> if you pass in
     * pointers that are not <code>NULL</code>.  Note however that because
     * keys map to multiple values, the key for this node may still be in
     * use by another node.  If this is the case, <code>NULL</code> is
     * returned in <code>key_out</code>.  An example of how to use this
     * method follows: <p>
     *
     * <pre>
     *
     *     mm->remove_back(1, &key_out, &value_out);
     *     if (key_out) {
     *         free(key_out);
     *     }
     *     free(value_out);
     *
     * </pre>
     *
     * The primary sort criterion is the user's <code>compare</code>
     * function, of course, but the <code>older_is_greater</code> boolean
     * is used as a secondary sort criterion.  When a key maps to many
     * values, the <code>older_is_greater</code> boolean determines how to
     * sort the older nodes relative to the newer nodes.  See
     * <code>get_back()</code> for more details. <p>
     *
     * @param[in] older_is_greater whether older nodes are greater
     * than newer nodes
     * @param[out] key_out pointer for the key stored in the node
     * @param[out] value_out pointer for the value stored in the node
     * @return whether the node was removed
     */
    virtual int
    remove_back(int older_is_greater,
                void** key_out,
                void** value_out)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_multimap__remove_back(&local,
                                        obj,
                                        older_is_greater,
                                        key_out,
                                        value_out);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the front node according to sort order.  This operation is
     * O(1).  If the multimap is empty, <code>NULL</code> is returned.
     * This class does not have to search the multimap in order to find
     * the front node because it indirectly keeps a pointer to the front
     * node up to date. <p>
     *
     * The primary sort criterion is the user's <code>compare</code>
     * function, of course, but the <code>older_is_greater</code> boolean
     * is used as a secondary sort criterion.  When a key maps to many
     * values, the <code>older_is_greater</code> boolean determines how to
     * sort the older nodes relative to the newer nodes. <p>
     *
     * The way the <code>older_is_greater</code> boolean works in this
     * implementation is that each key maps to a linked list of values.
     * The <code>compare</code> function only determines the relative
     * order of the keys.  To order the values, this implementation always
     * pushes newer values onto the front of the list.  Thus, the newest
     * value is always at the front of the list, and the oldest value is
     * always at the back of the list.  If <code>older_is_greater</code>
     * is true, the sort order is from front to back.  If
     * <code>older_is_greater</code> is false, the sort order is from back
     * to front. <p>
     *
     * For example, if you use this class to implement a priority queue
     * and want to get the node with the lowest priority that has been
     * waiting the longest, you would call this method with
     * <code>older_is_greater</code> set to <code>0</code>. <p>
     *
     * @param[in] older_is_greater whether older nodes are greater
     * than newer nodes
     * @return front node
     * @see get_next()
     * @see get_prev()
     */
    virtual cgul_multimap_node_cxx*
    get_front(int older_is_greater) const
    {
        cgul_multimap_node_t rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_multimap__get_front(&local, obj, older_is_greater);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (cgul_multimap_node_cxx*)rv;
    }

    /**
     * Return the back node according to sort order.  This operation is
     * O(1).  If the multimap is empty, <code>NULL</code> is returned.
     * This class does not have to search the multimap in order to find
     * the back node because it indirectly keeps a pointer to the back
     * node up to date.
     *
     * The primary sort criterion is the user's <code>compare</code>
     * function, of course, but the <code>older_is_greater</code> boolean
     * is used as a secondary sort criterion.  When a key maps to many
     * values, the <code>older_is_greater</code> boolean determines how to
     * sort the older nodes relative to the newer nodes. <p>
     *
     * The way the <code>older_is_greater</code> boolean works in this
     * implementation is that each key maps to a linked list of values.
     * The <code>compare</code> function only determines the relative
     * order of the keys.  To order the values, this implementation always
     * pushes newer values onto the front of the list.  Thus, the newest
     * value is always at the front of the list, and the oldest value is
     * always at the back of the list.  If <code>older_is_greater</code>
     * is true, the sort order is from front to back.  If
     * <code>older_is_greater</code> is false, the sort order is from back
     * to front. <p>
     *
     * For example, if you use this class to implement a priority queue
     * and want to get the node with the highest priority that has been
     * waiting the longest, you would call this method with
     * <code>older_is_greater</code> set to <code>1</code>. <p>
     *
     * @param[in] older_is_greater whether older nodes are greater
     * than newer nodes
     * @return back node
     * @see get_next()
     * @see get_prev()
     */
    virtual cgul_multimap_node_cxx*
    get_back(int older_is_greater) const
    {
        cgul_multimap_node_t rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_multimap__get_back(&local, obj, older_is_greater);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (cgul_multimap_node_cxx*)rv;
    }

    /**
     * Return the oldest node according to chronological order (i.e., the
     * order in which the nodes are inserted).  This operation is O(1).
     * If the multimap is empty, <code>NULL</code> is returned.  This
     * class does not have to search the multimap in order to find the
     * oldest node because it efficiently keeps a direct pointer to the
     * oldest node up to date. <p>
     *
     * The following example shows how to iterate over the entire multimap
     * in chronological order:
     *
     * <pre>
     *     cgul_multimap_node_cxx* n = mm->get_oldest();
     *     for ( ; n ; n = n->get_younger()) {
     *         ...
     *     }
     * </pre>
     *
     * @return oldest node
     * @see get_younger()
     */
    virtual cgul_multimap_node_cxx*
    get_oldest() const
    {
        cgul_multimap_node_t rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_multimap__get_oldest(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (cgul_multimap_node_cxx*)rv;
    }

    /**
     * Set the oldest node in the multimap to be <code>n</code>.  Calling
     * this method has the potential to confuse iterators and should be
     * handled with roughly the same level of caution as calling
     * <code>remove_node()</code>. <p>
     *
     * This method could be used, for example, if your code expires the
     * oldest node in the multimap, and you want to force <code>n</code> to be
     * the next node to expire. <p>
     *
     * @param[in] n node
     * @see set_youngest()
     */
    virtual void
    set_oldest(cgul_multimap_node_cxx* n)
    {
        cgul_exception_t local = NULL;
        cgul_multimap__set_oldest(&local, obj, (cgul_multimap_node_t)n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the youngest node according to chronological order (i.e.,
     * the order in which the nodes are inserted).  This operation is
     * O(1).  If the multimap is empty, <code>NULL</code> is returned.
     * This class does not have to search the multimap in order to find
     * the youngest node because it efficiently keeps a direct pointer to
     * the youngest node up to date. <p>
     *
     * The following example shows how to iterate over the entire multimap
     * in reverse chronological order:
     *
     * <pre>
     *     cgul_multimap_node_cxx* n = mm->get_youngest();
     *     for ( ; n ; n = n->get_older()) {
     *         ...
     *     }
     * </pre>
     *
     * @return youngest node
     * @see get_older()
     */
    virtual cgul_multimap_node_cxx*
    get_youngest() const
    {
        cgul_multimap_node_t rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_multimap__get_youngest(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (cgul_multimap_node_cxx*)rv;
    }

    /**
     * Set the youngest node in the multimap to be <code>n</code>.
     * Calling this method has the potential to confuse iterators and
     * should be handled with roughly the same level of caution as calling
     * <code>remove_node()</code>. <p>
     *
     * This method could be used, for example, if your code expires the
     * least-recently used (LRU) node.  By calling
     * <code>set_youngest()</code> each time a node is used,
     * the least-recently used node will be the oldest node in the
     * multimap. <p>
     *
     * @param[in] n node
     * @see cgul_multimap__set_oldest()
     */
    virtual void
    set_youngest(cgul_multimap_node_cxx* n)
    {
        cgul_exception_t local = NULL;
        cgul_multimap__set_youngest(&local, obj, (cgul_multimap_node_t)n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method clears the multimap by removing each node individually.
     * Strictly as a convenience, this method is an exception to the rule
     * that cgul containers never free keys or values.  If you pass in
     * <code>keys_cache</code> or <code>values_cache</code> instances that
     * are not <code>NULL</code>, the keys or values will be put back on
     * their respective caches. <p>
     *
     * Another easy way to clear your multimap is to just keep removing nodes
     * from the front of the multimap freeing the key/value pairs as you go:
     *
     * <pre>
     *     void* key = NULL;
     *     void* value = NULL;
     *     while (mm->remove_front(0, &key, &value)) {
     *         if (key) {
     *             free(key);
     *         }
     *         free(value);
     *     }
     * </pre>
     *
     * @param[in] keys_cache keys cache
     * @param[in] values_cache values cache
     * @see remove_range()
     * @see cgul_cache_cxx::get_freer()
     */
    virtual void
    clear(cgul_cache_cxx* keys_cache = NULL,
          cgul_cache_cxx* values_cache= NULL)
    {
        cgul_exception_t local = NULL;
        cgul_multimap__clear(
            &local,
            obj,
            (keys_cache ? keys_cache->get_obj() : NULL),
            (values_cache ? values_cache->get_obj() : NULL));
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the total number of values stored in the multimap.
     *
     * @return size of the multimap
     */
    virtual unsigned long int
    get_size() const
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_multimap__get_size(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Swap the underlying data for this object and <code>rhs</code>.
     * For large multimaps, this should be much faster than trying to
     * do the same thing using removes and inserts.
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    swap(cgul_multimap_cxx& rhs)
    {
        cgul_exception_t local = NULL;
        cgul_multimap__swap(&local, this->obj, rhs.obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method performs a left fold of the multimap with the
     * combining function <code>f</code>.  <code>f</code> is called
     * once for each key in the multimap starting at the front of the
     * multimap and iterating forward to the end of the multimap. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * key.  The second parameter passed into <code>f</code> is the
     * client data which is where the result of the fold should be
     * accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldl_keys(fold_key_t f,
               void* data)
    {
        const void* key = NULL;
        cgul_rbtree_t t = NULL;
        cgul_rbtree_node_t curr = NULL;
        cgul_rbtree_node_t next = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_multimap__foldl_keys()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Get the underlying red-black tree that holds the keys. */
        t = cgul_multimap__get_tree(&local, obj);

        /* Iterate over the nodes. */
        curr = cgul_rbtree__get_front(&local, t);
        for ( ; curr ; curr = next) {
            /* Get next now in case f deletes curr. */
            next = cgul_rbtree_node__get_next(&local, curr);
            /* Call back. */
            key = cgul_rbtree_node__get_key(&local, curr);
            if (!(*f)(key, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a right fold of the multimap with the
     * combining function <code>f</code>.  <code>f</code> is called
     * once for each key in the multimap starting at the back of the
     * multimap and iterating backward to the front of the
     * multimap. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * key.  The second parameter passed into <code>f</code> is the
     * client data which is where the result of the fold should be
     * accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldr_keys(fold_key_t f,
               void* data)
    {
        const void* key = NULL;
        cgul_rbtree_t t = NULL;
        cgul_rbtree_node_t curr = NULL;
        cgul_rbtree_node_t prev = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_multimap__foldl_keys()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Get the underlying red-black tree that holds the keys. */
        t = cgul_multimap__get_tree(&local, obj);

        /* Iterate over the nodes. */
        curr = cgul_rbtree__get_back(&local, t);
        for ( ; curr ; curr = prev) {
            /* Get prev now in case f deletes curr. */
            prev = cgul_rbtree_node__get_prev(&local, curr);
            /* Call back. */
            key = cgul_rbtree_node__get_key(&local, curr);
            if (!(*f)(key, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a left fold of the multimap with the
     * combining function <code>f</code>.  <code>f</code> is called
     * once for each value in the multimap starting at the front of
     * the multimap and iterating forward to the end of the
     * multimap. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * value.  The second parameter passed into <code>f</code> is the
     * client data which is where the result of the fold should be
     * accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] older_is_greater whether older nodes are greater
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldl_values(int older_is_greater,
                 fold_value_t f,
                 void* data)
    {
        void* value = NULL;
        cgul_multimap_node_t curr = NULL;
        cgul_multimap_node_t next = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_multimap__foldl_values()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_multimap__get_front(&local, obj, older_is_greater);
        for ( ; curr ; curr = next) {
            /* Get next now in case f deletes curr. */
            next = cgul_multimap_node__get_next(&local, curr, older_is_greater);
            /* Call back. */
            value = cgul_multimap_node__get_value(&local, curr);
            if (!(*f)(value, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a right fold of the multimap with the
     * combining function <code>f</code>.  <code>f</code> is called
     * once for each value in the multimap starting at the back of the
     * multimap and iterating backward to the front of the
     * multimap. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * value.  The second parameter passed into <code>f</code> is the
     * client data which is where the result of the fold should be
     * accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] older_is_greater whether older nodes are greater
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldr_values(int older_is_greater,
                 fold_value_t f,
                 void* data)
    {
        void* value = NULL;
        cgul_multimap_node_t curr = NULL;
        cgul_multimap_node_t prev = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_multimap__foldr_values()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_multimap__get_back(&local, obj, older_is_greater);
        for ( ; curr ; curr = prev) {
            /* Get prev now in case f deletes curr. */
            prev = cgul_multimap_node__get_prev(&local, curr, older_is_greater);
            /* Call back. */
            value = cgul_multimap_node__get_value(&local, curr);
            if (!(*f)(value, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a left fold of the multimap with the
     * combining function <code>f</code>.  <code>f</code> is called
     * once for each key/value pair in the multimap starting at the
     * front of the multimap and iterating forward to the end of the
     * multimap. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * key.  The second parameter passed into <code>f</code> is the
     * current value.  The third parameter passed into <code>f</code>
     * is the client data which is where the result of the fold should
     * be accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] older_is_greater whether older nodes are greater
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldl_pairs(int older_is_greater,
                fold_pair_t f,
                void* data)
    {
        const void* key = NULL;
        void* value = NULL;
        cgul_multimap_node_t curr = NULL;
        cgul_multimap_node_t next = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_multimap__foldl_pairs()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_multimap__get_front(&local, obj, older_is_greater);
        for ( ; curr ; curr = next) {
            /* Get next now in case f deletes curr. */
            next = cgul_multimap_node__get_next(&local, curr, older_is_greater);
            /* Call back. */
            key = cgul_multimap_node__get_key(&local, curr);
            value = cgul_multimap_node__get_value(&local, curr);
            if (!(*f)(key, value, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a right fold of the multimap with the
     * combining function <code>f</code>.  <code>f</code> is called
     * once for each key/value pair in the multimap starting at the
     * back of the multimap and iterating backward to the front of the
     * multimap. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * key.  The second parameter passed into <code>f</code> is the
     * current value.  The third parameter passed into <code>f</code>
     * is the client data which is where the result of the fold should
     * be accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] older_is_greater whether older nodes are greater
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldr_pairs(int older_is_greater,
                fold_pair_t f,
                void* data)
    {
        const void* key = NULL;
        void* value = NULL;
        cgul_multimap_node_t curr = NULL;
        cgul_multimap_node_t prev = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_multimap__foldr_pairs()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_multimap__get_back(&local, obj, older_is_greater);
        for ( ; curr ; curr = prev) {
            /* Get prev now in case f deletes curr. */
            prev = cgul_multimap_node__get_prev(&local, curr, older_is_greater);
            /* Call back. */
            key = cgul_multimap_node__get_key(&local, curr);
            value = cgul_multimap_node__get_value(&local, curr);
            if (!(*f)(key, value, data)) {
                break;
            }
        }
    }

    /**
     * Traverse all nodes passing each node to the function
     * <code>f</code>. <p>
     *
     * The first parameter passed into <code>f</code> is the multimap
     * <code>this</code>.  The second paramenter passed into
     * <code>f</code> is the node for this iteration.  The third
     * parameter passed into <code>f</code> is the client data
     * <code>data</code>. <p>
     *
     * <code>f</code> is provided with a safe context in which it can
     * remove the node that is passed into <code>f</code> by calling
     * <code>remove_node()</code>. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * the traversal to continue. <p>
     *
     * The primary sort criterion is the user's <code>compare</code>
     * function, of course, but the <code>older_is_greater</code> boolean
     * is used as a secondary sort criterion.  When a key maps to many
     * values, the <code>older_is_greater</code> boolean determines how to
     * sort the older nodes relative to the newer nodes. <p>
     *
     * NOTE: It is not strictly necessary that you use
     * <code>traverse()</code> or <code>traverse_range()</code> in
     * order to iterate over the multimap elements.  In fact, I would
     * recommend that you use
     * <code>cgul_multimap_node_cxx::get_next()</code> for most of your
     * iteration needs.  If you need to remove nodes though, you
     * should probably use this method.  If you look at this method's
     * source however, you will see that the
     * <code>cgul_multimap_node_cxx</code> class provides all the public
     * methods you need to safely remove nodes while you are iterating
     * over the multimap, but you do have to be careful.
     *
     * @param[in] older_is_greater whether older nodes are greater
     * @param[in] f traversal callback function
     * @param[in] data client data passed to <code>f</code>
     * @see traverse_range()
     */
    virtual void
    traverse(int older_is_greater,
             traverse_t f,
             void* data)
    {
        traverse_range(NULL, NULL, older_is_greater, f, data);
    }

    /**
     * Traverse all nodes starting with <code>first</code> (inclusive) and
     * ending with <code>last</code> (inclusive) passing each node to the
     * function <code>f</code>.  If you know the first node, but are not
     * sure of the last node, just use <code>NULL</code> as the last node.
     * This will cause this method to iterate until it reaches the end of
     * the multimap.  You can than have <code>f</code> return 0 when it
     * determines that the last node has been reached. <p>
     *
     * The first parameter passed into <code>f</code> is the multimap
     * <code>this</code>.  The second paramenter passed into
     * <code>f</code> is the node for this iteration.  The third
     * parameter passed into <code>f</code> is the client data
     * <code>data</code>. <p>
     *
     * <code>f</code> is provided with a safe context in which it can
     * remove the node that is passed into <code>f</code> by calling
     * <code>remove_node()</code>. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * the traversal to continue. <p>
     *
     * If <code>first</code> is <code>NULL</code>, iteration starts at the
     * beginning of the multimap.  If <code>last</code> is <code>NULL</code>,
     * iteration stops at the end of the multimap. <p>
     *
     * The primary sort criterion is the user's <code>compare</code>
     * function, of course, but the <code>older_is_greater</code> boolean
     * is used as a secondary sort criterion.  When a key maps to many
     * values, the <code>older_is_greater</code> boolean determines how to
     * sort the older nodes relative to the newer nodes. <p>
     *
     * NOTE: It is not strictly necessary that you use
     * <code>traverse()</code> or <code>traverse_range()</code> in
     * order to iterate over the multimap elements.  In fact, I would
     * recommend that you use
     * <code>cgul_multimap_node_cxx::get_next()</code> for most of your
     * iteration needs.  If you need to remove nodes though, you
     * should probably use this method.  If you look at this method's
     * source however, you will see that the
     * <code>cgul_multimap_node_cxx</code> class provides all the public
     * methods you need to safely remove nodes while you are iterating
     * over the multimap, but you do have to be careful.
     *
     * @param[in] first first node in range
     * @param[in] last last node in range
     * @param[in] older_is_greater whether older nodes are greater
     * @param[in] f traversal callback function
     * @param[in] data client data passed to <code>f</code>
     * @see traverse()
     */
    virtual void
    traverse_range(cgul_multimap_node_cxx* first,
                   cgul_multimap_node_cxx* last,
                   int older_is_greater,
                   traverse_t f,
                   void* data)
    {
        cgul_multimap_node_cxx* curr = NULL;
        cgul_multimap_node_cxx* next = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_multimap__traverse_range()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */
        
        if (first == NULL) {
            first = (cgul_multimap_node_cxx*)
                        cgul_multimap__get_front(&local, obj, older_is_greater);
        }
        if (last == NULL) {
            last = (cgul_multimap_node_cxx*)
                       cgul_multimap__get_back(&local, obj, older_is_greater);
        }

        /* Iterate over the nodes. */
        for (curr = first ; curr ; curr = next) {
            /* Get next now in case f deletes curr. */
            next = (cgul_multimap_node_cxx*)
                       cgul_multimap_node__get_next(&local,
                                                    (cgul_multimap_node_t)curr,
                                                    older_is_greater);
            /* Call back. */
            if (!(*f)(this, curr, data)) {
                break;
            }
            /* We are done if we just processed the last node. */
            if (curr == last) {
                break;
            }
        }
    }

    /**
     * Get the underlying <code>cgul_multimap</code> object.
     *
     * @return underlying object
     */
    virtual cgul_multimap_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_multimap</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_multimap__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_multimap_t
    take_obj()
    {
        cgul_multimap_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_multimap_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_multimap__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_multimap_cxx();
    cgul_multimap_cxx(const cgul_multimap_cxx&);
    cgul_multimap_cxx& operator=(const cgul_multimap_cxx&);

    /** Underlying <code>cgul_multimap</code> object. */
    cgul_multimap_t obj;

};

#endif  /*  CGUL_MULTIMAP_CXX_H  */
