#ifndef CGUL_MULTIPLY64_H
#define CGUL_MULTIPLY64_H

/**
 * Portable multiplication that can produce 64-bit products.
 *
 * @file
 * @author Paul Serice
 * @brief portable multiplication that can produce 64-bit products
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/**
 * This function multiplies two 32-bit unsigned integers
 * <code>x</code> and <code>y</code> to produce a 64-bit unsigned
 * product using only 32-bit arithmetic.  The high 32-bits of the
 * 64-bit product will be stored in <code>*high</code> if
 * <code>high</code> is not <code>NULL</code>, and the low 32-bits of
 * the 64-bit product will be stored in <code>*low</code> if
 * <code>low</code> is not <code>NULL</code>. <p>
 *
 * The following <code>printf()</code> format specification can be
 * used to convert <code>*high</code> and <code>*low</code> into a
 * proper hexadecimal string because it accounts for the significance
 * of the leading zeros in <code>*low</code>: <p>
 *
 * <pre>
 *     printf("%#0x%08x", *high, *low);
 * </pre>
 *
 * This function takes about 2.5 times as long to run as native 64-bit
 * multiplication.  Thus, it should only be used if your compiler does
 * not support a 64-bit type.  To conditionally use this code, include
 * the "cgul_int.h" header and test if <code>CGUL_UINT64</code> is not
 * defined. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] x first 32-bit multiplicand
 * @param[in] y second 32-bit multiplicand
 * @param[out] high high 32-bits of the product of x and y
 * @param[out] low low 32-bits of the product of x and y
 */
extern CGUL_EXPORT void
cgul_multiply64(cgul_exception_t* cex,
                unsigned int x,
                unsigned int y,
                unsigned int* high,
                unsigned int* low);

CGUL_END_C

#endif  /*  CGUL_MULTIPLY64_H  */
