#ifndef CGUL_MULTIPLY64_CXX_H
#define CGUL_MULTIPLY64_CXX_H

/**
 * C++ wrapper for <code>cgul_multiply64</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_multiply64</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_multiply64.h"

/**
 * This namespace provides the C++ bindings for
 * <code>cgul_multiply64</code>.  The main purpose of this class is to
 * convert the C-style function calls and exception handling in
 * <code>cgul_multiply64</code> into C++-style function calls and
 * exception handling.
 *
 * @brief C++ bindings for <code>cgul_multiply64</code>
 */
namespace cgul_multiply64_cxx {

    /**
     * This function multiplies two 32-bit unsigned integers
     * <code>x</code> and <code>y</code> to produce a 64-bit unsigned
     * product using only 32-bit arithmetic.  The high 32-bits of the
     * 64-bit product will be stored in <code>*high</code> if
     * <code>high</code> is not <code>NULL</code>, and the low 32-bits of
     * the 64-bit product will be stored in <code>*low</code> if
     * <code>low</code> is not <code>NULL</code>. <p>
     *
     * The following <code>printf()</code> format specification can be
     * used to convert <code>*high</code> and <code>*low</code> into a
     * proper hexadecimal string because it accounts for the significance
     * of the leading zeros in <code>*low</code>: <p>
     *
     * <pre>
     *     printf("%#0x%08x", *high, *low);
     * </pre>
     *
     * This function takes about 2.5 times as long to run as native 64-bit
     * multiplication.  Thus, it should only be used if your compiler does
     * not support a 64-bit type.  To conditionally use this code, include
     * the "cgul_int.h" header and test if <code>CGUL_UINT64</code> is not
     * defined. <p>
     *
     * @param[in] x first 32-bit multiplicand
     * @param[in] y second 32-bit multiplicand
     * @param[out] high high 32-bits of the product of x and y
     * @param[out] low low 32-bits of the product of x and y
     */
    inline void
    multiply64(unsigned int x,
               unsigned int y,
               unsigned int* high,
               unsigned int* low)
    {
        cgul_exception_t local = NULL;
        cgul_multiply64(&local, x, y, high, low);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

};

#endif  /*  CGUL_MULTIPLY64_CXX_H  */
