#ifndef CGUL_RANDOM_BBS_H
#define CGUL_RANDOM_BBS_H

/**
 * Blum-Blum-Shub (BBS) pseudo-random number generator.  This class
 * has not been audited for security or cryptography purposes.  It
 * should only be used for general statistical purposes.  Because the
 * BBS PRNG only generates one bit per iteration, it is the slowest of
 * the PRNGs in cgul.  On the other hand, a correctly coded BBS PRNG
 * has the best theoretical guarantee for generating random numbers.
 *
 * @file
 * @author Paul Serice
 * @brief PRNG (Blum-Blum-Shub)
 */

#include "cgul_common.h"
#include "cgul_big_integer.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_random_bbs</code> instance. */
typedef struct cgul_random_bbs* cgul_random_bbs_t;

/**
 * Create a new <code>cgul_random_bbs</code> object seeding it with
 * the value returned by <code>time()</code>, i.e., the 32-bit seconds
 * since the epoch.  The caller is responsible for freeing the object
 * by calling <code>cgul_random_bbs__delete()</code>.  If memory
 * cannot be allocated, <code>NULL</code> is returned, and an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_random_bbs</code> instance
 */
extern CGUL_EXPORT cgul_random_bbs_t
cgul_random_bbs__new(cgul_exception_t* cex);

/**
 * Create a new <code>cgul_random_bbs</code> object seeding it with
 * the value the user passes in <code>seed</code>.  The caller is
 * responsible for freeing the object by calling
 * <code>cgul_random_bbs__delete()</code>.  If memory cannot be
 * allocated, <code>NULL</code> is returned, and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] seed seed value
 * @return new <code>cgul_random_bbs</code> instance
 */
extern CGUL_EXPORT cgul_random_bbs_t
cgul_random_bbs__new_from_seed(cgul_exception_t* cex,
                               cgul_big_integer_t seed);

/**
 * This method frees all internally allocated memory.  Do not attempt
 * to dereference <code>r</code> after calling this method.
 *
 * @param[in] r <code>cgul_random_bbs</code> instance
 */
extern CGUL_EXPORT void
cgul_random_bbs__delete(cgul_random_bbs_t r);

/**
 * The next element in the random number sequence is returned in
 * <code>next</code>.  Because Blum-Blum-Shub is a bit generator, the
 * random numbers returned are only 1-bit wide.  It is possible to
 * generate larger random numbers by calling
 * <code>cgul_random_bbs__next_in_range()</code> which will assemble
 * the larger numbers from multiple iterations of the generator.
 *
 * @param[in] cex c-style exception
 * @param[in] r <code>cgul_random_bbs</code> instance
 * @param[out] next next element in the random number sequence
 * @see cgul_random_bbs__next_in_range()
 */
extern CGUL_EXPORT void
cgul_random_bbs__next(cgul_exception_t* cex,
                      cgul_random_bbs_t r,
                      cgul_big_integer_t next);

/**
 * If the value returned by <code>cgul_random_bbs__next()</code> is
 * not the exact range that is needed, this method can limit or extend
 * the return value to [0, range_max).  If <code>range_max</code> is
 * larger than the number of bits that can be generated in one
 * iteration, multiple iterations will be used.
 *
 * @param[in] cex c-style exception
 * @param[in] r <code>cgul_random_bbs</code> instance
 * @param[out] next next element in the random number sequence
 * @param[in] range_max exclusive limit on the maximum value returned
 */
extern CGUL_EXPORT void
cgul_random_bbs__next_in_range(cgul_exception_t* cex,
                               cgul_random_bbs_t r,
                               cgul_big_integer_t next,
                               cgul_big_integer_t range_max);

CGUL_END_C

#endif  /*  CGUL_RANDOM_BBS_H  */
