#ifndef CGUL_RANDOM_LCG256_H
#define CGUL_RANDOM_LCG256_H

/**
 * 256-bit linear congruential pseudo-random number generator.
 *
 * @file
 * @author Paul Serice
 * @brief 256-bit PRNG (LCG)
 */

#include "cgul_common.h"
#include "cgul_big_integer.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_random_lcg256</code> instance. */
typedef struct cgul_random_lcg256* cgul_random_lcg256_t;

/**
 * Create a new <code>cgul_random_lcg256</code> object seeding it with
 * the value returned by <code>time()</code>, i.e., the 32-bit seconds
 * since the epoch.  The caller is responsible for freeing the object
 * by calling <code>cgul_random_lcg256__delete()</code>.  If memory
 * cannot be allocated, <code>NULL</code> is returned, and an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_random_lcg256</code> instance
 */
extern CGUL_EXPORT cgul_random_lcg256_t
cgul_random_lcg256__new(cgul_exception_t* cex);

/**
 * Create a new <code>cgul_random_lcg256</code> object seeding it with
 * the value the user passes in <code>seed</code>.  The caller is
 * responsible for freeing the object by calling
 * <code>cgul_random_lcg256__delete()</code>.  If memory cannot be
 * allocated, <code>NULL</code> is returned, and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] seed seed value
 * @return new <code>cgul_random_lcg256</code> instance
 */
extern CGUL_EXPORT cgul_random_lcg256_t
cgul_random_lcg256__new_from_seed(cgul_exception_t* cex,
                               cgul_big_integer_t seed);

/**
 * This method frees all internally allocated memory.  Do not attempt
 * to dereference <code>r</code> after calling this method.
 *
 * @param[in] r <code>cgul_random_lcg256</code> instance
 */
extern CGUL_EXPORT void
cgul_random_lcg256__delete(cgul_random_lcg256_t r);

/**
 * The next element in the random number sequence is returned in
 * <code>next</code>.  The random numbers returned are 256 bits wide.
 * It is possible to generate larger random numbers by calling
 * <code>cgul_random_lcg256__next_in_range()</code> which will
 * assemble the larger numbers from multiple iterations of the
 * generator. <p>
 *
 * The most common LCG implementation is to use a modulus that is a
 * power of 2 allowing the modulo operation to be performed with a
 * simple bit mask or intrinsic modulo 2 arithmetic (depending on the
 * data type).  The cgul_random_lcg256 implementation is no different;
 * however, this results in the well-known problem that LCGs are
 * non-random in their low-order bits.  Specifically, the nth bit has
 * a period of 2^n.  Thus, the least-significant bit only has a period
 * of 2 meaning it always just toggles between 0 and 1. <p>
 *
 * The way to get around this problem is to discard the low-order
 * bits.  This can be done by shifting to the right the value returned
 * by this function or by using
 * <code>cgul_random_lcg256__next_in_range()</code> which returns
 * numbers based much more heavily on the high order bits. <p>
 *
 * @note Do not use the value returned by this function without
 * compensating for the non-random low-order bits (as described
 * above). <p>
 *
 * @param[in] cex c-style exception
 * @param[in] r <code>cgul_random_lcg256</code> instance
 * @param[out] next next element in the random number sequence
 * @see cgul_random_lcg256__next_in_range()
 */
extern CGUL_EXPORT void
cgul_random_lcg256__next(cgul_exception_t* cex,
                         cgul_random_lcg256_t r,
                         cgul_big_integer_t next);

/**
 * If the value returned by <code>cgul_random_lcg256__next()</code> is
 * not the exact range that is needed, this method can limit or extend
 * the return value to [0, range_max).  If <code>range_max</code> is
 * larger than the number of bits that can be generated in one
 * iteration, multiple iterations will be used.  If the range needs to
 * be extended beyond what can be generated in a single iteration, the
 * Blum-Blum-Shub generator in <code>cgul_random_bbs</code> is
 * probably a better choice. <p>
 *
 * It is highly recommended that this method be used when mapping the
 * random numbers to a smaller range because (like most LCGs), the low
 * order bits of this LCG are known to repeat much more quickly than
 * all the bits taken as a whole.  This method takes this problem into
 * account to return a suitable pseudo-random number. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] r <code>cgul_random_lcg256</code> instance
 * @param[out] next next element in the random number sequence
 * @param[in] range_max exclusive limit on the maximum value returned
 */
extern CGUL_EXPORT void
cgul_random_lcg256__next_in_range(cgul_exception_t* cex,
                                  cgul_random_lcg256_t r,
                                  cgul_big_integer_t next,
                                  cgul_big_integer_t range_max);

CGUL_END_C

#endif  /*  CGUL_RANDOM_LCG256_H  */
