#ifndef CGUL_RBTREE_H
#define CGUL_RBTREE_H

/**
 * Red-black tree implementation. <p>
 *
 * @file
 * @author Paul Serice
 * @brief red-black tree implementation
 */

/*
 * NOTE: In order to make modifications to this class, you are going
 * to need to have a thorough understanding of both 2-3-4 trees and
 * red-black trees.
 */

#include "cgul_common.h"
#include "cgul_cache.h"
#include "cgul_exception.h"
#include "cgul_rbtree_node.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_rbtree</code> instance. */
typedef struct cgul_rbtree* cgul_rbtree_t;

/**
 * This typedef is the interface the client must define in order for
 * <code>cgul_rbtree</code> to sort your keys as they are inserted.
 * When you define your comparison function, you should compare
 * <code>key1</code> to <code>key2</code> and return less than zero,
 * zero, or greater than zero if <code>key1</code> is less than, equal
 * to, or greater than <code>key2</code>.
 *
 * @param[in] key1 first key
 * @param[in] key2 second key
 * @return result of comparison
 */
typedef int (*cgul_rbtree__compare_t)(const void* key1,
                                      const void* key2);

/**
 * This typedef is the interface for the combining function used by
 * the following methods: <p>
 *
 * <pre>
 *     <code>cgul_rbtree__foldl_keys()</code>
 *     <code>cgul_rbtree__foldr_keys()</code>
 * </pre>
 *
 * The client must not alter <code>key</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] key key
 * @param[in] data client data
 * @return whether to continue
 */
typedef int (*cgul_rbtree__fold_key_t)(cgul_exception_t* cex,
                                       const void* key,
                                       void* data);

/**
 * This typedef is the interface for the combining function used by
 * the following methods: <p>
 *
 * <pre>
 *     <code>cgul_rbtree__foldl_values()</code>
 *     <code>cgul_rbtree__foldr_values()</code>
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @param[in] value value
 * @param[in] data client data
 * @return whether to continue
 */
typedef int (*cgul_rbtree__fold_value_t)(cgul_exception_t* cex,
                                         void* value,
                                         void* data);

/**
 * This typedef is the interface for the combining function used by
 * the following methods: <p>
 *
 * <pre>
 *     <code>cgul_rbtree__foldl_pairs()</code>
 *     <code>cgul_rbtree__foldr_pairs()</code>
 * </pre>
 *
 * The client must not alter <code>key</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] key key
 * @param[in] value value
 * @param[in] data client data
 * @return whether to continue
 */
typedef int (*cgul_rbtree__fold_pair_t)(cgul_exception_t* cex,
                                        const void* key,
                                        void* value,
                                        void* data);

/**
 * This typedef is the interface for the callback function used by the
 * following methods: <p>
 *
 * <pre>
 *     <code>cgul_rbtree__traverse()</code>
 *     <code>cgul_rbtree__traverse_range()</code>
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] n node
 * @param[in] data client data
 * @return whether to continue
 */
typedef int (*cgul_rbtree__traverse_t)(cgul_exception_t* cex,
                                       cgul_rbtree_t t,
                                       cgul_rbtree_node_t n,
                                       void* data);

/**
 * Create a new <code>cgul_rbtree</code> object.  <code>compare</code>
 * is the comparison function that allows the tree to maintain its
 * sorted order.  The caller is responsible for freeing the object by
 * calling <code>cgul_rbtree__delete()</code>.  If memory cannot be
 * allocated, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * The <code>cgul_rbtree</code> object that is returned does not allow
 * you to index into it as though it were an array.  This saves you
 * two <code>unsigned long</code> values for each node in the tree.
 * It also saves you about 15% for inserts and deletes.  If in doubt,
 * use this constructor. <p>
 *
 * This class does not take ownership of the inserted keys/value
 * pairs.  Instead, the client is responsible for freeing the
 * key/value pairs only after their nodes have been permanently
 * removed from the <code>cgul_rbtree</code>.  The
 * <code>cgul_rbtree__traverse()</code> method can be used to remove
 * each node safely before calling
 * <code>cgul_rbtree__delete()</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] compare comparison function
 * @return new <code>cgul_rbtree</code> instance
 * @see cgul_rbtree__compare_t
 */
extern CGUL_EXPORT cgul_rbtree_t
cgul_rbtree__new(cgul_exception_t* cex,
                 cgul_rbtree__compare_t compare);

/**
 * Create a new <code>cgul_rbtree</code> object.  <code>compare</code>
 * is the comparison function that allows the tree to maintain its
 * sorted order.  The caller is responsible for freeing the object by
 * calling <code>cgul_rbtree__delete()</code>.  If memory cannot be
 * allocated, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * The <code>cgul_rbtree</code> object that is returned allows you to
 * index into it as though it were an array.  This costs you two
 * <code>unsigned long</code> values for each node in the tree.  It
 * also costs you about 15% for inserts and deletes.  You should only
 * use this constructor if you know you want to be able to call
 * <code>cgul_rbtree__find_at()</code> or
 * <code>cgul_rbtree__remove_at()</code>. <p>
 *
 * This class does not take ownership of the inserted keys/value
 * pairs.  Instead, the client is responsible for freeing the
 * key/value pairs only after their nodes have been permanently
 * removed from the <code>cgul_rbtree</code>.  The
 * <code>cgul_rbtree__traverse()</code> method can be used to remove
 * each node safely before calling
 * <code>cgul_rbtree__delete()</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] compare comparison function
 * @return new <code>cgul_rbtree</code> instance
 * @see cgul_rbtree__compare_t
 */
extern CGUL_EXPORT cgul_rbtree_t
cgul_rbtree__new_with_indexing(cgul_exception_t* cex,
                               cgul_rbtree__compare_t compare);

/**
 * This method frees all internally allocated memory.  This does not
 * include the key/value pairs stored in the tree.  The client is
 * responsible for freeing the key/value pairs when the client thinks
 * it is convenient and safe to do so; however, the client must
 * understand that freeing a key before removing its node
 * invalidates the data structure. <p>
 *
 * As a convenience, you may want to call
 * <code>cgul_rbtree__clear()</code> before calling this method in
 * order to properly put the keys and values back on their respective
 * <code>cgul_cache</code> objects (if <code>cgul_cache</code>
 * objects are being used) before you delete the tree.  If
 * <code>cgul_cache</code> objects are not being used, the client
 * needs to arrange some other mechanism to free the keys or
 * values. <p>
 *
 * Do not try to access <code>t</code> after calling this method. <p>
 *
 * @param[in] t <code>cgul_rbtree</code> instance
 * @see cgul_rbtree__free_keys()
 * @see cgul_rbtree__free_values()
 */
extern CGUL_EXPORT void
cgul_rbtree__delete(cgul_rbtree_t t);

/**
 * This method calls <code>free()</code> on all the keys in the tree
 * <code>t</code>.  Because this is such a common operation, it is an
 * exception to the rule that cgul containers never free keys.  This
 * method should only ever be called immediately before calling
 * <code>cgul_rbtree__delete()</code> because it otherwise invalidates
 * the tree.  Because this function is basically an extension of
 * <code>cgul_rbtree__delete()</code>, it does not accept a
 * <code>cgul_exception</code> object as an input parameter.
 *
 * @param[in] t <code>cgul_rbtree</code> instance
 */
extern CGUL_EXPORT void
cgul_rbtree__free_keys(cgul_rbtree_t t);

/**
 * This method calls <code>free()</code> on all the values in the tree
 * <code>t</code>.  Because this is such a common operation, it is an
 * exception to the rule that cgul containers never free values.  This
 * method should only ever be called immediately before calling
 * <code>cgul_rbtree__delete()</code> because it otherwise invalidates
 * the tree.  Because this function is basically an extension of
 * <code>cgul_rbtree__delete()</code>, it does not accept a
 * <code>cgul_exception</code> object as an input parameter.
 *
 * @param[in] t <code>cgul_rbtree</code> instance
 */
extern CGUL_EXPORT void
cgul_rbtree__free_values(cgul_rbtree_t t);

/**
 * Get the size of the cache of nodes. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @return size of the cache of nodes
 */
extern CGUL_EXPORT unsigned long int
cgul_rbtree__get_cache_size(cgul_exception_t* cex,
                            cgul_rbtree_t t);

/**
 * Set the size of the cache of nodes. <p>
 *
 * For efficiency, the <code>cgul_rbtree</code> object can keep a
 * cache of nodes that can be reused.  In many situations, this can
 * greatly reduce the number of calls to <code>malloc()</code> which
 * can greatly improve performance and reduce memory
 * fragmentation. <p>
 *
 * You can release all the cached nodes and disable caching of nodes
 * by setting the cache reserve to 0 and the cache size to 0 (in that
 * order).  This is the default. <p>
 *
 * If an error occurs (while enlarging the cache), an exception is
 * thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] size of the cache of nodes
 */
extern CGUL_EXPORT void
cgul_rbtree__set_cache_size(cgul_exception_t* cex,
                            cgul_rbtree_t t,
                            unsigned long int size);

/**
 * Get the reserve limit of the cache of nodes. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @return reserve limit of the cache of nodes
 */
extern CGUL_EXPORT unsigned long int
cgul_rbtree__get_cache_reserve(cgul_exception_t* cex,
                               cgul_rbtree_t t);

/**
 * Set the reserve limit of the cache of nodes and guarantee that at
 * least that many nodes are allocated and waiting in reserve. <p>
 *
 * When you use this method to increase the reserve, you are
 * guaranteed that at least <code>reserve</code> count of nodes will
 * be allocated and reserved for future use. <p>
 *
 * This is useful because <code>cgul_rbtree__insert()</code> can
 * throw an exception if it cannot allocate a node to hold the
 * key/value pair.  Thus, by allocating and reserving a node now, we
 * know that <code>cgul_rbtree__insert()</code> will be able to run
 * without error later when the node is "unreserved".  Normally, this
 * is not an issue because you can almost always just insert at the
 * moment when it is requested, but if the <code>cgul_rbtree</code>
 * is busy at that moment (perhaps it is iterating over its nodes),
 * you'll have to queue insert (and remove) requests for later
 * execution. <p>
 *
 * To "unreserve" a node so that it can be used, you simply call this
 * method again with a value for <code>reserve</code> that is less
 * than the current reserve limit. <p>
 *
 * If an error occurs while allocating the reserved nodes, an
 * exception is thrown. <p>
 *
 * WARNING: If you reserve nodes without "unreserving" them, you will
 * introduce a memory leak that is difficult to track down because
 * this class is careful to free all of the cached nodes (including
 * the reserved nodes) when it is deleted, but the reserved nodes will
 * just sit around in memory until that time.  Thus, after reserving
 * nodes by calling this method, you must remember to timely set the
 * reserve level back to zero in order to make those nodes
 * available. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] reserve of the cache of nodes
 * @see cgul_cache__set_reserve
 */
extern CGUL_EXPORT void
cgul_rbtree__set_cache_reserve(cgul_exception_t* cex,
                               cgul_rbtree_t t,
                               unsigned long int reserve);

/**
 * Return 1 if the tree is empty; otherwise, return 0.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @return whether the tree is empty
 */
extern CGUL_EXPORT int
cgul_rbtree__is_empty(cgul_exception_t* cex,
                      cgul_rbtree_t t);

/**
 * Insert <code>key</code> into the tree <code>t</code>.  If
 * <code>key</code> already exists, another node with the same key is
 * not created.  The new or pre-existing node is returned in
 * <code>*n</code> if <code>n</code> is not <code>NULL</code>. <p>
 *
 * If a matching node is not found, <code>1</code> is returned because
 * a new node can and will be inserted.  If a matching node is found,
 * <code>0</code> is returned because the insertion failed because the
 * key already exists.<p>
 *
 * If an error occurs, 0 is returned, <code>*n</code> is set to
 * <code>NULL</code> (if <code>n</code> is not <code>NULL</code>), and
 * an exception is thrown. <p>
 *
 * It is important to understand that the <code>cgul_rbtree</code>
 * class (like all containers in the cgul library) does not take
 * ownership of <code>key</code>.  It also does not attempt to make a
 * copy of the thing pointed to by <code>key</code>.  This gives the
 * user complete control over the lifetime of the key and gives a real
 * performance boost in many cases; however, this means that
 * <code>key</code> must not be invalidated while it is still being
 * used by this class. <p>
 *
 * This method operates as it does so that you only have to search the
 * tree once when doing an insert.  Logically, you would usually like
 * to do a find to see if the tree holds your key.  If it does, you
 * use the value associated with that key in some way.  If it does
 * not, you then want to insert a new key/value pair into the tree.
 * The problem with divorcing the find from the insert is that it then
 * requires you to search the tree twice.  Once to see if the key
 * exists and once to do the insertion. <p>
 *
 * To prevent having to search the tree twice, you can just call this
 * method to handle both inserting and updating nodes.  It will search
 * the tree once, and it will always return a pointer to the correct
 * node.  You can then use <code>cgul_rbtree_node__get_value()</code>
 * or <code>cgul_rbtree_node__set_value()</code> to get or set the
 * value indexed by "key". <p>
 *
 * TIP: You might find it convenient to use <code>cgul_string</code>
 * to build up a string that you use as a key.  You can then use
 * <code>cgul_string__get_value()</code> [yes, get_value()] and pass
 * the string to <code>cgul_rbtree__insert()</code>.  If the insert
 * succeeds, you can inform the <code>cgul_string</code> object that
 * it is no longer the owner of the underlying C-style string by
 * calling <code>cgul_string__take_value()</code> [yes, take_value()].
 * An example is as follows: <p>
 *
 * <pre>
 *
 *     int insert_rv = 0;
 *     cgul_rbtree_node_t n = NULL;
 *
 *     char* w = cgul_string__get_value(word);
 *     insert_rv = cgul_rbtree__insert(t, w, &n);
 *     if (*cex) {
 *         goto out;
 *     }
 *     if (insert_rv) {
 *         cgul_string__take_value(word);
 *         . . .
 *     } else {
 *         . . .
 *     }
 *
 * </pre>
 *
 * NOTE: Inserting a key/value pair does not cause this class to take
 * ownership of <code>key</code> or <code>value</code>.  The client
 * must make sure the values pointed to by <code>key</code> and
 * <code>value</code> are not invalidated during the life of the newly
 * created node.  If the client needs to delete the <code>key</code>
 * or the <code>value</code> or both, the client needs to manually
 * arrange for this. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] key key
 * @param[out] n node that already existed or was newly created
 * @return 1 if a new node was inserted; 0 if a node with the same key
 * already exists
 */
extern CGUL_EXPORT int
cgul_rbtree__insert(cgul_exception_t* cex,
                    cgul_rbtree_t t,
                    const void* key,
                    cgul_rbtree_node_t* n);

/**
 * Find the node indexed by <code>key</code>.  If the key does not
 * exist, <code>NULL</code> is returned; otherwise, the node
 * associated with the key is returned.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] key key
 * @return node associated with the key or <code>NULL</code>
 */
extern CGUL_EXPORT cgul_rbtree_node_t
cgul_rbtree__find(cgul_exception_t* cex,
                  cgul_rbtree_t t,
                  const void* key);

/**
 * Find the node indexed by position.  This lets you use a
 * <code>cgul_rbtree</code> instances like you would an array.
 * Indexing into <code>cgul_rbtree</code> is a logarithmic function
 * which will be slower than an array, but you get all the advantages
 * of a balanced binary tree including fast insertion and deletion.
 * So, if you ever feel the need to shift the elements of an array,
 * this method might be for you. <p>
 *
 * If <code>index</code> is out of range, <code>NULL</code> is
 * returned; otherwise, the node returned is the same node that would
 * be returned if you had an array <code>t</code> that held all the
 * nodes in the tree in sorted order, and you asked for
 * <code>t[index]</code>. <p>
 *
 * This method throws an exception if <code>t</code> was not created
 * with <code>cguil_rbtree__new_with_indexing()</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] index index into sorted keys of node to return
 * @return node associated with <code>index</code>
 * @see cgul_rbtree__find_rank()
 */
extern CGUL_EXPORT cgul_rbtree_node_t
cgul_rbtree__find_at(cgul_exception_t* cex,
                     cgul_rbtree_t t,
                     unsigned long int index);

/**
 * Return whether the rank of the key <code>key</code> could be
 * determined.  If the rank could be determined it is returned in
 * <code>*rank</code> (if <code>rank</code> is not
 * <code>NULL</code>). <p>
 *
 * The rank is the number of keys less than <code>key</code>.  It can
 * be used as the index parameter to <code>cgul_rbtree__find_at()</code>
 * in order to lookup <code>key</code> by index. <p>
 *
 * This method throws an exception if <code>t</code> was not created
 * with <code>cguil_rbtree__new_with_indexing()</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] key key
 * @param[out] rank rank of key
 * @return whether the key exists in the tree
 * @see cgul_rbtree__find_at()
 */
extern CGUL_EXPORT int
cgul_rbtree__find_rank(cgul_exception_t* cex,
                       cgul_rbtree_t t,
                       const void* key,
                       unsigned long int* rank);

/**
 * Find the node having the largest key less than or equal to the
 * search key <code>search_key</code> and return it.  If such a key
 * does not exist, <code>NULL</code> is returned instead.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] search_key search key
 * @return floor or <code>NULL</code>
 */
extern CGUL_EXPORT cgul_rbtree_node_t
cgul_rbtree__find_floor(cgul_exception_t* cex,
                        cgul_rbtree_t t,
                        const void* search_key);

/**
 * Find the node having the smallest key greater than or equal to the
 * search key <code>search_key</code> and return it.  If such a key
 * does not exist, <code>NULL</code> is returned instead.
 *
 * @param[in] cex c-style exception
 * @param[in] rbtree <code>cgul_rbtree</code> instance
 * @param[in] search_key search key
 * @return ceiling or <code>NULL</code>
 */
extern CGUL_EXPORT cgul_rbtree_node_t
cgul_rbtree__find_ceiling(cgul_exception_t* cex,
                        cgul_rbtree_t rbtree,
                        const void* search_key);

/**
 * This function lets you search the tree using arbitrary client data
 * rather than a specific key.  The comparison function
 * <code>compare</code> will be called with <code>key1</code> set to
 * the current node which is of type <code>cgul_rbtree_node_t</code>
 * and <code>key2</code> set to <code>data</code>.  The comparison
 * function should return less than one if you want to move to the
 * current node's left child.  It should return greater than one if
 * you want to move to the current node's right child.  It should
 * return zero, if you have found what you are looking for.  If your
 * comparison function returns zero, that is the node that will be
 * returned; otherwise, <code>NULL</code> is returned.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] compare comparison function
 * @param[in] data data passed to <code>compare</code> in lieu of a second key
 * @return node that was found or <code>NULL</code> if no node could be found
 */
extern CGUL_EXPORT cgul_rbtree_node_t
cgul_rbtree__find_special(cgul_exception_t* cex,
                          cgul_rbtree_t t,
                          cgul_rbtree__compare_t compare,
                          const void* data);

/**
 * Return the first node according to sort order.  This operation is
 * O(1).  If the tree is empty, <code>NULL</code> is returned.  This
 * class does not have to search the tree in order to find the front
 * node because it efficiently keeps a direct pointer to the front
 * node up to date. <p>
 *
 * The following example shows how to iterate over the entire tree in
 * sort order: <p>
 *
 * <pre>
 *     cgul_rbtree_node_t n = cgul_rbtree__get_front(cex, t);
 *     for ( ; n ; n = cgul_rbtree_node__get_next(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @return front node
 */
extern CGUL_EXPORT cgul_rbtree_node_t
cgul_rbtree__get_front(cgul_exception_t* cex,
                       cgul_rbtree_t t);

/**
 * Return the last node according to sort order.  This operation is
 * O(1).  If the tree is empty, <code>NULL</code> is returned.  This
 * class does not have to search the tree in order to find the back
 * node because it efficiently keeps a direct pointer to the back node
 * up to date. <p>
 *
 * The following example shows how to iterate over the entire tree in
 * reverse sort order: <p>
 *
 * <pre>
 *     cgul_rbtree_node_t n = cgul_rbtree__get_back(cex, t);
 *     for ( ; n ; n = cgul_rbtree_node__get_prev(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @return back node
 */
extern CGUL_EXPORT cgul_rbtree_node_t
cgul_rbtree__get_back(cgul_exception_t* cex,
                      cgul_rbtree_t t);

/**
 * Return the oldest node according to chronological order (i.e., the
 * order in which the nodes are inserted).  This operation is O(1).
 * If the tree is empty, <code>NULL</code> is returned.  This class
 * does not have to search the tree in order to find the oldest node
 * because it efficiently keeps a direct pointer to the oldest node up
 * to date. <p>
 *
 * The following example shows how to iterate over the entire tree in
 * chronological order:
 *
 * <pre>
 *     cgul_rbtree_node_t n = cgul_rbtree__get_oldest(cex, t);
 *     for ( ; n ; n = cgul_rbtree_node__get_younger(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @return oldest node
 * @see cgul_rbtree_node__get_younger()
 */
extern CGUL_EXPORT cgul_rbtree_node_t
cgul_rbtree__get_oldest(cgul_exception_t* cex,
                        cgul_rbtree_t t);

/**
 * Set the oldest node in the tree to be <code>n</code>.  Calling this
 * method has the potential to confuse iterators and should be handled
 * with roughly the same level of caution as calling
 * <code>cgul_rbtree__remove_node()</code>. <p>
 *
 * This method could be used, for example, if your code expires the
 * oldest node in the tree, and you want to force <code>n</code> to be
 * the next node to expire. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] n node
 * @see cgul_rbtree__set_youngest()
 */
extern CGUL_EXPORT void
cgul_rbtree__set_oldest(cgul_exception_t* cex,
                        cgul_rbtree_t t,
                        cgul_rbtree_node_t n);

/**
 * Return the youngest node according to chronological order (i.e.,
 * the order in which the nodes are inserted).  This operation is
 * O(1).  If the tree is empty, <code>NULL</code> is returned.  This
 * class does not have to search the tree in order to find the
 * youngest node because it efficiently keeps a direct pointer to the
 * youngest node up to date. <p>
 *
 * The following example shows how to iterate over the entire tree in
 * reverse chronological order:
 *
 * <pre>
 *     cgul_rbtree_node_t n = cgul_rbtree__get_youngest(cex, t);
 *     for ( ; n ; n = cgul_rbtree_node__get_older(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @return youngest node
 * @see cgul_rbtree_node__get_older()
 */
extern CGUL_EXPORT cgul_rbtree_node_t
cgul_rbtree__get_youngest(cgul_exception_t* cex,
                          cgul_rbtree_t t);

/**
 * Set the youngest node in the tree to be <code>n</code>.  Calling
 * this method has the potential to confuse iterators and should be
 * handled with roughly the same level of caution as calling
 * <code>cgul_rbtree__remove_node()</code>. <p>
 *
 * This method could be used, for example, if your code expires the
 * least-recently used (LRU) node.  By calling
 * <code>cgul_rbtree__set_youngest()</code> each time a node is used,
 * the least-recently used node will be the oldest node in the
 * tree. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] n node
 * @see cgul_rbtree__set_oldest()
 */
extern CGUL_EXPORT void
cgul_rbtree__set_youngest(cgul_exception_t* cex,
                          cgul_rbtree_t t,
                          cgul_rbtree_node_t n);

/**
 * Remove the node in <code>t</code> associated with <code>key_in</code>.
 * The key and value pointers stored in the node that is to be removed
 * will be returned in <code>key_out</code> and <code>value_out</code>
 * if you pass in pointers that are not <code>NULL</code>.  This
 * method returns 1 if the node was removed; it returns 0
 * otherwise. <p>
 *
 * It is almost always a mistake to naively call this method while
 * iterating over the tree.  See
 * <code>cgul_rbtree__remove_node()</code> for details. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] key_in key associated with the node to be removed
 * @param[out] key_out pointer for the key stored in the node
 * @param[out] value_out pointer for the value stored in the node
 * @return whether the node was removed
 */
extern CGUL_EXPORT int
cgul_rbtree__remove(cgul_exception_t* cex,
                    cgul_rbtree_t t,
                    const void* key_in,
                    void** key_out,
                    void** value_out);

/**
 * Remove <code>node</code> from the tree <code>t</code>.  The key and
 * value pointers stored in the node that is to be removed will be
 * returned in <code>key_out</code> and <code>value_out</code> if you
 * pass in pointers that are not <code>NULL</code>. <p>
 *
 * There is no speed advantage to using this method instead of
 * <code>cgul_rbtree__remove()</code>.  The reason is that removing a
 * node has the potential to unbalance the tree requiring the code to
 * descend the tree even though we already have the node in hand. <p>
 *
 * It is almost always a mistake to naively call this method while
 * iterating over the tree.  The problem is that calling this
 * method invalidates the node making it impossible to call
 * <code>cgul_rbtree_node__get_next()</code> afterward.  The
 * solution is simple.  Just call
 * <code>cgul_rbtree_node__get_next()</code> before calling this
 * method:
 *
 * <pre>
 *
 *     curr = cgul_rbtree__find(&local, t, key);
 *     for ( ; curr ; curr = next) {
 *         next = cgul_rbtree_node__get_next(&local, curr);
 *         cgul_rbtree__remove_node(&local,
 *                                  t,
 *                                  curr,
 *                                  &key_out,
 *                                  &value_out);
 *         free(key_out);
 *         free(value_out);
 *     }
 *
 * </pre>
 *
 * Alternatively, you can use
 * <code>cgul_rbtree__remove_range()</code>,
 * <code>cgul_rbtree__traverse()</code>, or
 * <code>cgul_rbtree__traverse_range()</code>. <p>
 *
 * If you use one of the traversal methods, it is safe to call this
 * method naively.  For example, if your tree stores
 * <code>cgul_string</code> objects for both its keys and values,
 * another way to remove nodes is as follows: <p>
 *
 * <pre>
 *
 *     static int
 *     remove_node_cb(cgul_rbtree_t t,
 *                    cgul_rbtree_node_t node,
 *                    void* data)
 *     {
 *         cgul_string_t key = NULL;
 *         cgul_string_t value = NULL;
 *         cgul_exception_t local = NULL;
 *         cgul_rbtree__remove_node(&local,
 *                                  t,
 *                                  node,
 *                                  (void**)&key,
 *                                  (void**)&value);
 *         cgul_string__delete(key);
 *         cgul_string__delete(value);
 *         return 1;
 *     }
 *
 *     first = cgul_rbtree__find(&local, t, key1);
 *     last = cgul_rbtree__find(&local, t, key2);
 *     if (first && last) {
 *         cgul_rbtree__traverse_range(&local,
 *                                     t,
 *                                     first,
 *                                     last,
 *                                     &remove_node_cb,
 *                                     NULL);
 *     }
 *
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] n node to be removed
 * @param[out] key_out pointer for the key stored in the node
 * @param[out] value_out pointer for the value stored in the node
 */
extern CGUL_EXPORT void
cgul_rbtree__remove_node(cgul_exception_t* cex,
                         cgul_rbtree_t t,
                         cgul_rbtree_node_t n,
                         void** key_out,
                         void** value_out);

/**
 * This method lets you remove a node based on its sorted order
 * <code>index</code>.  If the index is out of bounds, 0 is returned;
 * otherwise 1 is returned.  If you are interested in getting your
 * hands on the key and value pointers stored in the node that is to
 * be removed, they will be returned in <code>key_out</code> and
 * <code>value_out</code> if you pass in pointers that are not
 * <code>NULL</code>.  <code>index</code> is zero-based. <p>
 *
 * This method throws an exception if <code>t</code> was not created
 * with <code>cguil_rbtree__new_with_indexing()</code>. <p>
 *
 * It is almost always a mistake to naively call this method while
 * iterating over the tree.  See
 * <code>cgul_rbtree__remove_node()</code> for details. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] index index of the sorted key associated with the node
 * to be removed
 * @param[out] key_out pointer for the key stored in the node
 * @param[out] value_out pointer for the value stored in the node
 * @return whether the node was removed
 */
extern CGUL_EXPORT int
cgul_rbtree__remove_at(cgul_exception_t* cex,
                       cgul_rbtree_t t,
                       unsigned long int index,
                       void** key_out,
                       void** value_out);

/**
 * As with all remove operations,
 * <code>cgul_rbtree__remove_front()</code> has to descend the tree
 * making sure the tree is balanced for the removal.  The performance
 * advantage to using this method is that it does not have to call the
 * comparison function to decide whether it should take the right or
 * left child of each node.  It knows to always take the left child. <p>
 *
 * The key and value pointers stored in the node that is to be removed
 * will be returned in <code>key_out</code> and <code>value_out</code>
 * if you pass in pointers that are not <code>NULL</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[out] key_out pointer for the key stored in the node
 * @param[out] value_out pointer for the value stored in the node
 * @return whether the node was removed
 */
extern CGUL_EXPORT int
cgul_rbtree__remove_front(cgul_exception_t* cex,
                          cgul_rbtree_t t,
                          void** key_out,
                          void** value_out);

/**
 * As with all remove operations,
 * <code>cgul_rbtree__remove_back()</code> has to descend the tree
 * making sure the tree is balanced for the removal.  The performance
 * advantage to using this method is that it does not have to call the
 * comparison function to decide whether it should take the right or
 * left child of each node.  It knows to always take the right
 * child. <p>
 *
 * The key and value pointers stored in the node that is to be removed
 * will be returned in <code>key_out</code> and <code>value_out</code>
 * if you pass in pointers that are not <code>NULL</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[out] key_out pointer for the key stored in the node
 * @param[out] value_out pointer for the value stored in the node
 * @return whether the node was removed
 */
extern CGUL_EXPORT int
cgul_rbtree__remove_back(cgul_exception_t* cex,
                         cgul_rbtree_t t,
                         void** key_out,
                         void** value_out);

/**
 * This method removes nodes in the range <code>first</code>
 * (inclusive) to <code>last</code> (inclusive).  Strictly as a
 * convenience, this method is an exception to the rule that cgul
 * containers never free keys or values.  If you pass in
 * <code>keys_cache</code> or <code>values_cache</code> instances that
 * are not <code>NULL</code>, the keys or values will be put back on
 * their respective caches. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] first first node in range
 * @param[in] last last node in range
 * @param[in] keys_cache keys cache
 * @param[in] values_cache values cache
 * @return whether any nodes were removed.
 * @see cgul_cache__get_freer()
 */
extern CGUL_EXPORT int
cgul_rbtree__remove_range(cgul_exception_t* cex,
                          cgul_rbtree_t t,
                          cgul_rbtree_node_t first,
                          cgul_rbtree_node_t last,
                          cgul_cache_t keys_cache,
                          cgul_cache_t values_cache);

/**
 * This method clears the tree by removing each node individually.
 * Strictly as a convenience, this method is an exception to the rule
 * that cgul containers never free keys or values.  If you pass in
 * <code>keys_cache</code> or <code>values_cache</code> instances that
 * are not <code>NULL</code>, the keys or values will be put back on
 * their respective caches. <p>
 *
 * Another easy way to clear your tree is to just keep removing nodes
 * from the front of the tree freeing the key/value pairs as you go:
 *
 * <pre>
 *     void* key = NULL;
 *     void* value = NULL;
 *     while (cgul_rbtree__remove_front(&local, tree, &key, &value)) {
 *         free(key);
 *         free(value);
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] keys_cache keys cache
 * @param[in] values_cache values cache
 * @see cgul_rbtree__remove_range
 * @see cgul_cache__get_freer()
 */
extern CGUL_EXPORT void
cgul_rbtree__clear(cgul_exception_t* cex,
                   cgul_rbtree_t t,
                   cgul_cache_t keys_cache,
                   cgul_cache_t values_cache);

/**
 * Return the total number of elements stored in <code>t</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @return size of the tree
 */
extern CGUL_EXPORT unsigned long int
cgul_rbtree__get_size(cgul_exception_t* cex,
                      cgul_rbtree_t t);

/**
 * This function performs a basic sanity check on the tree and returns
 * 1 if the tree passes and 0 if the tree fails.  For now, it checks
 * to see that there are never two red nodes without at least one
 * intervening black node.  It also checks to make sure that the black
 * depth of all nodes are within one level of each other.  There are
 * many unbalanced trees that satisfy this criteria however.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @return 1 if passed basic sanity check; otherwise, 0
 */
extern CGUL_EXPORT int
cgul_rbtree__is_ok(cgul_exception_t* cex,
                   cgul_rbtree_t t);

/**
 * Swap the underlying data for <code>t1</code> and <code>t2</code>.
 * For large trees, this should be much faster than trying to do the
 * same thing using removes and inserts.
 *
 * @param[in] cex c-style exception
 * @param[in] t1 first <code>cgul_rbtree</code> instance
 * @param[in] t2 second <code>cgul_rbtree</code> instance
 */
extern CGUL_EXPORT void
cgul_rbtree__swap(cgul_exception_t* cex,
                  cgul_rbtree_t t1,
                  cgul_rbtree_t t2);

/**
 * This method performs a left fold of the tree <code>t</code> with
 * the combining function <code>f</code>.  <code>f</code> is called
 * once for each key in the tree starting at the front of the tree and
 * iterating forward to the end of the tree. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current key.  The third parameter passed into
 * <code>f</code> is the client data which is where the result of the
 * fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_rbtree__foldl_keys(cgul_exception_t* cex,
                        cgul_rbtree_t t,
                        cgul_rbtree__fold_key_t f,
                        void* data);

/**
 * This method performs a right fold of the tree <code>t</code> with
 * the combining function <code>f</code>.  <code>f</code> is called
 * once for each key in the tree starting at the back of the tree and
 * iterating backward to the front of the tree. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current key.  The third parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_rbtree__foldr_keys(cgul_exception_t* cex,
                        cgul_rbtree_t t,
                        cgul_rbtree__fold_key_t f,
                        void* data);

/**
 * This method performs a left fold of the tree <code>t</code> with
 * the combining function <code>f</code>.  <code>f</code> is called
 * once for each value in the tree starting at the front of the tree
 * and iterating forward to the end of the tree. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current value.  The third parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_rbtree__foldl_values(cgul_exception_t* cex,
                          cgul_rbtree_t t,
                          cgul_rbtree__fold_value_t f,
                          void* data);

/**
 * This method performs a right fold of the tree <code>t</code> with
 * the combining function <code>f</code>.  <code>f</code> is called
 * once for each value in the tree starting at the back of the tree
 * and iterating backward to the front of the tree. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current value.  The third parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_rbtree__foldr_values(cgul_exception_t* cex,
                          cgul_rbtree_t t,
                          cgul_rbtree__fold_value_t f,
                          void* data);

/**
 * This method performs a left fold of the tree <code>t</code> with
 * the combining function <code>f</code>.  <code>f</code> is called
 * once for each key/value pair in the tree starting at the front of
 * the tree and iterating forward to the end of the tree.<p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current key.  The third parameter passed into
 * <code>f</code> is the current value.  The fourth parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_rbtree__foldl_pairs(cgul_exception_t* cex,
                         cgul_rbtree_t t,
                         cgul_rbtree__fold_pair_t f,
                         void* data);

/**
 * This method performs a right fold of the tree <code>t</code> with
 * the combining function <code>f</code>.  <code>f</code> is called
 * once for each key/value pair in the tree starting at the back of
 * the tree and iterating backward to the front of the tree. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current key.  The third parameter passed into
 * <code>f</code> is the current value.  The fourth parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_rbtree__foldr_pairs(cgul_exception_t* cex,
                         cgul_rbtree_t t,
                         cgul_rbtree__fold_pair_t f,
                         void* data);

/**
 * Traverse all nodes passing each node to the function
 * <code>f</code>. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the rbtree <code>t</code>.  The third paramenter
 * passed into <code>f</code> is the node for this iteration.  The
 * fourth parameter passed into <code>f</code> is the client data
 * <code>data</code>. <p>
 *
 * <code>f</code> is provided with a safe context in which it can
 * remove the node that is passed into <code>f</code> by calling
 * <code>cgul_rbtree__remove_node()</code>. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * the traversal to continue. <p>
 *
 * NOTE: It is not strictly necessary that you use
 * <code>cgul_rbtree__traverse()</code> or
 * <code>cgul_rbtree__traverse_range()</code> in order to iterate over
 * the rbtree elements.  In fact, I would recommend that you use
 * <code>cgul_rbtree_node__get_next()</code> for most of your iteration
 * needs.  If you need to remove nodes though, you should probably use
 * this method.  If you look at this method's source however, you will
 * see that the <code>cgul_rbtree_node</code> class provides all the
 * public methods you need to safely remove nodes while you are
 * iterating over the rbtree, but you do have to be careful.
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] f traversal callback function
 * @param[in] data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_rbtree__traverse(cgul_exception_t* cex,
                      cgul_rbtree_t t,
                      cgul_rbtree__traverse_t f,
                      void* data);

/**
 * Traverse all nodes starting with <code>first</code> (inclusive) and
 * ending with <code>last</code> (inclusive) passing each node to the
 * function <code>f</code>.  If you know the first node, but are not
 * sure of the last node, just use <code>NULL</code> as the last node.
 * This will cause this method to iterate until it reaches the end of
 * the rbtree.  You can than have <code>f</code> return 0 when it
 * determines that the last node has been reached. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the rbtree <code>t</code>.  The third paramenter
 * passed into <code>f</code> is the node for this iteration.  The
 * fourth parameter passed into <code>f</code> is the client data
 * <code>data</code>. <p>
 *
 * <code>f</code> is provided with a safe context in which it can
 * remove the node that is passed into <code>f</code> by calling
 * <code>cgul_rbtree__remove_node()</code>. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * the traversal to continue. <p>
 *
 * If <code>first</code> is <code>NULL</code>, iteration starts at the
 * beginning of the rbtree.  If <code>last</code> is <code>NULL</code>,
 * iteration stops at the end of the rbtree. <p>
 *
 * NOTE: It is not strictly necessary that you use
 * <code>cgul_rbtree__traverse()</code> or
 * <code>cgul_rbtree__traverse_range()</code> in order to iterate over
 * the rbtree elements.  In fact, I would recommend that you use
 * <code>cgul_rbtree_node__get_next()</code> for most of your
 * iteration needs.  If you need to remove nodes though, you should
 * probably use this method.  If you look at this method's source
 * however, you will see that the <code>cgul_rbtree_node</code> class
 * provides all the public methods you need to safely remove nodes
 * while you are iterating over the rbtree, but you do have to be
 * careful.
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] first first node in range
 * @param[in] last last node in range
 * @param[in] f traversal callback function
 * @param[in] data client data passed to <code>f</code>
 * @see cgul_rbtree__traverse
 */
extern CGUL_EXPORT void
cgul_rbtree__traverse_range(cgul_exception_t* cex,
                            cgul_rbtree_t t,
                            cgul_rbtree_node_t first,
                            cgul_rbtree_node_t last,
                            cgul_rbtree__traverse_t f,
                            void* data);

CGUL_END_C

#endif  /*  CGUL_RBTREE_H  */
