#ifndef CGUL_ROLLING_HASH_CXX_H
#define CGUL_ROLLING_HASH_CXX_H

/**
 * C++ wrapper for <code>cgul_rolling_hash</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_rolling_hash</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_rolling_hash.h"

/**
 * This class provides the C++ bindings for <code>cgul_rolling_hash</code>.
 * The main purpose of this class is to convert the C-style function
 * calls and exception handling in <code>cgul_rolling_hash</code> into
 * C++-style function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_rolling_hash</code>
 */
class cgul_rolling_hash_cxx {

public:

    /**
     * This function calculates the hash value for <code>length</code>
     * bytes in the block <code>block</code>.  The main purpose of
     * this function is to generate the initial hash value that can
     * then be passed into <code>roll()</code>, but it can also be
     * used as a stand-alone hash function. <p>
     *
     * @param[in] block block of binary data
     * @param[in] block_length length of <code>block</code> in bytes
     * @return hash value
     * @see roll()
     * @see cgul_hash_cxx::hash_block()
     */
    static cgul_uint32_t
    hash(const char* block,
         size_t block_length)
    {
        cgul_uint32_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_rolling_hash__hash(&local, block, block_length);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Peform a rolling hash calculation on the hash value
     * <code>hv</code> by surgically removing the part of the hash
     * value associated with the old character <code>c_old</code> that
     * is being removed and adding the part of the hash value
     * associated with the new character <code>c_new</code> that is
     * being added.  The initial value for <code>hv</code> should be
     * obtained by calling <code>hash()</code>.  The value for
     * <code>block_length</code> should be same as what originally was
     * passed into <code>hash()</code>. <p>
     *
     * @param[in] hv old hash value
     * @param[in] c_old old character
     * @param[in] c_new new character
     * @param[in] block_length length of the block in bytes
     * @return new hash value
     * @see hash()
     */
    static cgul_uint32_t
    roll(cgul_uint32_t hv,
         const char c_old,
         const char c_new,
         size_t block_length)
    {
        cgul_uint32_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_rolling_hash__roll(&local, hv, c_old, c_new, block_length);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

};

#endif  /*  CGUL_ROLLING_HASH_CXX_H  */
