#ifndef CGUL_SHUFFLE_CXX_H
#define CGUL_SHUFFLE_CXX_H

/**
 * C++ wrapper for <code>cgul_shuffle</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_shuffle</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_shuffle.h"

/**
 * This class provides the C++ bindings for <code>cgul_shuffle</code>.
 * The main purpose of this class is to convert the C-style function
 * calls and exception handling into C++-style function calls and
 * exception handling.
 *
 * @brief C++ bindings for <code>cgul_shuffle</code>
 */
class cgul_shuffle_cxx {

public:

    /**
     * This type definition defines the interface to a pseudo-random
     * number generator (PRNG) which the shuffle functions use to generate
     * pseudo-random numbers that are used to shuffle an array.  The
     * quality of the shuffle depends entirely upon the quality of the
     * pseudo-random numbers returned. <p>
     *
     * Functions that implement this interface will be passed the same
     * client data <code>client_data</code> that is passed into the
     * shuffle function.  The return value should be a pseudo-random
     * integer in the half-open interval <code>[0, range_max)</code>.  If
     * an error occurs when this function is called, the client should
     * propagate the error via <code>cex</code>. <p>
     *
     * Note that this interface closely resembles the signature of the
     * <code>cgul_random_*__next_in_range()</code> functions.  So closely
     * in fact, that it is usually not necessary to write a special
     * callback to implement this function.  Instead, this interface can
     * be satisfied by casting <code>cgul_random_*__next_in_range()</code>
     * to have <code>next_in_range_t</code> function type
     * and pass in your PRNG object as the client data. <p>
     */
    typedef cgul_shuffle__next_in_range_t next_in_range_t;

    /**
     * Perform an in-place Fisher-Yates shuffle on the array
     * <code>v</code> having <code>v_count</code> arbitrary elements of
     * size <code>element_size</code> (in bytes).  If <code>v</code> holds
     * integers or pointers, one of the other shuffle functions tailored
     * to the specific type will perform better.  The quality of the
     * shuffle depends entirely on the quality of the random numbers
     * returned by the <code>next_in_range</code> callback which is passed
     * the client data <code>client_data</code> unmodified.  Because this
     * function assumes a 32-bit PRNG, <code>v_count</code> can indicate
     * no more than (2^32 - 1) elements.  If an error occurs, an exception
     * is thrown.
     *
     * @param[in] v array of elements to shuffle
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] element_size size of each element in <code>v</code> (in bytes)
     * @param[in] next_in_range PRNG callback
     * @param[in] client_data client data for the PRNG callback
     */
    static void
    cgul_shuffle(void* v,
                 size_t v_count,
                 size_t element_size,
                 next_in_range_t next_in_range,
                 void* client_data)
    {
        cgul_exception_t local = NULL;
        ::cgul_shuffle(&local,
                       v,
                       v_count,
                       element_size,
                       next_in_range,
                       client_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place Fisher-Yates shuffle on the array
     * <code>v</code> having <code>v_count</code> <code>void*</code>
     * elements.  The quality of the shuffle depends entirely on the
     * quality of the random numbers returned by the
     * <code>next_in_range</code> callback which is passed the client data
     * <code>client_data</code> unmodified.  Because this function assumes
     * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
     * - 1) elements.  If an error occurs, an exception is thrown.
     *
     * @param[in] v array of elements to shuffle
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] next_in_range PRNG callback
     * @param[in] client_data client data for the PRNG callback
     */
    static void
    cgul_shuffle_pointers(void** v,
                          size_t v_count,
                          next_in_range_t next_in_range,
                          void* client_data)
    {
        cgul_exception_t local = NULL;
        ::cgul_shuffle_pointers(&local,
                                v,
                                v_count,
                                next_in_range,
                                client_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place Fisher-Yates shuffle on the array
     * <code>v</code> having <code>v_count</code> elements of type
     * <code>signed char</code>.  The quality of the shuffle depends
     * entirely on the quality of the random numbers returned by the
     * <code>next_in_range</code> callback which is passed the client data
     * <code>client_data</code> unmodified.  Because this function assumes
     * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
     * - 1) elements.  If an error occurs, an exception is thrown.
     *
     * @param[in] v array of elements to shuffle
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] next_in_range PRNG callback
     * @param[in] client_data client data for the PRNG callback
     */
    static void
    cgul_shuffle_schar(signed char* v,
                       size_t v_count,
                       next_in_range_t next_in_range,
                       void* client_data)
    {
        cgul_exception_t local = NULL;
        ::cgul_shuffle_schar(&local, v, v_count, next_in_range, client_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place Fisher-Yates shuffle on the array
     * <code>v</code> having <code>v_count</code> elements of type
     * <code>unsigned char</code>.  The quality of the shuffle depends
     * entirely on the quality of the random numbers returned by the
     * <code>next_in_range</code> callback which is passed the client data
     * <code>client_data</code> unmodified.  Because this function assumes
     * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
     * - 1) elements.  If an error occurs, an exception is thrown.
     *
     * @param[in] v array of elements to shuffle
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] next_in_range PRNG callback
     * @param[in] client_data client data for the PRNG callback
     */
    static void
    cgul_shuffle_uchar(unsigned char* v,
                       size_t v_count,
                       next_in_range_t next_in_range,
                       void* client_data)
    {
        cgul_exception_t local = NULL;
        ::cgul_shuffle_uchar(&local, v, v_count, next_in_range, client_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place Fisher-Yates shuffle on the array
     * <code>v</code> having <code>v_count</code> elements of type
     * <code>short</code>.  The quality of the shuffle depends entirely on
     * the quality of the random numbers returned by the
     * <code>next_in_range</code> callback which is passed the client data
     * <code>client_data</code> unmodified.  Because this function assumes
     * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
     * - 1) elements.  If an error occurs, an exception is thrown.
     *
     * @param[in] v array of elements to shuffle
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] next_in_range PRNG callback
     * @param[in] client_data client data for the PRNG callback
     */
    static void
    cgul_shuffle_short(short int* v,
                       size_t v_count,
                       next_in_range_t next_in_range,
                       void* client_data)
    {
        cgul_exception_t local = NULL;
        ::cgul_shuffle_short(&local, v, v_count, next_in_range, client_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place Fisher-Yates shuffle on the array
     * <code>v</code> having <code>v_count</code> elements of type
     * <code>unsigned short</code>.  The quality of the shuffle depends
     * entirely on the quality of the random numbers returned by the
     * <code>next_in_range</code> callback which is passed the client data
     * <code>client_data</code> unmodified.  Because this function assumes
     * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
     * - 1) elements.  If an error occurs, an exception is thrown.
     *
     * @param[in] v array of elements to shuffle
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] next_in_range PRNG callback
     * @param[in] client_data client data for the PRNG callback
     */
    static void
    cgul_shuffle_ushort(unsigned short int* v,
                        size_t v_count,
                        next_in_range_t next_in_range,
                        void* client_data)
    {
        cgul_exception_t local = NULL;
        ::cgul_shuffle_ushort(&local, v, v_count, next_in_range, client_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place Fisher-Yates shuffle on the array
     * <code>v</code> having <code>v_count</code> elements of type
     * <code>int</code>.  The quality of the shuffle depends entirely on
     * the quality of the random numbers returned by the
     * <code>next_in_range</code> callback which is passed the client data
     * <code>client_data</code> unmodified.  Because this function assumes
     * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
     * - 1) elements.  If an error occurs, an exception is thrown.
     *
     * @param[in] v array of elements to shuffle
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] next_in_range PRNG callback
     * @param[in] client_data client data for the PRNG callback
     */
    static void
    cgul_shuffle_int(int* v,
                     size_t v_count,
                     next_in_range_t next_in_range,
                     void* client_data)
    {
        cgul_exception_t local = NULL;
        ::cgul_shuffle_int(&local, v, v_count, next_in_range, client_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place Fisher-Yates shuffle on the array
     * <code>v</code> having <code>v_count</code> elements of type
     * <code>unsigned int</code>.  The quality of the shuffle depends
     * entirely on the quality of the random numbers returned by the
     * <code>next_in_range</code> callback which is passed the client data
     * <code>client_data</code> unmodified.  Because this function assumes
     * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
     * - 1) elements.  If an error occurs, an exception is thrown.
     *
     * @param[in] v array of elements to shuffle
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] next_in_range PRNG callback
     * @param[in] client_data client data for the PRNG callback
     */
    static void
    cgul_shuffle_uint(unsigned int* v,
                      size_t v_count,
                      next_in_range_t next_in_range,
                      void* client_data)
    {
        cgul_exception_t local = NULL;
        ::cgul_shuffle_uint(&local, v, v_count, next_in_range, client_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place Fisher-Yates shuffle on the array
     * <code>v</code> having <code>v_count</code> elements of type
     * <code>long</code>.  The quality of the shuffle depends entirely on
     * the quality of the random numbers returned by the
     * <code>next_in_range</code> callback which is passed the client data
     * <code>client_data</code> unmodified.  Because this function assumes
     * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
     * - 1) elements.  If an error occurs, an exception is thrown.
     *
     * @param[in] v array of elements to shuffle
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] next_in_range PRNG callback
     * @param[in] client_data client data for the PRNG callback
     */
    static void
    cgul_shuffle_long(long int* v,
                      size_t v_count,
                      next_in_range_t next_in_range,
                      void* client_data)
    {
        cgul_exception_t local = NULL;
        ::cgul_shuffle_long(&local, v, v_count, next_in_range, client_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place Fisher-Yates shuffle on the array
     * <code>v</code> having <code>v_count</code> elements of type
     * <code>unsigned long</code>.  The quality of the shuffle depends
     * entirely on the quality of the random numbers returned by the
     * <code>next_in_range</code> callback which is passed the client data
     * <code>client_data</code> unmodified.  Because this function assumes
     * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
     * - 1) elements.  If an error occurs, an exception is thrown.
     *
     * @param[in] v array of elements to shuffle
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] next_in_range PRNG callback
     * @param[in] client_data client data for the PRNG callback
     */
    static void
    cgul_shuffle_ulong(unsigned long int* v,
                       size_t v_count,
                       next_in_range_t next_in_range,
                       void* client_data)
    {
        cgul_exception_t local = NULL;
        ::cgul_shuffle_ulong(&local, v, v_count, next_in_range, client_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place Fisher-Yates shuffle on the array
     * <code>v</code> having <code>v_count</code> elements of type
     * <code>cgul_int8_t</code>.  The quality of the shuffle depends
     * entirely on the quality of the random numbers returned by the
     * <code>next_in_range</code> callback which is passed the client data
     * <code>client_data</code> unmodified.  Because this function assumes
     * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
     * - 1) elements.  If an error occurs, an exception is thrown.
     *
     * @param[in] v array of elements to shuffle
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] next_in_range PRNG callback
     * @param[in] client_data client data for the PRNG callback
     */
    static void
    cgul_shuffle_int8(cgul_int8_t* v,
                      size_t v_count,
                      next_in_range_t next_in_range,
                      void* client_data)
    {
        cgul_exception_t local = NULL;
        ::cgul_shuffle_int8(&local, v, v_count, next_in_range, client_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place Fisher-Yates shuffle on the array
     * <code>v</code> having <code>v_count</code> elements of type
     * <code>cgul_uint8_t</code>.  The quality of the shuffle depends
     * entirely on the quality of the random numbers returned by the
     * <code>next_in_range</code> callback which is passed the client data
     * <code>client_data</code> unmodified.  Because this function assumes
     * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
     * - 1) elements.  If an error occurs, an exception is thrown.
     *
     * @param[in] v array of elements to shuffle
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] next_in_range PRNG callback
     * @param[in] client_data client data for the PRNG callback
     */
    static void
    cgul_shuffle_uint8(cgul_uint8_t* v,
                       size_t v_count,
                       next_in_range_t next_in_range,
                       void* client_data)
    {
        cgul_exception_t local = NULL;
        ::cgul_shuffle_uint8(&local, v, v_count, next_in_range, client_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place Fisher-Yates shuffle on the array
     * <code>v</code> having <code>v_count</code> elements of type
     * <code>cgul_int16_t</code>.  The quality of the shuffle depends
     * entirely on the quality of the random numbers returned by the
     * <code>next_in_range</code> callback which is passed the client data
     * <code>client_data</code> unmodified.  Because this function assumes
     * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
     * - 1) elements.  If an error occurs, an exception is thrown.
     *
     * @param[in] v array of elements to shuffle
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] next_in_range PRNG callback
     * @param[in] client_data client data for the PRNG callback
     */
    static void
    cgul_shuffle_int16(cgul_int16_t* v,
                       size_t v_count,
                       next_in_range_t next_in_range,
                       void* client_data)
    {
        cgul_exception_t local = NULL;
        ::cgul_shuffle_int16(&local, v, v_count, next_in_range, client_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place Fisher-Yates shuffle on the array
     * <code>v</code> having <code>v_count</code> elements of type
     * <code>cgul_uint16_t</code>.  The quality of the shuffle depends
     * entirely on the quality of the random numbers returned by the
     * <code>next_in_range</code> callback which is passed the client data
     * <code>client_data</code> unmodified.  Because this function assumes
     * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
     * - 1) elements.  If an error occurs, an exception is thrown.
     *
     * @param[in] v array of elements to shuffle
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] next_in_range PRNG callback
     * @param[in] client_data client data for the PRNG callback
     */
    static void
    cgul_shuffle_uint16(cgul_uint16_t* v,
                        size_t v_count,
                        next_in_range_t next_in_range,
                        void* client_data)
    {
        cgul_exception_t local = NULL;
        ::cgul_shuffle_uint16(&local, v, v_count, next_in_range, client_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place Fisher-Yates shuffle on the array
     * <code>v</code> having <code>v_count</code> elements of type
     * <code>cgul_int32_t</code>.  The quality of the shuffle depends
     * entirely on the quality of the random numbers returned by the
     * <code>next_in_range</code> callback which is passed the client data
     * <code>client_data</code> unmodified.  Because this function assumes
     * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
     * - 1) elements.  If an error occurs, an exception is thrown.
     *
     * @param[in] v array of elements to shuffle
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] next_in_range PRNG callback
     * @param[in] client_data client data for the PRNG callback
     */
    static void
    cgul_shuffle_int32(cgul_int32_t* v,
                       size_t v_count,
                       next_in_range_t next_in_range,
                       void* client_data)
    {
        cgul_exception_t local = NULL;
        ::cgul_shuffle_int32(&local, v, v_count, next_in_range, client_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform an in-place Fisher-Yates shuffle on the array
     * <code>v</code> having <code>v_count</code> elements of type
     * <code>cgul_uint32_t</code>.  The quality of the shuffle depends
     * entirely on the quality of the random numbers returned by the
     * <code>next_in_range</code> callback which is passed the client data
     * <code>client_data</code> unmodified.  Because this function assumes
     * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
     * - 1) elements.  If an error occurs, an exception is thrown.
     *
     * @param[in] v array of elements to shuffle
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] next_in_range PRNG callback
     * @param[in] client_data client data for the PRNG callback
     */
    static void
    cgul_shuffle_uint32(cgul_uint32_t* v,
                        size_t v_count,
                        next_in_range_t next_in_range,
                        void* client_data)
    {
        cgul_exception_t local = NULL;
        ::cgul_shuffle_uint32(&local, v, v_count, next_in_range, client_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    #ifdef CGUL_INT64
    /**
     * Perform an in-place Fisher-Yates shuffle on the array
     * <code>v</code> having <code>v_count</code> elements of type
     * <code>cgul_int64_t</code>.  The quality of the shuffle depends
     * entirely on the quality of the random numbers returned by the
     * <code>next_in_range</code> callback which is passed the client data
     * <code>client_data</code> unmodified.  Because this function assumes
     * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
     * - 1) elements.  If an error occurs, an exception is thrown.
     *
     * @param[in] v array of elements to shuffle
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] next_in_range PRNG callback
     * @param[in] client_data client data for the PRNG callback
     */
    static void
    cgul_shuffle_int64(cgul_int64_t* v,
                       size_t v_count,
                       next_in_range_t next_in_range,
                       void* client_data)
    {
        cgul_exception_t local = NULL;
        ::cgul_shuffle_int64(&local, v, v_count, next_in_range, client_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }
    #endif /* CGUL_INT64 */

    #ifdef CGUL_UINT64
    /**
     * Perform an in-place Fisher-Yates shuffle on the array
     * <code>v</code> having <code>v_count</code> elements of type
     * <code>cgul_uint64_t</code>.  The quality of the shuffle depends
     * entirely on the quality of the random numbers returned by the
     * <code>next_in_range</code> callback which is passed the client data
     * <code>client_data</code> unmodified.  Because this function assumes
     * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
     * - 1) elements.  If an error occurs, an exception is thrown.
     *
     * @param[in] v array of elements to shuffle
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] next_in_range PRNG callback
     * @param[in] client_data client data for the PRNG callback
     */
    static void
    cgul_shuffle_uint64(cgul_uint64_t* v,
                        size_t v_count,
                        next_in_range_t next_in_range,
                        void* client_data)
    {
        cgul_exception_t local = NULL;
        ::cgul_shuffle_uint64(&local, v, v_count, next_in_range, client_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
}
    #endif /* CGUL_UINT64 */

};

#endif  /*  CGUL_SHUFFLE_CXX_H  */
