#ifndef CGUL_STRING_H
#define CGUL_STRING_H

/**
 * Simple string class.  The idea behind this class is to not co-opt
 * the role of c-style strings.  Rather the idea is that you can use
 * this class to safely generate c-style strings and then use
 * <code>cgul_string__take_value()</code> (or maybe
 * <code>cgul_string__get_value()</code>) once your string has the
 * desired value.  You can then use the resulting <code>char*</code>
 * with any of the standard c library calls that require a string.
 *
 * @file
 * @author Paul Serice
 * @brief string class
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_hchar.h"
#include "cgul_wchar.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_string</code> instance. */
typedef struct cgul_string* cgul_string_t;

/**
 * Create a new <code>cgul_string</code> object that holds an empty
 * string.  The caller is responsible for freeing the object by
 * calling <code>cgul_string__delete()</code>.  If memory cannot be
 * allocated, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * The following three methods implement the "cacheable" interface
 * defined by <code>cgul_cache</code>:
 *
 * <pre>
 *
 *     cgul_string__new()
 *     cgul_string__delete()
 *     cgul_string__clear()
 *
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_string</code> instance
 */
extern CGUL_EXPORT cgul_string_t
cgul_string__new(cgul_exception_t* cex);

/**
 * Create a new <code>cgul_string</code> object that is initialized
 * with a copy of <code>value</code>.  The caller is responsible for
 * freeing the object by calling <code>cgul_string__delete()</code>.
 * If memory cannot be allocated, <code>NULL</code> is returned, and
 * an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] value initial value
 * @return new <code>cgul_string</code> instance
 */
extern CGUL_EXPORT cgul_string_t
cgul_string__new_from_cstring(cgul_exception_t* cex,
                              const char* value);

/**
 * This method frees all internally allocated memory.  Do not try to
 * use <code>s</code> after calling this method.
 *
 * @param[in] s <code>cgul_string</code> instance
 */
extern CGUL_EXPORT void
cgul_string__delete(cgul_string_t s);

/**
 * This method returns a pointer to the internal C-style string.  You
 * are allowed to manually set individual characters in the string.
 * Just remember to call
 * <code>cgul_string__externally_adjusted()</code> afterward.  Compare
 * this method to <code>cgul_string__take_value()</code>.
 * 
 * @param[in] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @return underlying c-style string
 * @see cgul_string__externally_adjusted
 */
extern CGUL_EXPORT char*
cgul_string__get_value(cgul_exception_t* cex,
                       cgul_string_t s);

/**
 * This method lets you take ownership of the underlying C-style
 * string which is returned.  The client is responsible for calling
 * <code>free()</code> on the pointer returned. <p>
 *
 * Just prior to taking ownership, the internal state of
 * <code>s</code> is reset.  If an error occurs when attempting to
 * reset <code>s</code>, <code>s</code> is left unaltered,
 * <code>NULL</code> is returned, and an exception is thrown. <p>
 *
 * One use of this function is to build up a string using
 * <code>cgul_string</code> that you then pass to
 * <code>cgul_hash</code> or <code>cgul_rbtree</code>.  Compare this
 * function with <code>cgul_string__get_value()</code>. <p>
 * 
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @return underlying c-style string
 */
extern CGUL_EXPORT char*
cgul_string__take_value(cgul_exception_t* cex,
                        cgul_string_t s);

/**
 * Set the value of the underlying c-style string by copying
 * characters from <code>value</code> to <code>s</code>.  The caller
 * is responsible for making sure <code>value</code> is terminated
 * with a <code>NULL</code> character.  This method does not uses
 * <code>strcpy()</code> or <code>memcpy()</code>.  This means that
 * <code>value</code> and <code>s</code> can overlap.
 * 
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] value new value
 */
extern CGUL_EXPORT void
cgul_string__set_value(cgul_exception_t* cex,
                       cgul_string_t s,
                       const char* value);

/**
 * This method returns the character in <code>s</code> indicated by
 * <code>index</code> where <code>index</code> is zero-based.  If
 * <code>index</code> is out of bounds, <code>(char)0</code> is
 * returned and an exception is thrown.
 * 
 * @param[in] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] index index into <code>s</code>
 * @return character at <code>index</code>
 */
extern CGUL_EXPORT char
cgul_string__get_value_at(cgul_exception_t* cex,
                          cgul_string_t s,
                          size_t index);

/**
 * This method sets the character in <code>s</code> indicated by
 * <code>index</code> where <code>index</code> is zero-based to
 * <code>c</code>.  If <code>index</code> is out of bounds, an
 * exception is thrown.
 * 
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] index index into <code>s</code>
 * @param[in] c character
 */
extern CGUL_EXPORT void
cgul_string__set_value_at(cgul_exception_t* cex,
                          cgul_string_t s,
                          size_t index,
                          char c);

/**
 * This method is equivalent to calling
 * <code>cgul_string__set_value()</code> where <code>value</code> is
 * an empty string.  This method is designed to be used as the
 * <code>cgul_cache_restructor_t</code> method.  This means it is not
 * allowed to throw an exception.
 *
 * @param[in] s <code>cgul_string</code> instance
 */
extern CGUL_EXPORT void
cgul_string__clear(cgul_string_t s);

/**
 * This method converts the string to a <code>signed char</code>.
 * Typically, "base" will be set to 0 to allow conversion for decimal,
 * hexadecimal, and octal based on the leading character in the
 * string.  The string cannot have any extraneous characters.  If it
 * does, the conversion will fail.  On success, the converted value is
 * returned; otherwise, 0 is returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] base base to use for the conversion
 * @return converted value
 * @see cgul_string__get_value_at()
 */
extern CGUL_EXPORT signed char
cgul_string__to_schar(cgul_exception_t* cex,
                      cgul_string_t s,
                      int base);

/**
 * This method converts the string to an <code>unsigned char</code>.
 * Typically, <code>base</code> will be set to 0 to allow conversion
 * for decimal, hexadecimal, and octal based on the leading characters
 * in the string.  The string cannot have any extraneous characters.
 * If it does, the conversion will fail.  On success, the converted
 * value is returned; otherwise, 0 is returned and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] base base to use for the conversion
 * @return converted value
 * @see cgul_string__get_value_at()
 */
extern CGUL_EXPORT unsigned char
cgul_string__to_uchar(cgul_exception_t* cex,
                      cgul_string_t s,
                      int base);

/**
 * This method converts the string to a <code>short</code>.
 * Typically, "base" will be set to 0 to allow conversion for decimal,
 * hexadecimal, and octal based on the leading character in the
 * string.  The string cannot have any extraneous characters.  If it
 * does, the conversion will fail.  On success, the converted value is
 * returned; otherwise, 0 is returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] base base to use for the conversion
 * @return converted value
 */
extern CGUL_EXPORT short int
cgul_string__to_short(cgul_exception_t* cex,
                      cgul_string_t s,
                      int base);

/**
 * This method converts the string to an <code>unsigned short</code>.
 * Typically, <code>base</code> will be set to 0 to allow conversion
 * for decimal, hexadecimal, and octal based on the leading characters
 * in the string.  The string cannot have any extraneous characters.
 * If it does, the conversion will fail.  On success, the converted
 * value is returned; otherwise, 0 is returned and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] base base to use for the conversion
 * @return converted value
 */
extern CGUL_EXPORT unsigned short int
cgul_string__to_ushort(cgul_exception_t* cex,
                       cgul_string_t s,
                       int base);

/**
 * This method converts the string to an <code>int</code>.  Typically,
 * "base" will be set to 0 to allow conversion for decimal,
 * hexadecimal, and octal based on the leading character in the
 * string.  The string cannot have any extraneous characters.  If it
 * does, the conversion will fail.  On success, the converted value is
 * returned; otherwise, 0 is returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] base base to use for the conversion
 * @return converted value
 */
extern CGUL_EXPORT int
cgul_string__to_int(cgul_exception_t* cex,
                    cgul_string_t s,
                    int base);

/**
 * This method converts the string to an <code>unsigned</code>.
 * Typically, <code>base</code> will be set to 0 to allow conversion
 * for decimal, hexadecimal, and octal based on the leading characters
 * in the string.  The string cannot have any extraneous characters.
 * If it does, the conversion will fail.  On success, the converted
 * value is returned; otherwise, 0 is returned and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] base base to use for the conversion
 * @return converted value
 */
extern CGUL_EXPORT unsigned int
cgul_string__to_unsigned(cgul_exception_t* cex,
                         cgul_string_t s,
                         int base);

/**
 * This method converts the string to a <code>long</code>.  Typically,
 * <code>base</code> will be set to 0 to allow conversion for decimal,
 * hexadecimal, and octal based on the leading characters in the
 * string.  The string cannot have any extraneous characters.  If it
 * does, the conversion will fail.  On success, the converted value is
 * returned; otherwise, 0 is returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] base base to use for the conversion
 * @return converted value
 */
extern CGUL_EXPORT long int
cgul_string__to_long(cgul_exception_t* cex,
                     cgul_string_t s,
                     int base);

/**
 * This method converts the string to an <code>unsigned long</code>.
 * Typically, <code>base</code> will be set to 0 to allow conversion
 * for decimal, hexadecimal, and octal based on the leading characters
 * in the string.  The string cannot have any extraneous characters.
 * If it does, the conversion will fail.  On success, the converted
 * value is returned; otherwise, 0 is returned and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] base base to use for the conversion
 * @return converted value
 */
extern CGUL_EXPORT unsigned long int
cgul_string__to_unsigned_long(cgul_exception_t* cex,
                              cgul_string_t s,
                              int base);

/**
 * This method converts the string to a <code>double</code>.  The
 * string cannot have any extraneous characters.  If it does, the
 * conversion will fail.  On success, the converted value is returned;
 * otherwise, 0.0 is returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @return converted value
 */
extern CGUL_EXPORT double
cgul_string__to_double(cgul_exception_t* cex,
                       cgul_string_t s);

/**
 * Appends <code>value</code> to <code>s</code>.  The implementation
 * of this method is such that it is possible for <code>value</code>
 * to be a substring of <code>s</code>.  If an error occurs, an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] value value to append
 */
extern CGUL_EXPORT void
cgul_string__append(cgul_exception_t* cex,
                    cgul_string_t s,
                    const char* value);

/**
 * Append a block of non-NUL characters.  The caller should verify
 * that there are no embedded NUL characters in the string.  If an
 * error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] block block to append
 * @param[in] block_size size of block to append
 */
extern CGUL_EXPORT void
cgul_string__append_block(cgul_exception_t* cex,
                          cgul_string_t s,
                          const void* block,
                          size_t block_size);

/**
 * This function is a thin wrapper around
 * <code>cgul_string__append_vsprintf()</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param format format string
 * @param ... variable number of arguments that satisfy <code>format</code>
 * @return number of characters appended
 */
extern CGUL_EXPORT int
cgul_string__append_sprintf(cgul_exception_t* cex,
                            cgul_string_t s,
                            const char* format,
                            ...)
#if defined(__GNUC__) && !defined(__MINGW_H)
    __attribute__ ((format (printf, 3, 4)))
#endif
;

/**
 * This function substitutes the conversion specifications in the
 * <code>format</code> string and appends them to <code>s</code>.  If
 * an error occurs, an exception is thrown. <p>
 *
 * The format string is the same as what you pass into
 * <code>sprintf()</code> or <code>snprintf()</code>.  The following
 * length modifiers are honored: l, h, z. <p>
 *
 * Any string associated with a "%s" conversion specification is
 * assumed to be a UTF-8 string.  The main difference that this makes
 * is that, if a field width is specified, it specifies the number of
 * characters, not bytes, in the field. <p>
 *
 * The "%ls" conversion specification is honored, but instead of
 * converting a string composed of <code>wchar_t</code> characters to
 * a multi-byte string encoded according to the current locale, it
 * converts a string composed of Unicode UTF-32
 * <code>cgul_wchar_t</code> characters to a UTF-8 string.  The "%ls"
 * conversions also differ in that, if a precision is given, it
 * determines the number of characters appended not the number of
 * bytes. <p>
 *
 * The "%lc" conversion specification is honored, but like the "%ls"
 * conversion specification, it converts from UTF-32 to UTF-8
 * exclusively.  Also, it expects the associated argument to be
 * <code>cgul_wchar_t</code>, not <code>wchar_t</code>. <p>
 *
 * The GNU "%m" conversion specification is honored.  This conversion
 * specification appends the result of
 * <code>strerror(errno)</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param format format string
 * @param args variable number of arguments that satisfy <code>format</code>
 * @return number of characters appended
 */
extern CGUL_EXPORT int
cgul_string__append_vsprintf(cgul_exception_t* cex,
                             cgul_string_t s,
                             const char* format,
                             va_list args);

/**
 * Append <code>c</code> to <code>s</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] c character to append
 */
extern CGUL_EXPORT void
cgul_string__append_char(cgul_exception_t* cex,
                         cgul_string_t s,
                         char c);

/**
 * Append the UTF-16 character <code>utf16</code> to <code>s</code> as
 * a UTF-8 sequence.  Because <code>utf16</code> could be a surrogate
 * pair, to avoid reading off the end of the sequence, the number of
 * <code>cgul_hchar_t</code> elements (i.e., 16-bit elements) in the
 * <code>utf16</code> buffer must be specified using
 * <code>utf16_length</code>.  If more than one UTF-16 character is
 * present in the <code>utf16</code> buffer, only the first character
 * is appended.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] utf16 UTF-16 character to append
 * @param[in] utf16_length length of <code>utf16</code> in
 * <code>cgul_hchar_t</code> elements
 */
extern CGUL_EXPORT void
cgul_string__append_hchar(cgul_exception_t* cex,
                          cgul_string_t s,
                          cgul_hchar_t* utf16,
                          size_t utf16_length);

/**
 * Append the UTF-32 character <code>utf32</code> to <code>s</code> as
 * a UTF-8 sequence.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] utf32 UTF-32 character to append
 */
extern CGUL_EXPORT void
cgul_string__append_wchar(cgul_exception_t* cex,
                          cgul_string_t s,
                          cgul_wchar_t utf32);

/**
 * Append <code>value</code> to <code>s</code> in decimal format.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] value to append
 */
extern CGUL_EXPORT void
cgul_string__append_decimal(cgul_exception_t* cex,
                            cgul_string_t s,
                            long int value);

/**
 * Append <code>value</code> to <code>s</code> in decimal format.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] value to append
 */
extern CGUL_EXPORT void
cgul_string__append_unsigned_decimal(cgul_exception_t* cex,
                                     cgul_string_t s,
                                     size_t value);

/**
 * Append <code>value</code> to <code>s</code> in octal format.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] value to append
 */
extern CGUL_EXPORT void
cgul_string__append_octal(cgul_exception_t* cex,
                          cgul_string_t s,
                          size_t value);

/**
 * Append <code>value</code> to <code>s</code> in hexadecimal format
 * using "abcdef".
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] value to append
 * @param[in] caps whether to capitalize the hex digits
 */
extern CGUL_EXPORT void
cgul_string__append_hexadecimal(cgul_exception_t* cex,
                                cgul_string_t s,
                                size_t value,
                                int caps);

/**
 * Append <code>value</code> to <code>s</code>.  "precision"
 * determines the number of significant digits that are printed.
 * Zeros after the decimal point at the end of number are not printed.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] value to append
 * @param[in] precision precision
 */
extern CGUL_EXPORT void
cgul_string__append_double(cgul_exception_t* cex,
                           cgul_string_t s,
                           double value,
                           int precision);

/**
 * Append <code>ptr</code> to <code>s</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] ptr pointer to append
 */
extern CGUL_EXPORT void
cgul_string__append_pointer(cgul_exception_t* cex,
                            cgul_string_t s,
                            const void* ptr);

/**
 * Inserts <code>value</code> into <code>s</code> at
 * <code>index</code> which can range from 0 to the length (inclusive)
 * of <code>s</code>.  If <code>index</code> is out of bounds or if
 * memory cannot be allocated, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] index index in <code>s</code> where <code>value</code> is inserted
 * @param[in] value value to insert
 * @param[in] value_length length of value
 */
extern CGUL_EXPORT void
cgul_string__insert(cgul_exception_t* cex,
                    cgul_string_t s,
                    size_t index,
                    const char* value,
                    size_t value_length);

/**
 * Inserts <code>value</code> into <code>s</code> at
 * <code>index</code> which can range from 0 to the length (inclusive)
 * of <code>s</code>.  If <code>index</code> is out of bounds or if
 * memory cannot be allocated, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] index index in <code>s</code> where <code>value</code> is inserted
 * @param[in] value value to insert
 */
extern CGUL_EXPORT void
cgul_string__insert_char(cgul_exception_t* cex,
                         cgul_string_t s,
                         size_t index,
                         char value);

/**
 * Pad the current string in <code>s</code> such that it is at least
 * <code>width</code> characters (not bytes) wide using the
 * <code>pad</code> character.  <code>s</code> will be right aligned.
 * Note that this method assumes that <code>s</code> is in UTF-8
 * format and counts characters accordingly.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] width padding width
 * @param[in] pad character to use for padding
 */
extern CGUL_EXPORT void
cgul_string__pad_right(cgul_exception_t* cex,
                       cgul_string_t s,
                       size_t width,
                       cgul_wchar_t pad);

/**
 * Pad the current string in <code>s</code> such that it is at least
 * <code>width</code> characters (not bytes) wide using the
 * <code>pad</code> character.  <code>s</code> will be left aligned.
 * Note that this method assumes that <code>s</code> is in UTF-8
 * format and counts characters accordingly.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] width padding width
 * @param[in] pad character to use for padding
 */
extern CGUL_EXPORT void
cgul_string__pad_left(cgul_exception_t* cex,
                      cgul_string_t s,
                      size_t width,
                      cgul_wchar_t pad);

/**
 * Trim the string by removing leading and trailing white-space where
 * white-space is definded by <code>cgul_char__isspace()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 */
extern CGUL_EXPORT void
cgul_string__trim(cgul_exception_t* cex,
                  cgul_string_t s);

/**
 * Trim the string by removing leading white-space where white-space
 * is definded by <code>cgul_char__isspace()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 */
extern CGUL_EXPORT void
cgul_string__trim_left(cgul_exception_t* cex,
                       cgul_string_t s);

/**
 * Trim the string by removing trailing white-space where white-space
 * is definded by <code>cgul_char__isspace()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 */
extern CGUL_EXPORT void
cgul_string__trim_right(cgul_exception_t* cex,
                        cgul_string_t s);

/**
 * Reverse the order of the characters in <code>s</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 */
extern CGUL_EXPORT void
cgul_string__reverse(cgul_exception_t* cex,
                     cgul_string_t s);

/**
 * This method returns the length of the string <code>s</code>in
 * bytes.  If <code>s</code> is a UTF-8 string, call
 * <code>cgul_unicode__get_char_count()</code> to get the length of
 * the string in characters.
 *
 * @param[in] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @return length of the string
 */
extern CGUL_EXPORT size_t
cgul_string__get_length(cgul_exception_t* cex,
                        cgul_string_t s);

/**
 * This method can be used to set the length of <code>s</code>.  It
 * can only be used to make <code>s</code> shorter.  If it is used in
 * an attempt to make <code>s</code> longer, an exception is thrown,
 * and zero is returned.  If the length of <code>s</code> is
 * successfully changed, the previous length of the string is
 * returned.
 *
 * @param[in] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] length new, shorter length of the string
 * @return previous length of the string
 */
extern CGUL_EXPORT size_t
cgul_string__set_length(cgul_exception_t* cex,
                        cgul_string_t s,
                        size_t length);

/**
 * If you manually alter the underlying buffer by modifying the
 * <code>char*</code> string returned by
 * <code>cgul_string__get_value()</code>, you'll need to manually have
 * the class reset its internal state.
 *
 * @param[in] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @see cgul_string__get_value
 */
extern CGUL_EXPORT void
cgul_string__externally_adjusted(cgul_exception_t* cex,
                                 cgul_string_t s);

/**
 * Trim the string by returning excess capacity to the heap.
 *
 * @param[in] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 */
extern CGUL_EXPORT void
cgul_string__trim_capacity(cgul_exception_t* cex,
                           cgul_string_t s);

/**
 * Get the minimum capacity.
 *
 * @param[in] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @return minimum capacity
 */
extern CGUL_EXPORT size_t
cgul_string__get_minimum_capacity(cgul_exception_t* cex,
                                  cgul_string_t s);

/**
 * Set the minimum capacity.  The default value is <code>128</code>
 * which means there is always capacity for 127 characters plus one
 * NUL-terminator.  Decreasing the capacity will decrease the amount
 * of space used and (typically) also decrease performance.  If an
 * error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] minimum_capacity minimum capacity
 */
extern CGUL_EXPORT void
cgul_string__set_minimum_capacity(cgul_exception_t* cex,
                                  cgul_string_t s,
                                  size_t minimum_capacity);

/**
 * Perform a case-significant string comparison of <code>s1</code> and
 * <code>s2</code> using the current locale.  Return -1, 0, or 1 if
 * <code>s1</code> is less than, equal to, or greater than
 * <code>s2</code> respectively. <p>
 *
 * Note that this method can throw an exception if either string
 * contains a character that is not part of the current locale's
 * collating sequence. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] s1 left-hand side
 * @param[in] s2 right-hand side
 * @return -1, 0, or 1 if <code>s1</code> is less than, equal to, or
 * greater than <code>s2</code> respectively
 * @see cgul_char__strcmp()
 */
extern CGUL_EXPORT int
cgul_string__compare(cgul_exception_t* cex,
                     cgul_string_t s1,
                     cgul_string_t s2);

/**
 * Perform a case-insensitive string comparison of string
 * <code>s1</code> and <code>s2</code> using the current locale.
 * Return -1, 0, or 1 if <code>s1</code> is less than, equal to, or
 * greater than <code>s2</code> respectively. <p>
 *
 * Note that this method can throw an exception if either string
 * contains a character that is not part of the current locale's
 * collating sequence. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] s1 left-hand side
 * @param[in] s2 right-hand side
 * @return -1, 0, or 1 if <code>*s1</code> is less than, equal to, or
 * greater than <code>*s2</code> respectively
 * @see cgul_create_ivalue
 */
extern CGUL_EXPORT int
cgul_string__icompare(cgul_exception_t* cex,
                      cgul_string_t s1,
                      cgul_string_t s2);

/**
 * Swap the underlying data for <code>s1</code> and <code>s2</code>.
 * For large strings, this should be much faster than trying to do the
 * same thing using <code>cgul_string__set_value()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] s1 <code>cgul_string</code> instance
 * @param[in] s2 <code>cgul_string</code> instance
 */
extern CGUL_EXPORT void
cgul_string__swap(cgul_exception_t* cex,
                  cgul_string_t s1,
                  cgul_string_t s2);


CGUL_END_C

#endif  /*  CGUL_STRING_H  */
