#ifndef CGUL_TRIE_NODE_CXX_H
#define CGUL_TRIE_NODE_CXX_H

/**
 * C++ wrapper for <code>cgul_trie_node</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_trie_node</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_trie_node.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_trie_node</code> objects.  The main purpose of this
 * class is to convert the C-style function calls and exception
 * handling in <code>cgul_trie_node</code> into C++-style function
 * calls and exception handling.  This class must not be extended.
 *
 * @brief C++ bindings for <code>cgul_trie_node</code>
 */
class cgul_trie_node_cxx {

public:

    /*
     * IMPLEMENTATION NOTE:
     * ===================
     *
     * There are no real cgul_trie_node_cxx objects.  Instead, this
     * class is carefully implemented so that "cgul_trie_node"
     * objects (not "cgul_trie_node_cxx" objects) can be directly
     * used to invoke the methods of this class.
     *
     * This works because this class has no associated data _and_ no
     * virtual methods which would cause a vtable lookup, and vtables
     * are typically embedded in the object itself.
     *
     * In summary, do _not_ add data or virtual functions to this
     * class.
     */

    friend class cgul_trie_cxx;

    /**
     * Return the key stored in this node.  The key is used to determine
     * the sort order of the node.  If you want to set the value of the
     * key, there is no method to do so because changing the value of a
     * key requires changing the node's position in the
     * <code>cgul_trie</code>.  Thus, if you need to change the value of a
     * key, you should delete the node from the <code>cgul_trie</code> and
     * insert a new node with the new value for the key.
     *
     * @return key
     */
    const char*
    get_key() const
    {
        const char* rv = NULL;
        cgul_trie_node_t n = (cgul_trie_node_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_trie_node__get_key(&local, n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the value stored in this node.
     *
     * @return value
     */
    void*
    get_value() const
    {
        void* rv = NULL;
        cgul_trie_node_t n = (cgul_trie_node_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_trie_node__get_value(&local, n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the value stored in this node.
     *
     * @param[in] value
     */
    void
    set_value(void* value)
    {
        cgul_trie_node_t n = (cgul_trie_node_t)this;
        cgul_exception_t local = NULL;
        cgul_trie_node__set_value(&local, n, value);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the previous node.  Together with
     * <code>get_next()</code>, this method lets you traverse the trie
     * in sorted order. If there is no previous node, this method
     * returns <code>NULL</code>.
     *
     * @return previous node
     */
    cgul_trie_node_cxx*
    get_prev() const
    {
        cgul_trie_node_cxx* rv = NULL;
        cgul_trie_node_t n = (cgul_trie_node_t)this;
        cgul_exception_t local = NULL;
        rv = (cgul_trie_node_cxx*)cgul_trie_node__get_prev(&local, n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the next node.  Together with <code>get_prev()</code>,
     * this method lets you traverse the trie in sorted order.  If
     * there is no next node, this method returns <code>NULL</code>.
     *
     * @return next node
     */
    cgul_trie_node_cxx*
    get_next() const
    {
        cgul_trie_node_cxx* rv = NULL;
        cgul_trie_node_t n = (cgul_trie_node_t)this;
        cgul_exception_t local = NULL;
        rv = (cgul_trie_node_cxx*)cgul_trie_node__get_next(&local, n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the next older node.  Together with
     * <code>get_younger()</code>, this method lets you traverse the
     * trie in chronological order. If there is no older node, this
     * method returns <code>NULL</code>. <p>
     *
     * <code>get_older()</code> and <code>get_younger()</code> are
     * faster than <code>get_prev()</code> and <code>get_next()</code>
     * because the latter pair of functions has to walk the trie to
     * find the previous and next nodes in sort order, but the former
     * pair has direct pointers to the younger and older nodes. <p>
     *
     * The following example shows how to iterate over the entire trie
     * in reverse chronological order.  Notice that you have to start
     * with the youngest node when calling <code>get_older()</code>:
     *
     * <pre>
     *     cgul_trie_node_cxx* n = t->get_youngest();
     *     for ( ; n ; n = n->get_older()) {
     *         ...
     *     }
     * </pre>
     *
     * @return older node
     */
    cgul_trie_node_cxx*
    get_older() const
    {
        cgul_trie_node_cxx* rv = NULL;
        cgul_trie_node_t n = (cgul_trie_node_t)this;
        cgul_exception_t local = NULL;
        rv = (cgul_trie_node_cxx*)cgul_trie_node__get_older(&local, n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the next younger node.  Together with
     * <code>get_older()</code>, this method lets you traverse the
     * trie in chronological order. If there is no younger node, this
     * method returns <code>NULL</code>. <p>
     *
     * <code>get_older()</code> and <code>get_younger()</code> are
     * faster than <code>get_prev()</code> and <code>get_next()</code>
     * because the latter pair of functions has to walk the trie to
     * find the previous and next nodes in sort order, but the former
     * pair has direct pointers to the younger and older nodes. <p>
     *
     * The following example shows how to iterate over the entire trie
     * in chronological order.  Notice that you have to start with the
     * oldest node when calling <code>get_younger()</code>:
     *
     * <pre>
     *     cgul_trie_node_cxx* n = t->get_oldest();
     *     for ( ; n ; n = n->get_younger()) {
     *         ...
     *     }
     * </pre>
     *
     * @return younger node
     */
    cgul_trie_node_cxx*
    get_younger() const
    {
        cgul_trie_node_cxx* rv = NULL;
        cgul_trie_node_t n = (cgul_trie_node_t)this;
        cgul_exception_t local = NULL;
        rv = (cgul_trie_node_cxx*)cgul_trie_node__get_younger(&local, n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the underlying <code>cgul_trie_node</code> object.
     *
     * @return underlying object
     */
    cgul_trie_node_t
    get_obj() const
    {
        /* Because of the special implementation of this class, "this"
         * is in fact the underlying object.  (See the comments at the
         * top of the file.) */
        return (cgul_trie_node_t)this;
    }

private:

    /* Hide these. */
    cgul_trie_node_cxx();
    cgul_trie_node_cxx(const cgul_trie_node_cxx&);
    ~cgul_trie_node_cxx();
    cgul_trie_node_cxx& operator=(const cgul_trie_node_cxx&);

};


#endif  /*  CGUL_TRIE_NODE_CXX_H  */
