#include "cgul_bdf_p.h"
#include "cgul_bdf_glyph_p.h"
#include "cgul_crlf_file.h"
#include "cgul_defensive.h"
#include "cgul_libc.h"
#include "cgul_rbtree.h"
#include "cgul_rbtree_node.h"
#include "cgul_split.h"
#include "cgul_strtok.h"
#include "cgul_unicode.h"
#include "cgul_vector.h"


/**
 * Value used for the encoding to flag a glyph as being unencoded.
 * This happens when the BDF has the following ENCODING line for the
 * glyph:
 *
 * <pre>
 *     ENCODING -1
 * </pre>
 */
static const cgul_wchar_t CGUL_BDF__UNENCODED = ~(cgul_wchar_t)0;

/**
 * This method is used as the comparison function for the
 * <code>bdf->glyphs</code> red-black tree.  It returns less than
 * zero, zero, or greater than zero if <code>key1</code> is less than,
 * equal to, or greater than <code>key2</code> respectively.  The keys
 * in the <code>bdf->glyphs</code> tree are pointers to
 * <code>cgul_wchar_t</code> code points for the stored glyph.
 *
 * @param[in] key1 first key
 * @param[in] key2 second key
 * @return comparison of <code>key1</code> with <code>key2</code>
 */
static int
cgul_bdf__compare_encodings(const void* key1,
                            const void* key2)
{
    int rv = 0;

    cgul_wchar_t encoding1 = *(cgul_wchar_t*)key1;
    cgul_wchar_t encoding2 = *(cgul_wchar_t*)key2;

    if (encoding1 < encoding2 ) {
        rv = -1;
    } else if (encoding1 > encoding2) {
        rv = 1;
    }

    return rv;
}

/**
 * This method converts the hexadecimal <code>digit</code> as an ASCII
 * character into an integer.  If <code>digit</code> is not a valid
 * hexadecimal character, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] digit hexadecimal digit as ASCII character
 * @return value of <code>digit</code> as an integer
 */
static int
cgul_bdf__hex_digit_to_int(cgul_exception_t* cex,
                           unsigned char digit)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    switch (digit) {
    case '0':
        rv = 0;
        break;
    case '1':
        rv = 1;
        break;
    case '2':
        rv = 2;
        break;
    case '3':
        rv = 3;
        break;
    case '4':
        rv = 4;
        break;
    case '5':
        rv = 5;
        break;
    case '6':
        rv = 6;
        break;
    case '7':
        rv = 7;
        break;
    case '8':
        rv = 8;
        break;
    case '9':
        rv = 9;
        break;
    case 'a':
    case 'A':
        rv = 10;
        break;
    case 'b':
    case 'B':
        rv = 11;
        break;
    case 'c':
    case 'C':
        rv = 12;
        break;
    case 'd':
    case 'D':
        rv = 13;
        break;
    case 'e':
    case 'E':
        rv = 14;
        break;
    case 'f':
    case 'F':
        rv = 15;
        break;
    default:
        *cex = cgul_exception__new("Invalid hex digit: %c", digit);
        goto out;
    }

 out:
    return rv;
}

/**
 * This method verifies that all the currently loaded BDFs were loaded
 * without index files.  If any BDFs were loaded with an index file,
 * this method throws an exception that includes
 * <code>func_name</code> as part of the error message.
 *
 * @param[in,out] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] func_name function name
 */
static void
cgul_bdf__verify_no_index_files(cgul_exception_t* cex,
                                cgul_bdf_t bdf,
                                const char* func_name)
{
    cgul_list_node_t node = NULL;
    cgul_bdf_files_t files = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(func_name);

    /* Iterate over all the BDF files that are currently loaded. */
    node = cgul_list__get_front(cex, bdf->files);
    for ( ; node ; node = cgul_list_node__get_next(cex, node)) {
        /* Get the files for this node. */
        files = (cgul_bdf_files_t)cgul_list_node__get_value(cex, node);
        /* Make sure this BDF file does not have an index file. */
        if (files && files->index_file_name) {
            *cex = cgul_exception__new(
                       "%s: %s: All BDF files must be loaded "
                       "without an index file.  This can be done by calling "
                       "cgul_bdf__load_explicit() with the index file set "
                       "to NULL.",
                       func_name,
                       files->bdf_file_name);
            goto out;
        }
    }

 out:
    return;
}

/**
 * This method attempts to removes the glyph for <code>encoding</code>
 * from the cache of glyphs for <code>bdf</code>.  If successful,
 * <code>1</code> is returned; otherwise, <code>0</code> is
 * returned. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] encoding encoding
 */
static int
cgul_bdf__remove_glyph(cgul_exception_t* cex,
                       cgul_bdf_t bdf,
                       cgul_wchar_t encoding)
{
    int rv = 0;
    void* value = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);

    /* Attempt to remove the glyph for "encoding". */
    rv = cgul_rbtree__remove(cex, bdf->glyphs, &encoding, NULL, &value);
    if (!rv) {
        /* There was no glyph for this encoding. */
        goto out;
    }

    /* Delete the glyph. */
    assert(value);
    cgul_bdf_glyph__delete((cgul_bdf_glyph_t)value);

 out:
    return rv;
}

/**
 * If an index file is being used and <code>bdf->cache_size</code> is
 * greater than <code>0</code>, this method removes the least-recently
 * used (LRU) glyphs from <code>bdf->glyphs</code> such that no more
 * than <code>bdf->cache_size</code> glyphs are cached.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 */
static void
cgul_bdf__expire_glyphs(cgul_exception_t* cex,
                        cgul_bdf_t bdf)
{
    cgul_bdf_glyph_t glyph = NULL;
    cgul_rbtree_node_t oldest = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* It only makes sense to expire glyphs if all the BDF files that
     * have been loaded were loaded with an associated BDF index file;
     * otherwise, there will be no way to reload some of the glyphs
     * after they expire. */
    if (!bdf->allow_cache_expiration) {
        goto out;
    }

    /* By convention, a cache size of zero disables expiring of glyphs. */
    if (bdf->cache_size == 0) {
        goto out;
    }

    /* Remove glyphs while the tree is larger than the cache size. */
    while (cgul_rbtree__get_size(cex, bdf->glyphs) > bdf->cache_size) {
        /* Determine the oldest glyph. */
        oldest = cgul_rbtree__get_oldest(cex, bdf->glyphs);
        glyph = (cgul_bdf_glyph_t)cgul_rbtree_node__get_value(cex, oldest);
        /* Remove the oldest glyph from bdf->glyphs. */
        cgul_bdf__remove_glyph(cex, bdf, glyph->encoding);
    }

 out:
    return;
}

/**
 * This method adds <code>glyph</code> to the internal cache of glyphs
 * held by <code>bdf</code>.  If the insertion is a success
 * <code>1</code> is returned.  If the insertion failed (typically
 * because a glyph with the same encoding is already present),
 * <code>0</code> is returned.  If an error occurs, <code>0</code> is
 * returned, and an exception is thrown. <p>
 *
 * If <code>cgul_bdf__set_cache_size()</code> has been used to limit
 * the number of glyphs cached by <code>bdf</code>, it is possible
 * that adding one glyph will cause another to be removed.  To disable
 * this behavior, do not use an index file when
 * <code>cgul_bdf__load()</code> or
 * <code>cgul_bdf__load_explicit()</code> is called or simply make
 * sure the cache size is set to <code>0</code> which allows the cache
 * to grow without limit. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] glyph glyph
 * @return whether <code>glyph</code> could be inserted
 */
static int
cgul_bdf__add_glyph(cgul_exception_t* cex,
                    cgul_bdf_t bdf,
                    cgul_bdf_glyph_t glyph)
{
    int rv = 0;
    cgul_rbtree_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(glyph);

    /* Insert the encoding into the tree. */
    rv = cgul_rbtree__insert(cex,
                             bdf->glyphs,
                             (void*)&glyph->encoding,
                             &node);
    if (*cex) {
        goto out;
    }

    /* Do not add "glyph" if it is a duplicate. */
    if (!rv) {
        goto out;
    }

    /* Set the glyph to be the value of the newly inserted encoding. */
    cgul_rbtree_node__set_value(cex, node, glyph);

    /* Expire old glyphs. */
    cgul_bdf__expire_glyphs(cex, bdf);

 out:
    return rv;
}

/**
 * This method returns the next line in <code>f</code> stripping out
 * COMMENT lines.  If EOF is reached, <code>NULL</code> is returned.
 * If an error occurs, <code>NULL</code> is returned, and an exception
 * is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] f BDF file
 * @return next line of <code>f</code>
 */
static const char*
cgul_bdf__next_line(cgul_exception_t* cex,
                    cgul_bdf_t bdf,
                    cgul_crlf_file_t f)
{
    const char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);

    /* Get the next line from the file. */
    while ((rv = cgul_crlf_file__get_line(cex, f)) != NULL) {
        /* Return this line if it isn't a comment. */
        if (strncmp(rv, "COMMENT", 7) != 0) {
            break;
        }
    }

 out:

    /* Return NULL if this method throws an exception. */
    if (*cex && rv) {
        /* Do not free rv.  It is owned by the cgul_crlf_file object. */
        rv = NULL;
    }

    return rv;
}

/**
 * This method splits <code>line</code> into its component parts and
 * returns a <code>cgul_vector</code> of C-style strings of each
 * part.  The caller is responsible for calling <code>free()</code> on
 * each string in the returned vector and for calling
 * <code>cgul_vector__delete()</code> on the vector itself.  If an
 * error occurs, <code>NULL</code> is returned, and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] line line of text from the BDF file
 * @return vector of each component of <code>line</code>
 */
static cgul_vector_t
cgul_bdf__split_line(cgul_exception_t* cex,
                     const char* line)
{
    char* part = NULL;
    cgul_vector_t rv = NULL;
    cgul_exception_t local = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(line);

    /* Allocate the vector that we return. */
    rv = cgul_vector__new(cex);
    if (*cex) {
        goto out;
    }

    /* Split the line copying each part into "rv". */
    cgul_split(cex, line, " \t", 0 /* allow_empty_parts */, rv);

 out:

    /* Clean up if an exception is thrown. */
    if (*cex && rv) {
        while ((part = (char*)cgul_vector__pop_back(&local, rv)) != NULL) {
            free(part);
        }
        rv = NULL;
    }

    return rv;
}

/**
 * This method is designed to free the individual parts in the
 * <code>cgul_vector</code> returned by
 * <code>cgul_bdf__split_line()</code>.  Because it will only ever be
 * called in the "out" section of other methods, there is no need to
 * force the caller to pass in a pointer to a
 * <code>cgul_exception_t</code> object.
 */
static void
cgul_bdf__free_line_parts(cgul_vector_t parts_vector)
{
    void* part = NULL;
    cgul_exception_t local = NULL;

    assert(parts_vector);

    while ((part = cgul_vector__pop_back(&local, parts_vector)) != NULL) {
        free(part);
    }
}

/**
 * Process the STARTFONT line.  If an error occurs, an exception is
 * thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] f BDF file
 * @param[in] bdf_file_name bdf file name
 */
static void
cgul_bdf__process_startfont_line(cgul_exception_t* cex,
                                 cgul_bdf_t bdf,
                                 cgul_crlf_file_t f,
                                 const char* bdf_file_name)
{
    unsigned long int current_line = 0;
    const char** parts = NULL;
    size_t parts_count = 0;
    cgul_vector_t parts_vector = NULL;
    const char* line = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);
    assert(bdf_file_name);

#if 0
    /* Read the first line of the file.  We do not use
     * cgul_bdf__next_line() here because comments are only allowed
     * between the STARTFONT and ENDFONT lines. */
    line = cgul_crlf_file__get_line(cex, f);
#else
    /* I have run into fonts that put COMMENT lines above the
     * STARTFONT line. */
    line = cgul_bdf__next_line(cex, bdf, f);
#endif
    if (*cex) {
        goto out;
    }

    /* Get the current line. */
    current_line = cgul_crlf_file__get_line_count(cex, f);

    /* Check for EOF. */
    if (line == NULL) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Unexpected end of file.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Split the line. */
    parts_vector = cgul_bdf__split_line(cex, line);
    if (*cex) {
        goto out;
    }

    /* Get the underlying parts array and its size. */
    parts = (const char**)cgul_vector__get_array(cex, parts_vector);
    parts_count = cgul_vector__get_size(cex, parts_vector);

    /* Validate the line. */
    if ((parts_count != 2) || (strcmp(parts[0], "STARTFONT") != 0)) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid or missing STARTFONT.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

 out:

    /* Clean up. */
    if (parts_vector) {
        /* Free each part. */
        cgul_bdf__free_line_parts(parts_vector);
        /* Free the vector. */
        cgul_vector__delete(parts_vector);
    }

    return;
}

/**
 * Process the FONT line.  If an error occurs, an exception is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] f BDF file
 * @param[in] bdf_file_name bdf file name
 */
static void
cgul_bdf__process_font_line(cgul_exception_t* cex,
                            cgul_bdf_t bdf,
                            cgul_crlf_file_t f,
                            const char* bdf_file_name)
{
    unsigned long int current_line = 0;
    const char* line = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);
    assert(bdf_file_name);

    /* Get the FONT line. */
    line = cgul_bdf__next_line(cex, bdf, f);
    if (*cex) {
        goto out;
    }

    /* Get the current line. */
    current_line = cgul_crlf_file__get_line_count(cex, f);

    /* Check for EOF. */
    if (line == NULL) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Unexpected end of file.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Verify the FONT line. */
    if (strncmp(line, "FONT ", 5) != 0) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid or missing FONT.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* It is not clear to me how to merge the FONT lines when multiple
     * fonts are loaded.  Instead, the name of the first font is
     * saved.  The others are discarded. */
    if (!bdf->font_name) {
        bdf->font_name = cgul_libc__strdup(cex, &line[5]);
    }

 out:
    return;
}

/**
 * Process the SIZE line.  If an error occurs, an exception is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] f BDF file
 * @param[in] bdf_file_name bdf file name
 */
static void
cgul_bdf__process_size_line(cgul_exception_t* cex,
                            cgul_bdf_t bdf,
                            cgul_crlf_file_t f,
                            const char* bdf_file_name)
{
    unsigned int font_size = 0;
    unsigned int dpi_x = 0;
    unsigned int dpi_y = 0;
    unsigned long int current_line = 0;
    const char* line = 0;
    const char** parts = NULL;
    size_t parts_count = 0;
    cgul_vector_t parts_vector = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);
    assert(bdf_file_name);

    /* Allocate a temporary string object. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Get the SIZE line. */
    line = cgul_bdf__next_line(cex, bdf, f);
    if(*cex) {
        goto out;
    }

    /* Get the current line. */
    current_line = cgul_crlf_file__get_line_count(cex, f);

    /* Check for EOF. */
    if (line == NULL) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Unexpected end of file.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Split the line. */
    parts_vector = cgul_bdf__split_line(cex, line);
    if (*cex) {
        goto out;
    }

    /* Get the underlying parts array and its size. */
    parts = (const char**)cgul_vector__get_array(cex, parts_vector);
    parts_count = cgul_vector__get_size(cex, parts_vector);

    /* Validate the line. */
    if ((parts_count != 4) || (strcmp(parts[0], "SIZE") != 0)) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid or missing SIZE.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Convert font_size and merge it with previously loaded BDFs. */
    cgul_string__set_value(cex, s, parts[1]);
    font_size = cgul_string__to_unsigned(cex, s, 0 /* base */);
    if (font_size > bdf->font_size) {
        bdf->font_size = font_size;
    }

    /* Convert dpi_x and merge it with the previously loaded BDFs.  */
    cgul_string__set_value(cex, s, parts[2]);
    dpi_x = cgul_string__to_unsigned(cex, s, 0 /* base */);
    if (dpi_x > bdf->dpi_x) {
        bdf->dpi_x = dpi_x;
    }

    /* Convert dpi_y and merge it with the previously loaded BDFs. */
    cgul_string__set_value(cex, s, parts[3]);
    dpi_y = cgul_string__to_unsigned(cex, s, 0 /* base */);
    if (dpi_y > bdf->dpi_y) {
        bdf->dpi_y = dpi_y;
    }

    /* Check for errors. */
    if (*cex) {
        /* Free the pending exception. */
        cgul_exception__delete(*cex);
        /* Throw a more descriptive one. */
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid font SIZE.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

 out:

    /* Clean up. */
    if (parts_vector) {
        /* Free each part. */
        cgul_bdf__free_line_parts(parts_vector);
        /* Free the vector. */
        cgul_vector__delete(parts_vector);
    }
    if (s) {
        cgul_string__delete(s);
    }

    return;
}

/**
 * Process the FONTBOUNDINGBOX line.  If an error occurs, an exception
 * is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] f BDF file
 * @param[in] bdf_file_name bdf file name
 */
static void
cgul_bdf__process_fontboundingbox_line(cgul_exception_t* cex,
                                       cgul_bdf_t bdf,
                                       cgul_crlf_file_t f,
                                       const char* bdf_file_name)
{
    /* We try to make the font bounding box as large as possible when
     * merging fonts.  Because of this, bb_width and bb_height start
     * out small and grow larger and large; whereas, bb_x_offset and
     * bb_y_offset start out large and get smaller and smaller. */
    unsigned int bb_width = 0;
    unsigned int bb_height = 0;
    int bb_x_offset = 0;
    int bb_y_offset = 0;
    unsigned long int current_line = 0;
    const char* line = 0;
    const char** parts = NULL;
    size_t parts_count = 0;
    cgul_vector_t parts_vector = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);
    assert(bdf_file_name);

    /* Allocate a temporary string object. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Get the FONTBOUNDINGBOX line. */
    line = cgul_bdf__next_line(cex, bdf, f);
    if (*cex) {
        goto out;
    }

    /* Get the current line. */
    current_line = cgul_crlf_file__get_line_count(cex, f);

    /* Check for EOF. */
    if (line == NULL) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Unexpected end of file.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Split the line. */
    parts_vector = cgul_bdf__split_line(cex, line);
    if (*cex) {
        goto out;
    }

    /* Get the underlying parts array and its size. */
    parts = (const char**)cgul_vector__get_array(cex, parts_vector);
    parts_count = cgul_vector__get_size(cex, parts_vector);

    /* Validate the line. */
    if ((parts_count != 5) || (strcmp(parts[0], "FONTBOUNDINGBOX") != 0)) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid or missing FONTBOUNDINGBOX.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Convert bb_width and merge with previously loaded BDFs. */
    cgul_string__set_value(cex, s, parts[1]);
    bb_width = cgul_string__to_unsigned(cex, s, 0 /* base */);
    if (bb_width > bdf->bb_width) {
        bdf->bb_width = bb_width;
    }

    /* Convert bb_height and merge with previously loaded BDFs. */
    cgul_string__set_value(cex, s, parts[2]);
    bb_height = cgul_string__to_unsigned(cex, s, 0 /* base */);
    if (bb_height > bdf->bb_height) {
        bdf->bb_height = bb_height;
    }

    /* Convert bb_x_offset and merge with previously loaded BDFs. */
    cgul_string__set_value(cex, s, parts[3]);
    bb_x_offset = cgul_string__to_int(cex, s, 0 /* base */);
    if (bb_x_offset < bdf->bb_x_offset) {
        bdf->bb_x_offset = bb_x_offset;
    }

    /* Convert bb_y_offset and merge with previously loaded BDFs. */
    cgul_string__set_value(cex, s, parts[4]);
    bb_y_offset = cgul_string__to_int(cex, s, 0 /* base */);
    if (bb_y_offset < bdf->bb_y_offset) {
        bdf->bb_y_offset = bb_y_offset;
    }

    /* Check for errors. */
    if (*cex) {
        /* Free the pending exception. */
        cgul_exception__delete(*cex);
        /* Throw a more descriptive one. */
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid font FONTBOUNDINGBOX.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

 out:

    /* Clean up. */
    if (parts_vector) {
        /* Free each part. */
        cgul_bdf__free_line_parts(parts_vector);
        /* Free the vector. */
        cgul_vector__delete(parts_vector);
    }
    if (s) {
        cgul_string__delete(s);
    }

    return;
}

/**
 * Process the PROPERTY line.  If an error occurs, an exception is
 * thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] f BDF file
 * @param[in] bdf_file_name bdf file name
 */
static void
cgul_bdf__process_property_line(cgul_exception_t* cex,
                                cgul_bdf_t bdf,
                                cgul_crlf_file_t f,
                                const char* bdf_file_name)
{
    unsigned int ascent = 0;
    unsigned int descent = 0;
    unsigned long int current_line = 0;
    const char* line = 0;
    const char** parts = NULL;
    size_t parts_count = 0;
    cgul_vector_t parts_vector = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);
    assert(bdf_file_name);

    /* Allocate a temporary string object. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Get the property line. */
    line = cgul_bdf__next_line(cex, bdf, f);
    if (*cex) {
        goto out;
    }

    /* Get the current line. */
    current_line = cgul_crlf_file__get_line_count(cex, f);

    /* Check for EOF. */
    if (line == NULL) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Unexpected end of file.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Split the line. */
    parts_vector = cgul_bdf__split_line(cex, line);
    if (*cex) {
        goto out;
    }

    /* Get the underlying parts array and its size. */
    parts = (const char**)cgul_vector__get_array(cex, parts_vector);
    parts_count = cgul_vector__get_size(cex, parts_vector);

    /* For now, we are only interested in the FONT_ASCENT,
     * FONT_DESCENT, and DEFAULT_CHAR properties.  We skip the
     * rest. */
    if (parts_count == 2) {

        /* Remember the font ascent. */
        if (strcmp(parts[0], "FONT_ASCENT") == 0) {
            /* Convert ascent. */
            cgul_string__set_value(cex, s, parts[1]);
            ascent = cgul_string__to_unsigned(cex, s, 0);
            if (*cex) {
                /* Free the pending exception. */
                cgul_exception__delete(*cex);
                /* Throw a more descriptive one. */
                *cex = cgul_exception__new("%s:%lu:"
                                           "Invalid FONT_ASCENT: %s",
                                           bdf_file_name,
                                           current_line,
                                           parts[1]);
                goto out;
            }
            /* Merge ascent with previously loaded BDFs.*/
            if (ascent > bdf->ascent) {
                bdf->ascent = ascent;
                /* Remember that we fond the FONT_ASCENT property. */
                bdf->ascent_found = 1;
            }
        }

        /* Remember the font descent. */
        else if (strcmp(parts[0], "FONT_DESCENT") == 0) {
            /* Convert descent. */
            cgul_string__set_value(cex, s, parts[1]);
            descent = cgul_string__to_unsigned(cex, s, 0);
            if (*cex) {
                /* Free the pending exception. */
                cgul_exception__delete(*cex);
                /* Throw a more descriptive one. */
                *cex = cgul_exception__new("%s:%lu:"
                                           "Invalid FONT_DESCENT: %s",
                                           bdf_file_name,
                                           current_line,
                                           parts[1]);
                goto out;
            }
            /* Merge descent with previously loaded BDFs. */
            if (descent > bdf->descent) {
                bdf->descent = descent;
                /* Remember that we fond the FONT_DESCENT property. */
                bdf->descent_found = 1;
            }
        }

        /* Remember the default character.  Just skip the line if the
         * bdf file specifies a -1. */
        else if (   (strcmp(parts[0], "DEFAULT_CHAR") == 0)
                 && (strcmp(parts[1], "-1") != 0))
        {
            /* Use this default character only if we don't already
             * have one from one of the previously loaded BDFs. */
            if (!bdf->default_glyph) {
                /* Convert default character.*/
                cgul_string__set_value(cex, s, parts[1]);
                bdf->default_char = cgul_string__to_unsigned(cex, s, 0);
                if (*cex) {
                    /* Free the pending exception. */
                    cgul_exception__delete(*cex);
                    /* Throw a more descriptive one. */
                    *cex = cgul_exception__new("%s:%lu:"
                                               "Invalid DEFAULT_CHAR: %s",
                                               bdf_file_name,
                                               current_line,
                                               parts[1]);
                    goto out;
                }
                /* Remember that we fond the FONT_DESCENT property. */
                bdf->default_char_found = 1;
            }
        }

    }

 out:

    /* Clean up. */
    if (parts_vector) {
        /* Free each part. */
        cgul_bdf__free_line_parts(parts_vector);
        /* Free the vector. */
        cgul_vector__delete(parts_vector);
    }
    if (s) {
        cgul_string__delete(s);
    }

    return;
}

/**
 * Process the PROPERTIES section.  If an error occurs, an exception
 * is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] f BDF file
 * @param[in] bdf_file_name bdf file name
 */
static void
cgul_bdf__process_properties_section(cgul_exception_t* cex,
                                     cgul_bdf_t bdf,
                                     cgul_crlf_file_t f,
                                     const char* bdf_file_name)
{
    unsigned long i = 0;
    unsigned long i_max = 0;
    unsigned long int current_line = 0;
    const char* line = 0;
    const char** parts = NULL;
    size_t parts_count = 0;
    cgul_vector_t parts_vector = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);
    assert(bdf_file_name);

    /* Allocate a temporary string object. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Get the STARTPROPERTIES line. */
    line = cgul_bdf__next_line(cex, bdf, f);
    if (*cex) {
        goto out;
    }

    /* Get the current line. */
    current_line = cgul_crlf_file__get_line_count(cex, f);

    /* Check for EOF. */
    if (line == NULL) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Unexpected end of file.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Split the line. */
    parts_vector = cgul_bdf__split_line(cex, line);
    if (*cex) {
        goto out;
    }

    /* Get the underlying parts array and its size. */
    parts = (const char**)cgul_vector__get_array(cex, parts_vector);
    parts_count = cgul_vector__get_size(cex, parts_vector);

    /* Validate the line. */
    if ((parts_count != 2) || (strcmp(parts[0], "STARTPROPERTIES") != 0)) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid or missing STARTPROPERTIES.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Get the count of the number of properties. */
    cgul_string__set_value(cex, s, parts[1]);
    i_max = cgul_string__to_unsigned(cex, s, 0 /* base */);
    if (*cex) {
        /* Free the pending exception. */
        cgul_exception__delete(*cex);
        /* Throw a more descriptive one. */
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid STARTPROPERTIES.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Iterate over each property. */
    for (i = 0 ; i < i_max ; ++i) {
        cgul_bdf__process_property_line(cex, bdf, f, bdf_file_name);
        if (*cex) {
            goto out;
        }
    }

    /* Get the ENDPROPERTIES line. */
    line = cgul_bdf__next_line(cex, bdf, f);
    if (*cex) {
        goto out;
    }

    /* Get the current line. */
    current_line = cgul_crlf_file__get_line_count(cex, f);

    /* Check for EOF. */
    if (line == NULL) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Unexpected end of file.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Verify the ENDPROPERTIES line. */
    if (strcmp(line, "ENDPROPERTIES") != 0) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid or missing ENDPROPERTIES.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

 out:

    /* Clean up. */
    if (parts_vector) {
        /* Free each part. */
        cgul_bdf__free_line_parts(parts_vector);
        /* Free the vector. */
        cgul_vector__delete(parts_vector);
    }
    if (s) {
        cgul_string__delete(s);
    }

    return;
}

/**
 * Process the CHARS line.  If an error occurs, an exception is
 * thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] f BDF file
 * @param[in] bdf_file_name bdf file name
 */
static void
cgul_bdf__process_chars_line(cgul_exception_t* cex,
                             cgul_bdf_t bdf,
                             cgul_crlf_file_t f,
                             const char* bdf_file_name)
{
    unsigned long int current_line = 0;
    const char* line = 0;
    const char** parts = NULL;
    size_t parts_count = 0;
    cgul_vector_t parts_vector = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);
    assert(bdf_file_name);

    /* Allocate a temporary string object. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Get the CHARS line. */
    line = cgul_bdf__next_line(cex, bdf, f);
    if (*cex) {
        goto out;
    }

    /* Get the current line. */
    current_line = cgul_crlf_file__get_line_count(cex, f);

    /* Check for EOF. */
    if (line == NULL) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Unexpected end of file.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Split the line. */
    parts_vector = cgul_bdf__split_line(cex, line);
    if (*cex) {
        goto out;
    }

    /* Get the underlying parts array and its size. */
    parts = (const char**)cgul_vector__get_array(cex, parts_vector);
    parts_count = cgul_vector__get_size(cex, parts_vector);

    /* Validate the line. */
    if ((parts_count != 2) || (strcmp(parts[0], "CHARS") != 0)) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid or missing CHARS.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Convert the character count.  It is important that
     * bdf->char_count hold the correct number of characters in the
     * _current_ BDF file because it is used by
     * cgul_bdf__process_character_sections() to iterate over the
     * STARTCHAR sections.  If you need to know the total number of
     * characters in the merged font, call cgul_bdf__get_char_count()
     * which knows how to calculate this value (if possible). */
    cgul_string__set_value(cex, s, parts[1]);
    bdf->char_count = cgul_string__to_unsigned_long(cex, s, 0 /* base */);
    if (*cex) {
        /* Free the pending exception. */
        cgul_exception__delete(*cex);
        /* Throw a more descriptive one. */
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid font CHARS.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

 out:

    /* Clean up. */
    if (parts_vector) {
        /* Free each part. */
        cgul_bdf__free_line_parts(parts_vector);
        /* Free the vector. */
        cgul_vector__delete(parts_vector);
    }
    if (s) {
        cgul_string__delete(s);
    }

    return;
}

/**
 * Process the STARTCHAR line initializing the glyph <code>g</code>.
 * If an error occurs, an exception is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] f BDF file
 * @param[out] g glyph
 * @param[in] bdf_file_name bdf file name
 */
static void
cgul_bdf__process_startchar_line(cgul_exception_t* cex,
                                 cgul_bdf_t bdf,
                                 cgul_crlf_file_t f,
                                 cgul_bdf_glyph_t g,
                                 const char* bdf_file_name)
{
    unsigned long int current_line = 0;
    const char* line = 0;
    const char** parts = NULL;
    size_t parts_count = 0;
    cgul_vector_t parts_vector = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);
    assert(g);
    assert(bdf_file_name);

    /* Get the STARTCHAR line. */
    line = cgul_bdf__next_line(cex, bdf, f);
    if (*cex) {
        goto out;
    }

    /* Get the current line. */
    current_line = cgul_crlf_file__get_line_count(cex, f);

    /* Check for EOF. */
    if (line == NULL) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Unexpected end of file.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Split the line. */
    parts_vector = cgul_bdf__split_line(cex, line);
    if (*cex) {
        goto out;
    }

    /* Get the underlying parts array and its size. */
    parts = (const char**)cgul_vector__get_array(cex, parts_vector);
    parts_count = cgul_vector__get_size(cex, parts_vector);

    /* Validate the line. */
    if (   (parts_count < 1)
        || (parts_count > 2)
        || (strcmp(parts[0], "STARTCHAR") != 0))
    {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid or missing STARTCHAR.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Remember where to find this glyph in the BDF file. */
    g->file_offset = cgul_crlf_file__get_line_offset(cex, f);

#if 0
    /*
     * NOTE: It is common for the character labels to be longer than
     * 14 characters even though this technically violates the BDF
     * standard.
     */

    /* The character's label can be no more than 14 characters long. */
    if ((parts_count == 2) && strlen(parts[1]) > 14) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Label for STARTCHAR is too long.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }
#endif

    /* If a descriptive name for this character was given, set it. */
    if (parts_count == 2) {
        cgul_bdf_glyph__set_descriptive_name(cex, g, parts[1]);
    }

 out:

    /* Clean up. */
    if (parts_vector) {
        /* Free each part. */
        cgul_bdf__free_line_parts(parts_vector);
        /* Free the vector. */
        cgul_vector__delete(parts_vector);
    }

    return;
}

/**
 * Process the ENCODING line initializing the glyph <code>g</code>.
 * If an error occurs, an exception is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] f BDF file
 * @param[out] g glyph
 * @param[in] bdf_file_name bdf file name
 */
static void
cgul_bdf__process_encoding_line(cgul_exception_t* cex,
                                cgul_bdf_t bdf,
                                cgul_crlf_file_t f,
                                cgul_bdf_glyph_t g,
                                const char* bdf_file_name)
{
    unsigned long int current_line = 0;
    const char* line = 0;
    unsigned long int encoding = 0;
    const char** parts = NULL;
    size_t parts_count = 0;
    cgul_vector_t parts_vector = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);
    assert(g);
    assert(bdf_file_name);

    /* Allocate a temporary string object. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Get the ENCODING line. */
    line = cgul_bdf__next_line(cex, bdf, f);
    if (*cex) {
        goto out;
    }

    /* Get the current line. */
    current_line = cgul_crlf_file__get_line_count(cex, f);

    /* Check for EOF. */
    if (line == NULL) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Unexpected end of file.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Split the line. */
    parts_vector = cgul_bdf__split_line(cex, line);
    if (*cex) {
        goto out;
    }

    /* Get the underlying parts array and its size. */
    parts = (const char**)cgul_vector__get_array(cex, parts_vector);
    parts_count = cgul_vector__get_size(cex, parts_vector);

    /* Validate the line. */
    if (   (parts_count < 2)
        || (parts_count > 3)
        || (strcmp(parts[0], "ENCODING") != 0))
    {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid or missing ENCODING.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Handle the "ENCODING -1" form which means the glyph is not
     * encoded.  The idea behind this type of glyph is that the font
     * server can use the glyph's descriptive name to dynamically map
     * the glyph to an encoding.  The BDF spec allows us to ignore
     * these glyphs, and that is what we do. */
    if ((parts_count == 2) && (strcmp(parts[1], "-1") == 0)) {
        /* Flag as unencoded. */
        cgul_bdf_glyph__set_encoding(cex, g, CGUL_BDF__UNENCODED);
        goto out;
    }

    /* If we have 3 parameters, the middle one must be a -1. */
    if ((parts_count == 3) && (strcmp(parts[1], "-1") != 0)) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid ENCODING.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* For backwards compatibility, the BDF spec allows "ENCODING -1
     * <encoding>" to mean the same as "ENCODING <encoding>".  So we
     * need to pick the correct parameter to convert. */
    if (parts_count == 2) {
        cgul_string__set_value(cex, s, parts[1]);
    } else if (parts_count == 3) {
        cgul_string__set_value(cex, s, parts[2]);
    } else {
        *cex = cgul_exception__new("Internal Error: "
                                   "cgul_bdf__process_encoding_line: "
                                   "Unexpected parts count.");
        goto out;
    }

    /* Convert the encoding. */
    encoding = cgul_string__to_unsigned(cex, s, 0 /* base */);
    if (*cex) {
        /* Free the pending exception. */
        cgul_exception__delete(*cex);
        /* Throw a more descriptive one. */
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid character encoding.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Remember the encoding. */
    cgul_bdf_glyph__set_encoding(cex, g, encoding);

 out:

    /* Clean up. */
    if (parts_vector) {
        /* Free each part. */
        cgul_bdf__free_line_parts(parts_vector);
        /* Free the vector. */
        cgul_vector__delete(parts_vector);
    }
    if (s) {
        cgul_string__delete(s);
    }

    return;
}

/**
 * Process the SWIDTH line initializing the glyph <code>g</code>.  If
 * an error occurs, an exception is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] f BDF file
 * @param[out] g glyph
 * @param[in] bdf_file_name bdf file name
 */
static void
cgul_bdf__process_swidth_line(cgul_exception_t* cex,
                              cgul_bdf_t bdf,
                              cgul_crlf_file_t f,
                              cgul_bdf_glyph_t g,
                              const char* bdf_file_name)
{
    unsigned long int current_line = 0;
    const char* line = 0;
    unsigned int swidth_x = 0;
    unsigned int swidth_y = 0;
    const char** parts = NULL;
    size_t parts_count = 0;
    cgul_vector_t parts_vector = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);
    assert(g);
    assert(bdf_file_name);

    /* Allocate a temporary string object. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Get the SWIDTH line. */
    line = cgul_bdf__next_line(cex, bdf, f);
    if (*cex) {
        goto out;
    }

    /* Get the current line. */
    current_line = cgul_crlf_file__get_line_count(cex, f);

    /* Check for EOF. */
    if (line == NULL) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Unexpected end of file.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Split the line. */
    parts_vector = cgul_bdf__split_line(cex, line);
    if (*cex) {
        goto out;
    }

    /* Get the underlying parts array and its size. */
    parts = (const char**)cgul_vector__get_array(cex, parts_vector);
    parts_count = cgul_vector__get_size(cex, parts_vector);

    /* Validate the line. */
    if ((parts_count != 3) || (strcmp(parts[0], "SWIDTH") != 0)) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid or missing SWIDTH.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Convert swidth_x. */
    cgul_string__set_value(cex, s, parts[1]);
    swidth_x = cgul_string__to_unsigned(cex, s, 0 /* base */);

    /* Convert swidth_y. */
    cgul_string__set_value(cex, s, parts[2]);
    swidth_y = cgul_string__to_unsigned(cex, s, 0 /* base */);

    /* Check for errors. */
    if (*cex) {
        /* Free the pending exception. */
        cgul_exception__delete(*cex);
        /* Throw a more descriptive one. */
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid character swidth.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Remember the swidth. */
    cgul_bdf_glyph__set_swidth_x(cex, g, swidth_x);
    cgul_bdf_glyph__set_swidth_y(cex, g, swidth_y);

 out:

    /* Clean up. */
    if (parts_vector) {
        /* Free each part. */
        cgul_bdf__free_line_parts(parts_vector);
        /* Free the vector. */
        cgul_vector__delete(parts_vector);
    }
    if (s) {
        cgul_string__delete(s);
    }

    return;
}

/**
 * Process the DWIDTH line initializing the glyph <code>g</code>.  If
 * an error occurs, an exception is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] f BDF file
 * @param[out] g glyph
 * @param[in] bdf_file_name bdf file name
 */
static void
cgul_bdf__process_dwidth_line(cgul_exception_t* cex,
                              cgul_bdf_t bdf,
                              cgul_crlf_file_t f,
                              cgul_bdf_glyph_t g,
                              const char* bdf_file_name)
{
    unsigned long int current_line = 0;
    const char* line = 0;
    unsigned int dwidth_x = 0;
    unsigned int dwidth_y = 0;
    const char** parts = NULL;
    size_t parts_count = 0;
    cgul_vector_t parts_vector = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);
    assert(g);
    assert(bdf_file_name);

    /* Allocate a temporary string object. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Get the DWIDTH line. */
    line = cgul_bdf__next_line(cex, bdf, f);
    if (*cex) {
        goto out;
    }

    /* Get the current line. */
    current_line = cgul_crlf_file__get_line_count(cex, f);

    /* Check for EOF. */
    if (line == NULL) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Unexpected end of file.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Split the line. */
    parts_vector = cgul_bdf__split_line(cex, line);
    if (*cex) {
        goto out;
    }

    /* Get the underlying parts array and its size. */
    parts = (const char**)cgul_vector__get_array(cex, parts_vector);
    parts_count = cgul_vector__get_size(cex, parts_vector);

    /* Validate the line. */
    if ((parts_count != 3) || (strcmp(parts[0], "DWIDTH") != 0)) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid or missing DWIDTH",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Convert dwidth_x. */
    cgul_string__set_value(cex, s, parts[1]);
    dwidth_x = cgul_string__to_unsigned(cex, s, 0 /* base */);

    /* Convert dwidth_y. */
    cgul_string__set_value(cex, s, parts[2]);
    dwidth_y = cgul_string__to_unsigned(cex, s, 0 /* base */);

    /* Check for errors. */
    if (*cex) {
        /* Free the pending exception. */
        cgul_exception__delete(*cex);
        /* Throw a more descriptive one. */
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid character dwidth.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Remember the dwidth. */
    cgul_bdf_glyph__set_dwidth_x(cex, g, dwidth_x);
    cgul_bdf_glyph__set_dwidth_y(cex, g, dwidth_y);

 out:

    /* Clean up. */
    if (parts_vector) {
        /* Free each part. */
        cgul_bdf__free_line_parts(parts_vector);
        /* Free the vector. */
        cgul_vector__delete(parts_vector);
    }
    if (s) {
        cgul_string__delete(s);
    }

    return;
}

/**
 * Process the BBX line initializing the glyph <code>g</code>.  If an
 * error occurs, an exception is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] f BDF file
 * @param[out] g glyph
 * @param[in] bdf_file_name bdf file name
 */
static void
cgul_bdf__process_bbx_line(cgul_exception_t* cex,
                           cgul_bdf_t bdf,
                           cgul_crlf_file_t f,
                           cgul_bdf_glyph_t g,
                           const char* bdf_file_name)
{
    unsigned long int current_line = 0;
    const char* line = 0;
    unsigned int glyph_bb_width = 0;
    unsigned int glyph_bb_height = 0;
    unsigned int glyph_bb_x_offset = 0;
    unsigned int glyph_bb_y_offset = 0;
    const char** parts = NULL;
    size_t parts_count = 0;
    cgul_vector_t parts_vector = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);
    assert(g);
    assert(bdf_file_name);

    /* Allocate a temporary string object. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }


    /* Get the BBX line. */
    line = cgul_bdf__next_line(cex, bdf, f);
    if (*cex) {
        goto out;
    }

    /* Get the current line. */
    current_line = cgul_crlf_file__get_line_count(cex, f);

    /* Check for EOF. */
    if (line == NULL) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Unexpected end of file.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Split the line. */
    parts_vector = cgul_bdf__split_line(cex, line);
    if (*cex) {
        goto out;
    }

    /* Get the underlying parts array and its size. */
    parts = (const char**)cgul_vector__get_array(cex, parts_vector);
    parts_count = cgul_vector__get_size(cex, parts_vector);

    /* Validate the line. */
    if ((parts_count != 5) || (strcmp(parts[0], "BBX") != 0)) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid or missing BBX.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Convert bb_width. */
    cgul_string__set_value(cex, s, parts[1]);
    glyph_bb_width = cgul_string__to_unsigned(cex, s, 0 /* base */);

    /* Convert bb__height. */
    cgul_string__set_value(cex, s, parts[2]);
    glyph_bb_height = cgul_string__to_unsigned(cex, s, 0 /* base */);

    /* Convert bb_x_offset. */
    cgul_string__set_value(cex, s, parts[3]);
    glyph_bb_x_offset = cgul_string__to_int(cex, s, 0 /* base */);

    /* Convert bb_y_offset. */
    cgul_string__set_value(cex, s, parts[4]);
    glyph_bb_y_offset = cgul_string__to_int(cex, s, 0 /* base */);

    /* Check for errors. */
    if (*cex) {
        /* Free the pending exception. */
        cgul_exception__delete(*cex);
        /* Throw a more descriptive one. */
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid character bbx.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Remember the bounding box. */
    cgul_bdf_glyph__set_bb_width(cex, g, glyph_bb_width);
    cgul_bdf_glyph__set_bb_height(cex, g, glyph_bb_height);
    cgul_bdf_glyph__set_bb_x_offset(cex, g, glyph_bb_x_offset);
    cgul_bdf_glyph__set_bb_y_offset(cex, g, glyph_bb_y_offset);

 out:

    /* Clean up. */
    if (parts_vector) {
        /* Free each part. */
        cgul_bdf__free_line_parts(parts_vector);
        /* Free the vector. */
        cgul_vector__delete(parts_vector);
    }
    if (s) {
        cgul_string__delete(s);
    }

    return;
}

/**
 * Process the BITMAP line.  If an error occurs, an exception is
 * thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] f BDF file
 * @param[in] bdf_file_name bdf file name
 */
static void
cgul_bdf__process_bitmap_line(cgul_exception_t* cex,
                              cgul_bdf_t bdf,
                              cgul_crlf_file_t f,
                              const char* bdf_file_name)
{
    unsigned long int current_line = 0;
    const char* line = 0;
    const char** parts = NULL;
    size_t parts_count = 0;
    cgul_vector_t parts_vector = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);
    assert(bdf_file_name);

    /* Get the BITMAP line. */
    line = cgul_bdf__next_line(cex, bdf, f);
    if (*cex) {
        goto out;
    }

    /* Get the current line. */
    current_line = cgul_crlf_file__get_line_count(cex, f);

    /* Check for EOF. */
    if (line == NULL) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Unexpected end of file.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Split the line. */
    parts_vector = cgul_bdf__split_line(cex, line);
    if (*cex) {
        goto out;
    }

    /* Get the underlying parts array and its size. */
    parts = (const char**)cgul_vector__get_array(cex, parts_vector);
    parts_count = cgul_vector__get_size(cex, parts_vector);

    /* Validate the line. */
    if ((parts_count != 1) || (strcmp(parts[0], "BITMAP") != 0)) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid or missing BITMAP.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

 out:

    /* Clean up. */
    if (parts_vector) {
        /* Free each part. */
        cgul_bdf__free_line_parts(parts_vector);
        /* Free the vector. */
        cgul_vector__delete(parts_vector);
    }

    return;
}

/**
 * Process a bitmap data line initializing the glyph <code>g</code>.
 * If an error occurs, an exception is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] f BDF file
 * @param[out] g glyph
 * @param[in] row row of glyph that data belongs to
 * @param[in] bdf_file_name bdf file name
 */
static void
cgul_bdf__process_bitmap_data_line(cgul_exception_t* cex,
                                   cgul_bdf_t bdf,
                                   cgul_crlf_file_t f,
                                   cgul_bdf_glyph_t g,
                                   unsigned int row,
                                   const char* bdf_file_name)
{
    size_t i = 0;
    size_t i_max = 0;
    unsigned long int current_line = 0;
    size_t bytes_wide = 0;
    unsigned int j = 0;
    int bits = 0;
    const char* line = 0;
    const char** parts = NULL;
    size_t parts_count = 0;
    cgul_vector_t parts_vector = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);
    assert(g);
    assert(bdf_file_name);

    /* Get the width of the glyph in bytes.  Remember that the data is
     * a bitmap so it is packed with 8 pixels per byte. */
    bytes_wide = cgul_bdf_glyph__get_bytes_wide(cex, g);

    /* Get the data line. */
    line = cgul_bdf__next_line(cex, bdf, f);
    if (*cex) {
        goto out;
    }

    /* Get the current line. */
    current_line = cgul_crlf_file__get_line_count(cex, f);

    /* Check for EOF. */
    if (line == NULL) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Unexpected end of file.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Split the line. */
    parts_vector = cgul_bdf__split_line(cex, line);
    if (*cex) {
        goto out;
    }

    /* Get the underlying parts array and its size. */
    parts = (const char**)cgul_vector__get_array(cex, parts_vector);
    parts_count = cgul_vector__get_size(cex, parts_vector);

    /* Validate the line. */
    if (parts_count != 1) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid or missing data.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* The data is specified in hex and is padded to the nearest byte.
     * Because two hex characters specify a byte, the length of the
     * line should be at least "bytes_wide * 2". */
    i_max = bytes_wide * 2;
    if (strlen(parts[0]) < i_max) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid or missing data.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Convert the data that describes the shape of the glyph. */
    while (i < i_max) {

        /* Get the upper-half of the byte and set the corresponding
         * dots in the glyph. */
        bits = cgul_bdf__hex_digit_to_int(cex, parts[0][i++]);
        cgul_bdf_glyph__set_dot(cex, g, j++, row, bits & 0x8);
        cgul_bdf_glyph__set_dot(cex, g, j++, row, bits & 0x4);
        cgul_bdf_glyph__set_dot(cex, g, j++, row, bits & 0x2);
        cgul_bdf_glyph__set_dot(cex, g, j++, row, bits & 0x1);

        /* Get the lower-half of the byte and set the corresponding
         * dots in the glyph. */
        bits = cgul_bdf__hex_digit_to_int(cex, parts[0][i++]);
        cgul_bdf_glyph__set_dot(cex, g, j++, row, bits & 0x8);
        cgul_bdf_glyph__set_dot(cex, g, j++, row, bits & 0x4);
        cgul_bdf_glyph__set_dot(cex, g, j++, row, bits & 0x2);
        cgul_bdf_glyph__set_dot(cex, g, j++, row, bits & 0x1);
    }

    /* Check for errors. */
    if (*cex) {
        /* Free the pending exception. */
        cgul_exception__delete(*cex);
        /* Throw a more descriptive one. */
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid or missing data.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

 out:

    /* Clean up. */
    if (parts_vector) {
        /* Free each part. */
        cgul_bdf__free_line_parts(parts_vector);
        /* Free the vector. */
        cgul_vector__delete(parts_vector);
    }

    return;
}

/**
 * Process the BITMAP section initializing the glyph <code>g</code>.
 * If an error occurs, an exception is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] f BDF file
 * @param[in,out] g glyph
 * @param[in] bdf_file_name bdf file name
 */
static void
cgul_bdf__process_bitmap_section(cgul_exception_t* cex,
                                 cgul_bdf_t bdf,
                                 cgul_crlf_file_t f,
                                 cgul_bdf_glyph_t g,
                                 const char* bdf_file_name)
{
    unsigned int i = 0;
    unsigned int i_max = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);
    assert(g);
    assert(bdf_file_name);

    /* Process the BITMAP line. */
    cgul_bdf__process_bitmap_line(cex, bdf, f, bdf_file_name);

    /* There is one line of raw data for each row in the bitmap. */
    i_max = cgul_bdf_glyph__get_bb_height(cex, g);
    for (i = 0 ; i < i_max ; ++i) {
        cgul_bdf__process_bitmap_data_line(cex, bdf, f, g, i, bdf_file_name);
    }

 out:
    return;
}

/**
 * Process the ENDCHAR line.  If an error occurs, an exception is
 * thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] f BDF file
 * @param[in] bdf_file_name bdf file name
 */
static void
cgul_bdf__process_endchar_line(cgul_exception_t* cex,
                               cgul_bdf_t bdf,
                               cgul_crlf_file_t f,
                               const char* bdf_file_name)
{
    unsigned long int current_line = 0;
    const char* line = 0;
    const char** parts = NULL;
    size_t parts_count = 0;
    cgul_vector_t parts_vector = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);
    assert(bdf_file_name);

    /* Get the ENDCHAR line. */
    line = cgul_bdf__next_line(cex, bdf, f);
    if (*cex) {
        goto out;
    }

    /* Get the current line. */
    current_line = cgul_crlf_file__get_line_count(cex, f);

    /* Check for EOF. */
    if (line == NULL) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Unexpected end of file.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Split the line. */
    parts_vector = cgul_bdf__split_line(cex, line);
    if (*cex) {
        goto out;
    }

    /* Get the underlying parts array and its size. */
    parts = (const char**)cgul_vector__get_array(cex, parts_vector);
    parts_count = cgul_vector__get_size(cex, parts_vector);

    /* Validate the line. */
    if ((parts_count != 1) || (strcmp(parts[0], "ENDCHAR") != 0)) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid or missing ENDCHAR.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

 out:

    /* Clean up. */
    if (parts_vector) {
        /* Free each part. */
        cgul_bdf__free_line_parts(parts_vector);
        /* Free the vector. */
        cgul_vector__delete(parts_vector);
    }

    return;
}

/**
 * Process a character section initializing the glyph <code>g</code>.
 * If an error occurs, an exception is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] f BDF file
 * @param[out] g glyph
 * @param[in] bdf_file_name bdf file name
 */
static void
cgul_bdf__process_character_section(cgul_exception_t* cex,
                                    cgul_bdf_t bdf,
                                    cgul_crlf_file_t f,
                                    cgul_bdf_glyph_t g,
                                    const char* bdf_file_name)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);
    assert(g);
    assert(bdf_file_name);

    /* Process the STARTCHAR line. */
    cgul_bdf__process_startchar_line(cex, bdf, f, g, bdf_file_name);

    /* Process the ENCODING line. */
    cgul_bdf__process_encoding_line(cex, bdf, f, g, bdf_file_name);

    /* Process the SWIDTH line. */
    cgul_bdf__process_swidth_line(cex, bdf, f, g, bdf_file_name);

    /* Process the DWIDTH line. */
    cgul_bdf__process_dwidth_line(cex, bdf, f, g, bdf_file_name);

    /* Process the BBX line. */
    cgul_bdf__process_bbx_line(cex, bdf, f, g, bdf_file_name);

    /* Process the BITMAP section. */
    cgul_bdf__process_bitmap_section(cex, bdf, f, g, bdf_file_name);

    /* Process the ENDCHAR line. */
    cgul_bdf__process_endchar_line(cex, bdf, f, bdf_file_name);

 out:
    return;
}

/**
 * Process the character sections.  If an error occurs, an exception
 * is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] f BDF file
 * @param[in] bdf_file_name bdf file name
 */
static void
cgul_bdf__process_character_sections(cgul_exception_t* cex,
                                     cgul_bdf_t bdf,
                                     cgul_crlf_file_t f,
                                     const char* bdf_file_name)
{
    int insert_rv = 0;
    unsigned long int i = 0;
    cgul_bdf_glyph_t glyph = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);
    assert(bdf_file_name);

    /* Iterate over each STARTCHAR ... ENDCHAR section. */
    for (i = 0 ; i < bdf->char_count ; ++i) {

        /* Create a new glyph object that will hold the glyph
         * described by this STARTCHAR ... ENDCHAR section and that
         * will be inserted into the bdf->glyphs tree as the value for
         * this encoding.  Note that "i" is not the encoding.  The
         * encoding is read from the BDF file. */
        glyph = cgul_bdf_glyph__new(cex);
        if (*cex) {
            goto out;
        }

        /* Initialize the glyph by reading the next character section.
         * Note that this loads the encoding into glyph->encoding. */
        cgul_bdf__process_character_section(cex, bdf, f, glyph, bdf_file_name);
        if(*cex) {
            goto out;
        }

        /* Check for an unencoded glyph.  These are the ones with an
         * encoding line that looks like "ENCODING -1". */
        if (glyph->encoding == CGUL_BDF__UNENCODED) {
            /* The BDF spec allows use to discard unencoded glyphs. */
            cgul_bdf_glyph__delete(glyph);
            glyph = NULL;
            continue;
        }

        /* Add the glyph to "bdf".  If a glyph for this same encoding
         * exists from apreviously loaded BDF, the new glyph will not
         * be inserted.  The first glyph is the one that will be used.
         * This is how the glyphs are merged when multiple BDFs are
         * used. */
        insert_rv = cgul_bdf__add_glyph(cex, bdf, glyph);
        if (*cex) {
            goto out;
        }

        /* Test if a glyph for this encoding already exists. */
        if (!insert_rv) {
            /* A glyph already exists for this encoding so delete the
             * new one that is not going to be used.*/
            cgul_bdf_glyph__delete(glyph);
            glyph = NULL;
            continue;
        }

        /* The insertion was a success. The bdf->glyphs tree now owns
         * the glyph. */
        glyph = NULL;
    }

 out:

    /* Clean up. */
    if (glyph) {
        cgul_bdf_glyph__delete(glyph);
    }

    return;
}

/**
 * Process the ENDFONT line.  If an error occurs, an exception is
 * thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] f BDF file
 * @param[in] bdf_file_name bdf file name
 */
static void
cgul_bdf__process_endfont_line(cgul_exception_t* cex,
                               cgul_bdf_t bdf,
                               cgul_crlf_file_t f,
                               const char* bdf_file_name)
{
    unsigned long int current_line = 0;
    const char* line = 0;
    const char** parts = NULL;
    size_t parts_count = 0;
    cgul_vector_t parts_vector = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);
    assert(bdf_file_name);

    /* Read what should be the last line of the file. */
    line = cgul_bdf__next_line(cex, bdf, f);
    if (*cex) {
        goto out;
    }

    /* Get the current line. */
    current_line = cgul_crlf_file__get_line_count(cex, f);

    /* Check for EOF. */
    if (line == NULL) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Unexpected end of file.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Split the line. */
    parts_vector = cgul_bdf__split_line(cex, line);
    if (*cex) {
        goto out;
    }

    /* Get the underlying parts array and its size. */
    parts = (const char**)cgul_vector__get_array(cex, parts_vector);
    parts_count = cgul_vector__get_size(cex, parts_vector);

    /* Validate the line. */
    if ((parts_count != 1) || (strcmp(parts[0], "ENDFONT") != 0)) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Invalid or missing ENDFONT.",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

 out:

    /* Clean up. */
    if (parts_vector) {
        /* Free each part. */
        cgul_bdf__free_line_parts(parts_vector);
        /* Free the vector. */
        cgul_vector__delete(parts_vector);
    }

    return;
}

/**
 * This method generates the default glyph which gets used when the
 * user requests the glyph for a character that is not present in the
 * BDF font.  If the BDF font has a DEFAULT_CHAR property, that
 * character is used to generate the default glyph; if not, a glyph is
 * generated that looks like an empty box.  In both cases, a new glyph
 * is generated and saved to <code>bdf->default_glyph</code>.  If an
 * error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 */
static void
cgul_bdf__generate_default_glyph(cgul_exception_t* cex,
                                 cgul_bdf_t bdf)
{
    int i = 0;
    int i_max = 0;
    unsigned int ascent = 0;
    unsigned int dwidth_x = 0;
    cgul_bdf_glyph_t real_glyph = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(bdf->default_glyph == NULL);

    /* Create the new glyph. */
    bdf->default_glyph = cgul_bdf_glyph__new(cex);
    if (*cex) {
        goto out;
    }

    /* If the BDF font specified a DEFAULT_CHAR property, try to look
     * up the glyph for that character. */
    if (bdf->default_char_found) {

        /* Try to find the glyph for the default character. */
        real_glyph =
            cgul_bdf__get_glyph(cex, bdf, bdf->default_char, 0);
        if (*cex) {
            goto out;
        }

        /* If we found the real glyph, copy it to bdf->default_glyph. */
        if (real_glyph) {
            cgul_bdf_glyph__copy(cex, bdf->default_glyph, real_glyph);
            goto out;
        }
    }

    /*
     * If you reach here, it is either because the BDF font did not
     * specify a DEFAULT_CHAR property or it did specify one, but it
     * does not exist.  Either way, we need to generate one from
     * scratch.
     */

    /* Set up default values for "ascent" and "dwidth_x" of the box we
     * are going to draw by using the ascent of the font.  We may be
     * able to do better however because the ascent of the font is
     * usually several pixels taller than the ascent for capital
     * letters and numerals. */
    ascent = cgul_bdf__get_ascent(cex, bdf);
    dwidth_x = (2 * ascent) / 3;
    if (ascent > 6) {
        ascent -= 3;
    }

    /* It looks better if we can match the ascent of the empty box
     * with the ascent of capital leters and numerals in the font.  So
     * we try to get the ascent and dwidth_x of the '8' character. */
    real_glyph = cgul_bdf__get_glyph(cex, bdf, (cgul_wchar_t)'8', 0);
    if (!real_glyph) {
        real_glyph = cgul_bdf__get_glyph(cex, bdf, (cgul_wchar_t)'A', 0);
    }
    if (*cex) {
        goto out;
    }

    /* If we found the glyph for either '8' or 'A', borrow its ascent
     * and dwidth_x. */
    if (real_glyph) {
        ascent = cgul_bdf_glyph__get_ascent(cex, real_glyph);
        dwidth_x = cgul_bdf_glyph__get_dwidth_x(cex, real_glyph);
    }

    /* Calculate and set the attributes for the glyph. */
    cgul_bdf_glyph__set_encoding(cex, bdf->default_glyph, (cgul_wchar_t)-1);
    cgul_bdf_glyph__set_swidth_x(cex, bdf->default_glyph, 0);
    cgul_bdf_glyph__set_swidth_y(cex, bdf->default_glyph, 0);
    cgul_bdf_glyph__set_dwidth_x(cex, bdf->default_glyph, dwidth_x);
    cgul_bdf_glyph__set_dwidth_y(cex, bdf->default_glyph, 0);
    cgul_bdf_glyph__set_bb_width(cex, bdf->default_glyph, dwidth_x - 2);
    cgul_bdf_glyph__set_bb_height(cex, bdf->default_glyph, ascent);
    cgul_bdf_glyph__set_bb_x_offset(cex, bdf->default_glyph, 1);
    cgul_bdf_glyph__set_bb_y_offset(cex, bdf->default_glyph, 0);
    if (*cex) {
        goto out;
    }

    /*
     * Draw the box.  Remember that (0,0) is the upper-left corner of
     * the bounding box.
     */

    /* Draw the horizontal lines.  You use "ascent - 1" below because
     * the set_dot() method takes the zero-based index that is being
     * set. */
    i_max = (int)dwidth_x - 2;
    for (i = 0 ; i < i_max ; ++i) {
        cgul_bdf_glyph__set_dot(cex, bdf->default_glyph, i, 0, 1);
        cgul_bdf_glyph__set_dot(cex, bdf->default_glyph, i, ascent - 1, 1);
    }

    /* Draw the vertical lines.  You use "dwidth_x - 3" below because
     * the set_dot() method takes the zero-based index that is being
     * set. */
    i_max = (int)ascent;
    for (i = 0 ; i < i_max ; ++i) {
        cgul_bdf_glyph__set_dot(cex, bdf->default_glyph, 0, i, 1);
        cgul_bdf_glyph__set_dot(cex, bdf->default_glyph, dwidth_x - 3, i, 1);
    }

 out:
    return;
}

/**
 * This method reads <code>index_file_name</code> to find the offset
 * in the BDF file for <code>encoding</code>.  If an offset is not
 * found, <code>0</code> will be returned; otherwise, the offset will
 * be returned.  Note that the structure of the BDF file is such that
 * no encoding can have an offset of zero.
 *
 * @param[in,out] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] index_file_name index file name
 * @param[in] encoding encoding
 * @return offset
 */
static unsigned long int
cgul_bdf__load_glyph_offset(cgul_exception_t* cex,
                            cgul_bdf_t bdf,
                            const char* index_file_name,
                            cgul_wchar_t encoding)
{
    unsigned long int rv = 0;
    const char* line = NULL;
    cgul_crlf_file_t f = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(index_file_name);

    /* Allocate a temporary string. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Open the index file. */
    f = cgul_crlf_file__new_from_fname(cex,
                                       index_file_name,
                                       1 /* is_block_buffered */);
    if (*cex) {
        goto out;
    }

    /*
     * Seek to the line that holds the offset for this glyph.  There
     * are a few things to know.
     *
     *   - Each line holds one offset and is exactly 11 characters
     *     long including the trailing '\n' character.
     *
     *   - The offsets are sorted by encoding starting with zero.
     *
     *   - Every encoding is preset up to the largest encoding in the
     *     BDF file.  Even encodings that do not have glyphs are
     *     present if they are in the range [0 - max_encoding].
     *
     *   - Encodings beyond the largest encoding in the BDF file are
     *     not in the index file.  This is done to try to limit the
     *     size of the index file.
     *
     * See cgul_bdf__write_index_file() to see exactly how the index file
     * is generated.
     */
    cgul_crlf_file__fseek(cex, f, encoding * 11, SEEK_SET);
    if (*cex) {
        goto out;
    }

    /* Get the offset as a string. */
    line = cgul_crlf_file__get_line(cex, f);
    if (!line) {
        /* If there is no line, it means "encoding" is larger than the
         * largest encoding in the BDF file which caused fseek() to
         * reach EOF which caused there to be no line. */
        goto out;
    }

    /* Convert from a string to an unsigned long.  Note that you must
     * use a base of 10 when converting because the offsets have
     * leading zeros which make them look like octal otherwise. */
    cgul_string__set_value(cex, s, line);
    rv = cgul_string__to_unsigned_long(cex, s, 10 /* base */);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (f) {
        cgul_crlf_file__delete(f);
    }
    if (s) {
        cgul_string__delete(s);
    }

    return rv;
}

/**
 * This method attempts to load the glyph from
 * <code>bdf_file_name</code> that is located at <code>offset</code>
 * in the BDF file.  If successful, the glyph is loaded into memory,
 * stored in <code>bdf->glyphs</code>, and finally returned;
 * otherwise, <code>NULL</code> is returned.  If an error occurs,
 * <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] bdf_file_name bdf file name
 * @param[in] index_file_name index file name
 * @param[in] offset offset
 * @return glyph located at <code>offset</code>
 */
static cgul_bdf_glyph_t
cgul_bdf__load_glyph(cgul_exception_t* cex,
                     cgul_bdf_t bdf,
                     const char* bdf_file_name,
                     const char* index_file_name,
                     unsigned long int offset)
{
    int add_rv = 0;
    cgul_crlf_file_t f = NULL;
    cgul_bdf_glyph_t g = NULL;
    cgul_bdf_glyph_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(bdf_file_name);
    assert(index_file_name);

    /* Create a glyph. */
    g = cgul_bdf_glyph__new(cex);
    if (*cex) {
        goto out;
    }

    /* Open the file. */
    f = cgul_crlf_file__new_from_fname(cex,
                                       bdf_file_name,
                                       1 /* is_block_buffered */);
    if (*cex) {
        goto out;
    }

    /* Seek to the start of the character. */
    cgul_crlf_file__fseek(cex, f, offset, SEEK_SET);
    if (*cex) {
        goto out;
    }

    /* Load the glyph from file. */
    cgul_bdf__process_character_section(cex, bdf, f, g, bdf_file_name);
    if (*cex) {
        /* Delete the current exception. */
        cgul_exception__delete(*cex);
        /* Throw a more meaningful exception. */
        *cex = cgul_exception__new(
                   "cgul_bdf__load_glyph: "
                   "BDF index file \"%s\" is out of sync with BDF \"%s\".",
                   index_file_name,
                   bdf_file_name);
        goto out;
    }

    /* Add the glyph to the tree. */
    add_rv = cgul_bdf__add_glyph(cex, bdf, g);
    if (*cex) {
        goto out;
    }

    /* Throw an exception if this is a duplicate. */
    if (!add_rv) {
        *cex = cgul_exception__new("cgul_bdf__load_glyph: "
                                   "Attempt to load duplicate glyph: %lu",
                                   (unsigned long int)g->encoding);
        goto out;
    }

    /* Success. */
    rv = g;

    /* "bdf" now owns g. */
    g = NULL;

 out:

    /* Clean up. */
    if (f) {
        cgul_crlf_file__delete(f);
    }
    if (g) {
        cgul_bdf_glyph__delete(g);
    }

    return rv;
}

/**
 * This method is a helper for <code>cgul_bdf__get_glyph()</code>
 * which calls it when it does not have the glyph for
 * <code>encoding</code> currently loaded in memory.  This function
 * tries to use the BDF index files to find the offset for the glyph
 * in the BDF file.  If that is a success, it will then load the glyph
 * into memory, store it in <code>bdf->glyphs</code>, and finally
 * return it; otherwise, <code>NULL</code> is returned.  If an error
 * occurs, <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] bdf <code>cgul_bdf</code> instance
 * @param[in] encoding encoding
 * @return glyph
 */
static cgul_bdf_glyph_t
cgul_bdf__get_glyph_from_file(cgul_exception_t* cex,
                              cgul_bdf_t bdf,
                              cgul_wchar_t encoding)
{
    cgul_bdf_files_t files = NULL;
    cgul_bdf_glyph_t rv = NULL;
    cgul_list_node_t node = NULL;
    unsigned long int offset = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(bdf->files);

    /* Iterate over the list of BDF files and their indexes. */
    node = cgul_list__get_front(cex, bdf->files);
    for ( ; node ; node = cgul_list_node__get_next(cex, node)) {

        /* Get the BDF file name and the BDF index file name. */
        files = (cgul_bdf_files_t)cgul_list_node__get_value(cex, node);
        assert(files);

        /* Make sure file names were specified. */
        if (!files->bdf_file_name || !files->index_file_name) {
            continue;
        }

        /* Load the offset for the glyph from the BDF index file. */
        offset = cgul_bdf__load_glyph_offset(cex,
                                             bdf,
                                             files->index_file_name,
                                             encoding);
        if (*cex) {
            goto out;
        }

        if (offset == 0) {
            /* The glyph is not present in this BDF.  Try the next BDF. */
            continue;
        }

        /* Load the glyph from the BDF file and store in bdf->glyphs. */
        rv = cgul_bdf__load_glyph(cex,
                                  bdf,
                                  files->bdf_file_name,
                                  files->index_file_name,
                                  offset);
        if (*cex) {
            goto out;
        }

        if (rv) {
            /* Success. */
            break;
        }

    }

 out:
    return rv;
}

cgul_bdf_t
cgul_bdf__new(cgul_exception_t* cex)
{
    cgul_bdf_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate the new cgul_bdf object. */
    rv = (cgul_bdf_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* By default we allow glyphs to expire (if the user also changes
     * the cache size to a positive value).  This value is set to
     * false when the client loads the first BDF file that does not
     * have an associated BDF index file. */
    rv->allow_cache_expiration = 1;

    /* Allocate the files list. */
    rv->files = cgul_list__new(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate the glyphs tree. */
    rv->glyphs = cgul_rbtree__new(cex, cgul_bdf__compare_encodings);
    if (*cex) {
        goto out;
    }

    /* Allocate the font list string. */
    rv->font_list = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (*cex && rv) {
        cgul_bdf__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_bdf__delete(cgul_bdf_t bdf)
{
    cgul_exception_t local = NULL;

    if (bdf) {

        /* Unload the font freeing all the glyphs as well as freeing
         * all the bdf and index file names stored in bdf->files. */
        cgul_bdf__unload(&local, bdf);

        /* Delete the font list string. */
        if (bdf->font_list) {
            cgul_string__delete(bdf->font_list);
        }

        /* Delete the glyphs tree. */
        if (bdf->glyphs) {
            cgul_rbtree__delete(bdf->glyphs);
        }

        /* Delete the files list. */
        if (bdf->files) {
            cgul_list__delete(bdf->files);
        }

        /* Free the object itself. */
        free(bdf);
    }
}

void
cgul_bdf__load(cgul_exception_t* cex,
               cgul_bdf_t bdf,
               const char* font_list)
{
    const char* font_name = NULL;
    const char* bdf_file_name = NULL;
    const char* index_file_name = NULL;
    cgul_string_t s1 = NULL;
    cgul_string_t s2 = NULL;
    cgul_strtok_t tok = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(font_list);

    /* Allocate tok, s1, and s2. */
    tok = cgul_strtok__new(cex);
    s1 = cgul_string__new(cex);
    s2 = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Load the tokenizer so it can split font_list. */
    cgul_strtok__set_value(cex, tok, font_list);
    if (*cex) {
        goto out;
    }

    /* Iterate over each font in font_list. */
    while ((font_name = cgul_strtok__get_next(cex, tok, "|", 0)) != NULL) {

        /*
         * Try use font_name exactly as given.
         */

        /* Generate the names of the BDF and BDF index file. */
        cgul_string__set_value(cex, s1, font_name);
        cgul_string__set_value(cex, s2, font_name);
        cgul_string__append_char(cex, s2, 'i');
        bdf_file_name = cgul_string__get_value(cex, s1);
        index_file_name = cgul_string__get_value(cex, s2);
        if (*cex) {
            goto out;
        }

        /* Try to load the BDF file along with the BDF index file.  If
         * no exception, the font loaded correctly.  Continue with the
         * next font in the list. */
        cgul_bdf__load_explicit(cex, bdf, bdf_file_name, index_file_name);
        if (*cex == NULL) {
            continue;
        }

        /* Quash the exception and try again. */
        cgul_exception__delete(*cex);
        *cex = NULL;

        /* Now try to load the BDF without the index file.  Again, if
         * no exception, the font loaded correctly.  Continue with the
         * next font in the list. */
        cgul_bdf__load_explicit(cex, bdf, bdf_file_name, NULL);
        if (*cex == NULL) {
            continue;
        }

        /* Quash the exception and try again. */
        cgul_exception__delete(*cex);
        *cex = NULL;

        /*
         * Try use font_name with ".bdf" appended.
         */

        /* Generate the names of the BDF and BDF index file. */
        cgul_string__set_value(cex, s1, font_name);
        cgul_string__append(cex, s1, ".bdf");
        cgul_string__set_value(cex, s2, font_name);
        cgul_string__append(cex, s2, ".bdfi");
        bdf_file_name = cgul_string__get_value(cex, s1);
        index_file_name = cgul_string__get_value(cex, s2);
        if (*cex) {
            goto out;
        }

        /* Try to load the BDF file along with the BDF index file.  If
         * no exception, the font loaded correctly.  Continue with the
         * next font in the list. */
        cgul_bdf__load_explicit(cex, bdf, bdf_file_name, index_file_name);
        if (*cex == NULL) {
            continue;
        }

        /* Quash the exception and try again. */
        cgul_exception__delete(*cex);
        *cex = NULL;

        /* Now try to load the BDF without the index file.  Again, if
         * no exception, the font loaded correctly.  Continue with the
         * next font in the list. */
        cgul_bdf__load_explicit(cex, bdf, bdf_file_name, NULL);
        if (*cex == NULL) {
            continue;
        }

        /* Quash the exception. */
        cgul_exception__delete(*cex);
        *cex = NULL;

        /* Throw a more meaningful exception. */
        *cex = cgul_exception__new(
                   "cgul_bdf__load: \"%s\": Unable to find font: %s",
                   font_list, font_name);
        goto out;
    }

 out:

    /* Clean up. */
    if (s2) {
        cgul_string__delete(s2);
    }
    if (s1) {
        cgul_string__delete(s1);
    }
    if (tok) {
        cgul_strtok__delete(tok);
    }

    return;
}

void
cgul_bdf__load_explicit(cgul_exception_t* cex,
                        cgul_bdf_t bdf,
                        const char* bdf_file_name,
                        const char* index_file_name)
{
    unsigned long int current_line = 0;
    cgul_crlf_file_t f = NULL;
    cgul_bdf_files_t files = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Do not assert index_file_name.  It is allowed to be NULL. */
    assert(bdf);
    assert(bdf_file_name);

    /* Allocate the "files" structure which is used for storing the
       BDF file name and the BDF index file name. */
    files = (cgul_bdf_files_t)cgul_libc__calloc(cex, 1, sizeof(*files));
    if (*cex) {
        goto out;
    }

    /* Open the BDF file.  We use cgul_crlf_file partly because it is
     * extremely efficient but mostly because it knows how to read
     * DOS, Mac, and Unix text files, and the BDF standard states that
     * BDF files can either be DOS or Unix text files. */
    f = cgul_crlf_file__new_from_fname(cex,
                                       bdf_file_name,
                                       1 /* is_block_buffered */);
    if (*cex) {
        goto out;
    }

    /* Remember the name of the file.  This is required when loading
     * glyphs on demand. */
    files->bdf_file_name = cgul_libc__strdup(cex, bdf_file_name);
    if (*cex) {
        goto out;
    }

    /* Test the index file if the caller specified one. */
    if (!index_file_name) {
        /* If no index file is specified, disable cache expiration;
         * otherwise, it will not be possible to reload an arbitrary
         * glyph after it expires. */
        bdf->allow_cache_expiration = 0;
    } else {
        /* Make sure we can open the index file. */
        FILE* tmp = fopen(index_file_name, "rb");
        if (!tmp) {
            *cex = cgul_exception__new("cgul_bdf__load: "
                                       "Unable to open BDF index file: %s",
                                       index_file_name);
            goto out;
        }
        /* Close the temporary file. */
        fclose(tmp);
        /* Remember the index_file_name. */
        files->index_file_name = cgul_libc__strdup(cex, index_file_name);
        if (*cex) {
            goto out;
        }
    }

    /* Add "files" to the list of files.  Note that if an error occurs
     * below, we just leave "files" on the list and do not bother
     * trying to unmerge.  It is documented in the comments to this
     * function that an error leaves "bdf" in an undefined state.  The
     * client needs to delete "bdf" and start over. */
    cgul_list__push_back(cex, bdf->files, files);
    if (*cex) {
        goto out;
    }

    /* "files" is now owned by the bdf->files list. */
    files = NULL;

    /* Process the STARTFONT line. */
    cgul_bdf__process_startfont_line(cex, bdf, f, bdf_file_name);

    /* Process the FONT line. */
    cgul_bdf__process_font_line(cex, bdf, f, bdf_file_name);

    /* Process the SIZE line. */
    cgul_bdf__process_size_line(cex, bdf, f, bdf_file_name);

    /* Process the FONTBOUNDINGBOX line. */
    cgul_bdf__process_fontboundingbox_line(cex, bdf, f, bdf_file_name);

    /* Process the properties section. */
    cgul_bdf__process_properties_section(cex, bdf, f, bdf_file_name);

    /* Check for errors. */
    if (*cex) {
        goto out;
    }

    /* Get the current line. */
    current_line = cgul_crlf_file__get_line_count(cex, f);

    /* Throw an exception if the font ascent was not found in the
     * properties section. */
    if (!bdf->ascent_found) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Unable to find FONT_ASCENT property",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Throw an exception if the font descent was not found in the
     * properties section. */
    if (!bdf->descent_found) {
        *cex = cgul_exception__new("%s:%lu:"
                                   "Unable to find FONT_DESCENT property",
                                   bdf_file_name,
                                   current_line);
        goto out;
    }

    /* Process the CHARS line. */
    cgul_bdf__process_chars_line(cex, bdf, f, bdf_file_name);

    /* If we don't have an index file, load all the glyphs now. */
    if (!index_file_name) {

        /* Process the character sections. */
        cgul_bdf__process_character_sections(cex, bdf, f, bdf_file_name);

        /* Process the ENDFONT line. */
        cgul_bdf__process_endfont_line(cex, bdf, f, bdf_file_name);

    }

    /* Check for errors. */
    if (*cex) {
        goto out;
    }

    /* Generate the default glyph. */
    if (!bdf->default_glyph) {
        cgul_bdf__generate_default_glyph(cex, bdf);
        if (*cex) {
            goto out;
        }
    }

    /* Append to the font list. */
    if (cgul_string__get_length(cex, bdf->font_list) > 0) {
        cgul_string__append_char(cex, bdf->font_list, '|');
    }
    cgul_string__append(cex, bdf->font_list, bdf_file_name);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (f) {
        cgul_crlf_file__delete(f);
    }
    if (files) {
        if (files->bdf_file_name) {
            free(files->bdf_file_name);
        }
        if (files->index_file_name) {
            free(files->index_file_name);
        }
        free(files);
    }

    return;
}

void
cgul_bdf__unload(cgul_exception_t* cex,
                 cgul_bdf_t bdf)
{
    void* value = NULL;
    cgul_bdf_files_t files = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);

    /* Remove and delete all the glyphs from the bdf->glyphs tree.
     * Do not delete the bdf->glyphs tree itself however. */
    if (bdf->glyphs) {
        /* Delete each glyph from the tree.  Note that the "key"
         * for this "key" value pair is really just a pointer to
         * "value->encoding" so you must not try to free it
         * because it will be freed when we delete the value. */
        while (cgul_rbtree__remove_back(cex,
                                        bdf->glyphs,
                                        NULL,
                                        &value))
        {
            /* Delete the value. */
            cgul_bdf_glyph__delete((cgul_bdf_glyph_t)value);
        }
    }

    /* font_name */
    if (bdf->font_name) {
        free(bdf->font_name);
        bdf->font_name = NULL;
    }

    /* Free the names of the BDF files and their index files.  Do not
     * delete the bdf->files list itself however. */
    if (bdf->files) {
        while ((value = cgul_list__pop_back(cex, bdf->files)) != NULL) {
            files = (cgul_bdf_files_t)value;
            if (files) {
                /* Free bdf_file_name. */
                if (files->bdf_file_name) {
                    free(files->bdf_file_name);
                }
                /* Free index_file_name. */
                if (files->index_file_name) {
                    free(files->index_file_name);
                }
                /* Free the structure itself. */
                free(files);
            }
        }
    }

    /* Reset the font list string. */
    if (bdf->font_list) {
        cgul_string__set_value(cex, bdf->font_list, "");
    }

    bdf->font_size = 0;
    bdf->dpi_x = 0;
    bdf->dpi_y = 0;
    bdf->bb_width = 0;
    bdf->bb_height = 0;
    bdf->bb_x_offset = 0;
    bdf->bb_y_offset = 0;
    bdf->ascent = 0;
    bdf->descent = 0;
    bdf->ascent_found = 0;
    bdf->descent_found = 0;
    bdf->char_count = 0;
    bdf->default_char = 0;
    bdf->default_char_found = 0;

    /* bdf->cache -- Do not reset bdf->cache size to zero.  The size
     * of the cache does not depend on which BDF file is loaded. */

    bdf->allow_cache_expiration = 1;

    /* default_glyph */
    if (bdf->default_glyph) {
        cgul_bdf_glyph__delete(bdf->default_glyph);
        bdf->default_glyph = NULL;
    }

 out:
    return;
}

unsigned long int
cgul_bdf__get_cache_size(cgul_exception_t* cex,
                         cgul_bdf_t bdf)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);

    rv = bdf->cache_size;

 out:
    return rv;
}

void
cgul_bdf__set_cache_size(cgul_exception_t* cex,
                         cgul_bdf_t bdf,
                         unsigned long int cache_size)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);

    /* Set the new cache size. */
    bdf->cache_size = cache_size;

    /* Expire glyphs based on the new cache size. */
    cgul_bdf__expire_glyphs(cex, bdf);

 out:
    return;
}

void
cgul_bdf__get_utf8_extents(cgul_exception_t* cex,
                           cgul_bdf_t bdf,
                           const char* utf8,
                           size_t utf8_length,
                           int substitute,
                           unsigned int* width,
                           unsigned int* height)
{
    cgul_wchar_t* utf32 = NULL;
    size_t utf32_length = 0;
    cgul_wstring_t utf32_wstring = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Do not assert "width" or "height" because they can be NULL. */
    assert(bdf);
    assert(utf8);

    /* Allocate a temporary UTF-32 string. */
    utf32_wstring = cgul_wstring__new(cex);
    if (*cex) {
        goto out;
    }

    /* Convert from UTF-8 to UTF-32 so we can use the code in
     * cgul_bdf__get_utf32_extents() to calculate the extents. */
    cgul_unicode__mbstowcs(cex, utf8, utf8_length, 1, utf32_wstring);
    if (*cex) {
        goto out;
    }

    /* Unpack the UTF-32 string. */
    utf32 = cgul_wstring__get_value(cex, utf32_wstring);
    utf32_length = cgul_wstring__get_length(cex, utf32_wstring);

    /* Calculate the extents. */
    cgul_bdf__get_utf32_extents(cex, bdf,
                                 utf32, utf32_length,
                                 substitute,
                                 width, height);

 out:

    /* Clean up. */
    if (utf32_wstring) {
        cgul_wstring__delete(utf32_wstring);
    }

    return;
}

/**
 * This method is a helper for
 * <code>cgul_bdf__get_utf32_extents()</code> so that the code that
 * handles EOL can all reside in one place even though it is needed in
 * the main loop to handle embedded EOL characters and it is needed
 * again after the main loop to handle strings that are missing their
 * trailing EOL character.  Because this method is just a place holder
 * for code that is common to multiple sections of
 * <code>cgul_bdf__get_utf32_extents()</code>, it is tightly coupled
 * with that method and can only really be understood in the context
 * of that method. <p>
 *
 * This method adds the height of the line as given by
 * <code>height_font</code> to <code>height_max</code>. <p>
 *
 * It then adjusts <code>width_line</code> by subtracting
 * <code>dwidth_x</code> from <code>width_line</code> which returns
 * the line width to the value it had before the last character on the
 * line was encountered.  It then adds <code>right_edge</code> to the
 * line width to get the exact number of pixels needed for the last
 * character.  This way, the final width of the line should be
 * exact. <p>
 *
 * It then sets <code>width_max</code> to <code>width_line</code> if
 * the new line is wider than the previous maximum line width. <p>
 *
 * Finally, it sets <code>dwidth</code>, <code>width_line</code>, and
 * <code>right_edge</code> to zero so that the next line can be
 * started without any artifacts from the line we just finished.
 *
 * @param[in,out] cex c-style exception
 * @param[in] height_font height of the font (i.e., space between lines)
 * @param[in,out] height_max maximum height of the text
 * @param[in,out] width_max maximum width of the text
 * @param[in,out] width_line width of the current line of text
 * @param[in,out] dwidth_x dwidth_x of the last character on the current line
 * @param[in,out] right_edge pixels from the origin to the right edge of the last character on the current line
 */
static void
cgul_bdf__get_utf32_extents__handle_eol(cgul_exception_t* cex,
                                        unsigned int height_font,
                                        unsigned int* height_max,
                                        unsigned int* width_max,
                                        unsigned int* width_line,
                                        unsigned int* dwidth_x,
                                        int* right_edge)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(height_max);
    assert(width_max);
    assert(width_line);
    assert(dwidth_x);

    /* Add this line to the total height of the text. */
    *height_max = cgul_defensive__add_unsigned(cex, *height_max, height_font);
    if (*cex) {
        goto out;
    }

    /* Fix the amount of space needed for the last character.
     * We subtract by dwidth_x to compensate for always adding
     * it in cgul_bdf__get_utf32_extents(). */
    if (*right_edge > (int)*dwidth_x) {
        *width_line = *width_line - *dwidth_x + *right_edge;
    }

    /* Remember the width of the longest line. */
    *width_max = CGUL_MAX(*width_max, *width_line);

    /* Reset the line-oriented values that must not carry over
     * (in case we get two '\n' characters in a row). */
    *dwidth_x = 0;
    *width_line = 0;
    *right_edge = 0;

 out:
    return;
}

void
cgul_bdf__get_utf32_extents(cgul_exception_t* cex,
                            cgul_bdf_t bdf,
                            const cgul_wchar_t* utf32,
                            size_t utf32_length,
                            int substitute,
                            unsigned int* width,
                            unsigned int* height)
{
    size_t i = 0;
    int right_edge = 0;
    int bb_x_offset = 0;
    unsigned int bb_width = 0;
    unsigned int dwidth_x = 0;
    unsigned int width_line = 0;
    unsigned int width_max = 0;
    unsigned int height_max = 0;
    unsigned int height_font = 0;
    cgul_wchar_t wc = CGUL_WCHAR__NUL;
    cgul_bdf_glyph_t glyph = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Do not assert "width" or "height" because they can be NULL. */
    assert(bdf);
    assert(utf32);

    /* Get the height of the font. */
    height_font = cgul_bdf__get_height(cex, bdf);

    /* Iterate over each wide character in "utf32". */
    for (i = 0 ; i < utf32_length ; ++i) {

        /* Get the next UTF-32 character. */
        wc = utf32[i];

        /* If this is an EOL character, update the final values and
         * continue with the first character on the next line. */
        if (wc == (cgul_wchar_t)'\n') {
            cgul_bdf__get_utf32_extents__handle_eol(cex,
                                                    height_font, &height_max,
                                                    &width_max, &width_line,
                                                    &dwidth_x, &right_edge);
            if (*cex) {
                goto out;
            }
            continue;
        }

        /* Get the glyph for wc. */
        glyph = cgul_bdf__get_glyph(cex, bdf, wc, substitute);
        if (*cex) {
            goto out;
        }
        if (glyph == NULL) {
            *cex = cgul_exception__new("cgul_bdf__get_utf32_extents: "
                                       "no glyph for character: %lu",
                                       (unsigned long)wc);
            goto out;
        }

        /* Get information about this glyph. */
        dwidth_x = cgul_bdf_glyph__get_dwidth_x(cex, glyph);
        bb_x_offset = cgul_bdf_glyph__get_bb_x_offset(cex, glyph);
        bb_width = cgul_bdf_glyph__get_bb_width(cex, glyph);

        /* The first character on the line, may extend into the
         * character cell to its left.  Add this amount to
         * width_line. */
        if (width_line == 0) {
            if (bb_x_offset < 0) {
                width_line += -bb_x_offset;
            }
        }

        /* Add the dwidth_x for this character to width_line.  The
         * dwidth_x is the spacing between the origin of this
         * character and the origin of the next character.  If you are
         * worried that this is might not be right for the last
         * character on the line, you are right.  We use the the
         * "right_edge" variable (below) to correct this. */
        width_line += dwidth_x;

        /* The last character on the line, may extend into the
         * character cell to its right, or another problem is that the
         * last character on the line might not use all of the
         * horizontal space in its own character cell.  So, we
         * calculate exactly how many pixels from the origin to the
         * right edge of this character.  We cannot do anything with
         * this value yet because this character might not be the last
         * character on the line.  So, we just remember the value for
         * use later. */
        right_edge = bb_x_offset + (int)bb_width;
        /* Make sure right_edge is not negative. */
        if (right_edge < 0) {
            right_edge = 0;
        }

    }

    /* Some input strings might be missing their trailing EOL
     * character, but we still need to handle this case just like it
     * was any other end of line.  */
    if (wc != (cgul_wchar_t)'\n') {
        cgul_bdf__get_utf32_extents__handle_eol(cex,
                                                height_font, &height_max,
                                                &width_max, &width_line,
                                                &dwidth_x, &right_edge);
        if (*cex) {
            goto out;
        }
    }

    /*
     * Success
     */

    /* Return the width. */
    if (width) {
        *width = width_max;
    }

    /* Return the height. */
    if (height) {
        *height = height_max;
    }

 out:
    return;
}

cgul_bdf_glyph_t
cgul_bdf__get_glyph(cgul_exception_t* cex,
                    cgul_bdf_t bdf,
                    cgul_wchar_t encoding,
                    int substitute)
{
    cgul_bdf_glyph_t rv = NULL;
    cgul_rbtree_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);

    /* Try to lookup "encoding" in the glyphs cache. */
    node = cgul_rbtree__find(cex, bdf->glyphs, &encoding);
    if (node) {
        /* If we found a glyph, return it. */
        rv = (cgul_bdf_glyph_t)cgul_rbtree_node__get_value(cex, node);
        assert(rv);
        /* If cache expiration is enabled, make sure this node is
         * marked as the most-recently used. */
        if (bdf->allow_cache_expiration) {
            cgul_rbtree__set_youngest(cex, bdf->glyphs, node);
        }
        /* Success. */
        goto out;
    }

    /* Try to load the glyph using one of the index files. */
    rv = cgul_bdf__get_glyph_from_file(cex, bdf, encoding);
    if (*cex) {
        goto out;
    }
    if (rv) {
        /* Success.  Note that old glyphs will be expired when the
         * method above calls cgul_bdf__add_glyph().  Also note that
         * we do not need to explicitly mark the new glyph as being
         * the most-recently used/youngest because newly inserted
         * glyphs are always the youngest. */
        goto out;
    }

    /* If we didn't find a glyph and the user wants to substitute the
     * default glyph for the missing glyph, return the default
     * glyph. */
    if (substitute) {
        rv = bdf->default_glyph;
    }

 out:
    return rv;
}

void
cgul_bdf__shift_right(cgul_exception_t* cex,
                      cgul_bdf_t bdf)
{
    int right_shift = 0;
    int x_offset_new = 0;
    cgul_bdf_glyph_t glyph = NULL;
    cgul_rbtree_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(bdf->files);
    assert(bdf->glyphs);

    /* This method works exclusively on the glyphs in the cache so
     * make sure all of the BDF files were loaded without an index
     * file in order to force all the glyphs into the "glyphs"
     * cache. */
    cgul_bdf__verify_no_index_files(cex, bdf, "cgul_bdf__shift_right");
    if (*cex) {
        goto out;
    }

    /* We only need to shift right if the font's bounding box has a
     * negative x-offset. */
    if (bdf->bb_x_offset < 0) {

        /* Determine the number of pixels used to shift each glyph right. */
        right_shift = -(bdf->bb_x_offset);

        /* Iterate over each glyph. */
        node = cgul_rbtree__get_front(cex, bdf->glyphs);
        for ( ; node ; node = cgul_rbtree_node__get_next(cex, node)) {

            /* Get the glyph. */
            glyph = (cgul_bdf_glyph_t)cgul_rbtree_node__get_value(cex, node);
            if (!glyph) {
                continue;
            }

            /* If the bounding box doesn't have a height, it is
             * effectively empty so there is nothing to shift right.
             * These types of glyphs are usually used for spacing, and
             * it is important that they have a bounding box with a
             * width, height, x-offset, and y-offset all set to
             * zero. */
            if (cgul_bdf_glyph__get_bb_height(cex, glyph) == 0) {
                /* Make sure the other parameters of the box are all
                 * zeros.  This will prevent calls to pad_left() from
                 * giving the bounding box a non-zero width. */
                cgul_bdf_glyph__set_bb_width(cex, glyph, 0);
                cgul_bdf_glyph__set_bb_x_offset(cex, glyph, 0);
                cgul_bdf_glyph__set_bb_y_offset(cex, glyph, 0);
                continue;
            }

            /*
             * Handle all other glyphs.
             */

            /* Calculate the new x-offset. */
            x_offset_new =
                cgul_bdf_glyph__get_bb_x_offset(cex, glyph) + right_shift;

            /* Set the new x-offset. */
            cgul_bdf_glyph__set_bb_x_offset(cex, glyph, x_offset_new);
        }

        /* Remember that the entire font has been shifted right. */
        bdf->bb_x_offset = 0;
    }

 out:
    return;
}

void
cgul_bdf__pad_left(cgul_exception_t* cex,
                   cgul_bdf_t bdf)
{
    int x = 0;
    int y = 0;
    int width = 0;
    int height = 0;
    int x_offset = 0;
    int dot = 0;
    cgul_bdf_glyph_t glyph = NULL;
    cgul_bdf_glyph_t tmp = NULL;
    cgul_rbtree_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);

    /* This method works exclusively on the glyphs in the cache so
     * make sure all of the BDF files were loaded without an index
     * file in order to force all the glyphs into the "glyphs"
     * cache. */
    cgul_bdf__verify_no_index_files(cex, bdf, "cgul_bdf__pad_left");
    if (*cex) {
        goto out;
    }

    /* Allocate a temporary glyph. */
    tmp = cgul_bdf_glyph__new(cex);
    if (*cex) {
        goto out;
    }

    /* Iterate over each glyph. */
    node = cgul_rbtree__get_front(cex, bdf->glyphs);
    for ( ; node ; node = cgul_rbtree_node__get_next(cex, node)) {

        /* Get the glyph. */
        glyph = (cgul_bdf_glyph_t)cgul_rbtree_node__get_value(cex, node);
        if (!glyph) {
            continue;
        }

        /* Get the x-offset for this glyph's bounding box. */
        x_offset = cgul_bdf_glyph__get_bb_x_offset(cex, glyph);

        /* Sanity check to make sure the caller has called
         * shift_right() before calling this method. */
        if (x_offset < 0) {
            *cex = cgul_exception__new(
                       "Client failed to call cgul_bdf__shift_right() "
                       "before calling cgul_bdf__pad_left().");
            goto out;
        }

        /* There is nothing to do if the x-offset is already zero. */
        if (x_offset == 0) {
            continue;
        }

        /* Make a copy of the glyph that we can modify. */
        cgul_bdf_glyph__copy(cex, tmp, glyph);
        if (*cex) {
            goto out;
        }

        /* Force the x-offset of tmp to zero. */
        cgul_bdf_glyph__set_bb_x_offset(cex, tmp, 0);

        /* Increase the width of the glyph so that we can add padding
         * to compensate for forcing the x-offset to zero.  Note that
         * the cgul_bdf_glyph__set_bb_width() method knows to allocate
         * a new empty bitmap buffer when the width changes size. */
        width = (int)cgul_bdf_glyph__get_bb_width(cex, glyph) + x_offset;
        cgul_bdf_glyph__set_bb_width(cex, tmp, width);
        if (*cex) {
            goto out;
        }

        /* Generate the new glyph that has the padding on the left. */
        height = (int)cgul_bdf_glyph__get_bb_height(cex, glyph);
        for (y = 0 ; y < height ; ++y) {
            for (x = 0 ; x < width ; ++x) {

                /* Add the padding. */
                if (x < x_offset) {
                    cgul_bdf_glyph__set_dot(cex, tmp, x, y, 0);
                    continue;
                }

                /* Copy the pixel from the original glyph into place. */
                dot = cgul_bdf_glyph__get_dot(cex, glyph, x - x_offset, y);
                cgul_bdf_glyph__set_dot(cex, tmp, x, y, dot);
            }
        }

        /* Copy our results back to "glyph." */
        cgul_bdf_glyph__copy(cex, glyph, tmp);
        if (*cex) {
            goto out;
        }

    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_bdf_glyph__delete(tmp);
    }

    return;
}

extern void
cgul_bdf__write(cgul_exception_t* cex,
                cgul_bdf_t bdf,
                const char* bdf_file_name)
{
    FILE* f = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(bdf_file_name);

    /* Make sure all the BDF files were loaded without an index file;
     * otherwise, it is too hard to count all the glyphs in the merged
     * BDF. */
    cgul_bdf__verify_no_index_files(cex, bdf, "cgul_bdf__write");
    if (*cex) {
        goto out;
    }

    /* Open the file for writing text. */
    f = fopen(bdf_file_name, "w");
    if (!f) {
        *cex = cgul_exception__new("cgul_bdf__write: "
                                   "Unable to open file: %s",
                                   bdf_file_name);
        goto out;
    }

    /* Write the BDF to the open file. */
    cgul_bdf__write_file(cex, bdf, f);

 out:

    if (f) {
        fclose(f);
    }

    return;
}

extern void
cgul_bdf__write_file(cgul_exception_t* cex,
                     cgul_bdf_t bdf,
                     FILE* f)
{
    cgul_rbtree_node_t node = NULL;
    cgul_bdf_glyph_t glyph = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);

    /* Make sure all the BDF files were loaded without an index file;
     * otherwise, it is too hard to count all the glyphs in the merged
     * BDF.  This also makes sure all the glyphs are loaded in the
     * bdf->glyphs tree from the start and never unloaded. */
    cgul_bdf__verify_no_index_files(cex, bdf, "cgul_bdf__write_file");
    if (*cex) {
        goto out;
    }

    /* Make sure the font has a name. */
    if (!bdf->font_name || (bdf->font_name[0] == '\0')) {
        *cex = cgul_exception__new("cgul_bdf__write_file(): "
                                   "Missing xlfd font name.");
        goto out;
    }

    cgul_libc__fprintf(cex,
                       f,
                       "STARTFONT 2.1\n");

    cgul_libc__fprintf(cex,
                       f,
                       "FONT %s\n",
                       bdf->font_name);

    cgul_libc__fprintf(cex,
                       f,
                       "SIZE %u %u %u\n",
                       bdf->font_size,
                       bdf->dpi_x,
                       bdf->dpi_y);

    cgul_libc__fprintf(cex, f,
                       "FONTBOUNDINGBOX %u %u %d %d\n",
                       bdf->bb_width,
                       bdf->bb_height,
                       bdf->bb_x_offset,
                       bdf->bb_y_offset);

    /* Keep the STARTPROPERTIES count up to date! */
    cgul_libc__fprintf(cex, f, "STARTPROPERTIES %d\n", 2);
    cgul_libc__fprintf(cex, f, "FONT_ASCENT %u\n", bdf->ascent);
    cgul_libc__fprintf(cex, f, "FONT_DESCENT %u\n", bdf->descent);
    cgul_libc__fprintf(cex, f, "ENDPROPERTIES\n");
    cgul_libc__fprintf(cex, f, "CHARS %lu\n",
                       cgul_rbtree__get_size(cex, bdf->glyphs));

    if (*cex) {
        goto out;
    }

    /* Print the "STARTCHAR" sections. */
    node = cgul_rbtree__get_front(cex, bdf->glyphs);
    for ( ; node ; node = cgul_rbtree_node__get_next(cex, node)) {
        /* Get the glyph. */
        glyph = (cgul_bdf_glyph_t)cgul_rbtree_node__get_value(cex, node);
        if (glyph) {
            /* Write the glyph. */
            cgul_bdf_glyph__write_file(cex, glyph, f);
            if (*cex) {
                goto out;
            }
        }
    }

    /* ENDFONT */
    cgul_libc__fprintf(cex, f, "ENDFONT\n");

 out:
    return;
}

void
cgul_bdf__write_index(cgul_exception_t* cex,
                      cgul_bdf_t bdf,
                      const char* index_file_name)
{
    FILE* f = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(index_file_name);

    /* Index files are specific to one BDF.  So, it does not make any
     * sense to attempt to write an index file when the user has
     * merged multiple BDFs by calling cgul_bdf__load() more than
     * once. */
    if (cgul_list__get_size(cex, bdf->files) > 1) {
        *cex = cgul_exception__new(
                   "cgul_bdf__write_index: Attempt to write a new index "
                   "when multiple BDF files are loaded.  Call "
                   "cgul_bdf__load_explicit() with NULL as the BDF index "
                   "exactly once instead.");
        goto out;
    }

    /* Make sure the bdf was loaded without an index file in order to
     * force all the glyphs into the cache. */
    cgul_bdf__verify_no_index_files(cex, bdf, "cgul_bdf__write_index");
    if(*cex) {
        goto out;
    }

    /* Open the bdfi index file in binary mode. */
    f = fopen(index_file_name, "wb");
    if (!f) {
        *cex = cgul_exception__new("cgul_bdf__write_index: "
                                   "Unable to open file: %s",
                                   index_file_name);
        goto out;
    }

    /* Write the index file. */
    cgul_bdf__write_index_file(cex, bdf, f);

 out:

    if (f) {
        fclose(f);
    }

    return;
}

void
cgul_bdf__write_index_file(cgul_exception_t* cex,
                           cgul_bdf_t bdf,
                           FILE* f)
{
    cgul_wchar_t i = 0;
    cgul_wchar_t i_max = 0;
    unsigned long int offset = 0;
    cgul_bdf_glyph_t glyph = NULL;
    cgul_rbtree_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(f);

    /* Index files are specific to one BDF.  So, it does not make any
     * sense to attempt to write an index file when the user has
     * merged multiple BDFs by calling cgul_bdf__load() more than
     * once. */
    if (cgul_list__get_size(cex, bdf->files) > 1) {
        *cex = cgul_exception__new(
                   "cgul_bdf__write_index_file: Attempt to writing a new index "
                   "when multiple BDF files are loaded.  Call "
                   "cgul_bdf__load_explicit() with NULL as the BDF index "
                   "exactly once instead.");
        goto out;
    }

    /* Make sure the bdf was loaded without an index file in order to
     * force all the glyphs into the cache. */
    cgul_bdf__verify_no_index_files(cex, bdf, "cgul_bdf__write_index_file");
    if(*cex) {
        goto out;
    }

    /* Find the largest encoding in the font and assign it to i_max. */
    node = cgul_rbtree__get_back(cex, bdf->glyphs);
    if (!node) {
        goto out;
    }
    glyph = (cgul_bdf_glyph_t)cgul_rbtree_node__get_value(cex, node);
    assert(glyph);
    i_max = cgul_bdf_glyph__get_encoding(cex, glyph);

    /* Iterate over all glyphs in the range [0, i_max]. */
    for (i = 0 ; i <= i_max ; ++i) {
        glyph = cgul_bdf__get_glyph(cex, bdf, i, 0 /* substitute */);
        if (*cex) {
            goto out;
        }
        if (glyph) {
            /* Get the offset for this glyph. */
            offset = cgul_bdf_glyph__get_file_offset(cex, glyph);
        } else {
            /* Use an offset of zero to indicate that the glyph does
             * not exist.  This works because the format of a BDF file
             * is such that it is impossible for a STARTCHAR section
             * to appear as the first thing in a BDF file. */
            offset = 0;
        }
        /* If you change this line, update
         * cgul_bdf__load_glyph_offset() so it can read your
         * changes! */
        cgul_libc__fprintf(cex, f, "%010lu\n", offset);
    }

 out:
    return;
}

unsigned int
cgul_bdf__get_ascent(cgul_exception_t* cex,
                     cgul_bdf_t bdf)
{
    unsigned int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);

    rv = bdf->ascent;

 out:
    return rv;
}

unsigned int
cgul_bdf__get_descent(cgul_exception_t* cex,
                      cgul_bdf_t bdf)
{
    unsigned int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);

    rv = bdf->descent;

 out:
    return rv;
}

unsigned int
cgul_bdf__get_height(cgul_exception_t* cex,
                     cgul_bdf_t bdf)
{
    unsigned int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);

    rv = cgul_defensive__add_unsigned(cex, bdf->ascent, bdf->descent);

 out:
    return rv;
}

const char*
cgul_bdf__get_font_list(cgul_exception_t* cex,
                        cgul_bdf_t bdf)
{
    char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);

    rv = cgul_string__get_value(cex, bdf->font_list);

 out:
    return rv;
}

const char*
cgul_bdf__get_font_name(cgul_exception_t* cex,
                        cgul_bdf_t bdf)
{
    char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);

    rv = bdf->font_name;

 out:
    return rv;
}

unsigned int
cgul_bdf__get_font_size(cgul_exception_t* cex,
                        cgul_bdf_t bdf)
{
    unsigned int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);

    rv = bdf->font_size;

 out:
    return rv;
}

unsigned int
cgul_bdf__get_dpi_x(cgul_exception_t* cex,
                    cgul_bdf_t bdf)
{
    unsigned int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);

    rv = bdf->dpi_x;

 out:
    return rv;
}

unsigned int
cgul_bdf__get_dpi_y(cgul_exception_t* cex,
                    cgul_bdf_t bdf)
{
    unsigned int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);

    rv = bdf->dpi_y;

 out:
    return rv;
}

unsigned int
cgul_bdf__get_bb_width(cgul_exception_t* cex,
                       cgul_bdf_t bdf)
{
    unsigned int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);

    rv = bdf->bb_width;

 out:
    return rv;
}

unsigned int
cgul_bdf__get_bb_height(cgul_exception_t* cex,
                        cgul_bdf_t bdf)
{
    unsigned int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);

    rv = bdf->bb_height;

 out:
    return rv;
}

int
cgul_bdf__get_bb_x_offset(cgul_exception_t* cex,
                          cgul_bdf_t bdf)
{
    unsigned int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);

    rv = bdf->bb_x_offset;

 out:
    return rv;
}

int
cgul_bdf__get_bb_y_offset(cgul_exception_t* cex,
                          cgul_bdf_t bdf)
{
    unsigned int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);

    rv = bdf->bb_y_offset;

 out:
    return rv;
}

unsigned long int
cgul_bdf__get_char_count(cgul_exception_t* cex,
                         cgul_bdf_t bdf)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bdf);
    assert(bdf->files);

    /* If there is only one BDF file loaded, bdf->char_count hold the
     * correct count of characters even if cache expiration is
     * enabled. */
    if (cgul_list__get_size(cex, bdf->files) <= 1) {
        rv = bdf->char_count;
        goto out;
    }

    /*
     * If you reach here, it is because more than one BDF file has
     * been loaded.
     */

    /* Make sure all the BDF files were loaded without an index file;
     * otherwise, it is too hard to count all the glyphs in the merged
     * BDF.  This also makes sure all the glyphs are loaded in the
     * bdf->glyphs tree from the start and never unloaded. */
    cgul_bdf__verify_no_index_files(cex, bdf, "cgul_bdf__get_char_count");
    if (*cex) {
        goto out;
    }

    /* Return the count of the merged glyphs. */
    rv = cgul_rbtree__get_size(cex, bdf->glyphs);

 out:
    return rv;
}
