#ifndef CGUL_BDF_P_H
#define CGUL_BDF_P_H

#include "cgul_common.h"
#include "cgul_bdf.h"
#include "cgul_list.h"
#include "cgul_rbtree.h"
#include "cgul_string.h"

CGUL_BEGIN_C

/** Data encapsulated by the <code>cgul_bdf</code> class. */
struct cgul_bdf {

    /**
     * List of loaded BDF files.  Each value in the list is a pointer
     * to a <code>cgul_bdf_files</code> structure.
     */
    cgul_list_t files;

    /**
     * This string describes all the fonts that have been loaded and
     * merged.  It has the same format as the font list string passed
     * into <code>cgul_bdf__load()</code>.
     */
    cgul_string_t font_list;

    /** Name of the BDF font as given on the FONT line. */
    char* font_name;

    /** The size of the font in points. */
    unsigned int font_size;

    /** The x-resolution of the font in dots per inch. */
    unsigned int dpi_x;

    /** The y-resolution of the font in dots per inch. */
    unsigned int dpi_y;

    /** The width of the font's bounding box. */
    unsigned int bb_width;

    /** The height of the font's bounding box. */
    unsigned int bb_height;

    /** The x-offset of the lower-left corner of the font's bounding box. */
    int bb_x_offset;

    /** The y-offset of the lower-left corner of the font's bounding box. */
    int bb_y_offset;

    /** The ascent of the font. */
    unsigned int ascent;

    /** The descent of the font. */
    unsigned int descent;

    /** Whether the BDF file provides the font ascent. */
    int ascent_found;

    /** Whether the BDF file provides the font descent. */
    int descent_found;

    /**
     * Count of the number of characters in the _last_ BDF file loaded.
     * It is used by cgul_bdf__process_character_sections() to iterate
     * over the STARTCHAR sections. <p>
     *
     * Never access this value if you need to know the _total_ number
     * of characters in the font; instead, call
     * cgul_bdf__get_char_count().  It knows how to find the total
     * number of characters (if possible). <p>
     */
    unsigned long int char_count;

    /** Map of each character to its glyph. */
    cgul_rbtree_t glyphs;

    /**
     * When <code>glyphs</code> exceeds <code>cache_size</code>, the
     * least-recently used (LRU) glyph in the tree will be removed but
     * only if <code>allow_cache_expiration</code> is true and only if
     * <code>cache_size</code> is not zero. <p>
     *
     * If <code>cache_size</code> is set to zero, glyphs will never
     * expire. If <code>cache_size</code> is non-zero, the client must
     * not itself attempt to cache glyphs because the underlying
     * object can be removed at any point.  Instead, the client must
     * either make copies of the glyphs or set <code>cache_size</code>
     * to be large enough and to always query the
     * <code>cgul_bdf</code> object in order to get a valid
     * pointer. <p>
     */
    unsigned long int cache_size;

    /**
     * This boolean keeps track of whether glyphs can be expired from
     * the cache at all.  It defaults to true, but if the client loads
     * a BDF file that does not have an associated BDF index file,
     * this boolean is set to false; otherwise, glyphs from such a BDF
     * file could never be reloaded after they expire.
     */
    int allow_cache_expiration;

    /**
     * The default characater.  If this is set by the BDF font, this
     * indicates the default glyph to use when one is missing.
     */
    cgul_wchar_t default_char;

    /** Whether the default character has been found. */
    int default_char_found;

    /**
     * The default glyph to use when the font does not have a glyph
     * for a particular character.  This will usually look like an
     * empty box. */
    cgul_bdf_glyph_t default_glyph;

};

/**
 * Structure that holds one BDF file name along with an optional BDF
 * index file name.  Pointers to these structures are held in the
 * <code>cgul_bdf::files</code> list.
 */
typedef struct {

    /** BDF file name. */
    char* bdf_file_name;

    /** BDF index file name. */
    char* index_file_name;

} cgul_bdf_files, *cgul_bdf_files_t;

CGUL_END_C

#endif  /*  CGUL_BDF_P_H  */
