#ifndef CGUL_CRC32_P_H
#define CGUL_CRC32_P_H

#include "cgul_common.h"
#include "cgul_crc32.h"

CGUL_BEGIN_C

/*
 * NOTE: The C standard guarantees that the "unsigned long" type is at
 * least 32 bits wide.
 */

/**
 * Data encapsulated by the <code>cgul_crc32</code> class.
 */
struct cgul_crc32 {

    /**
     * "g" is the Generator Polynomial (with the most significant bit
     * removed).  The common value for g is:
     *
     * 802.3-Frames: x^32 + x^26 + x^23 + x^22 + x^16 + x^12
     *                    + x^11 + x^10 + x^8  + x^7 + x^5
     *                    + x^4 + x^2 + x + 1
     *
     *       ==> g = 0x04c11db7
     *
     * The convention when dealing with the reflected CRC algorithm is
     * to specify the reflection of the real generator polynomial.
     * (For an example, take a look at the implementation of FCS-32 in
     * RFC 1662.)  Because of this convention, if you are using the
     * reflected CRC algorithm, the common value you should set g to
     * is 0xedb88320 (which the reflection of 0x04c11db7).
     */
    unsigned long g;

    /**
     * "init" is used to initialize the accumulator register.  By
     * initializing the accumulator to a non-zero value, you prevent a
     * common blind spot caused by invalid leading zeros.  By
     * convention, the "init" value is almost always 0xffffffff.
     */
    unsigned long init;

    /**
     * "final_xor" is the final value to XOR the calculated CRC with.
     * This prevents a common blind spot caused by invalid trailing
     * zeros.  By convention, this almost always 0xffffffff.
     */
    unsigned long final_xor;

    /**
     * Some hardware uses the most significant bit as a parity bit.
     * Thus, the hardware transmits the least significant bits first
     * starting with bit0.  Instead of reflecting the incoming bits
     * and calculating the checksum normally, the bits are fed through
     * the CRC shift register as they arrive.  To compensate, the
     * algorithm to calculate the CRC is different, and the CRC
     * calculated is actually the reflection of the real CRC.  FCS-32
     * (RFC 1662) is an example of a reflected CRC.
     */
    int reflected;

    /**
     * We may need to append the size to the byte stream that gets
     * CRCed.
     */
    int append_size;

    /**
     * We need to keep a count of the bytes we have processed so that
     * we can append it if necessary.
     */
    size_t size;

    /**
     * We may need to XOR the size with sizeXOR before appending.
     */
    size_t size_xor;

    /**
     * The lookup table used by this object.  The table will be
     * generated only once and shared among all other CRC32 objects
     * that use the same table.  Do not try to free this pointer
     * because it points to a lookup table saved in the
     * cgul_crc32__tables tree.
     */
    unsigned long* table;

    /**
     * We need a general purpose register where we can accumulate XOR
     * operations.  This variable is initialized with "init" above.
     */
    unsigned long accumulator;
};

/**
 * This structure keeps track of all the tables we have already
 * generated.
 */
typedef struct cgul_crc32__tables_key {
    unsigned long g;
    int reflected;
    unsigned long* table;
    size_t ref_count;
} *cgul_crc32__tables_key_t;

CGUL_END_C

#endif  /*  CGUL_CRC32_P_H  */
