#ifndef CGUL_CRLF_FILE_P_H
#define CGUL_CRLF_FILE_P_H

#include "cgul_common.h"
#include "cgul_crlf.h"
#include "cgul_crlf_file.h"
#include "cgul_string.h"

CGUL_BEGIN_C

/**
 * The type of source from which the file is read.  It is done like
 * this to avoid the hassle of setting up a polymorphic type.  This
 * should be the only class that has to deal with this.
 */
typedef enum cgul_crlf_file_source {

    /**
     * The source is unknown.
     */
    CGUL_CRLF_FILE_SOURCE__UNKNOWN = 0,

    /**
     * The source is a block-buffered file.
     */
    CGUL_CRLF_FILE_SOURCE__BLOCK_BUFFERED_FILE,

    /**
     * The source is a character-buffered file.
     */
    CGUL_CRLF_FILE_SOURCE__CHAR_BUFFERED_FILE,

    /**
     * The source is memory.
     */
    CGUL_CRLF_FILE_SOURCE__MEMORY

} cgul_crlf_file_source_t;

/**
 * Data encapsulated by the <code>cgul_crlf_file</code> class.
 */
struct cgul_crlf_file {

    /**
     * Type of source from which lines of text are read.
     */
    cgul_crlf_file_source_t source_type;

    /**
     * Name of the currently opened file.  If a <code>FILE*</code> was
     * passed in instead of a file name, <code>"FILE"</code> will be
     * used.  If memory was passed in instead of a file name,
     * <code>"MEMORY"</code> will be used.
     */
    char* fname;

    /**
     * File from which lines of text are read.
     */
    FILE* f;

    /**
     * Boolean that describes whether this class owns <code>f</code>.
     * If and only if this class owns <code>f</code>, will it be
     * closed when the object is deleted.
     */
    int we_own_f;

    /**
     * Underlying filter responsible for converting lines of text from
     * DOS, Mac, or Unix EOL convention for block-buffered files
     * (including memory-based files which can easily be made to
     * simulate block-buffered files with a simple memcpy()).
     */
    cgul_crlf_t crlf;

    /**
     * The memory buffer supplied by the user that holds the "file"
     * when the data source is memory.  Note that bytes are copied out
     * of this buffer into the (generally) smaller <code>buf</code>
     * which is then fed into <code>crlf</code> for conversion.
     */
    const char* mem;

    /**
     * Size of the <code>mem</code> buffer.
     */
    size_t msize;

    /**
     * Current offset into <code>mem</code>.  This marks the beginning
     * of the next block that will be copied to <code>buf</code> and
     * then fed into <code>crlf</code> for memory-based files.
     */
    size_t moffset;

    /**
     * The block buffer we feed to <code>cgul_crlf__convert()</code>.
     */
    unsigned char* buf;

    /**
     * Size of the <code>buf</code> block buffer.
     */
    size_t bsize;

    /**
     * String used when calling cgul_get_line_relaxed().
     */
    cgul_string_t s;

    /**
     * Remember whether we have finished.
     */
    int is_finished;

};

CGUL_END_C

#endif  /*  CGUL_CRLF_FILE_P_H  */
