#include "cgul_event_router_p.h"
#include "cgul_libc.h"
#include "cgul_string.h"

/** Singleton */
static cgul_event_router_t cgul_event_router__instance;

/**
 * This method is a callback for <code>cgul_list__traverse()</code>.
 * It is used to free the data stored in the node <code>n</code> of
 * the second-level container. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] l second-level container
 * @param[in] n node in <code>h</code>
 * @param[in] data client data (not used)
 * @return true in order to continue traversing <code>h</code>
 */
static int
cgul_event_router__level2_free_node(cgul_exception_t* cex,
                                    cgul_list_t l,
                                    cgul_list_node_t n,
                                    void* data)
{
    int rv = 0;
    cgul_event_router__listener_t listener = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    listener = (cgul_event_router__listener_t)cgul_list_node__get_value(cex, n);
    free(listener);
    rv = 1;

 out:
    return rv;
}

/**
 * This method is a callback for <code>cgul_rbtree__traverse()</code>.
 * It is used to free the key and value stored in the node
 * <code>n</code> of the first-level container. <p>
 *
 * Note that this method calls <code>cgul_list__traverse()</code> to
 * free the values in the nodes of the second-level container that
 * itself is held as the value of <code>n</code>. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] t first-level container
 * @param[in] n node in <code>h</code>
 * @param[in] data client data (not used)
 * @return true in order to continue traversing <code>t</code>
 */
static int
cgul_event_router__level1_free_node(cgul_exception_t* cex,
                                    cgul_rbtree_t t,
                                    cgul_rbtree_node_t n,
                                    void* data)
{
    int rv = 0;
    char* key = NULL;
    cgul_list_t level2 = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(n);

    /* Unpack the node. */
    key = (char*)cgul_rbtree_node__get_key(cex, n);
    level2 = (cgul_list_t)cgul_rbtree_node__get_value(cex, n);

    /* Free all the nodes in the second-level container. */
    cgul_list__traverse(cex,
                        level2,
                        &cgul_event_router__level2_free_node,
                        NULL);

    cgul_list__delete(level2);
    free(key);

    rv = 1;

 out:
    return rv;
}

/**
 * This method inserts or gets the second-level container associated
 * with the <code>rendezvous</code> string.  This method performs both
 * operations because <code>cgul_rbtree_tree_insert()</code> performs
 * both operations in order to improve performance by only searching
 * the tree once. <p>
 *
 * The first-level container maps rendezvous strings to a second-level
 * container that holds the callback functions and their client data.
 * Getting the second-level container is tantamount to getting all the
 * listeners for a particular rendezvous string. <p>
 *
 * If <code>rendezvous_out</code> is not <code>NULL</code>, it will
 * point to the copy of the <code>rendezvous</code> string that is
 * owned by the first-level container.  If an error occurs and
 * <code>rendezvous_out</code> is not <code>NULL</code>, the string it
 * points to will be set to <code>NULL</code>. <p>
 *
 * This method returns a newly inserted second-level container.  If an
 * error occurs, an exception is thrown, and <code>NULL</code> is
 * returned. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] er event router
 * @param[in] rendezvous rendezvous string
 * @param[out] rendezvous_out copy of the rendezvous string owned by
 * the level-one container
 * @return second-level container
 */
static cgul_list_t
cgul_event_router__level1_insert(cgul_exception_t* cex,
                                 cgul_event_router_t er,
                                 const char* rendezvous,
                                 char** rendezvous_out)
{
    int insert_rv = 0;
    cgul_string_t key = NULL;
    cgul_rbtree_node_t n = NULL;
    cgul_list_t rv = NULL;
    cgul_exception_t local = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(er);
    assert(rendezvous);

    /* Create a copy of "rendezvous" to try to insert into the
     * first-level container. */
    key = cgul_string__new_from_cstring(cex, rendezvous);
    if (*cex) {
        goto out;
    }

    /* Insert or get the existing node for the key. */
    insert_rv = cgul_rbtree__insert(cex,
                                    er->level1,
                                    cgul_string__get_value(cex, key),
                                    &n);
    if (*cex) {
        goto out;
    }

    /* If we couldn't insert a new key, it means the key already exists. */
    if (!insert_rv) {
        /* Return rendezvous_out. */
        if (rendezvous_out) {
            *rendezvous_out = (char*)cgul_rbtree_node__get_key(cex, n);
        }
        /* Return the second-level container. */
        rv = (cgul_list_t)cgul_rbtree_node__get_value(cex, n);
        goto out;
    }

    /* Create the second-level container to add as the "value" for
     * this key. */
    rv = cgul_list__new(cex);
    if (*cex) {
        goto out;
    }

    /* Finish mapping "key" to "rv". */
    cgul_rbtree_node__set_value(cex, n, rv);

    /* Return rendezvous_out. */
    if (rendezvous_out) {
        *rendezvous_out = cgul_string__get_value(cex, key);
    }

    /* Take ownership of the key. */
    cgul_string__take_value(cex, key);

 out:

    /* Clean up. */
    if (*cex && insert_rv) {
        /* Roll back the insertion. */
        cgul_rbtree__remove(&local, er->level1, rendezvous, NULL, NULL);
        /* Free the second-level container. */
        if (rv) {
            cgul_list__delete(rv);
            rv = NULL;
        }
        /* Clear rendezvous_out. */
        if (rendezvous_out) {
            *rendezvous_out = NULL;
        }
        /* We do not free the rendezvous string here because it is
         * still owned by "key" which we are about to delete. */
    }
    /* Always delete the key. */
    if (key) {
        cgul_string__delete(key);
    }

    return rv;
}

/**
 * This method removes the second-level container associated with
 * <code>rendezvous</code> from the first-level container.  This code
 * in this method has been factored out because it is needed in more
 * than one place. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] er event router
 * @param[in] rendezvous rendezvous string
 */
static void
cgul_event_router__level1_remove(cgul_exception_t* cex,
                                 cgul_event_router_t er,
                                 const char* rendezvous)
{
    void* key = NULL;
    void* level2 = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(er);
    assert(rendezvous);

    /* Remove the second-level container from the first-level container. */
    if (cgul_rbtree__remove(cex, er->level1, rendezvous, &key, &level2)) {

        /* Free all the nodes in the second-level container. */
        cgul_list__traverse(cex,
                            (cgul_list_t)level2,
                            &cgul_event_router__level2_free_node,
                            NULL);

        /* Delete the second-level container. */
        cgul_list__delete((cgul_list_t)level2);

        /* Free our copy of the rendezvous string. */
        free(key);
    }

 out:
    return;

}

/**
 * This method searches the first-level container for the second-level
 * container associated with <code>rendezvous</code> string.  If the
 * second-level container is found, it is returned; otherwise,
 * <code>NULL</code> is returned. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] er event router
 * @param[in] rendezvous rendezvous string
 * @return second-level container
 */
static cgul_list_t
cgul_event_router__get_level2(cgul_exception_t* cex,
                              cgul_event_router_t er,
                              const char* rendezvous)
{
    cgul_list_t rv = NULL;
    cgul_rbtree_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(er);
    assert(rendezvous);

    n = cgul_rbtree__find(cex, er->level1, rendezvous);
    if (n) {
        rv = (cgul_list_t)cgul_rbtree_node__get_value(cex, n);
        assert(rv);
    }

 out:
    return rv;
}

/**
 * This method is the constructor for the
 * <code>cgul_event_router</code> class.  It is private as part of
 * the singleton implementation.  Use
 * <code>cgul_event_router__get_instance()</code> instead.  If memory
 * cannot be allocated, <code>NULL</code> is returned, and an
 * exception is thrown. <p>
 *
 * @note If necessary, you can make this method part of the public
 * interface for this class.  You would typically want to do this if a
 * single instance of <code>cgul_event_router</code> is not
 * sufficient for your needs. <p>
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_event_router</code> instance
 */
static cgul_event_router_t
cgul_event_router__new(cgul_exception_t* cex)
{
    cgul_event_router_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = (cgul_event_router_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    rv->level1 = cgul_rbtree__new(cex, (cgul_rbtree__compare_t)&strcmp);

 out:

    /* Clean up partially constructed objects. */
    if (*cex && rv) {
        cgul_event_router__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_event_router_t
cgul_event_router__get_instance(cgul_exception_t* cex)
{
    cgul_event_router_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (cgul_event_router__instance == NULL) {
        cgul_event_router__instance = cgul_event_router__new(cex);
    }
    rv = cgul_event_router__instance;

 out:
    return rv;
}

void
cgul_event_router__delete(cgul_event_router_t er)
{
    cgul_exception_t local = NULL;

    if (er) {

        /* We are responsible for freeing each "value" stored in the
         * first-level container.  Each "value" stored in the
         * first-level container is a second-level container (which
         * holds the listeners for each event). */
        if (er->level1) {

            /* Free all the nodes in the first-level container which
             * in turn frees all the nodes in the second-level
             * container as well as deleting the second-level
             * containers themselves. */
            cgul_rbtree__traverse(&local,
                                  er->level1,
                                  &cgul_event_router__level1_free_node,
                                  NULL);

            /* Delete the first-level container. */
            cgul_rbtree__delete(er->level1);
        }

        /* Free this instance. */
        free(er);

        /* Prepare for the next call to cgul_event_router__get_instance(). */
        cgul_event_router__instance = NULL;

    }
}


cgul_event_router__listener_t
cgul_event_router__add_listener(cgul_exception_t* cex,
                                cgul_event_router_t er,
                                const char* rendezvous,
                                cgul_event_router__callback_t cb,
                                const void* data)
{
    cgul_event_router__listener_t rv = NULL;
    char* rendezvous_out = NULL;
    cgul_list_t level2 = NULL;
    cgul_list_node_t level2_node = NULL;
    cgul_event_router__listener_t listener = NULL;
    cgul_exception_t local = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(er);
    assert(rendezvous);
    assert(cb);

    /* Get the existing or insert a new second-level container for the
     * rendezvous string.  This is all done in one call because the
     * underlying implementation is a cgul_rbtree which is also
     * capable of doing the same which improves performance because
     * the tree only needs to be searched once. */
    level2 = cgul_event_router__level1_insert(cex,
                                              er,
                                              rendezvous,
                                              &rendezvous_out);
    if (*cex) {
        goto out;
    }

    /* Create the listener object. */
    listener = (cgul_event_router__listener_t)
        cgul_libc__calloc(cex, 1, sizeof(*listener));
    if (*cex) {
        goto out;
    }

    /* Add the new listener to the back of the second-level container. */
    level2_node = cgul_list__push_back(cex, level2, listener);
    if (*cex) {
        goto out;
    }

    /* Initialize the listener. */
    listener->rendezvous = rendezvous_out;
    listener->cb = cb;
    listener->data = data;
    listener->level2 = level2;
    listener->level2_node = level2_node;

    /* Success. */
    rv = (cgul_event_router__listener_t)listener;

 out:

    /* Rollback if errors occurred. */
    if (*cex) {
        /* Rollback the insertion into the second-level container. */
        if (level2_node) {
            cgul_list__pop_back(&local, level2);
        }
        /* Remove the second-level container, if it is empty. */
        if (level2 && (cgul_list__get_size(&local, level2) == 0)) {
            cgul_event_router__level1_remove(&local, er, rendezvous);
        }
        /* Free the listener pointer. */
        if (listener) {
            free(listener);
        }
    }

    return rv;
}

void
cgul_event_router__remove_listener(cgul_exception_t* cex,
                                   cgul_event_router_t er,
                                   cgul_event_router__listener_t listener)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(er);
    assert(listener);

    /* Remove "listener" from the second-level container. */
    cgul_list__remove_node(cex, listener->level2, listener->level2_node, NULL);

    /* If this rendezvous string has no more listeners, remove it from
     * the first-level container. */
    if (cgul_list__get_size(cex, listener->level2) == 0) {
        cgul_event_router__level1_remove(cex, er, listener->rendezvous);
    }

    /* Free the listener. */
    free(listener);

 out:
    return;
}

void
cgul_event_router__send_event(cgul_exception_t* cex,
                              cgul_event_router_t er,
                              const char* rendezvous,
                              const void* event)
{
    int cb_rv = 1;
    size_t i = 0;
    size_t cb_count = 0;
    cgul_list_t all = NULL;
    cgul_list_t level2 = NULL;
    cgul_list_node_t n = NULL;
    cgul_event_router__minimal_listener_t v = NULL;
    cgul_event_router__listener_t listener = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(er);
    assert(rendezvous);

    /* Get the list of listeners for the "all" event. */
    all = cgul_event_router__get_level2(cex, er, "all");

    /* Get the list of listeners for this specific event. */
    level2 = cgul_event_router__get_level2(cex, er, rendezvous);

    /* Get the total number of listeners. */
    cb_count = 0;
    if (all) {
        cb_count += cgul_list__get_size(cex, all);
    }
    if (level2) {
        cb_count += cgul_list__get_size(cex, level2);
    }

    /* If there are no listeners, we are done. */
    if (cb_count == 0) {
        goto out;
    }

    /* Allocate an array of cgul_event_router__minimal_listener
     * structures.  This array is used to hold copies of everything
     * you need to make all the callbacks.
     *
     * Do not attempt to cache v across calls to this method because
     * that would prevent the callbacks from recursively sending
     * events because there would then be only one "v". */
    v = (cgul_event_router__minimal_listener_t)
            cgul_libc__malloc(cex, cb_count * sizeof(*v));
    if (*cex) {
        goto out;
    }

    /* In order for the entire class to be re-entrant, this method
     * copies all of the relevant callback information from the
     * second-level containers _before_ it starts to make the
     * callbacks.  This makes it possible for the callback functions
     * to do things like remove themselves or recursively send events
     * because nothing will be iterating over the main data structures
     * when the callbacks are being made. */
    if (all) {
        n = cgul_list__get_front(cex, all);
        for (i = 0 ; n ; ++i, n = cgul_list_node__get_next(cex, n)) {
            listener = (cgul_event_router__listener_t)
                           cgul_list_node__get_value(cex, n);
            v[i].cb = listener->cb;
            v[i].data = listener->data;
        }
    }
    if (level2) {
        n = cgul_list__get_front(cex, level2);
        for ( ; n ; ++i, n = cgul_list_node__get_next(cex, n)) {
            listener = (cgul_event_router__listener_t)
                           cgul_list_node__get_value(cex, n);
            v[i].cb = listener->cb;
            v[i].data = listener->data;
        }
    }

    /* We are no longer iterating over the second-level container so
     * it is safe to start making callbacks. */
    for (i = 0 ; i < cb_count ; ++i) {
        cb_rv = (*(v[i].cb))(rendezvous, event, v[i].data);
        if (!cb_rv) {
            break;
        }
    }

 out:

    if (v) {
        free(v);
    }

    return;
}

void
cgul_event_router__set_data(cgul_exception_t* cex,
                            cgul_event_router_t er,
                            cgul_event_router__listener_t listener,
                            const void* data)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(er);
    assert(listener);

    /* Set the data. */
    listener->data = data;

 out:
    return;
}

size_t
cgul_event_router__get_listener_count(cgul_exception_t* cex,
                                      cgul_event_router_t er,
                                      const char* rendezvous)
{
    size_t rv = 0;
    cgul_list_t level2 = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(er);
    assert(rendezvous);

    /* Fetch the second-level container specified by the rendezvous string. */
    level2 = cgul_event_router__get_level2(cex, er, rendezvous);
    if (level2 == NULL) {
        goto out;
    }

    /* The size of the container is equivalent to the listener count. */
    rv = cgul_list__get_size(cex, level2);

 out:
    return rv;
}
