#ifndef CGUL_EVENT_ROUTER_P_H
#define CGUL_EVENT_ROUTER_P_H

#include "cgul_common.h"
#include "cgul_event_router.h"
#include "cgul_list.h"
#include "cgul_rbtree.h"

CGUL_BEGIN_C

/**
 * Data encapsulated by the <code>cgul_event_router</code> class.
 */
struct cgul_event_router {

    /**
     * This is the first-level container that maps a rendezvous string
     * to the second-level container that holds the list of listeners
     * waiting for the event associated with the rendezvous string.
     */
    cgul_rbtree_t level1;

};

/**
 * This structure is used for holding listeners and their client data
 * so that they can be sent events.
 *
 * IMPORTANT: This structure contains pointers to both the
 * second-level container and the node in the second-level container.
 * This makes it possible to find a particular callback in O(1) time.
 * Thus, you should never have to do a linear search through the
 * second-level container trying to find the callback.
 */
struct cgul_event_router__listener {

    /**
     * Copy of the rendezvous string.  This pointer is owned by the
     * first-level container.  Do not free it.
     */
    const char* rendezvous;

    /**
     * Registered callback.  This is just a function pointer.
     */
    cgul_event_router__callback_t cb;

    /**
     * Client data for <code>listener</code>.  The client owns this
     * chunck of memory.  Do not free it.
     */
    const void* data;

    /**
     * The second-level container where this listener can be found.
     * This just a copy of the pointer held owned by the first-level
     * container.  Do not delete it.
     */
    cgul_list_t level2;

    /**
     * The node in the second-level container where this listener can
     * be found.  This just a copy of the pointer held owned by the
     * second-level container.  Do not delete it.
     */
    cgul_list_node_t level2_node;

};

/**
 * In order to be re-entrant, this class copies the entire list of
 * callbacks to an array of structures of this type before making any
 * callbacks.
 */
typedef struct {

    /**
     * Registered listener.  This is just a function pointer.
     */
    cgul_event_router__callback_t cb;

    /**
     * Client data for <code>listener</code>.  The client owns this
     * chunck of memory.  Do not free it.
     */
    const void* data;

} *cgul_event_router__minimal_listener_t;

CGUL_END_C

#endif  /*  CGUL_EVENT_ROUTER_P_H  */
