#ifndef CGUL_HASH_NODE_P_H
#define CGUL_HASH_NODE_P_H

#include "cgul_common.h"
#include "cgul_hash_node.h"

CGUL_BEGIN_C

struct cgul_hash_node {

    /*
     * NOTE: Do not try to store a back-reference to the containing
     * cgul_hash.  It is very tempting to do so, but it makes
     * cgul_hash_swap() inefficient.  As it is currently implemented,
     * cgul_hash_swap() can simply do a shallow copy of the very few
     * data elments in the top-level "struct cgul_hash" structure.  If
     * you add a back-reference to this top-level cgul_hash structure
     * in each node, you'll have to iterate over the entire hash and
     * change each pointer to point to the new cgul_hash if you want
     * your cgul_hash_swap() method to do its job completely.
     * Instead, we just keep a few redundant copies of things out of
     * "struct cgul_hash".
     */

    /**
     * Key.
     */
    const void* key;

    /**
     * Value the user wants to store.  The user is responsible for
     * freeing anything he or she puts here.
     */
    const void* value;

    /**
     * Cache the hash value of <code>key</code> to avoid having to
     * recalculate it.  This should be safe as keys are not supposed
     * to change.
     */
    unsigned int hv;

    /**
     * Need to remember the index into the buckets so we can remove
     * the node when resizing.
     */
    unsigned int index;

    /**
     * The prev node in the same bucket.  Do not use this pointer for
     * general iteration; use the <code>older</code> and
     * <code>younger</code> pointers instead.  In fact
     * <code>cgul_hash_node__get_prev()</code> returns
     * <code>older</code> instead of returning this pointer.
     */
    cgul_hash_node_t prev;

    /**
     * The next node in the same bucket.  Do not use this pointer for
     * general iteration; use the <code>older</code> and
     * <code>younger</code> pointers instead.  In fact
     * <code>cgul_hash_node__get_next()</code> returns
     * <code>younger</code> instead of returning this pointer.
     */
    cgul_hash_node_t next;

    /**
     * Pointer to older node.  Together with <code>younger</code>,
     * <code>older</code> forms a doubly linked list that has nothing
     * to do with hash order; rather, the <code>older</code> node
     * points to the next older node.  If there is no older node,
     * <code>older</code> is <code>NULL</code>.  By calling
     * <code>cgul_hash_node__get_older()</code>, the user can iterate
     * over the nodes in reverse chronological order.
     */
    cgul_hash_node_t older;

    /**
     * Pointer to younger node.  Together with <code>older</code>,
     * <code>younger</code> forms a doubly linked list that has
     * nothing to do with hash order; rather, the <code>younger</code>
     * node points to the next younger node.  If there is no younger
     * node, <code>younger</code> is <code>NULL</code>.  By calling
     * <code>cgul_hash_node__get_younger()</code>, the user can
     * iterate over the nodes in chronological order.
     */
    cgul_hash_node_t younger;

};

/*
 * Private methods.
 */

/**
 * Create a new <code>cgul_hash_node</code> object.  The caller is
 * responsible for freeing the object by calling
 * <code>cgul_hash_node__delete()</code>.  If memory cannot be
 * allocated, <code>NULL</code> is returned and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_hash_node</code> instance
 */
extern cgul_hash_node_t
cgul_hash_node__new(cgul_exception_t* cex);

/**
 * This method frees all internally allocated memory.  Do not try to
 * access <code>n</code> after calling this method.
 *
 * @param[in] n <code>cgul_hash_node</code> instance
 */
extern CGUL_EXPORT void
cgul_hash_node__delete(cgul_hash_node_t n);

/**
 * Set the key for this node.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_hash_node</code> instance
 * @param[in] key key
 */
extern CGUL_EXPORT void
cgul_hash_node__set_key(cgul_exception_t* cex,
                        cgul_hash_node_t n,
                        const void* key);

CGUL_END_C

#endif  /*  CGUL_HASH_NODE_P_H  */
