#include "cgul_heap_sort.h"
#include "cgul_heap_primitives.h"
#include "cgul_libc.h"

void
cgul_heap_sort(cgul_exception_t* cex,
               void* v,
               size_t v_count,
               size_t element_size,
               cgul_heap_primitives__compare_t cf)
{
    void* scratch = NULL;
    void** pointers_array = NULL;
    char* tmp = NULL;
    size_t i = 0;
    size_t v_size = v_count * element_size;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(cf);

    /* Sanity checks. */
    if ((v_count <= 1) || (element_size <= 0)) {
        goto out;
    }

    /*
     * The general idea is to sort pointers that point into a copy of
     * "v" under the assumption that copying pointers during a sort
     * will be more efficient than copying entire structures during
     * the sort.  After sorting the pointers, we will copy individual
     * elements from the copy back into the original "v" based upon
     * the sorted pointers.
     */

    /* Create an exact copy of "v" and put it in "scratch". */
    scratch = cgul_libc__malloc(cex, v_size);
    if (*cex) {
        goto out;
    }
    memcpy(scratch, v, v_size);

    /* Generate an array of pointers into the scratch area. */
    pointers_array = (void**)cgul_libc__malloc(cex, sizeof(void*) * v_count);
    if (*cex) {
        goto out;
    }
    for (i = 0, tmp = (char*)scratch;
         i < v_count;
         ++i, tmp += element_size)
    {
        pointers_array[i] = tmp;
    }

    /* Sort the pointers. */
    cgul_heap_sort_pointers(cex, pointers_array, v_count, cf);
    if (*cex) {
        goto out;
    }

    /* Copy the elements back into "v" based upon the sorted pointers. */
    for (i = 0, tmp = (char*)v ;
         i < v_count ;
         ++i, tmp += element_size)
    {
        memcpy(tmp, pointers_array[i], element_size);
    }

 out:

    /* Clean up. */
    if (pointers_array) {
        free(pointers_array);
    }
    if (scratch) {
        free(scratch);
    }
}

void
cgul_heap_sort_pointers(cgul_exception_t* cex,
                        void** v,
                        size_t v_count,
                        cgul_heap_primitives__compare_t cf)
{
    void* tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(cf);

    /* Sanity checks. */
    if (v_count <= 1) {
        goto out;
    }

    /* Create the heap in-place from the unordered array. */
    cgul_heap_primitives__heapify(cex, v, v_count, cf);

    for ( ; v_count > 1 ; --v_count) {

        /* Swap the largest element to the back of the array. */
        tmp = v[v_count - 1];
        v[v_count - 1] = v[0];
        v[0] = tmp;

        /* Shift down the element that was just swapped to v[0].  Note
         * that swapping the largest remaining element into place at
         * the back of the array effectively shrinks the heap by one
         * element.  This is why (v_count - 1) is passed in as the
         * heap size and why we have to check to make sure the heap is
         * still not empty before making the call to shift_down().
         * Actually, we check to make sure that (v_count - 1) still
         * indicates there is more than one element in the heap
         * (because a one-element heap is already sorted). */
        if (v_count > 2) {
            cgul_heap_primitives__shift_down(
                cex,
                v,
                v_count - 1,  /* size */
                0,            /* index */
                1,            /* use_floyds_method */
                cf);
        }
    }

 out:
    return;
}
