#include "cgul_radix_sort.h"
#include "cgul_libc.h"

/*
 * Vocabulary: value vs. enumeration vs. key
 * ==========  =====     ===========     ===
 *
 * There are three similar, but subtly different concepts when dealing
 * with radix sort.  The typical academic description of "radix sort"
 * sorts a "value" by breaking it into "digits" and sorting on each
 * digit in turn.  This is basically what the code belows does, but it
 * adds two levels of abstraction to the idea of a "value."
 *
 * The first level of abstraction is the "enumeration."  The way it
 * works is that the user can pass in an "enumeration function" which
 * maps "values" to "enumerations."  We then sort by "enumeration" by
 * breaking it into "digits" and sorting on each digit in turn.  For
 * example, normally the code below sorts in ascending order, but by
 * providing an "enumeration function" the client can sort in
 * descending order simply by having the "enumerations" invert the
 * order of the "values."  When the client is allowed to not specify
 * an "enumeration function," the arrays being sorted are necessarily
 * of an integer type, and in this case, the "values" double as the
 * "enumerations."  This lets the user sort arrays of integers without
 * incurring overhead of the "enumeration function" callback.
 *
 * The second level of abstraction is the "key."  The code doesn't
 * really sort on "values" or on "enumerations"; rather, the code
 * really sorts by "key."  This extra level of abstraction is added so
 * that SIGNED values can be used as "enumerations" instead of just
 * UNSIGNED values.  This is explained more fully in the "How to radix
 * sort SIGNED values" section below, but suffice it to say that
 * signed "enumerations" have to be tweaked in a way that would be
 * inconvenient and difficult for the client to always get right.  So
 * instead, we do it for the client and utimately sort by "key"
 * instead of by "enumeration" (or by "value").  Also, the "key"
 * abstraction makes it easy to sort in descending order of the
 * "enumerations".  Because the order of the "keys" controls the order
 * of the sort, we just have to reverse the order of the "keys"
 * without having to reverse the order of the "enumerations."
 *
 * Do NOT confuse C-style enumerations with "enumerations" as defined
 * by this file.  They are not related.
 */

/*
 * NOTE: How to radix sort SIGNED values.
 * ====
 *
 * Sorting unsigned values is straight forward because each digit
 * sorts the same as every other digit, but when dealing with two's
 * complement, signed values, the most-significant bit should be
 * inverted so that negative values (with msb = 1) sort before
 * positive values with (msb = 0).  The easiest way to do this is to
 * just use addition.  For example, if we use a 4-bit radix,
 * RADIX_DIGITS is 0x10000.  If we shift RADIX_DIGITS right by one
 * bit, we get 0x1000.  If we add this value to any key, it will just
 * toggle the most-significant bit (because we will use RADIX_MASK to
 * remove any overflow).
 *
 * You may wonder about the other bits, but all the other bits are ok.
 * For example, for 4-bit radix, -1 = 0b1111 and -2 - 0b1110.  If you
 * look at these values as unsigned values (like the way the functions
 * below do), than -1 will sort before -2.  So, when the
 * most-significant bit is toggled, it causes all the negative values
 * to sort before all the positive values.  Then, within the negative
 * values, the other bits are just right for sorting them from
 * smallest to largest.
 */

/**
 * The number of bits in the radix.  This should be 16, 8, 4, 2, or 1.
 * If sorting 8-bit enumerations, the code below will automatically
 * reduce the radix bits to 8 if <code>RADIX_BITS</code> is set to 16.
 * Call <code>cgul_radix_sort__get_radix_info()</code> to get the
 * correct information related to the radix being used.  If you need
 * to change the default radix, RADIX_BITS should be the only value
 * that needs to be changed.
 */
#define RADIX_BITS 16

/**
 * Calculate the key associated with the element
 * <code>gk_arr[gk_index]</code> of the array being sorted.  If
 * <code>gk_ef</code> is not <code>NULL</code>, the client's
 * enumeration function <code>gk_ef</code> is called to map the
 * element to its enumeration.  If <code>gk_ef</code> is
 * <code>NULL</code>, the enumeration is taken to just be the element
 * itself.  This macro then shifts the enumeration by
 * <code>gk_bits_to_shift</code> bits and masks the result with
 * <code>gk_radix_mask</code> to isolate the radix digit that is being
 * used as the key.  If both <code>gk_is_last_pass</code> and
 * <code>gk_are_enumerations_signed</code> are true, this macro then
 * adjusts the key as described in the "How to radix sort SIGNED
 * values" comment at the top of this file.  (This macro uses the
 * prefix "gk" (short for "get key") to help avoid namespace
 * collisions because C macros are not hygienic.)
 *
 * @param[in] gk_ef enumeration function
 * @param[in] gk_arr array being sorted
 * @param[in] gk_index index of element in array being sorted
 * @param[in] gk_bits_to_shift number of bits to shift the enumeration
 * @param[in] gk_radix_digits number of unique digits in the radix
 * @param[in] gk_radix_mask radix mask
 * @param[in] gk_is_last_pass whether this is the last pass
 * @param[in] ascending_order whether to sort in ascending order
 * @param[in] gk_are_enumerations_signed whether enumerations are signed
 * @param[out] gk_key key associated with the element
 */
#define cgul_radix_sort__get_key(                                       \
    gk_ef,                                                              \
    gk_arr,                                                             \
    gk_index,                                                           \
    gk_bits_to_shift,                                                   \
    gk_radix_digits,                                                    \
    gk_radix_mask,                                                      \
    gk_is_last_pass,                                                    \
    gk_ascending_order,                                                 \
    gk_are_enumerations_signed,                                         \
    gk_key)                                                             \
do {                                                                    \
    /* Use the "gk_lc" prefix for our local variables. */               \
    size_t gk_lc_enumeration = 0;                                       \
                                                                        \
    if (gk_ef) {                                                        \
        /* Call the client's enumeration function "gk_ef" passing it */ \
        /* the pointer at "gk_index" in the "gk_arr" array.  The     */ \
        /* problem is that this macro is included in many functions  */ \
        /* where "gk_arr" holds integer values that are smaller than */ \
        /* pointers causing the compiler to complain.  To get around */ \
        /* this, the array is always cast to hold "void*" values     */ \
        /* even when it doesn't.  This is not generally safe, but in */ \
        /* the specific context of this macro, it is ok because this */ \
        /* part of the macro is only executed if "gk_arr" does in    */ \
        /* fact hold pointers.                                       */ \
        void** gk_lc_arr = (void**)gk_arr;                              \
        gk_lc_enumeration = (size_t)((gk_ef)((gk_lc_arr)[gk_index]));   \
    } else {                                                            \
        /* Use the value in the array as the enumeration. */            \
        gk_lc_enumeration = (size_t)((gk_arr)[gk_index]);               \
    }                                                                   \
                                                                        \
    /* Normally, the key is just the radix digit of the enumeration, */ \
    /* but see the comment at the top of the file for how this       */ \
    /* changes for signed enumerations on the last pass.             */ \
    (gk_key) = (gk_lc_enumeration >> (gk_bits_to_shift)) & (gk_radix_mask); \
    if ((gk_is_last_pass) && (gk_are_enumerations_signed)) {            \
        (gk_key) = ((gk_key) + ((gk_radix_digits) >> 1)) & (gk_radix_mask); \
    }                                                                   \
                                                                        \
    /* For descending order, we just have to reverse the order of    */ \
    /* the keys.  This can be done by just flipping each bit.        */ \
    if (!(gk_ascending_order)) {                                        \
        (gk_key) = ~(gk_key) & (gk_radix_mask);                         \
    }                                                                   \
                                                                        \
} while (0)

/**
 * Given the size <code>enumeration_size</code> (in bytes) of the
 * integer enumerations being sorted, determine the radix to use and
 * return the number of bits in the radix </code>radix_bits</code>,
 * the number of unique digits in the radix <code>radix_digits</code>
 * (i.e., the radix itself), and the mask <code>mask</code> needed to
 * isolate each digit after shifting the digit to the "ones"
 * place. <p>
 *
 * This function exists so that only the <code>RADIX_BITS</code>
 * manifest constant needs to be changed in order to try a different
 * radix; however, if the number of bits implied by
 * <code>enumeration_size</code> is less than <code>RADIX_BITS</code>,
 * this function reduces the radix to be suitable for the smaller
 * enumerations being sorted. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] enumeration_size size of the integer enumerations being sorted
 * @param[out] radix_bits number of bits in the radix
 * @param[out] radix_digits number of unique digits in the radix
 * @param[out] radix_mask mask used to isolate each digit
 */
static void
cgul_radix_sort__get_radix_info(cgul_exception_t* cex,
                                size_t enumeration_size,
                                size_t* radix_bits,
                                size_t* radix_digits,
                                size_t* radix_mask)
{
    size_t enumeration_bits = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(radix_bits);
    assert(radix_digits);
    assert(radix_mask);

    /* Initialize the radix information.  Fall back to an 8-bit radix
     * if RADIX_BITS is 16, but we are sorting 8-bit enumerations. */
    enumeration_bits = enumeration_size * CHAR_BIT;
    if (enumeration_bits < RADIX_BITS) {
        *radix_bits = enumeration_bits;
    } else {
        *radix_bits = RADIX_BITS;
    }

    /* Look up the radix information based on the number of bits in
     * the radix and return it to the caller. */
    switch (*radix_bits) {

    case 16:
        *radix_digits = 0x10000;
        *radix_mask   = 0x0ffff;
        break;

    case 8:
        *radix_digits = 0x00100;
        *radix_mask   = 0x000ff;
        break;

    case 4:
        *radix_digits = 0x00010;
        *radix_mask   = 0x0000f;
        break;

    case 2:
        *radix_digits = 0x00004;
        *radix_mask   = 0x00003;
        break;

    case 1:
        *radix_digits = 0x00002;
        *radix_mask   = 0x00001;
        break;

    default:
        *cex = cgul_exception__new("cgul_radix_sort__get_radix_info: "
                                   "invalid number of bits in radix: %lu",
                                   (unsigned long)radix_bits);
        goto out;
    }

 out:
    return;
}

/**
 * This function is one pass of a full radix sort.  It is heavily
 * parameterized so that it can handle many different cases.  It is
 * called by <code>cgul_radix_sort__common()</code>.  It performs a
 * "counting sort" on the first array <code>cnt_arr1</code> with the
 * sorted values ending up in the second array <code>cnt_arr2</code>.
 * The sort key for the counting sort is the radix digit of the
 * enumeration in the <code>cnt_place</code> place.  For example, if
 * <code>cnt_place</code> is <code>0</code>, the sort key is the
 * least-significant radix digit of the enumeration.  If
 * <code>cnt_place</code> is <code>1</code>, the sort key is the
 * second least-significant radix digit.  The number of elements in
 * the <code>cnt_arr1</code> should be <code>cnt_array_size</code> which
 * should also be the number of elements that the
 * <code>cnt_arr2</code> array can hold.  If an error occurs, an
 * exception is thrown. <p>
 *
 * If the "enumeration function" <code>cnt_ef</code> is not
 * <code>NULL</code>, <code>cnt_arr1</code> and <code>cnt_arr2</code>
 * should hold pointers to values.  This macro will at some point call
 * <code>cnt_ef</code> on each pointer in <code>cnt_arr1</code> to get
 * the enumeration associated with that pointer thereby allowing this
 * macro to sort the pointers in <code>cnt_arr1</code> in ascending
 * order based on the pointers' enumerations if
 * <code>cnt_ascending_order</code> is true or in descending order
 * otherwise.  The client's <code>cnt_ef</code> function is allowed to
 * returned signed or unsigned enumerations depending on whether
 * <code>cnt_are_enumerations_signed</code> is true or false
 * respectively.  The size of the enumeration in this case is
 * controlled by the signature for
 * <code>cgul_radix_sort__enumerate_t</code> which returns
 * <code>unsigned long</code> enumerations.  Thus, in this case,
 * <code>cnt_enumeration_size</code> should be <code>sizeof(unsigned
 * long)</code>. <p>
 *
 * If the "enumeration function" <code>cnt_ef</code> is
 * <code>NULL</code>, <code>cnt_arr1</code> and <code>cnt_arr2</code>
 * should hold integer values.  In this case, the values in
 * <code>cnt_arr1</code> double as the enumerations used for sorting,
 * and <code>cnt_enumeration_size</code> should equal
 * <code>sizeof(*arr1)</code>.  If
 * <code>cnt_are_enumerations_signed</code> is true, the enumerations
 * associated with each element are assumed to be signed; otherwise,
 * the enumerations are assumed to be unsigned.  By not specifying an
 * "enumeration function", the client gets optimized sorting for
 * arrays of integers that does not make any callbacks which is about
 * 10x faster than the version that does makes callbacks. <p>
 *
 * This macro also needs to know how to divide the enumerations into
 * radix digits so that it knows what to count.  The number of bits in
 * the radix is given by <code>cnt_radix_bits</code>.  The number of
 * unique digits in the radix (i.e., the radix itself) is given by
 * <code>cnt_radix_digits</code>, and the radix mask which is used to
 * isolate each digit of the radix is given by
 * <code>cnt_radix_mask</code>. <p>
 *
 * This chunk of code is defined as a macro in order to access C's
 * (limited) meta-programming facilities in order to be able to handle
 * efficiently sorting different types of arrays while minimizing code
 * duplication.  The meta-programming revolves around the
 * <code>cnt_arr1</code> and <code>cnt_arr2</code> arrays which can
 * have elements of any integer or pointer type. <p>
 *
 * This macro uses the "cnt" prefix (short for "count") to help avoid
 * namespace collisions because C macros are not hygienic. <p>
 *
 * @param[in,out] cnt_cex c-style exception
 * @param[in] cnt_arr1 first array which holds the elements to be sorted
 * @param[in] cnt_arr2 second array which holds the sorted elements on return
 * @param[in] cnt_array_size number of elements in each array
 * @param[in] cnt_enumeration_size size of the enumerations (in bytes)
 * @param[in] cnt_radix_bits number of bits in the radix
 * @param[in] cnt_radix_digits number of unique digits in the radix
 * @param[in] cnt_radix_mask mask used to isolate each digit
 * @param[in] cnt_ascending_order whether to sort in ascending order
 * @param[in] cnt_place place for the digits that will be used as keys
 * @param[in] cnt_ef enumeration function or <code>NULL</code>
 * @param[in] cnt_are_enumerations_signed whether the enumerations are signed
 */
#define cgul_radix_sort__counting_sort(                                 \
    cnt_cex,                                                            \
    cnt_arr1,                                                           \
    cnt_arr2,                                                           \
    cnt_array_size,                                                     \
    cnt_enumeration_size,                                               \
    cnt_radix_bits,                                                     \
    cnt_radix_digits,                                                   \
    cnt_radix_mask,                                                     \
    cnt_ascending_order,                                                \
    cnt_place,                                                          \
    cnt_ef,                                                             \
    cnt_are_enumerations_signed)                                        \
do {                                                                    \
    /* Use the "cnt_lc" prefix for our local variables. */              \
    size_t cnt_lc_i = 0;                                                \
    size_t cnt_lc_index = 0;                                            \
    size_t cnt_lc_key = 0;                                              \
    size_t cnt_lc_next = 0;                                             \
    size_t cnt_lc_curr_bits = 0;                                        \
    size_t cnt_lc_all_bits = 0;                                         \
                                                                        \
    /* Determine whether this is the last pass.  If all bits are     */ \
    /* accounted for, it is the last pass.  This needs to be known   */ \
    /* so that signed enumerations can be sorted differently from    */ \
    /* unsigned enumerations.  It is only the last pass that is      */ \
    /* different.  (See Knuth, Vol. 3, Section 5.2.5, Exercise 8,    */ \
    /* p. 177. with solution on p. 651.)                             */ \
    int cnt_lc_is_last_pass = 0;                                        \
                                                                        \
    /* The number of bits to shift based on the decimal place being  */ \
    /* used as the key.  This is initialized below.                  */ \
    size_t cnt_lc_bits_to_shift = 0;                                    \
                                                                        \
    /* Map each key (i.e., the digit) to the number of times it      */ \
    /* occurs.                                                       */ \
    size_t* cnt_lc_counts = NULL;                                       \
                                                                        \
    /* Map each key to its proper index when copying into the second */ \
    /* array.                                                        */ \
    size_t* cnt_lc_offsets = NULL;                                      \
                                                                        \
    assert(cnt_cex);                                                    \
    if (*(cnt_cex)) {                                                   \
        goto cnt_lc_out;                                                \
    }                                                                   \
                                                                        \
    assert(cnt_arr1);                                                   \
    assert(cnt_arr2);                                                   \
                                                                        \
    /* Sanity check. */                                                 \
    if ((cnt_array_size) <= 0) {                                        \
        goto cnt_lc_out;                                                \
    }                                                                   \
                                                                        \
    /* Sanity check.  If the user specifies a custom enumeration     */ \
    /* function "cnt_ef", it necessarily means the user is sorting   */ \
    /* an array of pointers to values (because the pointers get      */ \
    /* passed into the "cnt_ef" function to be mapped to an          */ \
    /* "enumeration").                                               */ \
    if ((cnt_ef) && (sizeof(*(cnt_arr1)) != sizeof(void*))) {           \
        goto cnt_lc_out;                                                \
    }                                                                   \
                                                                        \
    /* The number of bits needed to shift the radix digit for this  */  \
    /* pass so that it appears in the "ones" place.  This makes the */  \
    /* math easier.                                                 */  \
    cnt_lc_bits_to_shift = ((cnt_place) * cnt_radix_bits);              \
                                                                        \
    /* If all bits are accounted for, this is the last pass. */         \
    cnt_lc_curr_bits = cnt_lc_bits_to_shift + cnt_radix_bits;           \
    cnt_lc_all_bits = (cnt_enumeration_size) * CHAR_BIT;                \
    cnt_lc_is_last_pass = (cnt_lc_curr_bits == cnt_lc_all_bits);        \
                                                                        \
    /* Allocate. */                                                     \
    cnt_lc_counts = (size_t*)                                           \
        cgul_libc__calloc((cnt_cex),                                    \
                          cnt_radix_digits,                             \
                          sizeof(*cnt_lc_counts));                      \
    cnt_lc_offsets = (size_t*)                                          \
        cgul_libc__calloc((cnt_cex),                                    \
                          cnt_radix_digits,                             \
                          sizeof(*cnt_lc_offsets));                     \
    if (*(cnt_cex)) {                                                   \
        goto cnt_lc_out;                                                \
    }                                                                   \
                                                                        \
    /* Create a histogram for the frequency of each key.  This is    */ \
    /* the familiar "counting sort" (which is at the heart of "radix */ \
    /* sort") modified to handle both signed and unsigned            */ \
    /* enumerations.  (See the "Vocabulary: value vs. enumeration    */ \
    /* vs. key" comment at the top of this file.)                    */ \
    for (cnt_lc_i = 0 ; cnt_lc_i < (cnt_array_size) ; ++cnt_lc_i) {     \
                                                                        \
        /* Map cnt_arr1[cnt_lc_i] to its enumeration and extract the */ \
        /* key from the enumeration.  On return, "key" holds the     */ \
        /* key.                                                      */ \
        cgul_radix_sort__get_key((cnt_ef),                              \
                                 (cnt_arr1),                            \
                                 cnt_lc_i,                              \
                                 cnt_lc_bits_to_shift,                  \
                                 cnt_radix_digits,                      \
                                 cnt_radix_mask,                        \
                                 cnt_lc_is_last_pass,                   \
                                 (cnt_ascending_order),                 \
                                 (cnt_are_enumerations_signed),         \
                                 cnt_lc_key);                           \
                                                                        \
        /* Update the key's counts.  For signed enumerations and for */ \
        /* user-specified enumerations, it is critical to understand */ \
        /* that we are playing around with the ENUMERATION/KEY, not  */ \
        /* the VALUE in the array.  When we are copying values to    */ \
        /* cnt_arr2 (below), we calculate the KEY the exact same way */ \
        /* as we do here, and because the KEYS control the order     */ \
        /* which VALUES are copied and because the KEYS are in sort  */ \
        /* order, the VALUES will be copied in sort order (below).   */ \
        /* (See the "Vocabulary: value vs. enumeration vs. key"      */ \
        /* comment at the top of this file.)                         */ \
        ++cnt_lc_counts[cnt_lc_key];                                    \
    }                                                                   \
                                                                        \
    /* Calculate the offsets of where in cnt_arr2 to copy the first  */ \
    /* instance of each VALUE having the KEY associated with the     */ \
    /* radix digit being considered.  All values having KEY = 0 will */ \
    /* be copied over first.  Then, all values having KEY = 1 will   */ \
    /* be copied over next.  Etc.                                    */ \
    for (cnt_lc_i = 0 ; cnt_lc_i < cnt_radix_digits ; ++cnt_lc_i) {     \
        cnt_lc_offsets[cnt_lc_i] = cnt_lc_next;                         \
        cnt_lc_next += cnt_lc_counts[cnt_lc_i];                         \
    }                                                                   \
                                                                        \
    /* Copy each digit in the "cnt_arr1" array into place in the     */ \
    /* "cnt_arr2" array based on the indexes in the "cnt_lc_offsets" */ \
    /* array.                                                        */ \
    for (cnt_lc_i = 0 ; cnt_lc_i < (cnt_array_size) ; ++cnt_lc_i) {     \
                                                                        \
        /* Map cnt_arr1[cnt_lc_i] to its enumeration and extract the */ \
        /* key from the enumeration.  On return, "key" holds the     */ \
        /* key.                                                      */ \
        cgul_radix_sort__get_key((cnt_ef),                              \
                                 (cnt_arr1),                            \
                                 cnt_lc_i,                              \
                                 cnt_lc_bits_to_shift,                  \
                                 cnt_radix_digits,                      \
                                 cnt_radix_mask,                        \
                                 cnt_lc_is_last_pass,                   \
                                 (cnt_ascending_order),                 \
                                 (cnt_are_enumerations_signed),         \
                                 cnt_lc_key);                           \
                                                                        \
        /* Lookup the correct index for this value based on its key  */ \
        /* and post-increment the index so that the next value with  */ \
        /* the same key will be inserted immediately after this one. */ \
        cnt_lc_index = cnt_lc_offsets[cnt_lc_key]++;                    \
                                                                        \
        /* Copy the value from cnt_arr1 to cnt_arr2 placing it in    */ \
        /* its sort order.                                           */ \
        (cnt_arr2)[cnt_lc_index] = (cnt_arr1)[cnt_lc_i];                \
    }                                                                   \
                                                                        \
 cnt_lc_out:                                                            \
                                                                        \
    /* Clean up. */                                                     \
    if (cnt_lc_offsets) {                                               \
        free(cnt_lc_offsets);                                           \
    }                                                                   \
    if (cnt_lc_counts) {                                                \
        free(cnt_lc_counts);                                            \
    }                                                                   \
                                                                        \
} while (0)

/**
 * This macro is the entry point to the common code used by all the
 * radix sort functions.  It is parameterized to handle many different
 * cases.  The purpose of this macro is to call
 * <code>cgul_radix_sort__counting_sort()</code> once for each pass of
 * the radix sort algorithm.  Each pass of the radix sort algorithm
 * works on the next radix digit starting with the least-significant
 * radix digit.  (There are other radix sort algorithms that work
 * starting with the most-significant digit, but they have many
 * pitfalls.)  <p>
 *
 * Perform an in-place radix sort on the array <code>com_v</code>
 * which holds <code>com_v_count</code> instances of either integers
 * or pointers to values.  The integers or pointers to values in the
 * array should be of type <code>com_v_element_type</code>.  If an
 * error occurs, an exception is thrown. <p>
 *
 * If the "enumeration function" <code>com_ef</code> is not
 * <code>NULL</code>, this macro will at some point call
 * <code>com_ef</code> on each pointer in <code>com_v</code> to get
 * the enumeration associated with that pointer thereby allowing this
 * macro to sort the pointers in <code>com_v</code> in ascending order
 * based on the pointers' enumerations if
 * <code>com_ascending_order</code> is true or in descending order
 * otherwise.  The client's <code>com_ef</code> function is allowed to
 * returned signed or unsigned enumerations depending on whether
 * <code>com_are_enumerations_signed</code> is true or false
 * respectively. <p>
 *
 * If the "enumeration function" <code>com_ef</code> is
 * <code>NULL</code>, <code>com_v</code> should hold integer values of
 * type <code>com_v_element_type</code>.  In this case, the values in
 * <code>com_v</code> double as the enumerations used for sorting.  If
 * <code>com_are_enumerations_signed</code> is true, the enumerations
 * associated with each element are assumed to be signed; otherwise,
 * the enumerations are assumed to be unsigned.  This is the mechanism
 * that allows the client to directly sort an array of integers. <p>
 *
 * This chunk of code is defined as a macro in order to access C's
 * (limited) meta-programming facilities in order to be able to handle
 * efficiently sorting different types of arrays while minimizing code
 * duplication.  The meta-programming revolves around the
 * <code>com_v_element_type</code> parameter which specifies the type
 * of the elements being sorted. <p>
 *
 * This macro uses the "com" prefix (short for "common") to help avoid
 * namespace collisions because C macros are not hygienic. <p>
 *
 * @param[in, out] com_cex c-style exception
 * @param[in] com_v array of elements to sort
 * @param[in] com_v_count number of elements in <code>com_v</code>
 * @param[in] com_v_element_type type of the elements in <code>com_v</code>
 * @param[in] com_ascending_order whether to sort in ascending order
 * @param[in] com_ef enumeration function
 * @param[in] com_are_enumerations_signed whether the enumerations are signed
 */
#define cgul_radix_sort__common(                                        \
    com_cex,                                                            \
    com_v,                                                              \
    com_v_count,                                                        \
    com_v_element_type,                                                 \
    com_ascending_order,                                                \
    com_ef,                                                             \
    com_are_enumerations_signed)                                        \
do {                                                                    \
    /* Use the "com_lc" prefix for our local variables. */              \
    size_t com_lc_i = 0;                                                \
    size_t com_lc_i_max = 0;                                            \
    size_t com_lc_enumeration_size = 0;                                 \
    com_v_element_type* com_lc_buf = NULL;                              \
    com_v_element_type* com_lc_arr1 = NULL;                             \
    com_v_element_type* com_lc_arr2 = NULL;                             \
    com_v_element_type* com_lc_tmp = NULL;                              \
                                                                        \
    /* The number of bits in the radix. */                              \
    size_t com_lc_radix_bits = 0;                                       \
                                                                        \
    /* The number of unique digits in the radix (i.e., the radix     */ \
    /* itself).                                                      */ \
    size_t com_lc_radix_digits = 0;                                     \
                                                                        \
    /* The mask used to isolate each radix digit after shifiting it  */ \
    /* to the "ones" place.  Note that the "cnt_place" is zero-based */ \
    /* so the "ones" place is the same as "cnt_place == 0".          */ \
    size_t com_lc_radix_mask = 0;                                       \
                                                                        \
    /* The C compiler gets upset if you pass in NULL to a macro and  */ \
    /* the macro tries to invoke the function.  This doesn't work    */ \
    /* because the macro uses simple text substitution causing the   */ \
    /* C compiler to compile that "0u" is not a function.  To work   */ \
    /* around this, "ef_var" is declared so that "ef_var" is always  */ \
    /* passed into the macro instead of sometimes passing NULL.      */ \
    cgul_radix_sort__enumerate_t ef_var =                               \
        (cgul_radix_sort__enumerate_t)(com_ef);                         \
                                                                        \
    assert(com_cex);                                                    \
    if (*(com_cex)) {                                                   \
        goto com_lc_out;                                                \
    }                                                                   \
                                                                        \
    /* Sanity check. */                                                 \
    if ((com_v_count) <= 0) {                                           \
        goto com_lc_out;                                                \
    }                                                                   \
                                                                        \
    /* Sanity check. */                                                 \
    if (sizeof(*(com_v)) != sizeof(com_v_element_type)) {               \
        *(com_cex) = cgul_exception__new("cgul_radix_sort__common: "    \
                                       "size mismatch!");               \
        goto com_lc_out;                                                \
    }                                                                   \
                                                                        \
    /* Allocate the temporary buffer. */                                \
    com_lc_buf = (com_v_element_type*)                                  \
        cgul_libc__calloc((com_cex),                                    \
                          (com_v_count),                                \
                          sizeof(com_v_element_type));                  \
    if (*(com_cex)) {                                                   \
        goto com_lc_out;                                                \
    }                                                                   \
                                                                        \
    /* Determine the enumeration size.  If the client specified an   */ \
    /* enumeration function "com_ef", the enumerations are "unsigned */ \
    /* long" types (as returned by "com_ef"); otherwise, the         */ \
    /* elements being sorted double as the enumerations.             */ \
    if (ef_var) {                                                       \
        com_lc_enumeration_size = sizeof(unsigned long int*);           \
    } else {                                                            \
        com_lc_enumeration_size = sizeof(com_v_element_type);           \
    }                                                                   \
                                                                        \
    /* Initialize the radix information given the size of the        */ \
    /* enumerations to sort.  All of the code is parameterized to    */ \
    /* work with different radix values that are a reasonable power  */ \
    /* of 2.  The radix information basically determines how many    */ \
    /* will be processed on each pass of the radix sort algorithm.   */ \
    cgul_radix_sort__get_radix_info((com_cex),                          \
                                    (com_lc_enumeration_size),          \
                                    &com_lc_radix_bits,                 \
                                    &com_lc_radix_digits,               \
                                    &com_lc_radix_mask);                \
    if (*(com_cex)) {                                                   \
        goto com_lc_out;                                                \
    }                                                                   \
                                                                        \
    /* Do the radix sort by doing stable "counting sorts" that key   */ \
    /* off each digit in turn starting with the least significant    */ \
    /* digit.  Think of the enumeration as a composite key made up   */ \
    /* of subkeys having RADIX_BITS bits where each RADIX_BITS bits  */ \
    /* is one radix digit.  Then, you can get a final order where    */ \
    /* the values are sorted first by the most-significant byte,     */ \
    /* then by the second most-significant byte, etc. by using a     */ \
    /* stable sorting first by least-significant byte, then by the   */ \
    /* second least-significant byte etc.  This is similar to how    */ \
    /* a table in a GUI can be sorted by multiple columns if the     */ \
    /* table implements a stable sort.  You just click on each       */ \
    /* column in reverse order to how you want the final table       */ \
    /* sorted.                                                       */ \
    com_lc_arr1 = (com_v);                                              \
    com_lc_arr2 = com_lc_buf;                                           \
    com_lc_i_max = com_lc_enumeration_size * CHAR_BIT / com_lc_radix_bits; \
    for (com_lc_i = 0 ; com_lc_i < com_lc_i_max ; ++com_lc_i) {         \
                                                                        \
        /* Sort the first array ("com_lc_arr1") by the current       */ \
        /* radix digit storing the result in the second array        */ \
        /* ("com_lc_arr2").                                          */ \
        cgul_radix_sort__counting_sort(                                 \
            (com_cex),                                                  \
            com_lc_arr1,                                                \
            com_lc_arr2,                                                \
            (com_v_count),                                              \
            com_lc_enumeration_size,                                    \
            com_lc_radix_bits,                                          \
            com_lc_radix_digits,                                        \
            com_lc_radix_mask,                                          \
            (com_ascending_order),                                      \
            com_lc_i /* place */,                                       \
            ef_var /* enumeration function */,                          \
            (com_are_enumerations_signed));                             \
        if (*(com_cex)) {                                               \
            goto com_lc_out;                                            \
        }                                                               \
                                                                        \
        /* Swap the arrays. */                                          \
        com_lc_tmp = com_lc_arr1;                                       \
        com_lc_arr1 = com_lc_arr2;                                      \
        com_lc_arr2 = com_lc_tmp;                                       \
    }                                                                   \
                                                                        \
    /* Copy "com_lc_buf" to "com_v" if the for() loop above executed */ \
    /* an odd number of times.  This can happen, for example, if     */ \
    /* cgul_radix_sort_short() calls this macro and                  */ \
    /* com_lc_radix_bits is 16.                                      */ \
    if((com_lc_i & 1) == 1) {                                           \
        memcpy((com_v), com_lc_buf, sizeof(*(com_v)) * (com_v_count));  \
    }                                                                   \
                                                                        \
 com_lc_out:                                                            \
                                                                        \
    /* Clean up. */                                                     \
    if (com_lc_buf) {                                                   \
        free(com_lc_buf);                                               \
    }                                                                   \
                                                                        \
} while (0)

void
cgul_radix_sort(cgul_exception_t* cex,
                void* v,
                size_t v_count,
                size_t element_size,
                int ascending_order,
                cgul_radix_sort__enumerate_t ef,
                int are_enumerations_signed)
{
    void* scratch = NULL;
    void** pointers_array = NULL;
    char* tmp = NULL;
    size_t i = 0;
    size_t v_size = v_count * element_size;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(ef);

    /* Sanity checks. */
    if ((v_count <= 1) || (element_size <= 0)) {
        goto out;
    }

    /*
     * The general idea is to sort pointers that point into a copy of
     * "v" under the assumption that copying pointers during a sort
     * will be more efficient than copying entire structures during
     * the sort.  After sorting the pointers, we will copy individual
     * elements from the copy back into the original "v" based upon
     * the sorted pointers.
     */

    /* Create an exact copy of "v" and put it in "scratch". */
    scratch = cgul_libc__malloc(cex, v_size);
    if (*cex) {
        goto out;
    }
    memcpy(scratch, v, v_size);

    /* Generate an array of pointers into the scratch area. */
    pointers_array = (void**)cgul_libc__malloc(cex, sizeof(void*) * v_count);
    if (*cex) {
        goto out;
    }
    for (i = 0, tmp = (char*)scratch;
         i < v_count;
         ++i, tmp += element_size)
    {
        pointers_array[i] = tmp;
    }

    /* Sort the pointers. */
    cgul_radix_sort_pointers(cex,
                             pointers_array,
                             v_count,
                             ascending_order,
                             ef,
                             are_enumerations_signed);
    if (*cex) {
        goto out;
    }

    /* Copy the elements back into "v" based upon the sorted pointers. */
    for (i = 0, tmp = (char*)v ;
         i < v_count ;
         ++i, tmp += element_size)
    {
        memcpy(tmp, pointers_array[i], element_size);
    }

 out:

    /* Clean up. */
    if (pointers_array) {
        free(pointers_array);
    }
    if (scratch) {
        free(scratch);
    }
}

void
cgul_radix_sort_pointers(cgul_exception_t* cex,
                         void** v,
                         size_t v_count,
                         int ascending_order,
                         cgul_radix_sort__enumerate_t ef,
                         int are_enumerations_signed)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(ef);

    /* Sanity check. */
    if (v_count <= 1) {
        goto out;
    }

    /* Perform the radix sort using the user-specified
     * ordering/enumerations. */
    cgul_radix_sort__common(cex,
                            v,
                            v_count,
                            void* /* com_v_element_type */,
                            ascending_order,
                            ef /* enumeration function */,
                            are_enumerations_signed);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_radix_sort_schar(cgul_exception_t* cex,
                      signed char* v,
                      size_t v_count,
                      int ascending_order)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Sanity check. */
    if (v_count <= 1) {
        goto out;
    }

    /* Sort. */
    cgul_radix_sort__common(cex,
                            v,
                            v_count,
                            signed char,      /* com_v_element_type */
                            ascending_order,
                            NULL,             /* enumeration function */
                            1);               /* are_enumerations_signed */
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_radix_sort_uchar(cgul_exception_t* cex,
                      unsigned char* v,
                      size_t v_count,
                      int ascending_order)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Sanity check. */
    if (v_count <= 1) {
        goto out;
    }

    /* Sort. */
    cgul_radix_sort__common(cex,
                            v,
                            v_count,
                            unsigned char,    /* com_v_element_type */
                            ascending_order,
                            NULL,             /* enumeration function */
                            0);               /* are_enumerations_signed */
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_radix_sort_short(cgul_exception_t* cex,
                      short int* v,
                      size_t v_count,
                      int ascending_order)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Sanity check. */
    if (v_count <= 1) {
        goto out;
    }

    /* Sort. */
    cgul_radix_sort__common(cex,
                            v,
                            v_count,
                            short int,        /* com_v_element_type */
                            ascending_order,
                            NULL,             /* enumeration function */
                            1);               /* are_enumerations_signed */
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_radix_sort_ushort(cgul_exception_t* cex,
                       unsigned short int* v,
                       size_t v_count,
                       int ascending_order)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Sanity check. */
    if (v_count <= 1) {
        goto out;
    }

    /* Sort. */
    cgul_radix_sort__common(cex,
                            v,
                            v_count,
                            unsigned short int,  /* com_v_element_type */
                            ascending_order,
                            NULL,                /* enumeration function */
                            0);                  /* are_enumerations_signed */
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_radix_sort_int(cgul_exception_t* cex,
                    int* v,
                    size_t v_count,
                    int ascending_order)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Sanity check. */
    if (v_count <= 1) {
        goto out;
    }

    /* Sort. */
    cgul_radix_sort__common(cex,
                            v,
                            v_count,
                            int,              /* com_v_element_type */
                            ascending_order,
                            NULL,             /* enumeration function */
                            1);               /* are_enumerations_signed */
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_radix_sort_uint(cgul_exception_t* cex,
                      unsigned int* v,
                      size_t v_count,
                      int ascending_order)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Sanity check. */
    if (v_count <= 1) {
        goto out;
    }

    /* Sort. */
    cgul_radix_sort__common(cex,
                            v,
                            v_count,
                            unsigned int,     /* com_v_element_type */
                            ascending_order,
                            NULL,             /* enumeration function */
                            0);               /* are_enumerations_signed */
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_radix_sort_long(cgul_exception_t* cex,
                     long int* v,
                     size_t v_count,
                     int ascending_order)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Sanity check. */
    if (v_count <= 1) {
        goto out;
    }

    /* Sort. */
    cgul_radix_sort__common(cex,
                            v,
                            v_count,
                            long int,         /* com_v_element_type */
                            ascending_order,
                            NULL,             /* enumeration function */
                            1);               /* are_enumerations_signed */
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_radix_sort_ulong(cgul_exception_t* cex,
                      unsigned long int* v,
                      size_t v_count,
                      int ascending_order)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Sanity check. */
    if (v_count <= 1) {
        goto out;
    }

    /* Sort. */
    cgul_radix_sort__common(cex,
                            v,
                            v_count,
                            unsigned long int,  /* com_v_element_type */
                            ascending_order,
                            NULL,               /* enumeration function */
                            0);                 /* are_enumerations_signed */
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_radix_sort_int8(cgul_exception_t* cex,
                     cgul_int8_t* v,
                     size_t v_count,
                     int ascending_order)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Sanity check. */
    if (v_count <= 1) {
        goto out;
    }

    /* Sort. */
    cgul_radix_sort__common(cex,
                            v,
                            v_count,
                            cgul_int8_t,      /* com_v_element_type */
                            ascending_order,
                            NULL,             /* enumeration function */
                            1);               /* are_enumerations_signed */
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_radix_sort_uint8(cgul_exception_t* cex,
                      cgul_uint8_t* v,
                      size_t v_count,
                      int ascending_order)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Sanity check. */
    if (v_count <= 1) {
        goto out;
    }

    /* Sort. */
    cgul_radix_sort__common(cex,
                            v,
                            v_count,
                            cgul_uint8_t,     /* com_v_element_type */
                            ascending_order,
                            NULL,             /* enumeration function */
                            0);               /* are_enumerations_signed */
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_radix_sort_int16(cgul_exception_t* cex,
                      cgul_int16_t* v,
                      size_t v_count,
                      int ascending_order)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Sanity check. */
    if (v_count <= 1) {
        goto out;
    }

    /* Sort. */
    cgul_radix_sort__common(cex,
                            v,
                            v_count,
                            cgul_int16_t,     /* com_v_element_type */
                            ascending_order,
                            NULL,             /* enumeration function */
                            1);               /* are_enumerations_signed */
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_radix_sort_uint16(cgul_exception_t* cex,
                       cgul_uint16_t* v,
                       size_t v_count,
                       int ascending_order)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Sanity check. */
    if (v_count <= 1) {
        goto out;
    }

    /* Sort. */
    cgul_radix_sort__common(cex,
                            v,
                            v_count,
                            cgul_uint16_t,     /* com_v_element_type */
                            ascending_order,
                            NULL,              /* enumeration function */
                            0);                /* are_enumerations_signed */
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_radix_sort_int32(cgul_exception_t* cex,
                      cgul_int32_t* v,
                      size_t v_count,
                      int ascending_order)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Sanity check. */
    if (v_count <= 1) {
        goto out;
    }

    /* Sort. */
    cgul_radix_sort__common(cex,
                            v,
                            v_count,
                            cgul_int32_t,     /* com_v_element_type */
                            ascending_order,
                            NULL,             /* enumeration function */
                            1);               /* are_enumerations_signed */
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_radix_sort_uint32(cgul_exception_t* cex,
                       cgul_uint32_t* v,
                       size_t v_count,
                       int ascending_order)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Sanity check. */
    if (v_count <= 1) {
        goto out;
    }

    /* Sort. */
    cgul_radix_sort__common(cex,
                            v,
                            v_count,
                            cgul_uint32_t,     /* com_v_element_type */
                            ascending_order,
                            NULL,              /* enumeration function */
                            0);                /* are_enumerations_signed */
    if (*cex) {
        goto out;
    }

 out:
    return;
}

#ifdef CGUL_INT64
void
cgul_radix_sort_int64(cgul_exception_t* cex,
                      cgul_int64_t* v,
                      size_t v_count,
                      int ascending_order)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Sanity check. */
    if (v_count <= 1) {
        goto out;
    }

    /* Sort. */
    cgul_radix_sort__common(cex,
                            v,
                            v_count,
                            cgul_int64_t,     /* com_v_element_type */
                            ascending_order,
                            NULL,             /* enumeration function */
                            1);               /* are_enumerations_signed */
    if (*cex) {
        goto out;
    }

 out:
    return;
}
#endif /* CGUL_INT64 */

#ifdef CGUL_UINT64
void
cgul_radix_sort_uint64(cgul_exception_t* cex,
                       cgul_uint64_t* v,
                       size_t v_count,
                       int ascending_order)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Sanity check. */
    if (v_count <= 1) {
        goto out;
    }

    /* Sort. */
    cgul_radix_sort__common(cex,
                            v,
                            v_count,
                            cgul_uint64_t,     /* com_v_element_type */
                            ascending_order,
                            NULL,              /* enumeration function */
                            0);                /* are_enumerations_signed */
    if (*cex) {
        goto out;
    }

 out:
    return;
}
#endif /* CGUL_UINT64 */
