#include "cgul_strtok_p.h"
#include "cgul_libc.h"

cgul_strtok_t
cgul_strtok__new(cgul_exception_t* cex)
{
    cgul_strtok_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate the object. */
    rv = (cgul_strtok_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Remember that we are about to start the first pass. */
    rv->is_first_pass = 1;

 out:

    /* Clean up partially initialized object. */
    if (*cex && rv) {
        cgul_strtok__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_strtok__delete(cgul_strtok_t tok)
{
    if (tok) {

        /* value */
        if (tok->value) {
            free(tok->value);
        }

        /* Free the object itself. */
        free(tok);
    }
}

void
cgul_strtok__set_value(cgul_exception_t* cex,
                       cgul_strtok_t tok,
                       const char* value)
{
    char* tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(tok);

    /* Check the incoming "value". */
    if (value) {

        /* Duplicate "value". */
        tmp = cgul_libc__strdup(cex, value);
        if (*cex) {
            goto out;
        }

        /* Free the existing value. */
        if (tok->value) {
            free(tok->value);
        }

        /* Take ownership of the duplicate value. */
        tok->value = tmp;
        tmp = NULL;

        /* Reset the separator and the index values. */
        tok->separator = '\0';
        tok->start = 0;
        tok->end = 0;

        /* Calculate the new string length of "value". */
        tok->value_length = strlen(value);

        /* Remember that we are about to start the first pass. */
        tok->is_first_pass = 1;
    }

 out:

    /* Clean up. */
    if (tmp) {
        free(tmp);
    }

    return;
}

const char*
cgul_strtok__get_next(cgul_exception_t* cex,
                      cgul_strtok_t tok,
                      const char* separators,
                      int allow_empty_tokens)
{
    const char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(tok);
    assert(separators);

    /* Reset the trailing separator for this token. */
    tok->separator = '\0';

    /* If there was a previous token, move just past its end. */
    if (tok->is_first_pass) {
        tok->is_first_pass = 0;
    } else {
        tok->start = tok->end + 1;
        if (tok->start > tok->value_length) {
            goto out;
        }
    }

    /* Find the start of the token.  If we are not allowing empty
     * tokens, we have to telescope all the leading separators by
     * skipping passed them to the first non-separator.  */
    if (!allow_empty_tokens) {
        tok->start += strspn(&tok->value[tok->start], separators);
    }

    /* Find the end of the token, remember the trailing separator, and
     * nul-terminate the token. */
    tok->end = tok->start + strcspn(&tok->value[tok->start], separators);
    tok->separator = tok->value[tok->end];
    tok->value[tok->end] = '\0';

    /* Success. */
    if (tok->start < tok->value_length) {
        rv = &tok->value[tok->start];
    } else if (allow_empty_tokens && (tok->start == tok->value_length)) {
        rv = &tok->value[tok->start];
    }

 out:
    return rv;
}

char
cgul_strtok__get_separator(cgul_exception_t* cex,
                           cgul_strtok_t tok)
{
    char rv = '\0';

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(tok);

    rv = tok->separator;

 out:
    return rv;
}

size_t
cgul_strtok__get_start(cgul_exception_t* cex,
                       cgul_strtok_t tok)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(tok);

    rv = tok->start;

 out:
    return rv;
}

size_t
cgul_strtok__get_end(cgul_exception_t* cex,
                     cgul_strtok_t tok)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(tok);

    rv = tok->end;

 out:
    return rv;
}
