#include "cgul_substring_p.h"
#include "cgul_libc.h"
#include "cgul_big_integer.h"
#include "cgul_rolling_hash.h"

/**
 * Reset <code>ss</code> to its initial state.
 *
 * @param[in] cex c-style exception
 * @param[in] ss <code>cgul_substring</code> instance
 */
static void
cgul_substring__reset(cgul_exception_t* cex,
                      cgul_substring_t ss)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ss);

    /* ss */
    if (ss->ss) {
        free(ss->ss);
        ss->ss = NULL;
    }

    /* others */
    ss->ss_length = 0;
    ss->ss_hash = 0;

 out:
    return;
}

/**
 * Search the first <code>length</code> bytes in the block
 * <code>block</code> for the binary substring <code>ss</code> of
 * length <code>ss_length</code> in bytes.  If the substring is found,
 * a pointer to the first matching character in block is returned;
 * otherwise, <code>NULL</code> is returned. <p>
 *
 * This function is the same as
 * <code>cgul_substring__search_ez()</code> except it uses
 * pre-computed values making repeated searching more efficient.  The
 * pre-computed values should be obtained as follows: <p>
 *
 * <pre>
 *     ss_hash = cgul_rolling_hash__hash(cex, ss, ss_length);
 * </pre>
 *
 * The results are undefined if <code>ss_hash</code> is not consistent
 * with <code>ss</code>. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] ss binary substring
 * @param[in] ss_length substring length in bytes
 * @param[in] ss_hash substring hash
 * @param[in] block block of binary data to search
 * @param[in] block_length block length in bytes
 * @return pointer to first match or <code>NULL</code> if no match
 */
static const char*
cgul_substring__search_using_prevalues(cgul_exception_t* cex,
                                       const char* ss,
                                       size_t ss_length,
                                       cgul_uint32_t ss_hash,
                                       const char* block,
                                       size_t block_length)
{
    const char* rv = NULL;
    size_t i = 0;
    size_t i_max = 0;
    cgul_uint32_t block_hash = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Sanity check. */
    if (!ss || (ss_length <= 0) || !block || (block_length <= 0)) {
        goto out;
    }

    /* There can be no match if the block is smaller than the substring. */
    if (block_length < ss_length) {
        goto out;
    }

    /* Initialize the rolling hash to be the hash value of the first
     * ss_length bytes of the block. */
    block_hash = cgul_rolling_hash__hash(cex, block, ss_length);

    /* If the hashes match, explicitly compare substrings to make sure
     * this isn't just a hash collision. */
    if (ss_hash == block_hash) {
        /* If the substrings match exactly, return the starting address. */
        if (memcmp(ss, block, ss_length) == 0) {
            rv = block;
            goto out;
        }
    }

    /* Iterate over most of the characters in "block".  We have to
     * stop iterating whenever the leading edge of the rolling window
     * would extend beyond the end of the block. */
    i_max = block_length - ss_length;
    for (i = 0 ; i < i_max ; ++i) {

        /* Calculate the rolling hash by removing block[i] and adding
         * block[i + ss_length]. */
        block_hash =
            cgul_rolling_hash__roll(cex,
                                    block_hash,
                                    block[i],
                                    block[i + ss_length],
                                    ss_length);

        /* If the hashes match, explicitly compare substrings to make sure
         * this isn't just a hash collision. */
        if (ss_hash == block_hash) {
            /* If the substrings match exactly, return the starting address. */
            if (memcmp(ss, &block[i + 1], ss_length) == 0) {
                rv = &block[i + 1];
                goto out;
            }
        }
    }

 out:
    return rv;
}

const char*
cgul_substring__search_ez(cgul_exception_t* cex,
                          const char* ss,
                          size_t ss_length,
                          const char* block,
                          size_t block_length)
{
    const char* rv = NULL;
    cgul_uint32_t ss_hash = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Calculate the hash value for the substring. */
    ss_hash = cgul_rolling_hash__hash(cex, ss, ss_length);

    /* Perform the search using the pre-computed values. */
    rv = cgul_substring__search_using_prevalues(cex,
                                                ss,
                                                ss_length,
                                                ss_hash,block,
                                                block_length);

 out:
    return rv;
}

cgul_substring_t
cgul_substring__new(cgul_exception_t* cex)
{
    cgul_substring_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate the object. */
    rv = (cgul_substring_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* "Reset" the object to its initial state. */
    cgul_substring__reset(cex, rv);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up partially initialized objects. */
    if (*cex && rv) {
        cgul_substring__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_substring_t
cgul_substring__new_from_cstring(cgul_exception_t* cex,
                                 const char* substring)
{
    cgul_substring_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(substring);

    /* Allocate. */
    rv = cgul_substring__new(cex);
    if (*cex) {
        goto out;
    }

    /* Set the substring. */
    cgul_substring__set_value(cex, rv, substring, strlen(substring));
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up partially initialized objects. */
    if (*cex && rv) {
        cgul_substring__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_substring_t
cgul_substring__new_from_bstring(cgul_exception_t* cex,
                                 const char* substring,
                                 size_t length)
{
    cgul_substring_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(substring);

    /* Allocate. */
    rv = cgul_substring__new(cex);
    if (*cex) {
        goto out;
    }

    /* Set the substring. */
    cgul_substring__set_value(cex, rv, substring, length);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up partially initialized objects. */
    if (*cex && rv) {
        cgul_substring__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_substring__delete(cgul_substring_t ss)
{
    if (ss) {

        /* ss */
        if (ss->ss) {
            free(ss->ss);
        }

        /* Free the object itself. */
        free(ss);
    }
}

const char*
cgul_substring__get_value(cgul_exception_t* cex,
                          cgul_substring_t ss)
{
    const char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ss);

    rv = ss->ss;

 out:
    return rv;
}

void
cgul_substring__set_value(cgul_exception_t* cex,
                          cgul_substring_t ss,
                          const char* substring,
                          size_t length)
{
    struct cgul_substring tmp = {0};

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ss);

    /* Clear "tmp". */
    memset(&tmp, 0, sizeof(tmp));

    /* Sanity check. */
    if (!substring || (length <= 0)) {
        cgul_substring__reset(cex, ss);
        goto out;
    }

    /*
     * Make all changes to "tmp" to avoid damaging "ss" in case of exception.
     */

    /* ss.  Because substring is binary, you cannot use strdup() to
     * make a copy. */
    tmp.ss = (char*)cgul_libc__malloc(cex, length + 1);
    if (*cex) {
        goto out;
    }
    memcpy(tmp.ss, substring, length);
    tmp.ss[length] = '\0';

    /* others */
    tmp.ss_hash = cgul_rolling_hash__hash(cex, substring, length);
    if (*cex) {
        goto out;
    }

    /*
     * Success.
     */

    /* Carefully copy or take ownership of the values in "tmp". */
    if (ss->ss) {
        free(ss->ss);
    }
    ss->ss = tmp.ss;
    tmp.ss = NULL;

    ss->ss_length = length;
    ss->ss_hash = tmp.ss_hash;

 out:

    /* Clean up. */
    if (tmp.ss) {
        free(tmp.ss);
    }

    return;
}

size_t
cgul_substring__get_length(cgul_exception_t* cex,
                           cgul_substring_t ss)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ss);

    rv = ss->ss_length;

 out:
    return rv;
}

const char* 
cgul_substring__search(cgul_exception_t* cex,
                       cgul_substring_t ss,
                       const char* block,
                       size_t length)
{
    const char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ss);

    rv = cgul_substring__search_using_prevalues(cex,
                                                ss->ss,
                                                ss->ss_length,
                                                ss->ss_hash,
                                                block,
                                                length);

 out:
    return rv;
}
