#include "cgul_trie_node_p.h"
#include "cgul_libc.h"

/*
 * This method is used as the "restructor" for the "cgul_cache" of
 * nodes in the "cgul_trie".  "restructor" methods are defined in
 * "cgul_cache" to not throw exceptions.
 *
 * While the signature for this method does not exactly match the
 * signature defined for "restructor" methods, it can be safely
 * typecast to match.
 */
void
cgul_trie_node__initialize(cgul_trie_node_t n)
{
    assert(n);
    memset(n, 0, sizeof(*n));
}

cgul_trie_node_t
cgul_trie_node__new(cgul_exception_t* cex)
{
    cgul_trie_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate. */
    rv = (cgul_trie_node_t)cgul_libc__malloc(cex, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Initialize. */
    cgul_trie_node__initialize(rv);

 out:

    /* Clean up. */
    if (*cex && rv) {
        cgul_trie_node__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_trie_node__delete(cgul_trie_node_t n)
{
    if (n) {
        /* Free the node itself. */
        free(n);
    }
}

const char*
cgul_trie_node__get_key(cgul_exception_t* cex,
                        cgul_trie_node_t n)
{
    const char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = n->key;

 out:
    return rv;
}

void
cgul_trie_node__set_key(cgul_exception_t* cex,
                        cgul_trie_node_t n,
                        const char* key)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    n->key = key;

 out:
    return;
}

void*
cgul_trie_node__get_value(cgul_exception_t* cex,
                          cgul_trie_node_t n)
{
    void* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = n->value;

 out:
    return rv;
}

void
cgul_trie_node__set_value(cgul_exception_t* cex,
                          cgul_trie_node_t n,
                          void* value)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    n->value = value;

 out:
    return;
}

cgul_trie_node_t
cgul_trie_node__get_prev(cgul_exception_t* cex,
                         cgul_trie_node_t n)
{
    cgul_trie_node_t rv = NULL;
    cgul_trie_node_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    /*
     * Reverse sort order is determined by the following decreasing
     * list of priorities:
     *
     *    1) n->right
     *    2) n->middle
     *    3) n->key (i.e., found it)
     *    4) n->left
     *
     * Note that step 1 (i.e., n->right) is completed when the user
     * calls cgul_trie__get_back() and
     * cgul_trie_node__get_largest_node_in_subtree().
     */

    /* We start at step 4 (i.e., n->left) because the user passed in
     * the node "n" which means previous calls to this method reached
     * step 3 (i.e., n->key). */
    if (n->left) {
        rv = cgul_trie_node__get_largest_node_in_subtree(cex, n->left);
        goto out;
    }

    /* Walk back up the tree until you find an ancestor that
     * hasn't been exhausted yet. */
    for (curr = n ; curr && curr->parent ; curr = curr->parent) {

        /* curr is parent's right. */
        if (curr == curr->parent->right) {
            if (curr->parent->middle) {
                rv = cgul_trie_node__get_largest_node_in_subtree(
                         cex, curr->parent->middle);
                goto out;
            } else if (curr->parent->key) {
                rv = curr->parent;
                goto out;
            } else if (curr->parent->left) {
                /* This can happen if the user has removed keys from
                 * the tree which can result in the middle subtree
                 * being completely removed and "key" being set to
                 * NULL. */
                rv = cgul_trie_node__get_largest_node_in_subtree(
                         cex, curr->parent->left);
                goto out;
            }
            continue;
        }

        /* curr is parent's middle. */
        if (curr == curr->parent->middle) {
            if (curr->parent->key) {
                rv = curr->parent;
                goto out;
            } else if (curr->parent->left) {
                rv = cgul_trie_node__get_largest_node_in_subtree(
                         cex, curr->parent->left);
                goto out;                
            }
            continue;
        }
    }

    /* Special case for the root node.  Because the root node only
     * ever holds the empty string, it is the prefix of all other
     * strings in the trie.  If curr is the root node and (curr != n),
     * it means the user wants the previous node to some node that is
     * not the root node when all other ancestors have been exhausted.
     * In this case, the root node is the previous node provided the
     * curr->key (i.e., the empty string) has been set for the root
     * node. */
    if (!curr->parent && (curr != n)) {
        if (curr->key) {
            rv = curr;
            goto out;
        }
    }

 out:
    return rv;
}

cgul_trie_node_t
cgul_trie_node__get_next(cgul_exception_t* cex,
                         cgul_trie_node_t n)
{
    cgul_trie_node_t rv = NULL;
    cgul_trie_node_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    /*
     * Sort order is determined by the following decreasing list of
     * priorities:
     *
     *    1) n->left
     *    2) n->key (i.e., found it)
     *    3) n->middle
     *    4) n->right
     *
     * Note that step 1 (i.e., n->left) is completed when the user
     * calls cgul_trie__get_front() and
     * cgul_trie_node__get_smallest_node_in_subtree().
     */

    /* We start at step 3 (i.e., n->middle) because the user passed in
     * the node "n" which means previous calls to this method reached
     * step 2 (i.e., n->key). */
    if (n->middle) {
        rv = cgul_trie_node__get_smallest_node_in_subtree(cex, n->middle);
        goto out;
    }
    
    if (n->right) {
        rv = cgul_trie_node__get_smallest_node_in_subtree(cex, n->right);
        goto out;
    }

    /* Walk back up the tree until you find an ancestor that
     * hasn't been exhausted yet. */
    for (curr = n ; curr && curr->parent ; curr = curr->parent) {

        /* curr is parent's left. */
        if (curr == curr->parent->left) {
            if (curr->parent->key) {
                rv = curr->parent;
                goto out;
            } else if (curr->parent->middle) {
                rv = cgul_trie_node__get_smallest_node_in_subtree(
                         cex, curr->parent->middle);
                goto out;
            } else if (curr->parent->right) {
                /* This can happen if the user has removed keys from
                 * the tree which can result in the middle subtree
                 * being completely removed and "key" being set to
                 * NULL. */
                rv = cgul_trie_node__get_smallest_node_in_subtree(
                         cex, curr->parent->right);
                goto out;
            }
            continue;
        }

        /* curr is parent's middle. */
        if (curr == curr->parent->middle) {
            if (curr->parent->right) {
                rv = cgul_trie_node__get_smallest_node_in_subtree(
                         cex, curr->parent->right);
                goto out;
            }
            continue;
        }
    }

    /*
     * What about the root node?  Because the root node only ever
     * holds the empty string, it is the prefix of all other strings
     * in the trie.  Thus, while the root node can be the "front" node
     * for the entire trie, it cannot be the "next" node for any node
     * in the trie.
     */
    
 out:
    return rv;
}

cgul_trie_node_t
cgul_trie_node__get_older(cgul_exception_t* cex,
                          cgul_trie_node_t n)
{
    cgul_trie_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = n->older;

 out:
    return rv;
}

cgul_trie_node_t
cgul_trie_node__get_younger(cgul_exception_t* cex,
                          cgul_trie_node_t n)
{
    cgul_trie_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = n->younger;

 out:
    return rv;
}

cgul_trie_node_t
cgul_trie_node__get_smallest_node_in_subtree(cgul_exception_t* cex,
                                             cgul_trie_node_t n)
{
    cgul_trie_node_t rv = NULL;
    cgul_trie_node_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    /*
     * Sort order is determined by the following decreasing list of
     * priorities:
     *
     *    1) n->left
     *    2) n->key (i.e., found it)
     *    3) n->middle
     *    4) n->right
     */

    /* Descend the trie. */
    curr = n;
    while (curr) {

        if (curr->left) {
            curr = curr->left;
        } else if (curr->key) {
            rv = curr;
            break;
        } else if (curr->middle) {
            curr = curr->middle;
        } else if (curr->right) {
            curr = curr->right;
        } else {
            /* Special case to handle when the entire subtrie is
             * empty.  In this case, there is no smallest node because
             * the entire subtrie is empty.  The only time this should
             * occur is the special case of trie->root because of the
             * design decision to always have a trie->root node even
             * when the trie is empty.  At all other times, empty
             * nodes should be removed. */
            rv = NULL;
            break;
        }

    }

 out:
    return rv;
}

cgul_trie_node_t
cgul_trie_node__get_largest_node_in_subtree(cgul_exception_t* cex,
                                            cgul_trie_node_t n)
{
    cgul_trie_node_t rv = NULL;
    cgul_trie_node_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    /*
     * Reverse sort order is determined by the following decreasing
     * list of priorities:
     *
     *    1) n->right
     *    2) n->middle
     *    3) n->key (i.e., found it)
     *    4) n->left
     */

    /* Descend the trie. */
    curr = n;
    while (curr) {

        if (curr->right) {
            curr = curr->right;
        } else if (curr->middle) {
            curr = curr->middle;
        } else if (curr->key) {
            rv = curr;
            break;
        } else if (curr->left) {
            curr = curr->left;
        } else {
            /* Special case to handle when the entire subtrie is
             * empty.  In this case, there is no smallest node because
             * the entire subtrie is empty.  The only time this should
             * occur is the special case of trie->root because of the
             * design decision to always have a trie->root node even
             * when the trie is empty.  At all other times, empty
             * nodes should be removed. */
            rv = NULL;
            break;
        }

    }

 out:
    return rv;
}
