/**
 * This program uses <code>cgul_csv_cxx</code> to convert a CSV file
 * to <code>cgul_stanza</code> format.  It assumes the first row of
 * the CSV file contains the column headers.  It then produces one
 * stanza per row where the column headers map to the stanza keys and
 * the column values map to stanza values. <p>
 *
 * One use for this program is to convert CSV files into a format that
 * is easy to diff. <p>
 *
 * @file
 * @author Paul Serice
 * @brief csv to stanza converter
 */

#include <assert.h>
#include <iostream>
#include "cgul_cxx.h"

using namespace std;

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    unsigned long int i = 0;
    unsigned long int headers_count = 0;
    unsigned long int fields_count = 0;
    unsigned long int data_row_count = 0;
    const char* header = NULL;
    const char* value = NULL;
    cgul_csv_cxx csv;
    cgul_vector_cxx headers;

    try {

        /* Verify command line. */
        if (argc != 2) {
            fprintf(stderr, "Usage: main_csv_to_stanza <file.csv>\n");
            return 1;
        }

        /* Open the cvs file. */
        csv.open_fname(argv[1]);

        /* Read the headers. */
        csv.load_next();

        /* Remember the headers and their exact order so that the order of
         * the resulting stanza matches the order of the columns in the
         * csv file */
        headers_count = csv.get_number_of_fields();
        for (i = 0 ; i < headers_count ; ++i) {
            header = csv.take_value(i);
            headers.push_back(header);
        }

        /* Iterate over each data row. */
        while (csv.load_next()) {

            /* Increment the count of data rows. */
            ++data_row_count;

            /* Make sure this row has the expected number of columns. */
            fields_count = csv.get_number_of_fields();
            if (fields_count != headers_count) {
                throw cgul_exception_cxx(
                          "Data Row %lu: Expecting %lu columns but found %lu.",
                          data_row_count,
                          headers_count,
                          fields_count);
            }

            /* Print the stanza separator if necessary. */
            if (data_row_count > 1) {
                printf("\n");
            }

            /* Iterate over each column in the row. */
            for (i = 0 ; i < headers_count ; ++i) {

                /* Get the header for this column. */
                header = (const char*)headers.at(i);

                /* Get the value for this column. */
                value = (const char*)csv.lookup(i);

                /* Print the stanza line which could be multi-line text. */
                cgul_stanza_cxx::print(stdout, header, value);
            }
        }

    } catch (const cgul_exception_cxx& e) {
        cerr << "*** Error: " << e.get_message() << endl;
        rv = 1;
    }

    /*
     * Clean up.
     */

    /* Free each individual header. */
    for (;;) {
        header = (const char*)headers.pop_back();
        if (!header) {
            break;
        }
        free((void*)header);
    }

    return rv;
}
