#include <assert.h>
#include "cgul.h"

/**
 * Whether to print the unsorted and sorted vectors.
 */
#define ENABLE_PRINT 0

/**
 * This variable controls the range of the pseudo-random number
 * generator.  The numbers generated by the pseudo-random number
 * generator control how many numbers to sort as well as the actual
 * numbers that get sorted.
 */
static cgul_uint32_t LCG_MAX_RANGE = 10000000;

/**
 * Structure used to test <code>cgul_list__sort()</code>.  The
 * <code>x</code> value is set to a random number and is the only part
 * of the structure checked by the comparison function
 * <code>compare_points()</code>.  The <code>y</code> value is
 * incremented.  This is used to make sure the sort is stable.
 */
typedef struct point {
    unsigned long int x;
    unsigned long int y;
} *point_t;

static int
compare_points(point_t p1,
               point_t p2)
{
    int rv = 0;
    if (p1->x < p2->x) {
        rv = -1;
    } else if (p1->x > p2->x) {
        rv = 1;
    }
    return rv;
}

#if ENABLE_PRINT
static void
print_points(cgul_exception_t* cex,
             cgul_list_t lst)
{
    point_t p = NULL;
    cgul_list_node_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    curr = cgul_list__get_front(cex, lst);
    for ( ; curr ; curr = cgul_list_node__get_next(cex, curr)) {
        p = (point_t)cgul_list_node__get_value(cex, curr);
        printf("(%6lu, %6lu)\n", p->x, p->y);
    }

    puts("=====");

 out:
    return;
}
#endif

static void
insert_random_points(cgul_exception_t* cex,
                     cgul_list_t lst,
                     cgul_random_lcg32_t lcg,
                     cgul_uint32_t rmax)
{
    unsigned long int i = 0;
    unsigned long int i_max = 0;
    point_t p = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Get the next random number. */
    i_max = cgul_random_lcg32__next_in_range(cex, lcg, rmax);
    if (*cex) {
        goto out;
    }

    /* Create each point. */
    for (i = 0 ; i < i_max ; ++i) {

        /* Allocate the next point. */
        p = cgul_libc__calloc(cex, 1, sizeof(*p));
        if (*cex) {
            goto out;
        }

        /* Get the next random number. */
        p->x = cgul_random_lcg32__next_in_range(cex, lcg, rmax);
        if (*cex) {
            goto out;
        }

        /* Track the order in which the points are inserted so we can
         * check later to see if the sort is stable. */
        p->y = i;

        /* Insert the point on the list. */
#if 1
        cgul_list__push_back(cex, lst, p);
#else
        cgul_list__insert_sorted(cex, lst, p,
                                 (cgul_list__compare_t)compare_points);
#endif
        if (*cex) {
            goto out;
        }

        /* The list now owns the point. */
        p = NULL;
    }

 out:

    /* Clean up. */
    if (p) {
        free(p);
    }

    return;
}

static int
is_sorted_and_stable(cgul_exception_t* cex,
                     cgul_list_t lst)
{
    int rv = 0;
    unsigned long int prev_x;
    unsigned long int prev_y;
    point_t p = NULL;
    cgul_list_node_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Get the front of the list.  If the list is empty, it is sorted
     * and stable. */
    curr = cgul_list__get_front(cex, lst);
    if (!curr) {
        rv = 1;
        goto out;
    }

    /* Get the first values of x and y. */
    p = (point_t)cgul_list_node__get_value(cex, curr);
    prev_x = p->x;
    prev_y = p->y;

    /* Iterate over the remaining nodes. */
    curr = cgul_list_node__get_next(cex, curr);
    for ( ; curr ; curr = cgul_list_node__get_next(cex, curr)) {

        /* Get the current point. */
        p = (point_t)cgul_list_node__get_value(cex, curr);

        /* The x-value is a randomly generated number.  It is used to
         * test whether the list is sorted.  If the current value of x
         * is less than the previous value of x, the list is not
         * sorted. */
        if (p->x < prev_x) {
            goto out;
        }

        /* The y-value is the order the nodes were inserted.  It is
         * used to test for stability.  Stability can only be tested
         * when prev_x and p->x have the same value. */
        if ((prev_x == p->x) && (p->y < prev_y)) {
            goto out;
        }

        /* Remember the current point. */
        prev_x = p->x;
        prev_y = p->y;
    }

    /* Success. */
    rv = 1;

 out:
    return rv;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_uint32_t rmax = LCG_MAX_RANGE;
    cgul_list_t lst = NULL;
    cgul_random_lcg32_t lcg = NULL;
    cgul_exception_t local = NULL;

    /* Allocate. */
    lst = cgul_list__new(&local);
    lcg = cgul_random_lcg32__new(&local);
    if (local) {
        goto out;
    }

    /* Insert random points. */
    insert_random_points(&local, lst, lcg, rmax);

#if ENABLE_PRINT
    /* Print the points. */
    print_points(&local, lst);
#endif

    /* Sort. */
    cgul_list__sort(&local, lst, (cgul_list__compare_t)compare_points);

#if ENABLE_PRINT
    /* Print the points. */
    print_points(&local, lst);
#endif

    /* Test whether the list is in fact sorted and stable. */
    if (!is_sorted_and_stable(&local, lst)) {
        local = cgul_exception__new("Not sorted or not stable.");
        goto out;
    }

    printf("Total Points Sorted: %lu\n",
           (unsigned long)cgul_list__get_size(&local, lst));

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (lst) {
        while (!cgul_list__is_empty(&local, lst)) {
            free(cgul_list__pop_back(&local, lst));
        }
    }
    if (lcg) {
        cgul_random_lcg32__delete(lcg);
    }
    if (lst) {
        cgul_list__delete(lst);
    }

    return rv;
}
