/**
 * This program merges two sorted files writing the result to
 * <code>stdout</code>.  If you have access to POSIX sort, you can get
 * the same result by running "sort -m" instead.
 *
 * @file
 * @brief merge two sorted files
 */

#include "cgul.h"

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    const char* line1 = NULL;
    const char* line2 = NULL;
    cgul_crlf_file_t f1 = NULL;
    cgul_crlf_file_t f2 = NULL;
    cgul_exception_t local = NULL;

    /* Verify command line. */
    if (argc != 3) {
        printf("\nUsage: merge <file1> <file2>\n\n");
        goto out;
    }

    /* Open the sorted files. */
    f1 = cgul_crlf_file__new_from_fname(&local,
                                        argv[1],
                                        1 /* is_block_buffered */);
    f2 = cgul_crlf_file__new_from_fname(&local,
                                        argv[2],
                                        1 /* is_block_buffered */);
    if (local) {
        goto out;
    }

    /* Merge the sorted files writing the result to stdout. */
    for (;;) {

        /* If line1 has been consumed, get the next line from f1. */
        if (!line1) {
            line1 = cgul_crlf_file__get_line(&local, f1);
            if (!line1) {
                /* EOF */
                break;
            }
        }

        /* If line2 has been consumed, get the next line from f2. */
        if (!line2) {
            line2 = cgul_crlf_file__get_line(&local, f2);
            if (!line2) {
                /* EOF */
                break;
            }
        }

        /* Print the smaller line then and set it to NULL to mark it
         * as having been consumed. */
        if (strcmp(line1, line2) < 0) {
            printf("%s\n", line1);
            line1 = NULL;
        } else {
            printf("%s\n", line2);
            line2 = NULL;            
        }
    }

    /* There could still be a bullet in the chamber. */
    if (line1) {
        printf("%s\n", line1);
        line1 = NULL;
    }
    if (line2) {
        printf("%s\n", line2);
        line2 = NULL;
    }

    /* Print the remainder. */
    while ((line1 = cgul_crlf_file__get_line(&local, f1)) != NULL) {
        printf("%s\n", line1);
    }
    while ((line2 = cgul_crlf_file__get_line(&local, f2)) != NULL) {
        printf("%s\n", line2);
    }

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (f2) {
        cgul_crlf_file__delete(f2);
    }
    if (f1) {
        cgul_crlf_file__delete(f1);
    }

    return rv;
}
