/*
 * This program test cgul_mime by randomly generating multipart MIME
 * messages in memory and calling cgul_mime__parse() on the MIME
 * message in memory.  The client callback for cgul_mime__parse()
 * tries to reassemble the original MIME message.  If successful, the
 * process repeats.  If not successful, the MIME message is written to
 * file, and the program exits with an error.  You can then re-parse
 * (but not re-test (yet)) the MIME message in a debugger by
 * specifying the file on the command line.  So, the usage boils down
 * to the following:
 *
 *     1)  Test Random MIME:  ./main_mime_test
 *     2) Test MIME in File:  ./main_mime_test <file>
 */

#include "cgul.h"

/**
 * Manifest constants that allow you to tune the size of the various
 * parts of the MIME messages that are randomly created.
 */
#if 1
#define MAX_PREAMBLE_SIZE (20971520)
#define MAX_SECTIONS (101)
#define MAX_HEADER_SIZE (1048576)
#define MAX_BODY_SIZE (1048576)
#define MAX_EPILOGUE_SIZE (20971520)
#define MAX_PRODUCER_READ (16384)
#else
#define MAX_PREAMBLE_SIZE (0)
#define MAX_SECTIONS (1)
#define MAX_HEADER_SIZE (1048)
#define MAX_BODY_SIZE (1048)
#define MAX_EPILOGUE_SIZE (0)
#define MAX_PRODUCER_READ (16384)
#endif

/**
 * The EOL sequence to be used.
 */
#define EOL_SEQUENCE "\r\n"

/**
 * The data structure that describes each part in a multipart MIME message.
 */
typedef struct {

    /**
     * Pointer to the start of the header.
     */
    char* header;

    /**
     * Size of the header in bytes.
     */
    size_t header_size;

    /**
     * Pointer to the start of the body.
     */
    char* body;

    /**
     * Size of the body in bytes.
     */
    size_t body_size;

} *section_t;

/**
 * The data structure that describes a MIME message.
 */
typedef struct {

    /**
     * Size of the MIME message in bytes if converted from this data
     * structure into a block of contiguous memory using
     * <code>main__message__to_string()</code>.
     */
    size_t message_size;

    /**
     * EOL sequence to use.
     */
    char* eol;

    /**
     * Length of the eol sequence.
     */
    size_t eol_length;

    /**
     * Boundary string.
     */
    char* boundary;

    /**
     * Length of the <code>boundary</code> string.
     */
    size_t boundary_length;

    /**
     * Final boundary string.
     */
    char* final_boundary;

    /**
     * Length of the <code>final_boundary</code> string.
     */
    size_t final_boundary_length;

    /**
     * Pointer to the start of the preamble.
     */
    char* preamble;

    /**
     * Size of the preamble in bytes.
     */
    size_t preamble_size;

    /**
     * Linked list of <code>section</code> objects that describe each
     * part in the multipart MIME message.
     */
    cgul_list_t sections;

    /**
     * Pointer to the start of the epilogue.
     */
    char* epilogue;

    /**
     * Size of the epilogue in bytes.
     */
    size_t epilogue_size;

} *message_t;

/**
 * Client data for the callback that produces blocks for the MIME
 * parser.
 */
typedef struct producer_data {

    /**
     * Pointer to the start of the MIME message.
     */
    const char* message;

    /**
     * Size of the MIME message in bytes.
     */
    size_t message_size;

    /**
     * Index into the MIME message.
     */
    size_t index;

    /**
     * The maximum number of bytes the producer should pass back to
     * the parser on each iteration.
     */
    size_t max;

} *producer_data_t;

/**
 * Client data for the callback that consumes blocks from the MIME
 * parser.
 */
typedef struct consumer_data {

    /**
     * Data structure that describes the MIME message which should be
     * filled in by the callback as it receives each block.
     */
    message_t message;

    /**
     * String used to build up each part of the MIME message.
     */
    cgul_string_t s;

    /**
     * The previous section type.
     */
    cgul_mime__section_t prev_section;

} *consumer_data_t;

/**
 * Fill the block of memory <code>block</code> with
 * <code>block_size</code> bytes with numbers generated randomly with
 * the PRNG <code>prng</code>.  After the block is filled, it is
 * tested to make sure it does not contain the MIME boundary string
 * <code>boundary</code>.  If it does the process repeats until it
 * doesn't. <p>
 *
 * Note, this function does not add any "\r" or "\n" characters to the
 * block.  While this would be nice, an accidental EOL sequence in
 * just the right place could change the semantics of the entire MIME
 * message meaning that if <code>cgul_mime</code> parses the message
 * _correctly_ it is unlikely to match the original causing a spurious
 * failure. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] prng PRNG
 * @param[in] block block
 * @param[in] block_size size of block in bytes
 * @param[in] boundary MIME boundary string
 * @param[in] boundary_length length of the MIME boundary string
 */
static void
main__randomly_fill_block(cgul_exception_t* cex,
                          cgul_random_xorshift32_t prng,
                          char* block,
                          size_t block_size,
                          const char* boundary,
                          size_t boundary_length)
{
    size_t i = 0;
    const char* found = NULL;

    do {

        /* Fill the block being careful to introduce any "\r" or "\n"
         * characters because they could change the semantics of the
         * MIME message. */
        for (i = 0 ; i < block_size ; ++i) {
            do {
                block[i] = (char)cgul_random_xorshift32__next(cex, prng);
            } while((block[i] == '\r') || (block[i] == '\n'));
        }

        /* If we can find the boundary string in the block, the block
         * needs to be refilled. */
        found = cgul_substring__search_ez(cex,
                                          boundary,
                                          boundary_length,
                                          block,
                                          block_size);

    } while (found);
}

/**
 * Randomly generate a MIME boundary string and final_boundary string.
 * The boundary string and the final boundary string will both have
 * the EOL sequence <code>eol</code> prepended and appended.
 * Furthermore, both may have optional white space before the trailing
 * EOL.  Only the final boundary string will have a <code>"--"</code>
 * sequence after the boundary string but before the optional white
 * space.  The caller is responsible for calling <code>free()</code>
 * on both pointers returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] prng PRNG
 * @param[in] eol EOL sequence
 * @param[out] boundary boundary string
 * @param[out] final_boundary final boundary string
 */
static void
main__generate_boundary_strings(cgul_exception_t* cex,
                                cgul_random_xorshift32_t prng,
                                const char* eol,
                                char** boundary,
                                char** final_boundary)
{
    cgul_uint32_t x = 0;
    unsigned long int i = 0;
    unsigned long int i_max = 0;
    size_t length_before_whitespace = 0;
    cgul_string_t s1 = NULL;
    cgul_string_t s2 = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(prng);

    /* Allocate. */
    s1 = cgul_string__new(cex);
    s2 = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    do {

        /* Reset the strings. */
        cgul_string__set_value(cex, s1, "");
        cgul_string__set_value(cex, s2, "");
        if (*cex) {
            goto out;
        }

        /* Append the leading EOL sequence and "--" sequence to s1
         * which will become the boundary string. */
        cgul_string__append(cex, s1, eol);
        cgul_string__append(cex, s1, "--");

        /* Append the leading EOL sequence and "--" sequence to s2
         * which will become the final boundary string. */
        cgul_string__append(cex, s2, eol);
        cgul_string__append(cex, s2, "--");


        /* Add a random number of leading dash characters. */
        i_max = cgul_random_xorshift32__next_in_unit(cex, prng) * 32;
        for (i = 0 ; i < i_max ; ++i) {
            cgul_string__append_char(cex, s1, '-');
            cgul_string__append_char(cex, s2, '-');
            if (*cex) {
                goto out;
            }
        }

        /* Add some random hex digits. */
        i_max = cgul_random_xorshift32__next_in_unit(cex, prng) * 40;
        for (i = 0 ; i < i_max ; ++i) {
            /* Get the next 32-bit value. */
            x = cgul_random_xorshift32__next(cex, prng);
            /* For simplicity, only append the least-significant hex
             * digit, i.e., the least-significant 4 bits. */
            cgul_string__append_sprintf(cex, s1, "%x", (x & 0xf));
            cgul_string__append_sprintf(cex, s2, "%x", (x & 0xf));
            if (*cex) {
                goto out;
            }
        }

        /* For the final boundary string, append a trailing "--" sequence. */
        cgul_string__append(cex, s2, "--");
        if (*cex) {
            goto out;
        }

        /* Remember the length of the boundary string before we append
         * the optional white space. */
        length_before_whitespace = cgul_string__get_length(cex, s1);

        /* Add some random horizontal white space half of the time (by
         * skipping the loop unless i_max is >= 20). */
        i_max = cgul_random_xorshift32__next_in_unit(cex, prng) * 40;
        for (i = 20 ; i < i_max ; ++i) {
            /* Get the next 32-bit value. */
            x = cgul_random_xorshift32__next(cex, prng);
            /* Add a space or a tab. */
            if ((x & 0x1) == 0) {
                cgul_string__append_char(cex, s1, ' ');
                cgul_string__append_char(cex, s2, ' ');
            } else {
                cgul_string__append_char(cex, s1, '\t');
                cgul_string__append_char(cex, s2, '\t');
            }
            if (*cex) {
                goto out;
            }            
        }

    } while(length_before_whitespace < 5);

    /* Append trailing EOL sequences. */
    cgul_string__append(cex, s1, eol);
    cgul_string__append(cex, s2, eol);
    if (*cex) {
        goto out;
    }

    /* Success. */
    if (boundary) {
        *boundary = cgul_string__take_value(cex, s1);
    }
    if (final_boundary) {
        *final_boundary = cgul_string__take_value(cex, s2);
    }

 out:

    /* Clean up. */
    if (s2) {
        cgul_string__delete(s2);
    }
    if (s1) {
        cgul_string__delete(s1);
    }

    return;
}

/**
 * Delete the section <code>section</code> freeing all internally
 * allocated resources.  The caller must not attempt to dereference
 * <code>section</code> after calling this method.
 *
 * @param[in] section <code>section</code> instance
 */
static void
main__section__delete(section_t section)
{
    if (section) {

        /* body */
        if (section->body) {
            free(section->body);
        }

        /* header */
        if (section->header) {
            free(section->header);
        }

        /* Free the object itself. */
        free(section);
    }
}

/**
 * Create a new section.
 *
 * @param[in,out] cex c-style
 * @return new <code>section</code> instance
 */
static section_t
main__section__new(cgul_exception_t* cex)
{
    section_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate the object. */
    rv = cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up partially initialized objects. */
    if (*cex && rv) {
        main__section__delete(rv);
        rv = NULL;
    }
    
    return rv;
}

/**
 * Compare the first MIME section <code>section1</code> to the second
 * MIME section <code>section2</code>.  Return true if the sections
 * are equal; otherwise, return false.  Two MIME sections are
 * considered to be equal if the header and body contain data that are
 * byte-for-byte equal.  Note that this definition of equality permits
 * the two mime sections to have different boundary strings including
 * different white space after the boundary string but before the EOL
 * for the boundary line.
 *
 * @param[in] cex c-style exception
 * @param[in] section1 first MIME section
 * @param[in] section2 second MIME section
 * @return whether the MIME sections are equal
 */
static int
main__section__is_equal(cgul_exception_t* cex,
                        section_t section1,
                        section_t section2)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(section1);
    assert(section2);

    /* Compare the size of each header. */
    if (section1->header_size != section2->header_size) {
        goto out;
    }

    /* Compare the content of each header. */
    if (memcmp(section1->header, section2->header, section1->header_size)) {
        goto out;
    }

    /* Compare the size of each body. */
    if (section1->body_size != section2->body_size) {
        goto out;
    }

    /* Compare the content of each body. */
    if (memcmp(section1->body, section2->body, section1->body_size)) {
        goto out;
    }

    /* Success. */
    rv = 1;

 out:
    return rv;
}

/**
 * Common code for calculating the total number of bytes that the MIME
 * message represented by the data structure <code>msg</code> will
 * have after it has been expanded into one contiguous string.  The
 * calculated size is saved to <code>msg->message_size</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] msg <code>message</code> instance
 */
void
main__message__calculate_size(cgul_exception_t* cex,
                              message_t msg)
{
    unsigned long int i = 0;
    size_t total = 0;
    section_t section = NULL;
    cgul_list_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(msg);

    /* Increment the total size. */
    total += msg->preamble_size;

    /* Iterate over each section. */
    n = cgul_list__get_front(cex, msg->sections);
    for (i = 0 ; n ; n = cgul_list_node__get_next(cex, n), ++i) {

        /* Increment the total size to account for the boundary
         * separator.  If the preamble is empty, the first boundary
         * string does not have a leading EOL sequence. */
        if ((i == 0) && (msg->preamble_size == 0)) {
            total += msg->boundary_length;
            total -= msg->eol_length;
        } else {
            total += msg->boundary_length;
        }

        /* Get the current section. */
        section = cgul_list_node__get_value(cex, n);

        /* Increment the total size to account for the header,
         * back-to-back EOL sequences that separate the header from
         * the body, and the body.  As explained in
         * main__message__new_random__create_section(), the header is
         * either empty or it has at least one character followed by
         * an EOL sequence.  If the header is empty, the first EOL
         * sequences of the back-to-back pair comes from the boundary
         * line.  If the header is not empty, the first one comes from
         * the header itself. */
        total += section->header_size;
        total += msg->eol_length;
        total += section->body_size;
    }

    /* Increment the total to account for the final boundary
     * separator, but only if there was at least one part present in
     * the multipart message. */
    if (i > 0) {
        total += msg->final_boundary_length;
    }

    /* Increment the total size. */
    total += msg->epilogue_size;

    /* Success. */
    msg->message_size = total;

 out:
    return;
}

/**
 * This is a helper function for <code>main__message__delete()</code>
 * that frees all the sections on the list <code>sections</code>, but
 * it does not free the list itself.  Because it is invoked as part of
 * the <code>main__message__delete()</code> destructor, it does not
 * require a <code>cgul_exception</code> object.
 *
 * @param[in] sections sections
 */
static void
main__message__delete_sections(cgul_list_t sections)
{
    section_t section = NULL;
    cgul_exception_t local = NULL;

    if (sections) {

        /* Free each section. */
        while ((section = cgul_list__pop_back(&local, sections)) != NULL) {
            main__section__delete(section);
        }

    }
}

/**
 * Delete the message <code>msg</code> freeing all internally
 * allocated resources.  The caller must not attempt to dereference
 * <code>msg</code> after calling this method.
 *
 * @param[in] msg <code>message</code> instance
 */
static void
main__message__delete(message_t msg)
{
    if (msg) {

        /* eol */
        if (msg->eol) {
            free(msg->eol);
        }

        /* epilogue */
        if (msg->epilogue) {
            free(msg->epilogue);
        }

        /* sections */
        if (msg->sections) {
            main__message__delete_sections(msg->sections);
            cgul_list__delete(msg->sections);
        }

        /* preamble */
        if (msg->preamble) {
            free(msg->preamble);
        }

        /* final_boundary */
        if (msg->final_boundary) {
            free(msg->final_boundary);
        }

        /* final_boundary */
        if (msg->boundary) {
            free(msg->boundary);
        }

        /* Free the object itself. */
        free(msg);
    }
}

/**
 * Create a new, uninitialized message.
 *
 * @param[in,out] cex c-style
 * @return new <code>message</code> instance
 */
static message_t
main__message__new(cgul_exception_t* cex)
{
    message_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate the object. */
    rv = cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Add the linked list that will hold the sections. */
    rv->sections = cgul_list__new(cex);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up partially initialized objects. */
    if (*cex && rv) {
        main__message__delete(rv);
        rv = NULL;
    }
    
    return rv;
}

/**
 * This is a helper function for
 * <code>main__message__new_random()</code> that creates a random
 * preamble <code>msg->preamble</code>.  The size of the newly created
 * preamble will be saved to <code>msg->preamble_size</code>.  The
 * preamble itself will be initialized with random bytes from the PRNG
 * <code>prng</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] msg <code>message</code> instance
 * @param[in] prng PRNG
 */
static void
main__message__new_random__create_preamble(cgul_exception_t* cex,
                                           message_t msg,
                                           cgul_random_xorshift32_t prng)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Randomly determine the size of the preamble where 50% of the
     * time there is no preamble, i.e., the size is zero.  The other
     * 50% of the time the size of the preamble varies up to
     * MAX_PREAMBLE_SIZE.  To do this, a PRN is generated that is
     * double the size of MAX_PREAMBLE_SIZE.  If the value is in the
     * lower half, the preamble is empty.  If the value is in the
     * upper half, the PRN is scaled back down to zero (by subtracting
     * out the lower half), and the preamble gets that size. */
    msg->preamble_size
        = cgul_random_xorshift32__next_in_unit(cex, prng)
          * MAX_PREAMBLE_SIZE * 2;
    if (msg->preamble_size < MAX_PREAMBLE_SIZE) {
        msg->preamble_size = 0;
    } else {
        msg->preamble_size -= MAX_PREAMBLE_SIZE;
    }

    /* Allocate the preamble. */
    msg->preamble = cgul_libc__malloc(cex, msg->preamble_size + 1);
    if (*cex) {
        goto out;
    }
    msg->preamble[0] = '\0';

    /* Randomly initialize the preamble. */
    main__randomly_fill_block(cex,
                              prng,
                              msg->preamble,
                              msg->preamble_size,
                              msg->boundary + msg->eol_length,
                              msg->boundary_length - msg->eol_length);

 out:
    return;
}

/**
 * This is a helper function for
 * <code>main__message__new_random()</code> that creates a new, random
 * section and returns it.  The header and body of the section will be
 * initialized with random bytes from the PRNG <code>prng</code>.  The
 * caller is responsible for calling
 * <code>main__section__delete()</code> on the pointer returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] msg <code>message</code> instance
 * @param[in] prng PRNG
 * @return new, random section
 */
static section_t
main__message__new_random__create_section(cgul_exception_t* cex,
                                          message_t msg,
                                          cgul_random_xorshift32_t prng)
{
    size_t i = 0;
    size_t j = 0;
    section_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create the empty object that describes the next section. */
    rv = main__section__new(cex);
    if (*cex) {
        goto out;
    }

    do {
        /* Randomly determine the size of each header. */
        rv->header_size =
            cgul_random_xorshift32__next_in_unit(cex, prng) * MAX_HEADER_SIZE;
        
        /* Randomly determine the size of each body. */
        rv->body_size =
            cgul_random_xorshift32__next_in_unit(cex, prng) * MAX_BODY_SIZE;

    /* Both the header and the body cannot be empty.  (See the
     * "IMPORTANT" note below for why we don't just check
     * rv->header_size against zero.) */
    } while ((rv->header_size <= msg->eol_length) && (rv->body_size == 0));


    /* Allocate the header. */
    rv->header = cgul_libc__malloc(cex, rv->header_size + 1);
    if (*cex) {
        goto out;
    }
    rv->header[0] = '\0';

    /* IMPORTANT: If the header is not empty, it must always end with
     * an EOL sequence.  This is because the first of the back-to-back
     * EOL sequences that separate the header from the body is
     * considered to belong to the header which makes sense because
     * the header usually holds _lines_ of text, and the EOL sequence
     * for the last line of text in the header in the first EOL
     * sequence in the back-to-back pair.
     *
     * Also, cgul_mime__parse() always parses the MIME message this
     * way.  So in order to be able to verify the original MIME
     * message with the reconstructed MIME message, the original MIME
     * message needs to always have an EOL sequence at the end of
     * non-empty headers.
     *
     * Lastly, rv->header shouldn't just be an EOL sequence without
     * any other characters because this will result in 3 back-to-back
     * EOL sequences: one for the EOL sequence that terminates the
     * boundary string, one for the EOL sequence in rv->header, and
     * one for the EOL sequence that will be added by
     * main__message__to_string() when it appends the second of the
     * back-to-back EOL sequences that separate the header from the
     * body.  At any rate, two is all it takes because
     * cgul_mime__parse() correctly detects the first two as
     * indicating an empty header.  So in this case, the reconstructed
     * message will have an empty header, but the original message
     * will not be empty because it holds an EOL sequence. */
    if (rv->header_size <= msg->eol_length) {
        rv->header_size = 0;

    } else {

        /* Randomly initialize the header. */
        main__randomly_fill_block(cex,
                                  prng,
                                  rv->header,
                                  rv->header_size,
                                  msg->boundary + msg->eol_length,
                                  msg->boundary_length - msg->eol_length);

        /* Always terminate the header with an EOL sequence.  Note
         * that we have already checked above to make sure
         * (rv->header_size >= msg->eol_length) (when we are in the
         * "else" branch).  We do this copy by hand because strpy()
         * copies a trailing NUL-terminator which is not welcome in
         * this context, and strncpy() is not as portable as I would
         * like. */
        for (i = rv->header_size - msg->eol_length, j = 0;
             j < msg->eol_length;
             ++i, ++j)
        {
            rv->header[i] = msg->eol[j];
        }

    }

    /* Allocate the body. */
    rv->body = cgul_libc__malloc(cex, rv->body_size + 1);
    if (*cex) {
        goto out;
    }
    rv->body[0] = '\0';

    /* Randomly initialize the body. */
    main__randomly_fill_block(cex,
                              prng,
                              rv->body,
                              rv->body_size,
                              msg->boundary + msg->eol_length,
                              msg->boundary_length - msg->eol_length);

 out:

    /* Clean up. */
    if (*cex && rv) {
        main__section__delete(rv);
        rv = NULL;
    }

    return rv;
}

/**
 * This is a helper function for
 * <code>main__message__new_random()</code> that creates a random
 * epilogue <code>msg->epilogue</code>.  The size of the newly created
 * epilogue will be saved to <code>msg->epilogue_size</code>.  The
 * epilogue itself will be initialized with random bytes from the PRNG
 * <code>prng</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] msg <code>message</code> instance
 * @param[in] prng PRNG
 */
static void
main__message__new_random__create_epilogue(cgul_exception_t* cex,
                                           message_t msg,
                                           cgul_random_xorshift32_t prng)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Randomly determine the size of the epilogue where 50% of the
     * time there is no epilogue, i.e., the size is zero.  The other
     * 50% of the time the size of the epilogue varies up to
     * MAX_EPILOGUE_SIZE.  To do this, a PRN is generated that is
     * double the size of MAX_EPILOGUE_SIZE.  If the value is in the
     * lower half, the epilogue is empty.  If the value is in the
     * upper half, the PRN is scaled back down to zero (by subtracting
     * out the lower half), and the epilogue gets that size. */
    msg->epilogue_size
        = cgul_random_xorshift32__next_in_unit(cex, prng)
          * MAX_EPILOGUE_SIZE * 2;
    if (msg->epilogue_size < MAX_EPILOGUE_SIZE) {
        msg->epilogue_size = 0;
    } else {
        msg->epilogue_size -= MAX_EPILOGUE_SIZE;
    }

    /* Allocate the epilogue. */
    msg->epilogue = cgul_libc__malloc(cex, msg->epilogue_size + 1);
    if (*cex) {
        goto out;
    }
    msg->epilogue[0] = '\0';

    /* Randomly initialize the epilogue. */
    main__randomly_fill_block(cex,
                              prng,
                              msg->epilogue,
                              msg->epilogue_size,
                              msg->boundary + msg->eol_length,
                              msg->boundary_length - msg->eol_length);

 out:
    return;
}

/**
 * Create a new, random, MIME message.  The message returned is _not_
 * contiguous in memory.  Use <code>main__message__to_string()</code>
 * to convert this data structure into a MIME message that is
 * contiguous in memory.
 *
 * @param[in,out] cex c-style exception
 * @param[in] prng PRNG
 * @param[in] eol EOL sequence
 * @return new, random, MIME message
 */
static message_t
main__message__new_random(cgul_exception_t* cex,
                          cgul_random_xorshift32_t prng,
                          const char* eol)
{
    unsigned long int i = 0;
    unsigned long int section_count = 0;
    section_t section = NULL;
    message_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create an uninitialized object. */
    rv = main__message__new(cex);
    if (*cex) {
        goto out;
    }

    /* Duplicate the EOL sequence. */
    rv->eol = cgul_libc__strdup(cex, eol);
    if (*cex) {
        goto out;
    }
    rv->eol_length = strlen(rv->eol);

    /* Generate the boundary string.  Note that both the boundary
     * string and the final boundary string have the EOL sequence both
     * prepended and append.  Thus, when printing the first boundary
     * string, the EOL sequence will have to be manually skipped if
     * the preamble is empty.  */
    main__generate_boundary_strings(
        cex, prng, rv->eol, &rv->boundary, &rv->final_boundary);
    if (*cex) {
        goto out;
    }

    /* Get the length of the boundary string and the final boundary
     * string. */
    rv->boundary_length = strlen(rv->boundary);
    rv->final_boundary_length = strlen(rv->final_boundary);

    /* Randomly create rv->preamble and set rv->preamble_size. */
    main__message__new_random__create_preamble(cex, rv, prng);
    if (*cex) {
        goto out;
    }

    /* Randomly create the sections. */
    section_count
        = cgul_random_xorshift32__next_in_unit(cex, prng) * MAX_SECTIONS + 1;
    for (i = 0 ; i < section_count ; ++i) {

        /* Create the multipart section.  Unlike the preamble and the
         * epilogue, the section is not directly added to rv; instead,
         * it is returned. */
        section = main__message__new_random__create_section(cex, rv, prng);
        if (*cex) {
            goto out;
        }

        /* Add the section to the list of sections. */
        cgul_list__push_back(cex, rv->sections, section);
        if (*cex) {
            goto out;
        }

        /* The list now owns the section. */
        section = NULL;
    }

    /* Randomly create rv->epilogue and set rv->epilogue_size. */
    main__message__new_random__create_epilogue(cex, rv, prng);
    if (*cex) {
        goto out;
    }

    /* Calculate the size and set rv->message_size accordingly. */
    main__message__calculate_size(cex, rv);

 out:

    /* Clean up partially initialized objects. */
    if (*cex && rv) {
        main__message__delete(rv);
        rv = NULL;
    }
    if (section) {
        main__section__delete(section);
    }

    return rv;
}

/**
 * Return whether all the multipart MIME sections in the two MIME
 * messages <code>msg1</code> and <code>msg2</code> are equal.  This
 * function simply iterates over the sections in order and compares
 * each pair of sections by calling
 * <code>main__section__is_equal()</code>.  If
 * <code>main__section__is_equal()</code> returns fails for any pair
 * of sections, false is returned; otherwise, true is returned.
 *
 * @param[in] cex c-style exception
 * @param[in] msg1 first MIME message
 * @param[in] msg2 second MIME message
 * @return whether all the multipart MIME sections are equal
 */
static int
main__message__are_all_sections_equal(cgul_exception_t* cex,
                                      message_t msg1,
                                      message_t msg2)
{
    int rv = 0;
    size_t size1 = 0;
    size_t size2 = 0;
    section_t section1 = NULL;
    section_t section2 = NULL;
    cgul_list_node_t n1 = NULL;
    cgul_list_node_t n2 = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(msg1);
    assert(msg2);

    /* Make sure both messages have the same number of sections. */
    size1 = cgul_list__get_size(cex, msg1->sections);
    size2 = cgul_list__get_size(cex, msg2->sections);
    if (size1 != size2) {
        goto out;
    }

    /* Iterate over each pair of sections from msg1 and msg2. */
    n1 = cgul_list__get_front(cex, msg1->sections);
    n2 = cgul_list__get_front(cex, msg2->sections);
    while (n1 && n2) {

        /* Get the pair of sections. */
        section1 = cgul_list_node__get_value(cex, n1);
        section2 = cgul_list_node__get_value(cex, n2);

        /* Compare the pair of sections. */
        if (!main__section__is_equal(cex, section1, section2)) {
            goto out;
        }

        /* Get the next pair of nodes. */
        n1 = cgul_list_node__get_next(cex, n1);
        n2 = cgul_list_node__get_next(cex, n2);
    }

    /* Success. */
    rv = 1;

 out:
    return rv;
}

/**
 * Compare the first MIME message <code>msg1</code> to the second MIME
 * message <code>msg2</code>.  Return true if the messages are equal;
 * otherwise, return false.  Two MIME messages are considered to be
 * equal if the preamble, multipart sections, and epilogue contain
 * data that are byte-for-byte equal.  Note that this definition of
 * equality permits the two mime messages to have different boundary
 * strings including different white space after the boundary string
 * but before the EOL for the boundary line.
 *
 * @param[in] cex c-style exception
 * @param[in] msg1 first MIME message
 * @param[in] msg2 second MIME message
 * @return whether the MIME messages are equal
 */
static int
main__message__is_equal(cgul_exception_t* cex,
                        message_t msg1,
                        message_t msg2)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(msg1);
    assert(msg2);

    /* Compare the size of each preamble. */
    if (msg1->preamble_size != msg2->preamble_size) {
        goto out;
    }

    /* Compare the content of each preamble. */
    if (memcmp(msg1->preamble, msg2->preamble, msg1->preamble_size)) {
        goto out;
    }

    /* Compare the sections. */
    if (!main__message__are_all_sections_equal(cex, msg1, msg2)) {
        goto out;
    }

    /* Compare the size of each epilogue. */
    if (msg1->epilogue_size != msg2->epilogue_size) {
        goto out;
    }

    /* Compare the content of each epilogue. */
    if (memcmp(msg1->epilogue, msg2->epilogue, msg1->epilogue_size)) {
        goto out;
    }

    /* Success. */
    rv = 1;

 out:
    return rv;
}

/**
 * This function produces a full, in-memory MIME message from the data
 * structure <code>msg</code>.  A pointer to the in-memory MIME
 * message is returned in <code>*ptr</code> and the length in bytes of
 * the in-memory MIME message is returned in <code>*ptr_length</code>
 * provided <code>ptr_length</code> is not <code>NULL</code>.  The
 * caller is responsible for calling <code>free()</code> on
 * <code>*ptr</code>.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] msg <code>message</code> instance
 * @param[out] ptr pointer to the full MIME message contiguous in memory
 * @param[out] ptr_length length in bytes of the full MIME message
 */
static void
main__message__to_string(cgul_exception_t* cex,
                         message_t msg,
                         char** ptr,
                         size_t* ptr_length)
{
    unsigned long int i = 0;
    section_t section = NULL;
    cgul_string_t s = NULL;
    cgul_list_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(msg);

    /* Allocate. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Append the preamble. */
    cgul_string__append_block(cex, s, msg->preamble, msg->preamble_size);
    if (*cex) {
        goto out;
    }

    /* Iterate over each multipart section. */
    n = cgul_list__get_front(cex, msg->sections);
    for (i = 0 ; n ; n = cgul_list_node__get_next(cex, n), ++i) {

        /* Carefully append the MIME boundary separator.  If the
         * preamble is empty, the leading EOL sequence in the
         * msg->boundary should be skipped. */
        if ((i == 0) && (msg->preamble_size == 0)) {
            cgul_string__append_block(cex,
                                      s,
                                      msg->boundary + msg->eol_length,
                                      msg->boundary_length - msg->eol_length);
        } else {
            cgul_string__append_block(
                cex, s, msg->boundary, msg->boundary_length);
        }
        if (*cex) {
            goto out;
        }

        /* Get the current section. */
        section = cgul_list_node__get_value(cex, n);

        /* Append the header. */
        cgul_string__append_block(
            cex, s, section->header, section->header_size);
        if (*cex) {
            goto out;
        }

        /* Append the second of the back-to-back EOL sequences that
         * separates the header from the body.  As explained in
         * main__message__new_random__create_section(), the header is
         * either empty or it has at least one character followed by
         * an EOL sequence.  If the header is empty, the first EOL
         * sequences of the back-to-back pair comes from the boundary
         * line.  If the header is not empty, the first one comes from
         * the header itself. */
        cgul_string__append(cex, s, msg->eol);
        if (*cex) {
            goto out;
        }
        

        /* Append the body. */
        cgul_string__append_block(cex, s, section->body, section->body_size);
        if (*cex) {
            goto out;
        }
    }

    /* Append the final boundary string, but only if there was at
     * least one part present in the multipart message. */
    if (i > 0) {
        cgul_string__append_block(
            cex, s, msg->final_boundary, msg->final_boundary_length);
    }

    /* Append the epilogue. */
    cgul_string__append_block(cex, s, msg->epilogue, msg->epilogue_size);
    if (*cex) {
        goto out;
    }

    /* Make sure the pre-calculated size of the MIME message matches
     * the actual size. */
    if (msg->message_size != cgul_string__get_length(cex, s)) {
        *cex = cgul_exception__new(
                   "Internal error: "
                   "pre-calculated message size = %lu, but "
                   "actual message size = %lu\n",
                   (unsigned long)msg->message_size,
                   (unsigned long)cgul_string__get_length(cex, s));
        goto out;
    }

    /* Success. */
    if (ptr_length) {
        *ptr_length = cgul_string__get_length(cex, s);
    }
    if (ptr) {
        *ptr = cgul_string__take_value(cex, s);
    }

 out:

    /* Clean up. */
    if (s) {
        cgul_string__delete(s);
    }

    return;
}

/**
 * Convert the MIME message data structure <code>msg</code> into a
 * string and write it to the output file <code>fout</code>.  If an
 * error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] msg <code>message</code> instance
 * @param[in] fout output file
 */
static void
main__write_mime_to_file(cgul_exception_t* cex,
                         message_t msg,
                         FILE* fout)
{
    size_t s_length = 0;
    char* s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(msg);
    assert(fout);

    /* Convert the MIME message into a string. */
    main__message__to_string(cex, msg, &s, &s_length);
    if (*cex) {
        goto out;
    }

    /* Write the message to fout. */
    cgul_libc__save_file(cex, fout, s, s_length);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (s) {
        free(s);
    }

    return;
}

/**
 * Convert the MIME message data structure <code>msg</code> into a
 * string and write it to the file with name <code>fname</code>.  If
 * an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] msg <code>message</code> instance
 * @param[in] fname file name
 */
static void
main__write_mime_to_fname(cgul_exception_t* cex,
                          message_t msg,
                          const char* fname)
{
    FILE* f = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(msg);
    assert(fname);

    /* Open the file. */
    f = cgul_libc__fopen(cex, fname, "wb");
    if (*cex) {
        goto out;
    }

    /* Write the MIME message. */
    main__write_mime_to_file(cex, msg, f);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (f) {
        fclose(f);
    }

    return;
}

/**
 * This is the callback invoked by the <code>cgul_mime</code> parser
 * whenever it needs another block of bytes from the MIME message.
 * The request is fulfilled by copying bytes from the MIME message in
 * memory to the parser's block <code>block</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] block parser's block to be filled
 * @param[in] block_size size of block in bytes
 * @param[in] data client data
 */
static size_t
main__produce_blocks(cgul_exception_t* cex,
                     char* block,
                     size_t block_size,
                     void* data)
{
    size_t rv = 0;
    size_t bytes_remaining = 0;
    producer_data_t pdata = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(block);
    assert(data);

    /* Recover the client data. */
    pdata = (producer_data_t)data;

    assert(pdata->index <= pdata->message_size);

    /* Determine the byte rv to return to the parser. */
    bytes_remaining = pdata->message_size - pdata->index;
    rv = CGUL_MIN(block_size, bytes_remaining);
    rv = CGUL_MIN(rv, pdata->max);

    /* Copy the bytes into the parser's block. */
    if (rv > 0) {
        memcpy(block, &pdata->message[pdata->index], rv);

        /* Advance our index. */
        pdata->index += rv;
    }

 out:
    return rv;
}

/**
 * This is a helper function for <code>main__consume_blocks()</code>
 * that saves the section data in <code>cdata->s</code> whenever a
 * transition from one section type to another is detected.
 *
 * @param[in,out] cex c-style exception
 * @param[in] cdata consumer client data
 */
static void
main__consume_blocks__save_section(cgul_exception_t* cex,
                                   consumer_data_t cdata)
{
    section_t section_new = NULL;
    section_t section_existing = NULL;
    cgul_list_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cdata);

    /* Sanity check. */
    switch (cdata->prev_section) {

    /* Invalid */
    case CGUL_MIME__SECTION_INVALID:
        break;

    /* Preamble */
    case CGUL_MIME__SECTION_PREAMBLE:
        cdata->message->preamble_size = cgul_string__get_length(cex, cdata->s);
        cdata->message->preamble = cgul_string__take_value(cex, cdata->s);
        if (*cex) {
            goto out;
        }
        break;

    /* Header */
    case CGUL_MIME__SECTION_HEADER:
        /* Create a new section. */
        section_new = main__section__new(cex);
        /* Add the header to the section. */
        section_new->header_size = cgul_string__get_length(cex, cdata->s);
        section_new->header = cgul_string__take_value(cex, cdata->s);
        /* Add the section to the list of sections. */
        cgul_list__push_back(cex, cdata->message->sections, section_new);
        if (*cex) {
            goto out;
        }
        /* The list now owns the section. */
        section_new = NULL;
        break;

    /* Body */
    case CGUL_MIME__SECTION_BODY:
        /* Recover the current section. */
        n = cgul_list__get_back(cex, cdata->message->sections);
        if (!n) {
            *cex = cgul_exception__new(
                       "main__consume_blocks__save_section: "
                       "attempt to save body when no sections are active");
            goto out;
        }
        section_existing = cgul_list_node__get_value(cex, n);
        /* Make sure a body doesn't already exist. */
        if (section_existing->body) {
            *cex = cgul_exception__new(
                       "main__consume_blocks__save_section: "
                       "body already exists for section");
            goto out;
        }
        /* Add the body to the section. */
        section_existing->body_size = cgul_string__get_length(cex, cdata->s);
        section_existing->body = cgul_string__take_value(cex, cdata->s);
        if (*cex) {
            goto out;
        }
        break;

    /* Epilogue */
    case CGUL_MIME__SECTION_EPILOGUE:
        cdata->message->epilogue_size = cgul_string__get_length(cex, cdata->s);
        cdata->message->epilogue = cgul_string__take_value(cex, cdata->s);
        if (*cex) {
            goto out;
        }
        break;

    /* EOF */
    case CGUL_MIME__SECTION_EOF:
        break;

    /* Error */
    default:
        *cex = cgul_exception__new(
                   "main__consume_blocks__save_section: "
                   "unexpected section: %d", (int)cdata->prev_section);
        goto out;
    }

 out:

    /* Clean up. */
    if (section_new) {
        main__section__delete(section_new);
    }

    return;
}

/**
 * This is the callback invoked by the <code>cgul_mime</code> parser
 * whenever it has a block of parsed data ready.  This function than
 * uses the incoming blocks to recreate the MIME message.
 *
 * @param[in] cex c-style exception
 * @param[in] part current part in the multipart MIME message
 * @param[in] section current section in the part
 * @param[in] block parser's block to be filled
 * @param[in] block_size size of block in bytes
 * @param[in] data client data
 * @return whether to continue
*/
static int
main__consume_blocks(cgul_exception_t* cex,
                     unsigned long int part,
                     cgul_mime__section_t section,
                     char* block,
                     size_t block_size,
                     void* data)
{
    int rv = 0;
    consumer_data_t cdata = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(block);
    assert(data);

    /* Recover the client data. */
    cdata = (consumer_data_t)data;

    /* If we are starting a new section, save the contents of cdata->s
     * for the old section. */
    if (cdata->prev_section != section) {
        main__consume_blocks__save_section(cex, cdata);
        if (*cex) {
            goto out;
        }
    }

    /* Append the current block to cdata->s. */
    cgul_string__append_block(cex, cdata->s, block, block_size);
    if (*cex) {
        goto out;
    }

    /* Remember the section type. */
    cdata->prev_section = section;

    /* If this is EOF, update cdata->message->message_size to match
     * the changes we've made. */
    if (section == CGUL_MIME__SECTION_EOF) {
        main__message__calculate_size(cex, cdata->message);
    }

    /* Continue. */
    rv = 1;

 out:
    return rv;
}

/**
 * Common code needed by both <code>main__test_randomly()</code> and
 * <code>main__parse_file()</code> to parses a MIME message in memory.
 * This function sets up all the callback data need for the producer
 * and consumer callbacks and invokes <code>cgul_mime__parse()</code>.
 * The consumer for <code>cgul_mime__parse()</code> attempts to
 * reconstruct the original <code>message</code> data structure that
 * was used to create <code>msg</code>.  The reconstructed data
 * structure is then returned to the caller.  The caller is
 * responsible for calling <code>main__message__delete()</code> on the
 * pointer returned.  If an error occurs, an exception is thrown. <p>
 *
 * The caller also needs to pass in the boundary string
 * <code>boundary</code>, its length <code>boundary_length</code>, the
 * EOL sequence <code>eol</code>, and its length
 * <code>eol_length</code>.  Both the boundary string and the EOL
 * sequence are required parameters for
 * <code>cgul_mime__parse()</code>; however, the boundary string is
 * not passed directly to <code>cgul_mime__parse()</code> because (in
 * this program) it has a leading EOL sequence followed by a "--"
 * sequence.  Both of these must be removed before the boundary string
 * can be passed into <code>cgul_mime__parse()</code>.  Furthermore,
 * the boundary strings in this program also potentially have white
 * space before a trailing EOL sequence, and both the white space and
 * the trailing EOL sequence are also not allowed in the boundary
 * strings passed to <code>cgul_mime__parse()</code>. <p>
 *
 * In addition, the caller also needs to pass in the final boundary
 * string <code>final_boundary</code> and its length
 * <code>final_boundary_length</code>.  These are needed so that the
 * consumer callback for <code>cgul_mime__parse()</code> can attempt
 * to create an exact duplicate of the <code>message_t</code> data
 * structure that was used to create <code>msg</code>. <p>
 *
 * The PRNG <code>prng</code> is used to randomly select the largest
 * number of bytes that the producer is allowed to return to
 * <code>cgul_mime_parse()</code> when it asks the producer for more
 * data. <p>
 * 
 *
 * @param[in,out] cex c-style exception
 * @param[in] prng PRNG
 * @param[in] msg MIME message in memory
 * @param[in] msg_length length in bytes of the MIME message
 * @param[in] boundary boundary string
 * @param[in] boundary_length length in bytes of the boundary string
 * @param[in] final_boundary final boundary string
 * @param[in] final_boundary_length length in bytes of the final boundary string
 * @param[in] eol EOL sequence
 * @param[in] eol_length length in bytes of the EOL sequence
 * @return MIME message data structure generated by the conumser
 * callback for the <code>cgul_mime__parse()</code>
 */
static message_t
main__parse(cgul_exception_t* cex,
            cgul_random_xorshift32_t prng,
            const char* msg,
            size_t msg_length,
            const char* boundary,
            size_t boundary_length,
            const char* final_boundary,
            size_t final_boundary_length,
            const char* eol,
            size_t eol_length)
{
    message_t rv = NULL;
    /* Boundary string in format required by cgul_mime__parse(). */
    cgul_string_t b = NULL;
    struct producer_data pdata = {0};
    struct consumer_data cdata = {0};

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(msg);

    /* Allocate. */
    b = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Initialize the the client data for the producer callback by
     * converting the MIME message data structure msg1 to a real,
     * contiguous MIME message that is present in memory, and then
     * initializing the remaining parameters. */
    pdata.message = msg;
    pdata.message_size = msg_length;
    pdata.index = 0;
    pdata.max = cgul_random_xorshift32__next_in_unit(cex, prng)
                  * MAX_PRODUCER_READ + 1;
    if (*cex) {
        goto out;
    }

    /* Initialize the client data for the producer callback. */
    cdata.message = main__message__new(cex);
    cdata.s = cgul_string__new(cex);
    cdata.prev_section = CGUL_MIME__SECTION_INVALID;
    cdata.message->eol = cgul_libc__strdup(cex, eol);
    cdata.message->eol_length = eol_length;
    cdata.message->boundary = cgul_libc__strdup(cex, boundary);
    cdata.message->boundary_length = boundary_length;
    cdata.message->final_boundary = cgul_libc__strdup(cex, final_boundary);
    cdata.message->final_boundary_length = final_boundary_length;
    if (*cex) {
        goto out;
    }

    /* Create "b" that is a boundary string that is suitable for use
     * when calling cgul_mime__parse().  Note that cgul_mime__parse()
     * expects the boundary string to not have a leading EOL sequence,
     * to not have a leading "--" sequence, and to not have trailing
     * whitespace including the trailing EOL sequence (because this is
     * how the boundary string appears in the MIME Content-Type
     * header).  For convenience the boundary strings in this program
     * potentially have all of these extras.  So here we generate a
     * boundary string in the format required by cgul_mime__parse(). */
    if (boundary_length < eol_length + 2) {
        *cex = cgul_exception__new("main__parse: "
                                   "boundary string too short");
        goto out;
    }
    cgul_string__set_value(cex, b, boundary + eol_length + 2);
    cgul_string__trim_right(cex, b);
    if (*cex) {
        goto out;
    }

    /* Parse the MIME message in memory.  Note that we are using the
     * low-level cgul_mime__parse() instead of
     * cgul_mime__parse_memory() because we want to thoroughly test
     * the parser, and cgul_mime__parse_memory() would be too easy on
     * the parser because the parser wouldn't be required to cobble
     * together blocks like it has to do when reading a MIME message
     * from file. */
#if 1
    cgul_mime__parse(cex,
                     cgul_string__get_value(cex, b),
                     eol,
                     main__produce_blocks,
                     &pdata,
                     main__consume_blocks,
                     &cdata);
#else
    cgul_mime__parse_memory(cex,
                            msg,
                            msg_length,
                            cgul_string__get_value(cex, b),
                            eol,
                            main__consume_blocks,
                            &cdata);
#endif
    if (*cex) {
        goto out;
    }

    /* Take ownership of cdata.message so we can return it to the user. */
    rv = cdata.message;
    cdata.message = NULL;

 out:

    /* Clean up. */
    if (cdata.s) {
        cgul_string__delete(cdata.s);
    }
    if (cdata.message) {
        main__message__delete(cdata.message);
    }
    if (b) {
        cgul_string__delete(b);
    }

    return rv;
}

/**
 * Test <code>cgul_mime</code> by generating a random MIME message,
 * calling <code>cgul_mime__parse()</code> to parse the MIME message
 * building up a new MIME message, and finally comparing the new MIME
 * message to the MIME message that was parsed to make sure they are
 * the same.
 *
 * @param[in,out] cex c-style exception
 * @param[in] prng PRNG
 * @return whether the test completed successfully
 */
static int
main__test_randomly(cgul_exception_t* cex,
                    cgul_random_xorshift32_t prng)
{
    int rv = 0;
    char* ptr = NULL;
    size_t ptr_length = 0;
    message_t msg1 = NULL;
    message_t msg2 = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(prng);

    /* Create a new random MIME message data structure. */
    msg1 = main__message__new_random(cex, prng, EOL_SEQUENCE);
    if (*cex) {
        goto out;
    }

    /* Convert the msg1 data structure into a MIME message that is
     * contiguous in memory.  The MIME message will be in "ptr" when
     * this call returns. */
    main__message__to_string(cex, msg1, &ptr, &ptr_length);
    if (*cex) {
        goto out;
    }

    /* Parse the MIME message getting back what is hopefully a
     * reconstructed copy of the original MIME message. */
    msg2 = main__parse(cex,
                       prng,
                       ptr,
                       ptr_length,
                       msg1->boundary,
                       msg1->boundary_length,
                       msg1->final_boundary,
                       msg1->final_boundary_length,
                       msg1->eol,
                       msg1->eol_length);
    if (*cex) {
        goto out;
    }

    /* Verify that the message generated during parsing matches the
     * original message. */
    if (!main__message__is_equal(cex, msg1, msg2)) {
        const char* fname1 = "main_mime_test.master";
        const char* fname2 = "main_mime_test.parsed";

        /* Write the "master" and "parsed" MIME messages to file. */
        main__write_mime_to_fname(cex, msg1, fname1);
        main__write_mime_to_fname(cex, msg2, fname2);
        if (*cex) {
            goto out;
        }

        /* Throw an exception. */
        *cex = cgul_exception__new("main__test_randomly: "
                                   "Failure!  See files %s and %s.",
                                   fname1, fname2);
        goto out;
    }

    /* Success. */
    rv = 1;

 out:

    /* Clean up. */
    if (msg2) {
        main__message__delete(msg2);
    }
    if (ptr) {
        free(ptr);
    }
    if (msg1) {
        main__message__delete(msg1);
    }

    return rv;
}

/**
 * Load the MIME message from the file <code>file</code> into memory
 * and attempt to parse it using the boundary string
 * <code>boundary</code>.  If an error occurs, an exception is
 * thrown. <p>
 *
 * Note that this is one of the few places in the entire program where
 * the boundary string does not have a leading EOL sequence followed
 * by the "--" sequence.  This is because <code>boundary</code> is
 * passed in from the command-line, and this is the convention used
 * for the other command-line programs. <p>
 *
 * The EOL sequence <code>eol</code> needs to be passed in so that we
 * can convert <code>boundary</code> to match all the other boundary
 * strings used in this program. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] prng PRNG
 * @param[in] fin input file
 * @param[in] boundary boundary string
 * @param[in] eol EOL sequence
 */
static void
main__parse_file(cgul_exception_t* cex,
                 cgul_random_xorshift32_t prng,
                 FILE* fin,
                 const char* boundary,
                 const char* eol)
{
    size_t block_length = 0;
    char* block = NULL;
    message_t msg = NULL;
    const char* fname_out = NULL;
    cgul_string_t b = NULL;   /* boundary string */
    cgul_string_t fb = NULL;  /* final boundary string */

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate. */
    b = cgul_string__new(cex);
    fb = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    assert(fin);
    assert(boundary);
    assert(eol);

    /* Load the file into memory. */
    cgul_libc__load_file(cex, fin, &block, &block_length);
    if (*cex) {
        goto out;
    }

    /* "boundary" is passed in by the user from the command line and
     * does not have a leading EOL sequence nor a leading "--" unlike
     * all the other boundary strings in this program.  So here we
     * generate a boundary string that works with the rest of the
     * program. */
    cgul_string__set_value(cex, b, eol);
    cgul_string__append(cex, b, "--");
    cgul_string__append(cex, b, boundary);
    cgul_string__append(cex, b, eol);
    if (*cex) {
        goto out;
    }

    /* Do the same for the final boundary string. */
    cgul_string__set_value(cex, fb, eol);
    cgul_string__append(cex, fb, "--");
    cgul_string__append(cex, fb, boundary);
    cgul_string__append(cex, fb, "--");
    cgul_string__append(cex, fb, eol);
    if (*cex) {
        goto out;
    }

    /* Parse the MIME message. */
    msg = main__parse(cex,
                      prng,
                      block,
                      block_length,
                      cgul_string__get_value(cex, b),
                      cgul_string__get_length(cex, b),
                      cgul_string__get_value(cex, fb),
                      cgul_string__get_length(cex, fb),
                      eol,
                      strlen(eol));
    if (*cex) {
        goto out;
    }

    /* Write the result to file. */
    fname_out = "main_mime_test.parsed.2";
    main__write_mime_to_fname(cex, msg, fname_out);
    if (*cex) {
        goto out;
    }

    printf("Result written to %s.\n", fname_out);

 out:

    /* Clean up. */
    if (msg) {
        main__message__delete(msg);
    }
    if (block) {
        free(block);
    }
    if (fb) {
        cgul_string__delete(fb);
    }
    if (b) {
        cgul_string__delete(b);
    }

    return;
}

/**
 * Load the MIME message in the file with name <code>fname</code> into
 * memory and attempt to parse it using the boundary string
 * <code>boundary</code>.  If an error occurs, an exception is
 * thrown. <p>
 *
 * Note that this is one of the few places in the entire program where
 * the boundary string does not have a leading EOL sequence followed
 * by the "--" sequence.  This is because <code>boundary</code> is
 * passed in from the command-line, and this is the convention used
 * for the other command-line programs. <p>
 *
 * The EOL sequence <code>eol</code> needs to be passed in so that we
 * can convert <code>boundary</code> to match all the other boundary
 * strings used in this program. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] prng PRNG
 * @param[in] fname input file name
 * @param[in] boundary boundary string
 * @param[in] eol EOL sequence
 */
static void
main__parse_fname(cgul_exception_t* cex,
                  cgul_random_xorshift32_t prng,
                  const char* fname,
                  const char* boundary,
                  const char* eol)
{
    FILE* fin = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(fname);

    /* Open the file. */
    fin = cgul_libc__fopen(cex, fname, "rb");
    if (*cex) {
        goto out;
    }

    /* Parse the file. */
    main__parse_file(cex, prng, fin, boundary, eol);
    if (*cex) {
        goto out;
    }

 out:

    if (fin) {
        fclose(fin);
    }

    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    int test_rv = 0;
    unsigned long int count = 0;
    cgul_random_xorshift32_t prng = NULL;
    cgul_exception_t local = NULL;

    /* Allocate. */
    prng = cgul_random_xorshift32__new(&local);
    if (local) {
        goto out;
    }

    /* If a test fails, this program writes the MIME message that
     * caused the failure to file.  You can then re-parse (but not
     * re-test (yet)) by specifying the name of the file as the first
     * command-line parameters. */
    if (argc == 3) {
        /* NOTE: The boundary string here should not have a leading
         * EOL sequence nor a leading "--" sequence.  This is not true
         * for the rest of the program however. */
        main__parse_fname(&local,
                          prng,
                          argv[1],   /* File name. */
                          argv[2],   /* Boundary string. */
                          EOL_SEQUENCE);
        if (local) {
            goto out;
        }

    } else if (argc != 1) {
        fprintf(stderr, "Usage: main_mime_test\n");
        fprintf(stderr, "Usage: main_mime_test <file> <boundary>\n");

    } else {

        /* Test forever. */
        for (;;) {

            /* Test. */
            test_rv = main__test_randomly(&local, prng);
            if (local) {
                goto out;
            }

            /* Progress. */
            if (test_rv) {
                if ((++count & 0xf) == 0) {
                    printf("count = %#0lx\n", count);
                }
            }
        }
    }

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (prng) {
        cgul_random_xorshift32__delete(prng);
    }

    return rv;
}
