#include "cgul.h"

/**
 * This method prints the key/value pair for <code>key</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] stanza stanza
 * @param[in] key key
 */
static void
print_key_value_pair(cgul_exception_t* cex,
                     cgul_stanza_t stanza,
                     const char* key)
{
    const char* value = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);
    assert(key);

    /* Lookup the value. */
    value = cgul_stanza__lookup(cex, stanza, key);
    if (*cex) {
        goto out;
    }

    /* Carefully print the key/value pair. */
    cgul_libc__printf(cex, "%s ==> %s\n", key, value ? value : "<invalid key>");
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * Print each stanza from the stanza object <code>stanza</code>.  If
 * an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 */
static void
print_stanzas(cgul_exception_t* cex,
              cgul_stanza_t stanza)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);

    /* Iterate over each stanza. */
    while (cgul_stanza__load_next(cex, stanza)) {
        print_key_value_pair(cex, stanza, "foo");
        print_key_value_pair(cex, stanza, "bar");
        print_key_value_pair(cex, stanza, "baz");
        print_key_value_pair(cex, stanza, "hmm");
        cgul_libc__printf(cex, "====\n");
        if (*cex) {
            goto out;
        }
    }

 out:
    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    FILE* f = NULL;
    cgul_crlf_file_t crlf_file = NULL;
    cgul_stanza_t stanza = NULL;
    cgul_exception_t local = NULL;

    /* Allocate. */
    crlf_file = cgul_crlf_file__new(&local);
    stanza = cgul_stanza__new(&local);
    if (local) {
        goto out;
    }

    /* Open the stanza file and print its stanzas. */
    cgul_stanza__open_fname(
        &local, stanza, "main_stanza_input.txt", 1 /* is_block_buffered */);
    print_stanzas(&local, stanza);
    if (local) {
        goto out;
    }
    puts("****************************************************************");

    /* Re-open the stanza file and print its stanzas. */
    cgul_stanza__open_fname(
        &local, stanza, "main_stanza_input.txt", 1 /* is_block_buffered */);
    print_stanzas(&local, stanza);
    if (local) {
        goto out;
    }
    puts("****************************************************************");

    /* Re-open the stanza file and print its stanzas. */
    f = cgul_libc__fopen(&local, "main_stanza_input.txt", "r");
    cgul_stanza__open_file(&local, stanza, f, 1 /* is_block_buffered */);
    print_stanzas(&local, stanza);
    if (local) {
        goto out;
    }
    puts("****************************************************************");

    /* Re-open the stanza file and print its stanzas. */
    cgul_crlf_file__open_fname(&local,
                               crlf_file,
                               "main_stanza_input.txt",
                               1 /* is_block_buffered */);
    cgul_stanza__open_crlf_file(&local, stanza, crlf_file);
    print_stanzas(&local, stanza);
    if (local) {
        goto out;
    }
    puts("****************************************************************");

 out:

    /* Handle exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (f) {
        fclose(f);
    }
    if (stanza) {
        cgul_stanza__delete(stanza);
    }
    if (crlf_file) {
        cgul_crlf_file__delete(crlf_file);
    }

    return rv;
}
