#include "cgul.h"

/**
 * Application logging facility.
 */
#if 1
#  define APPLOG(...)                             \
    do {                                          \
        printf(__VA_ARGS__);                      \
    } while (0)
#else
#  define APPLOG(...)
#endif

/**
 * The number of test iterations.
 */
#define TOTAL_TEST_ITERATIONS (100)

/**
 * The minimum number of keys to generate.
 */
#define KEY_COUNT_MIN (0)

/**
 * The maximum number of keys to generate.
 */
#define KEY_COUNT_MAX (1000)

/**
 * The minimum string length of the keys to generate.
 */
#define KEY_LENGTH_MIN (0)

/**
 * The maximum string length of the keys to generate.
 */
#define KEY_LENGTH_MAX (31)

/**
 * The number of "find" tests to execute.
 */
#define FIND_TEST_COUNT (10)

/**
 * The number of "longest prefix of" tests to execute.
 */
#define LONGEST_PREFIX_OF_TEST_COUNT (5)

/**
 * The minimum number of prefix tests per removal.
 */
#define PREFIX_TEST_COUNT_MIN (0)

/**
 * The maximum number of prefix tests per removal.
 */
#define PREFIX_TEST_COUNT_MAX (100)

/**
 * The minimum string length of the prefixes to generate.
 */
#define PREFIX_LENGTH_MIN (0)

/**
 * The maximum string length of the prefixes to generate.
 */
#if 0
#  define PREFIX_LENGTH_MAX (KEY_LENGTH_MAX + (KEY_LENGTH_MAX >> 2) + 4)
#else
#  define PREFIX_LENGTH_MAX (5)
#endif

/**
 * The minimum number of floor tests to execute per removal.
 */
#define FLOOR_TEST_COUNT_MIN (0)

/**
 * The maximum number of floor tests to execute per removal.
 */
#define FLOOR_TEST_COUNT_MAX (100)

/**
 * The minimum number of ceiling tests to execute per removal.
 */
#define CEILING_TEST_COUNT_MIN (0)

/**
 * The maximum number of ceiling tests to execute per removal.
 */
#define CEILING_TEST_COUNT_MAX (FLOOR_TEST_COUNT_MAX)

/**
 * Return less than zero, zero, or greater than zero if
 * <code>key1</code> is less than, equal to, or greater than
 * <code>key2</code>.
 *
 * @param[in] key1 first key
 * @param[in] key2 second key
 * @return lexicographical comparison of the keys
 * @see main__compare_pointers_to_keys
 */
static int
main__compare_keys(const char* key1,
                   const char* key2)
{
    int rv = 0;
    size_t i = 0;

    /* Compare the parts of the keys where they both have characters. */
    for (i = 0 ; key1[i] && key2[i] ; ++i) {
        if (key1[i] < key2[i]) {
            rv = -1;
            goto out;
        } else if (key1[i] > key2[i]) {
            rv = 1;
            goto out;
        }
    }

    /* If you reach here, at least one key is a prefix of the other.
     * The lexicographical comparison of the keys now depends on which
     * is longer. */
    if ((key1[i] == 0) && (key2[i] == 0)) {
        rv = 0;
    } else if (key1[i] == 0) {
        rv = -1;
    } else {
        rv = 1;
    }

 out:
    return rv;
}

/**
 * Return less than zero, zero, or greater than zero if
 * <code>*key1</code> is less than, equal to, or greater than
 * <code>*key2</code>.
 *
 * @param[in] key1 first key
 * @param[in] key2 second key
 * @return lexicographical comparison of the keys
 * @see main__compare_pointers_to_keys
 */
static int
main__compare_pointers_to_keys(const char** key1,
                               const char** key2)
{
    return main__compare_keys(*key1, *key2);
}

/**
 * Return a string of length <code>length</code> having randomly
 * selected characters in the range <code>['a', 'z']</code>.  The PRNG
 * <code>prng</code> is used for randomly selecting the string's
 * content.  The client is responsible for calling <code>free()</code>
 * on the pointer returned.  If an error occurs, <code>NULL</code> is
 * returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] prng prng
 * @param[in] length length of string to return
 * @return randomly generated string
 */
static char*
main__generate_string_fixed_length(cgul_exception_t* cex,
                                   cgul_random_lfg32_t prng,
                                   size_t length)
{
    cgul_uint32_t n = 0;
    size_t i = 0;
    char* rv = NULL;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Generate. */
    for (i = 0 ; i < length ; ++i) {
        n = cgul_random_lfg32__next_in_range(cex, prng, 'z' - 'a' + 1);
        cgul_string__append_char(cex, tmp, 'a' + (char)n);
        if (*cex) {
            goto out;
        }
    }

    /* Success. */
    rv = cgul_string__take_value(cex, tmp);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (*cex && rv) {
        free(rv);
        rv = NULL;
    }
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

/**
 * Return a string having a length randomly selected from the range
 * <code>[length_min, length_max]</code> having randomly selected
 * characters in the range <code>['a', 'z']</code>.  The PRNG
 * <code>prng</code> is used for randomly selecting the string's
 * length and content.  The client is responsible for calling
 * <code>free()</code> on the pointer returned.  If an error occurs,
 * <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] prng prng
 * @param[in] length length of string to return
 * @return randomly generated string
 */
static char*
main__generate_string_random_length(cgul_exception_t* cex,
                                    cgul_random_lfg32_t prng,
                                    cgul_uint32_t length_min,
                                    cgul_uint32_t length_max)
{
    char* rv = NULL;
    cgul_uint32_t n = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Randomly select the length of the string to generate. */
    n = cgul_random_lfg32__next_in_range(
            cex, prng, length_max - length_min + 1);
    if (*cex) {
        goto out;
    }

    /* Generate the string. */
    rv = main__generate_string_fixed_length(cex, prng, length_min + n);
    if (*cex) {
        goto out;
    }

 out:
    return rv;
}

/**
 * Remove the first occurrence of the key <code>key</code> from the
 * sorted vector <code>v</code>.  The vector should be sorted
 * according to the comparison function <code>f</code>.  If the key is
 * found, it is removed, and <code>1</code> is returned; otherwise,
 * <code>0</code> is returned.
 *
 * @param[in] cex c-style exception
 * @param[in] v sorted vector
 * @param[in] key key to remove
 * @param[in] f key comparison function
 * @return whether the key could be removed
 */
static int
main__remove_key_from_vector(cgul_exception_t* cex,
                             cgul_vector_t v,
                             const char* key,
                             cgul_binary_search__compare_t f)
{
    int rv = 0;
    char** element = NULL;
    char** arr = NULL;
    unsigned long int arr_size = 0;
    unsigned long int index = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Do not assert "key" which allowed to be NULL. */
    assert(v);
    assert(f);

    /* Unpack the array. */
    arr = (char**)cgul_vector__get_array(cex, v);
    arr_size = cgul_vector__get_size(cex, v);

    /* Find the pointer to the element to remove. */
    element = (char**)cgul_binary_search(
        cex, &key, (const void**)arr, arr_size, sizeof(*arr), f);
    if (!element) {
        goto out;
    }

    /* Calculate the index for the element to remove. */
    index = (unsigned long int)(element - arr);

    /* Remove the element from the vector. */
    cgul_vector__erase(cex, v, index);
    if (*cex) {
        goto out;
    }

    /* Success. */
    rv = 1;

 out:
    return rv;
}

/**
 * Insert between <code>key_count_min</code> (inclusive) and
 * <code>key_count_max</code> (inclusive) randomly generated keys into
 * the trie and, for verification purposes, into the <code>keys</code>
 * vector.  The keys will be generated with random lengths between
 * <code>key_length_min</code> (inclusive) and
 * <code>key_length_max</code> (inclusive).  The client is responsible
 * for calling <code>free()</code> on all the keys.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] keys vector of keys inserted into the trie
 * @param[in] key_count_min minimum number of keys to insert
 * @param[in] key_count_max maximum number of keys to insert
 * @param[in] key_length_min minimum length of keys
 * @param[in] key_length_max maximum length of keys
 * @param[in] prng PRNG
 */
static void
main__insert_keys(cgul_exception_t* cex,
                  cgul_trie_t trie,
                  cgul_vector_t keys,
                  cgul_uint32_t key_count_min,
                  cgul_uint32_t key_count_max,
                  cgul_uint32_t key_length_min,
                  cgul_uint32_t key_length_max,
                  cgul_random_lfg32_t prng)
{
    int insert_rv = 0;
    cgul_uint32_t key_count = 0;
    unsigned long int i = 0;
    char* key = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(keys);
    assert(prng);

    /* Determine the number of keys to generate. */
    key_count = cgul_random_lfg32__next_in_range(
                    cex, prng, key_count_max - key_count_min + 1);
    if (*cex) {
        goto out;
    }
    key_count += key_count_min;

    /* Insert the randomly generated keys into the trie and for
     * verification purposes into the "keys" vector. */
    APPLOG("    key count: %lu\n", (unsigned long)key_count);
    for (i = 0 ; i < key_count ; ++i) {

        /* Generate the key. */
        key = main__generate_string_random_length(
                      cex, prng, key_length_min, key_length_max);
        if (*cex) {
            goto out;
        }

        /* Add the key to the keys vector.  This needs to be done
         * before adding the key to the trie because we need to set
         * "key" to NULL after ownership is transfered to the trie
         * which would prevent us from adding the key to the vector. */
        cgul_vector__push_back(cex, keys, key);
        if (*cex) {
            goto out;
        }

        /* Add the key to the trie. */
        insert_rv = cgul_trie__insert(cex, trie, key, NULL);
        if (*cex) {
            goto out;
        }
        if (!insert_rv) {
            /* The key is already in the trie.  Remove it from the
             * vector, and free it. */
            cgul_vector__pop_back(cex, keys);
            free(key);
            key = NULL;
        } else {
            /* The trie now owns the key. */
            key = NULL;
        }
    }

 out:

    /* Clean up. */
    if (key) {
        if (keys) {
            /* Make sure the "keys" vector does not hold a reference
             * to the key that is about to be deleted. */
            cgul_exception_t local = NULL;
            if (key == cgul_vector__get_back(&local, keys)) {
                cgul_vector__pop_back(&local, keys);
            }
        }
        free(key);
    }

    return;
}

/**
 * Verify that the trie <code>trie</code> is sorted by using forward
 * iteration to compare the keys in the trie in sort order with the
 * sorted <code>keys</code> vector.  If the trie is sorted
 * <code>1</code> is returned; otherwise, <code>0</code> is returned.
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] keys sorted vector of keys
 * @param[in] start_index starting index (inclusive)
 * @param[in] stop_index stopping index (exclusive)
 * @return whether the trie is sorted
 */
static int
main__verify_trie_is_sorted_using_forward_iteration(
    cgul_exception_t* cex,
    cgul_trie_t trie,
    cgul_vector_t keys,
    unsigned long int starting_index,
    unsigned long int stopping_index)
{
    int rv = 0;
    unsigned long int i = 0;
    const char* key1 = NULL;
    const char* key2 = NULL;
    cgul_trie_node_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(keys);
    assert(stopping_index <= cgul_vector__get_size(cex, keys));

    /* Make sure the sizes match. */
    if ((stopping_index - starting_index) != cgul_trie__get_size(cex, trie)) {
        goto out;
    }

    /* Compare each entry in "trie" and "keys". */
    curr = cgul_trie__get_front(cex, trie);
    for (i = starting_index;
         (i < stopping_index) && curr;
         ++i, curr = cgul_trie_node__get_next(cex, curr))
    {
        key1 = cgul_trie_node__get_key(cex, curr);
        key2 = cgul_vector__at(cex, keys, i);
        if (main__compare_keys(key1, key2) != 0) {
            goto out;
        }
    }

    /* Success. */
    rv = 1;

 out:
    return rv;
}

/**
 * Verify that the trie <code>trie</code> is sorted by using reverse
 * iteration to compare the keys in the trie in sort order with the
 * sorted <code>keys</code> vector.  If the trie is sorted
 * <code>1</code> is returned; otherwise, <code>0</code> is returned.
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] keys sorted vector of keys
 * @param[in] start_index starting index (inclusive)
 * @param[in] stop_index stopping index (exclusive)
 * @return whether the trie is sorted
 */
static int
main__verify_trie_is_sorted_using_reverse_iteration(
    cgul_exception_t* cex,
    cgul_trie_t trie,
    cgul_vector_t keys,
    unsigned long int starting_index,
    unsigned long int stopping_index)
{
    int rv = 0;
    unsigned long int i = 0;
    const char* key1 = NULL;
    const char* key2 = NULL;
    cgul_trie_node_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(keys);
    assert(stopping_index <= cgul_vector__get_size(cex, keys));

    /* Make sure the sizes match. */
    if ((stopping_index - starting_index) != cgul_trie__get_size(cex, trie)) {
        goto out;
    }

    /* Compare each entry in "trie" and "keys". */
    curr = cgul_trie__get_back(cex, trie);
    for (i = stopping_index;
         (i > 0) && curr;
         --i, curr = cgul_trie_node__get_prev(cex, curr))
    {
        key1 = cgul_trie_node__get_key(cex, curr);
        key2 = cgul_vector__at(cex, keys, i - 1);
        if (main__compare_keys(key1, key2) != 0) {
            goto out;
        }
    }

    /* Success. */
    rv = 1;

 out:
    return rv;
}

/**
 * This method will attempt to look up each key in the trie as
 * determined by the sorted list of keys stored in the
 * <code>keys</code> vector.  It then attempts to look up randomly
 * generated keys to verify that lookup failures are also reported
 * correctly.  The keys will be generated with random lengths between
 * <code>key_length_min</code> (inclusive) and
 * <code>key_length_max</code> (inclusive).  The PRNG
 * <code>prng</code> is used to randomly make the selections.  If an
 * error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] keys sorted list of all the keys currently in the trie
 * @param[in] key_length_min minimum length of keys
 * @param[in] key_length_max maximum length of keys
 * @param[in] test_count number of test iterations to execute
 * @param[in] prng PRNG
 */
static void
main__test_find(cgul_exception_t* cex,
                cgul_trie_t trie,
                cgul_vector_t keys,
                cgul_uint32_t key_length_min,
                cgul_uint32_t key_length_max,
                cgul_uint32_t test_count,
                cgul_random_lfg32_t prng)
{
    unsigned long int i = 0;
    unsigned long int i_max = 0;
    char* search_key = NULL;
    const char** keys_array = NULL;
    void* bsearch_rv = NULL;
    cgul_trie_node_t find_rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(keys);

    /* Sanity check. */
    if (test_count == 0) {
        goto out;
    }

    /* Get the "keys" array. */
    keys_array = (const char**)cgul_vector__get_array(cex, keys);

    /* Test to make sure we can find every key in the trie. */
    i_max = cgul_vector__get_size(cex, keys);
    for (i = 0 ; i < i_max ; ++i) {
        if (!cgul_trie__find(cex, trie, keys_array[i])) {
            *cex = cgul_exception__new(
                "main__test_find: Unable to find key: %s", keys_array[i]);
            goto out;
        }
    }

    /* Test for look up misses.  It is not guaranteed that randomly
     * generated keys are not in the trie.  So, if we get a hit, we
     * have to check the "keys" array to make sure it is not there. */
    for (i = 0 ; i < test_count ; ++i) {

        /* Generate the lookup key. */
        search_key = main__generate_string_random_length(
                         cex, prng, key_length_min, key_length_max);
        if (*cex) {
            goto out;
        }

        /* Look up the key in the trie. */
        find_rv = cgul_trie__find(cex, trie, search_key);

        /* Look up the key in the sorted "keys" vector. */
        bsearch_rv = cgul_binary_search_pointers(
            cex,
            search_key,
            (const void**)keys_array,
            i_max,
            (cgul_binary_search__compare_t)main__compare_keys);

        /* Check of the two lookups disagree. */
        if ((find_rv != NULL) != (bsearch_rv != NULL)) {
            *cex = cgul_exception__new(
                "main__test_find: lookup failure for string: %s", search_key);
            goto out;
        }

        /* Per-iteration clean up. */
        free(search_key);
        search_key = NULL;
    }

 out:

    /* Clean up. */
    if (search_key) {
        free(search_key);
    }

    return;
}

/**
 * This method will attempt to look up each array-style index in the
 * trie and compare the key that is found with the sorted list of keys
 * stored in the <code>keys</code> vector.  If an error occurs, an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] keys sorted list of all the keys currently in the trie
 */
static void
main__test_find_at(cgul_exception_t* cex,
                   cgul_trie_t trie,
                   cgul_vector_t keys)
{
    unsigned long int i = 0;
    const char* key = NULL;
    char** arr = NULL;
    unsigned long int arr_size = 0;
    cgul_trie_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(keys);

    /* Unpack the vector of keys. */
    arr = (char**)cgul_vector__get_array(cex, keys);
    arr_size = cgul_vector__get_size(cex, keys);

    /* Look up each key by index and compare with the "keys" array. */
    for (i = 0 ; i < arr_size ; ++i) {

        /* Find the node for key at index. */
        node = cgul_trie__find_at(cex, trie, i);
        if (*cex) {
            goto out;
        }
        if (!node) {
            *cex = cgul_exception__new(
                "main__test_find_at: "
                "Unable to find node for key at index: %lu",
                i);
            goto out;
        }

        /* Compare the found key with the expected key. */
        key = cgul_trie_node__get_key(cex, node);
        if (main__compare_keys(key, arr[i]) != 0) {
            *cex = cgul_exception__new(
                "main__test_find_at: "
                "Incorrect key for index %lu: "
                "(expected, actual) = (%s, %s)",
                i, arr[i], key);
            goto out;
        }

    }

    /* Do a lookup with an index that is out of bounds. */
    node = cgul_trie__find_at(cex, trie, arr_size);
    if (node) {
        *cex = cgul_exception__new(
            "main__test_find_at: "
            "Unexpectedly found nod for key at out-of-bounds index: %lu",
            arr_size);
        goto out;
    }

 out:
    return;
}

/**
 * Append the keys in <code>keys</code> that match the prefix
 * <code>prefix</code> to the vector of matches <code>matches</code>.
 * The valid range of indexes for the <code>keys</code> vector are
 * from <code>start_index</code> (inclusive) to
 * <code>stop_index</code> (exclusive).  If an error occurs, an
 * exception is thrown.
 *
 * @param[in,out] c-style exception
 * @param[in] keys keys against which to match
 * @param[in] start_index valid starting index for <code>keys</code> (inclusive)
 * @param[in] stop_index valid stopping index for <code>keys</code> (exclusive)
 * @param[in] prefix prefix to match against
 * @param[out] matches keys that match the prefix
 * @return list of keys that match prefix
 */
static void
main__append_prefix_matches(cgul_exception_t* cex,
                            cgul_vector_t keys,
                            unsigned long int start_index,
                            unsigned long int stop_index,
                            const char* prefix,
                            cgul_vector_t matches)
{
    unsigned long int i = 0;
    size_t prefix_length = 0;
    const char** keys_array = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(keys);
    assert(prefix);
    assert(stop_index <= cgul_vector__get_size(cex, keys));

    /* Calculate the length of the prefix. */
    prefix_length = strlen(prefix);

    /* Get the array of keys. */
    keys_array = (const char**)cgul_vector__get_array(cex, keys);

    /* Compare each key and prefix and append matches to "matches". */
    for (i = start_index ; i < stop_index ; ++i) {
        if (strncmp(prefix, keys_array[i], prefix_length) == 0) {
            cgul_vector__push_back(cex, matches, keys_array[i]);
            if (*cex) {
                goto out;
            }
        }
    }

 out:
    return;
}

/**
 * Randomly execute between <code>test_count_min</code> (inclusive)
 * and <code>test_count_max</code> (inclusive) tests that find all the
 * keys in the trie <code>trie</code> that have a randomly generated
 * prefix.  The prefixes will be generated with random lengths between
 * <code>prefix_length_min</code> (inclusive) and
 * <code>prefix_length_max</code> (inclusive).  The list of prefixes
 * returned by the trie is compared against the list generated by
 * inspecting the keys in the <code>keys</code> vector in the valid
 * range from <code>start_index</code> (inclusive) to
 * <code>stop_index</code> (exclusive).  The PRNG <code>prng</code> is
 * used to randomly make the selections.  If an error occurs, an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] keys keys used to verify results returned by the trie
 * @param[in] start_index valid starting index for <code>keys</code> (inclusive)
 * @param[in] stop_index valid stopping index for <code>keys</code> (exclusive)
 * @param[in] test_count_min minimum number of tests to execute
 * @param[in] test_count_max maximum number of tests to execute
 * @param[in] prng PRNG
 */
static void
main__test_find_by_prefix(cgul_exception_t* cex,
                          cgul_trie_t trie,
                          cgul_vector_t keys,
                          unsigned long int start_index,
                          unsigned long int stop_index,
                          cgul_uint32_t test_count_min,
                          cgul_uint32_t test_count_max,
                          cgul_uint32_t prefix_length_min,
                          cgul_uint32_t prefix_length_max,
                          cgul_random_lfg32_t prng)
{
    cgul_uint32_t i = 0;
    unsigned long int j = 0;
    cgul_uint32_t test_count = 0;
    char* prefix = NULL;
    char* expected_match = NULL;
    cgul_vector_t expected_matches = NULL;
    unsigned long int expected_matches_size = 0;
    const char** actual_matches = NULL;
    unsigned long int actual_matches_size = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(keys);
    assert(prng);
    assert(stop_index <= cgul_vector__get_size(cex, keys));

    /* Sanity check. */
    if (test_count_max - test_count_min <= 0) {
        goto out;
    }

    /* Allocate. */
    expected_matches = cgul_vector__new(cex);
    if (*cex) {
        goto out;
    }

    /* Randomly determine the number of prefix tests to execute. */
    test_count = cgul_random_lfg32__next_in_range(
                     cex, prng, test_count_max - test_count_min + 1);
    if (*cex) {
        goto out;
    }
    test_count += test_count_min;

    /* Run all the prefix tests. */
    for (i = 0 ; i < test_count ; ++i) {

        /* Randomly generate the prefix string. */
        prefix = main__generate_string_random_length(
                     cex, prng, prefix_length_min, prefix_length_max);
        if (*cex) {
            goto out;
        }

        /* Get the list of expected matches. */
        main__append_prefix_matches(
            cex, keys, start_index, stop_index, prefix, expected_matches);
        if (*cex) {
            goto out;
        }

        /* Sort the expected matches to comparing them easier. */
        cgul_vector__sort(cex,
                          expected_matches,
                          (cgul_merge_sort__compare_t)main__compare_keys);
        if (*cex) {
            goto out;
        }

        /* Get the list of actual prefix matches that the trie returns. */
        cgul_trie__find_keys_by_prefix(
            cex, trie, prefix, &actual_matches, &actual_matches_size);
        if (*cex) {
            goto out;
        }

        /* Make sure the actual number of matches agrees with the
         * expected number of matches. */
        expected_matches_size = cgul_vector__get_size(cex, expected_matches);
        if (actual_matches_size != expected_matches_size) {
            *cex = cgul_exception__new(
                 "main__test_find_by_prefix: incorrect number of matches: "
                 "(expected, actual) = (%lu, %lu)",
                 (unsigned long int)expected_matches_size,
                 (unsigned long int)actual_matches_size);
            goto out;
        }

        /* Compare each actual match with each expected match.  Both
         * lists are sorted. */
        for (j = 0 ; j < expected_matches_size ; ++j) {
            expected_match = cgul_vector__at(cex, expected_matches, j);
            if (main__compare_keys(actual_matches[j], expected_match) != 0) {
                *cex = cgul_exception__new(
                     "main__test_find_by_prefix: "
                     "(expected, actual) = (%s, %s)",
                     expected_match, actual_matches[j]);
                goto out;
            }
        }

        /*
         * Per-iteration clean up.
         */

        /* Clear the list of expected matches.  Do not free the keys
         * because they are owned by the trie. */
        cgul_vector__clear(cex, expected_matches);

        /* Free the actual matches. Do not free the keys because they
         * are owned by the trie.  */
        free(actual_matches);
        actual_matches = NULL;

        /* Free the prefix. */
        free(prefix);
        prefix = NULL;
    }

 out:

    /* Clean up. */
    if (actual_matches) {
        free(actual_matches);
    }
    if (prefix) {
        free(prefix);
    }
    if (expected_matches) {
        /* Do not free the keys because they are owned by the trie. */
        cgul_vector__delete(expected_matches);
    }

    return;
}

/**
 * The <code>keys</code> vector holds the sorted list of all the keys
 * currently in the trie <code>trie</code>.  This method will attempt
 * to find the key in the trie with the longest prefix of a randomly
 * generated string.  The keys will be generated with random lengths
 * between <code>key_length_min</code> (inclusive) and
 * <code>key_length_max</code> (inclusive).  The PRNG
 * <code>prng</code> is used to randomly make the selections.  If an
 * error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] keys sorted list of all the keys currently in the trie
 * @param[in] key_length_min minimum length of keys
 * @param[in] key_length_max maximum length of keys
 * @param[in] test_count number of test iterations to execute
 * @param[in] prng PRNG
 */
static void
main__test_find_longest_prefix_of(cgul_exception_t* cex,
                                  cgul_trie_t trie,
                                  cgul_vector_t keys,
                                  cgul_uint32_t key_length_min,
                                  cgul_uint32_t key_length_max,
                                  cgul_uint32_t test_count,
                                  cgul_random_lfg32_t prng)
{
    unsigned long int i = 0;
    unsigned long int j = 0;
    unsigned long int j_max = 0;
    char* complete_string = NULL;
    const char** keys_array = NULL;
    const char* find_rv = 0;
    const char* strstr_rv = NULL;
    const char* max_prefix = NULL;
    size_t max_prefix_length = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(keys);

    /* Sanity check. */
    if (test_count == 0) {
        goto out;
    }

    /* Get the "keys" array. */
    keys_array = (const char**)cgul_vector__get_array(cex, keys);

    /* Test for the key in the trie that is the longest prefix of a
     * randomly generated string. */
    for (i = 0 ; i < test_count ; ++i) {

        /* Generate the lookup key. */
        complete_string = main__generate_string_random_length(
                         cex, prng, key_length_min, key_length_max);
        if (*cex) {
            goto out;
        }

        /* Use the cgul_trie API to find the key in the trie that is
         * the longest prefix of the "complete_string." */
        find_rv = cgul_trie__find_longest_prefix_of(cex, trie, complete_string);

        /* Manually find the key in the trie that is the longest
         * prefix of the "complete_string". */
        max_prefix = NULL;
        max_prefix_length = 0;
        j_max = cgul_vector__get_size(cex, keys);
        for (j = 0 ; j < j_max ; ++j) {
            /* Test if keys_array[j] is a prefix of complete_string. */
            strstr_rv = strstr(complete_string, keys_array[j]);
            if (strstr_rv == complete_string) {
                /* If so, is it the longest prefix? */
                if (!max_prefix) {
                    max_prefix = keys_array[j];
                    max_prefix_length = strlen(max_prefix);
                } else {
                    size_t other_length = strlen(keys_array[j]);
                    if (max_prefix_length < other_length) {
                        max_prefix = keys_array[j];
                        max_prefix_length = other_length;
                    }
                }
            }
        }

        /* Compare the API result against the manual result. */
        if (find_rv != max_prefix) {
            *cex = cgul_exception__new(
                "main__test_find_longest_prefix_of: "
                "(expected, actual) = (%s, %s)",
                max_prefix, find_rv);
            goto out;
        }

        /* Per-iteration clean up. */
        free(complete_string);
        complete_string = NULL;
    }

 out:

    /* Clean up. */
    if (complete_string) {
        free(complete_string);
    }

    return;
}

/**
 * Randomly execute between <code>test_count_min</code> (inclusive)
 * and <code>test_count_max</code> (inclusive) tests to find the floor
 * in the trie <code>trie</code> for randomly generated search keys.
 * The sorted vector of keys <code>keys</code> is used to verify the
 * floor.  The search keys will be generated with random lengths
 * between <code>key_length_min</code> (inclusive) and
 * <code>key_length_max</code> (inclusive).  The PRNG
 * <code>prng</code> is used to randomly make the selections.  If an
 * error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] keys sorted list of all the keys currently in the trie
 * @param[in] key_length_min minimum length of keys
 * @param[in] key_length_max maximum length of keys
 * @param[in] test_count_min minimum number of test iterations to execute
 * @param[in] test_count_max maximum number of test iterations to execute
 * @param[in] prng PRNG
 */
static void
main__test_floor(cgul_exception_t* cex,
                 cgul_trie_t trie,
                 cgul_vector_t keys,
                 cgul_uint32_t key_length_min,
                 cgul_uint32_t key_length_max,
                 cgul_uint32_t test_count_min,
                 cgul_uint32_t test_count_max,
                 cgul_random_lfg32_t prng)
{
    cgul_uint32_t i = 0;
    cgul_uint32_t test_count = 0;
    char* search_key = NULL;
    const char* expected = NULL;
    const char* actual = NULL;
    cgul_trie_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Randomly determine the number of tests to execute. */
    test_count = cgul_random_lfg32__next_in_range(
                     cex, prng, test_count_max - test_count_min + 1);
    if (*cex) {
        goto out;
    }
    test_count += test_count_min;

    /* Execute each test. */
    for (i = 0 ; i < test_count ; ++i) {

        /* Randomly generate the search key. */
        search_key = main__generate_string_random_length(
                         cex, prng, key_length_min, key_length_max);
        if (*cex) {
            goto out;
        }

        /* Get the expected result. */
        expected = cgul_binary_search_floor_pointers(
            cex,
            search_key,
            (const void**)cgul_vector__get_array(cex, keys),
            cgul_vector__get_size(cex, keys),
            (cgul_binary_search__compare_t)main__compare_keys);

        /* Get the actual result. */
        node = cgul_trie__find_floor(cex, trie, search_key);
        if (node) {
            actual = cgul_trie_node__get_key(cex, node);
        } else {
            actual = NULL;
        }

        /* Compare expected and actual results.  Print an error if one
         * key is NULL but the other isn't or if they are both not
         * NULL but keys are not equal. */
        if (((!expected || !actual) && (expected != actual)) ||
            ((expected && actual) && (main__compare_keys(expected, actual))))
        {
            *cex = cgul_exception__new(
                "main__test_floor: "
                "(search_key, expected, actual): (%s, %s, %s)",
                search_key, expected, actual);
            goto out;
        }

        /* Per-iteration clean up. */
        free(search_key);
        search_key = NULL;
    }

 out:

    /* Clean up. */
    if (search_key) {
        free(search_key);
    }

    return;
}

/**
 * Randomly execute between <code>test_count_min</code> (inclusive)
 * and <code>test_count_max</code> (inclusive) tests to find the
 * ceiling in the trie <code>trie</code> for randomly generated search
 * keys.  The sorted vector of keys <code>keys</code> is used to
 * verify the ceiling.  The search keys will be generated with random
 * lengths between <code>key_length_min</code> (inclusive) and
 * <code>key_length_max</code> (inclusive).  The PRNG
 * <code>prng</code> is used to randomly make the selections.  If an
 * error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] keys sorted list of all the keys currently in the trie
 * @param[in] key_length_min minimum length of keys
 * @param[in] key_length_max maximum length of keys
 * @param[in] test_count_min minimum number of test iterations to execute
 * @param[in] test_count_max maximum number of test iterations to execute
 * @param[in] prng PRNG
 */
static void
main__test_ceiling(cgul_exception_t* cex,
                 cgul_trie_t trie,
                 cgul_vector_t keys,
                 cgul_uint32_t key_length_min,
                 cgul_uint32_t key_length_max,
                 cgul_uint32_t test_count_min,
                 cgul_uint32_t test_count_max,
                 cgul_random_lfg32_t prng)
{
    cgul_uint32_t i = 0;
    cgul_uint32_t test_count = 0;
    char* search_key = NULL;
    const char* expected = NULL;
    const char* actual = NULL;
    cgul_trie_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Randomly determine the number of tests to execute. */
    test_count = cgul_random_lfg32__next_in_range(
                     cex, prng, test_count_max - test_count_min + 1);
    if (*cex) {
        goto out;
    }
    test_count += test_count_min;

    /* Execute each test. */
    for (i = 0 ; i < test_count ; ++i) {

        /* Randomly generate the search key. */
        search_key = main__generate_string_random_length(
                         cex, prng, key_length_min, key_length_max);
        if (*cex) {
            goto out;
        }

        /* Get the expected result. */
        expected = cgul_binary_search_ceiling_pointers(
            cex,
            search_key,
            (const void**)cgul_vector__get_array(cex, keys),
            cgul_vector__get_size(cex, keys),
            (cgul_binary_search__compare_t)main__compare_keys);

        /* Get the actual result. */
        node = cgul_trie__find_ceiling(cex, trie, search_key);
        if (node) {
            actual = cgul_trie_node__get_key(cex, node);
        } else {
            actual = NULL;
        }

        /* Compare expected and actual results.  Print an error if one
         * key is NULL but the other isn't or if they are both not
         * NULL but keys are not equal. */
        if (((!expected || !actual) && (expected != actual)) ||
            ((expected && actual) && (main__compare_keys(expected, actual))))
        {
            *cex = cgul_exception__new(
                "main__test_ceiling: "
                "(search_key, expected, actual): (%s, %s, %s)",
                search_key, expected, actual);
            goto out;
        }

        /* Per-iteration clean up. */
        free(search_key);
        search_key = NULL;
    }

 out:

    /* Clean up. */
    if (search_key) {
        free(search_key);
    }

    return;
}

/**
 * Return a randomized copy of the keys in the <code>keys</code>
 * vector.  The PRNG <code>prng</code> is used to randomize the keys.
 * The caller must not attempt to free the keys in the vector returned
 * because they are shared with the <code>keys</code> vector and
 * likely owned by the trie at any rate.  If an error occurs,
 * <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] keys keys
 * @param[in] prng PRNG
 * @return randomized copy of keys
 */
static cgul_vector_t
main__randomize_keys(cgul_exception_t* cex,
                     cgul_vector_t keys,
                     cgul_random_lfg32_t prng)
{
    size_t i = 0;
    void** arr = NULL;
    size_t arr_size = 0;
    cgul_vector_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(keys);
    assert(prng);

    /* Allocate. */
    rv = cgul_vector__new(cex);
    if (*cex) {
        goto out;
    }

    /* Copy the keys to the new vector. */
    arr = cgul_vector__get_array(cex, keys);
    arr_size = cgul_vector__get_size(cex, keys);
    for (i = 0 ; i < arr_size ; ++i) {
        cgul_vector__push_back(cex, rv, arr[i]);
        if (*cex) {
            goto out;
        }
    }

    /* Randomize the copy of the keys. */
    cgul_vector__shuffle(
        cex,
        rv,
        (cgul_vector__next_in_range_t)cgul_random_lfg32__next_in_range,
        prng);

 out:

    /* Clean up. */
    if (*cex && rv) {
        /* Do not free the keys.  They are owned by some other object. */
        cgul_vector__delete(rv);
        rv = NULL;
    }

    return rv;
}

/**
 * Remove all the keys from the trie <code>trie</code> by randomizing
 * a copying the keys in the <code>sorted_keys</code> vector and using
 * the randomized copy to determine the order in which the keys are to
 * be remove.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] sorted_keys sorted list of keys
 * @param[in] prng PRNG
 */
static void
main__remove_keys(cgul_exception_t* cex,
                  cgul_trie_t trie,
                  cgul_vector_t sorted_keys,
                  cgul_random_lfg32_t prng)
{
    int remove_rv = 0;
    char* key_in = NULL;
    char* key_out = NULL;
    unsigned long int i = 0;
    unsigned long int i_max = 0;
    unsigned long int trie_size = 0;
    unsigned long int expected_size = 0;
    cgul_vector_t random_keys = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(sorted_keys);
    assert(prng);

    /* Randomize the list of keys. */
    random_keys = main__randomize_keys(cex, sorted_keys, prng);
    if (*cex) {
        goto out;
    }

    /* Verify the trie and the vector have the same size. */
    expected_size = cgul_vector__get_size(cex, random_keys);
    trie_size = cgul_trie__get_size(cex, trie);
    if (trie_size != expected_size) {
        *cex = cgul_exception__new("main__remove_keys: "
                                   "trie_size != expected_size: %lu != %lu",
                                   trie_size, expected_size);
        goto out;
    }

    /* Remove each key from the "random_keys" list. */
    i_max = cgul_vector__get_size(cex, random_keys);
    for (i = 0 ; i < i_max ; ++i) {

        /* Get the key to remove. */
        key_in = (char*)cgul_vector__at(cex, random_keys, i);

        /* Remove the key from the "sorted_keys" vector. */
        remove_rv = main__remove_key_from_vector(
            cex,
            sorted_keys,
            key_in,
            (cgul_binary_search__compare_t)main__compare_pointers_to_keys);
        if (!remove_rv) {
            *cex = cgul_exception__new(
                "main__remove_keys: Verification array missing key: %s",
                key_in);
            goto out;
        }

        /* Remove the key from the trie. */
        if (!cgul_trie__remove(cex, trie, key_in, &key_out, NULL)) {
            *cex = cgul_exception__new(
                "main__remove_keys: Missing key: %s", key_in);
            goto out;
        }
        free(key_out);
        key_out = NULL;

        /* Make sure the trie updates its size correctly. */
        --expected_size;
        trie_size = cgul_trie__get_size(cex, trie);
        if (trie_size != expected_size) {
            *cex = cgul_exception__new("main__remove_keys: "
                                       "trie_size != expected_size: %lu != %lu",
                                       trie_size, expected_size);
            goto out;
        }

        /* Test finding keys by prefix.  This functionality is the
         * main reason for using a trie.  Note that finding by prefix
         * is also tested before removing any keys in
         * main__execute_test(). */
        main__test_find_by_prefix(cex, trie, random_keys,
                                  i + 1, i_max,
                                  PREFIX_TEST_COUNT_MIN,
                                  PREFIX_TEST_COUNT_MAX,
                                  PREFIX_LENGTH_MIN,
                                  PREFIX_LENGTH_MAX,
                                  prng);
        if (*cex) {
            goto out;
        }

        /* Test finding the floor relative to a search key. */
        main__test_floor(cex, trie, sorted_keys,
                         KEY_LENGTH_MIN, KEY_LENGTH_MAX,
                         FLOOR_TEST_COUNT_MIN, FLOOR_TEST_COUNT_MAX,
                         prng);
        if (*cex) {
            goto out;
        }

        /* Test finding the ceiling relative to a search key. */
        main__test_ceiling(cex, trie, sorted_keys,
                         KEY_LENGTH_MIN, KEY_LENGTH_MAX,
                         CEILING_TEST_COUNT_MIN, CEILING_TEST_COUNT_MAX,
                         prng);
        if (*cex) {
            goto out;
        }

    }

 out:

    /* Clean up. */
    if (random_keys) {
        /* Do not free the keys.  They are owned by the trie. */
        cgul_vector__delete(random_keys);
    }

    return;
}

/**
 * Test the trie <code>trie</code> using values randomly generated by
 * the PRNG <code>prng</code>.  If an error occurs, an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie trie under test
 * @param[in] prng PRNG
 */
static void
main__execute_test(cgul_exception_t* cex,
                   cgul_trie_t trie,
                   cgul_random_lfg32_t prng)
{
    cgul_vector_t keys = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(prng);

    /* Allocate. */
    keys = cgul_vector__new(cex);
    if (*cex) {
        goto out;
    }

    /* Insert randomly generated keys. */
    APPLOG("    Inserting keys.\n");
    main__insert_keys(cex, trie, keys,
                      KEY_COUNT_MIN, KEY_COUNT_MAX,
                      KEY_LENGTH_MIN, KEY_LENGTH_MAX,
                      prng);
    if (*cex) {
        goto out;
    }

    /* Sort the "keys" verification array. */
    APPLOG("    Sorting verification array.\n");
    cgul_vector__sort(cex,
                      keys,
                      (cgul_merge_sort__compare_t)main__compare_keys);

    /* Verify the trie is still sorted. */
    APPLOG("    Verifying trie is sorted.\n");
    if (!main__verify_trie_is_sorted_using_forward_iteration(
            cex, trie, keys, 0, cgul_vector__get_size(cex, keys)))
    {
        *cex = cgul_exception__new("Forward Iteration: trie is not sorted!");
        goto out;
    }
    if (!main__verify_trie_is_sorted_using_reverse_iteration(
            cex, trie, keys, 0, cgul_vector__get_size(cex, keys)))
    {
        *cex = cgul_exception__new("Reverse Iteration: trie is not sorted!");
        goto out;
    }

    /* Test exact lookups. */
    APPLOG("    Testing find by exact key.\n");
    main__test_find(
        cex, trie, keys,
        KEY_LENGTH_MIN, KEY_LENGTH_MAX,
        FIND_TEST_COUNT,
        prng);
    if (*cex) {
        goto out;
    }

    /* Test array-style indexing. */
    APPLOG("    Testing find by array-style index.\n");
    main__test_find_at(cex, trie, keys);
    if (*cex) {
        goto out;
    }

    /* Test finding keys by prefix.  This functionality is the main
     * reason for using a trie.  Note that finding by prefix is also
     * tested after each remove in main__remove_keys(). */
    APPLOG("    Testing find by prefix.\n");
    main__test_find_by_prefix(cex, trie, keys,
                              0, cgul_vector__get_size(cex, keys),
                              PREFIX_TEST_COUNT_MIN,
                              PREFIX_TEST_COUNT_MAX,
                              PREFIX_LENGTH_MIN,
                              PREFIX_LENGTH_MAX,
                              prng);
    if (*cex) {
        goto out;
    }

    /* Test finding "longest prefix of". */
    APPLOG("    Testing find longest prefix of.\n");
    main__test_find_longest_prefix_of(
        cex, trie, keys,
        KEY_LENGTH_MIN, KEY_LENGTH_MAX,
        LONGEST_PREFIX_OF_TEST_COUNT,
        prng);
    if (*cex) {
        goto out;
    }

    /* Test finding the floor relative to a search key. */
    APPLOG("    Testing find floor.\n");
    main__test_floor(
        cex, trie, keys,
        KEY_LENGTH_MIN, KEY_LENGTH_MAX,
        FLOOR_TEST_COUNT_MIN, FLOOR_TEST_COUNT_MAX,
        prng);
    if (*cex) {
        goto out;
    }

    /* Test finding the ceiling relative to a search key. */
    APPLOG("    Testing find ceiling.\n");
    main__test_ceiling(
        cex, trie, keys,
        KEY_LENGTH_MIN, KEY_LENGTH_MAX,
        CEILING_TEST_COUNT_MIN, CEILING_TEST_COUNT_MAX,
        prng);
    if (*cex) {
        goto out;
    }

    /*
     * Randomly remove all keys, and after each key is removed, test
     * to make sure the trie is still correctly finding keys by prefix
     * and correctly find the floor and ceiling given a search key.
     */
    APPLOG("    Removing keys, finding by prefix, floor, and ceiling.\n");
    main__remove_keys(cex, trie, keys, prng);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (keys) {
        cgul_vector__delete(keys);
    }
    if (*cex && trie) {
        cgul_exception_t local = NULL;
        cgul_trie__clear(&local, trie, cgul_cache__get_freer(), NULL);
    }
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    unsigned long int i = 0;
    cgul_uint32_t seed = 0;
    cgul_trie_t trie = NULL;
    cgul_random_lfg32_t prng = NULL;
    cgul_exception_t local = NULL;

#if 1
    /* Use the time as the seed. */
    seed = (cgul_uint32_t)time(NULL);
#else
    /* Use specific seed value. */
    seed = 1425126242;
    APPLOG("===== REUSING SEED!!!\n");
#endif
    APPLOG("===== seed: %lu\n", (unsigned long)seed);

    /* Allocate. */
    trie = cgul_trie__new_with_indexing(&local);
    prng = cgul_random_lfg32__new_from_seed(&local, seed);
    if (local) {
        goto out;
    }

    /* Execute each test iteration. */
    for (i = 0 ; i < TOTAL_TEST_ITERATIONS ; ++i) {
        APPLOG("===== Test #%lu\n", i);
        main__execute_test(&local, trie, prng);
        if (local) {
            goto out;
        }
    }

 out:

    /* Handle exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        fprintf(stderr, "***  seed: %lu\n", (unsigned long)seed);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (prng) {
        cgul_random_lfg32__delete(prng);
    }
    if (trie) {
        cgul_trie__free_keys(trie);
        cgul_trie__delete(trie);
    }

    return rv;
}
