/*
 * Test the following methods using randomly generated keys:
 *
 *     cgul_trie__find_at()
 *     cgul_trie__find_rank()
 *     cgul_trie__remove_at()
 */

#include "cgul.h"

/**
 * Minimum number of keys to generate.
 */
#define KEY_COUNT_MIN (0)

/**
 * Maximum number of keys to generate.
 */
#define KEY_COUNT_MAX (1000000)

/**
 * Minimum key length when generating keys.
 */
#define KEY_LENGTH_MIN (0)

/**
 * Maximum key length when generating keys.
 */
#define KEY_LENGTH_MAX (31)

/**
 * Return less than zero, zero, or greater than zero if
 * <code>key1</code> is less than, equal to, or greater than
 * <code>key2</code>.
 *
 * @param[in] key1 first key
 * @param[in] key2 second key
 * @return lexicographical comparison of the keys
 * @see main__compare_pointers_to_keys
 */
static int
main__compare_keys(const char* key1,
                   const char* key2)
{
    int rv = 0;
    size_t i = 0;

    /* Compare the parts of the keys where they both have characters. */
    for (i = 0 ; key1[i] && key2[i] ; ++i) {
        if (key1[i] < key2[i]) {
            rv = -1;
            goto out;
        } else if (key1[i] > key2[i]) {
            rv = 1;
            goto out;
        }
    }

    /* If you reach here, at least one key is a prefix of the other.
     * The lexicographical comparison of the keys now depends on which
     * is longer. */
    if ((key1[i] == 0) && (key2[i] == 0)) {
        rv = 0;
    } else if (key1[i] == 0) {
        rv = -1;
    } else {
        rv = 1;
    }

 out:
    return rv;
}

/**
 * Return a string of length <code>length</code> having randomly
 * selected characters in the range <code>['a', 'z']</code>.  The PRNG
 * <code>prng</code> is used for randomly selecting the string's
 * content.  The client is responsible for calling <code>free()</code>
 * on the pointer returned.  If an error occurs, <code>NULL</code> is
 * returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] prng prng
 * @param[in] length length of string to return
 * @return randomly generated string
 */
static char*
main__generate_string_fixed_length(cgul_exception_t* cex,
                                   cgul_random_lfg32_t prng,
                                   size_t length)
{
    cgul_uint32_t n = 0;
    size_t i = 0;
    char* rv = NULL;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Generate. */
    for (i = 0 ; i < length ; ++i) {
        n = cgul_random_lfg32__next_in_range(cex, prng, 'z' - 'a' + 1);
        cgul_string__append_char(cex, tmp, 'a' + (char)n);
        if (*cex) {
            goto out;
        }
    }

    /* Success. */
    rv = cgul_string__take_value(cex, tmp);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (*cex && rv) {
        free(rv);
        rv = NULL;
    }
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

/**
 * Return a string having a length randomly selected from the range
 * <code>[length_min, length_max]</code> having randomly selected
 * characters in the range <code>['a', 'z']</code>.  The PRNG
 * <code>prng</code> is used for randomly selecting the string's
 * length and content.  The client is responsible for calling
 * <code>free()</code> on the pointer returned.  If an error occurs,
 * <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] prng prng
 * @param[in] length length of string to return
 * @return randomly generated string
 */
static char*
main__generate_string_random_length(cgul_exception_t* cex,
                                    cgul_random_lfg32_t prng,
                                    cgul_uint32_t length_min,
                                    cgul_uint32_t length_max)
{
    char* rv = NULL;
    cgul_uint32_t n = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Randomly select the length of the string to generate. */
    n = cgul_random_lfg32__next_in_range(
            cex, prng, length_max - length_min + 1);
    if (*cex) {
        goto out;
    }

    /* Generate the string. */
    rv = main__generate_string_fixed_length(cex, prng, length_min + n);
    if (*cex) {
        goto out;
    }

 out:
    return rv;
}

/**
 * Insert between <code>key_count_min</code> (inclusive) and
 * <code>key_count_max</code> (inclusive) randomly generated keys into
 * the trie <code>trie</code> and push a copy of each key in the order
 * generated onto the <code>keys</code> vector (which can be used
 * later for verification purposes).  The key are randomly generated
 * by the PRNG <code>prng</code> and will have length between
 * <code>key_length_min</code> (inclusive) and
 * <code>key_length_max</code>(inclusive).  The user is responsible
 * for calling <code>free()</code> on the keys inserted into the trie.
 * If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] prng PRNG
 * @param[in] keys keys
 * @param[in] key_count_min minimum key count
 * @param[in] key_count_max maximum key count
 * @param[in] key_length_min minimum key length
 * @param[in] key_length_max maximum key length
 */
static void
main__insert_keys(cgul_exception_t* cex,
                  cgul_trie_t trie,
                  cgul_vector_t keys,
                  cgul_random_lfg32_t prng,
                  cgul_uint32_t key_count_min,
                  cgul_uint32_t key_count_max,
                  cgul_uint32_t key_length_min,
                  cgul_uint32_t key_length_max)
{
    int insert_rv = 0;
    cgul_uint32_t i = 0;
    cgul_uint32_t key_count = 0;
    cgul_uint32_t key_length = 0;
    char* key = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Determine the number of keys to generate. */
    key_count = cgul_random_lfg32__next_in_range(
        cex, prng, key_count_max - key_count_min + 1);
    if (*cex) {
        goto out;
    }
    key_count += key_count_min;

#if 1
    printf("key_count = %lu\n", (unsigned long)key_count);
#endif

    /* Generate and insert the keys. */
    for (i = 0 ; i < key_count ; ++i) {

        /* Determine the length of the key to generate. */
        key_length = cgul_random_lfg32__next_in_range(
            cex, prng, key_length_max - key_length_min + 1);
        if (*cex) {
            goto out;
        }
        key_length += key_length_min;

        /* Randomly generate the next key. */
        key = main__generate_string_random_length(
            cex, prng, KEY_LENGTH_MIN, KEY_LENGTH_MAX);
        if (*cex) {
            goto out;
        }

        /* Try to insert the key which could fail if a key repeats. */
        insert_rv = cgul_trie__insert(cex, trie, key, NULL);
        if (*cex) {
            goto out;
        }
        if (!insert_rv) {
            free(key);
            key = NULL;
            continue;
        }

        /* Push the key onto the verification vector. */
        cgul_vector__push_back(cex, keys, key);
        if (*cex) {
            goto out;
        }

        /* The trie now owns the key. */
        key = NULL;

    }

 out:

    /* Clean up. */
    if (key) {
        free(key);
    }

    return;
}

/**
 * Test <code>cgul_trie__find_at()</code> by verifying the that the
 * keys in the <code>keys</code> vector exist in the trie
 * <code>trie</code> in the same order.  If an error occurs, an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] keys keys
 */
static void
main__test_find_at(cgul_exception_t* cex,
                   cgul_trie_t trie,
                   cgul_vector_t keys)
{
    unsigned long int i = 0;
    char* key = NULL;
    char** arr = NULL;
    unsigned long int arr_size = 0;
    cgul_trie_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(keys);

    /* Unpack the keys vector. */
    arr = (char**)cgul_vector__get_array(cex, keys);
    arr_size = cgul_vector__get_size(cex, keys);

    /* Verify that each key in "keys" exists in "trie" and in the same
     * order. */
    for (i = 0 ; i < arr_size ; ++i) {

        /* Find the node by index. */
        node = cgul_trie__find_at(cex, trie, i);
        if (!node) {
            *cex = cgul_exception__new(
                "main__test_find_at: Missing key: %lu", i);
            goto out;
        }

        /* Verify the keys match. */
        key = (char*)cgul_trie_node__get_key(cex, node);
        if (main__compare_keys(key, arr[i]) != 0) {
            *cex = cgul_exception__new(
                "main__test_find_at: "
                "Invalid key at index %lu: (expecting, actual) = (%lu, %lu)",
                i, (unsigned long int)*arr[i], (unsigned long int)*key);
            goto out;
        }
    }

    /* Verify lookup misses return NULL. */
    node = cgul_trie__find_at(cex, trie, arr_size);
    if (node) {
        key = (char*)cgul_trie_node__get_key(cex, node);
        *cex = cgul_exception__new(
            "main__test_find_at: "
            "Unexpectedly found key for out-of-bounds index: "
            "(index, key) = (%lu, %s)",
            arr_size, key);
        goto out;
    }

 out:
    return;
}

/**
 * Test <code>cgul_trie__find_rank()</code> by verifying the index of
 * each key in the <code>keys</code> vector equals the rank of that
 * same key in the trie <code>trie</code>.  If an error occurs, an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] keys keys
 */
static void
main__test_find_rank(cgul_exception_t* cex,
                     cgul_trie_t trie,
                     cgul_vector_t keys)
{
    int rank_rv = 0;
    unsigned long int i = 0;
    unsigned long int rank = 0;
    char** arr = NULL;
    unsigned long int arr_size = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(keys);

    /* Unpack the keys vector. */
    arr = (char**)cgul_vector__get_array(cex, keys);
    arr_size = cgul_vector__get_size(cex, keys);

    /* Verify that each key in "keys" exists in "trie" and in the same
     * order. */
    for (i = 0 ; i < arr_size ; ++i) {

        /* Find the rank of the key. */
        rank_rv = cgul_trie__find_rank(cex, trie, arr[i], &rank);
        if (!rank_rv) {
            *cex = cgul_exception__new(
                "main__test_find_rank: Missing key: %lu", i);
            goto out;
        }

        /* Verify the index of the key match the rank. */
        if (rank != i) {
            *cex = cgul_exception__new(
                "main__test_find_at: Invalid rank for key: "
                "(key, expecting, actual) = (%s, %lu, %lu)",
                arr[i], i, rank);
            goto out;
        }
    }

 out:
    return;
}

/**
 * Randomly remove all keys from the trie <code>trie</code> using
 * <code>cgul_trie__remove_at()</code>.  The order in which keys are
 * removed is determined by the PRNG <code>prng</code>.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] PRNG prng
 */
static void
main__test_remove_at(cgul_exception_t* cex,
                     cgul_trie_t trie,
                     cgul_random_lfg32_t prng)
{
    int remove_rv = 0;
    cgul_uint32_t index = 0;
    char* key = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    while (!cgul_trie__is_empty(cex, trie)) {

        /* Generate the index of the next key to remove. */
        index = cgul_random_lfg32__next_in_range(
            cex, prng, cgul_trie__get_size(cex, trie));

        /* Remove this key from the trie. */
        remove_rv = cgul_trie__remove_at(cex, trie, index, &key, NULL);
        if (!remove_rv) {
            *cex = cgul_exception__new(
                "main__test_remove_at: "
                "Unable to remove key at index: %lu",
                (unsigned long int)index);
            goto out;
        }

        /* Free the key. */
        free(key);
        key = NULL;

    }

 out:
    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_uint32_t seed = 0;
    cgul_trie_t trie = NULL;
    cgul_vector_t keys = NULL;
    cgul_random_lfg32_t prng = NULL;
    cgul_exception_t local = NULL;

#if 1
    /* Use the time as the seed. */
    seed = (cgul_uint32_t)time(NULL);
#else
    /* Use specific seed value. */
    seed = 1425126242;
    printf("===== REUSING SEED!!!\n");
#endif
    printf("===== seed: %lu\n", (unsigned long)seed);

    /* Allocate. */
    keys = cgul_vector__new(&local);
    prng = cgul_random_lfg32__new_from_seed(&local, seed);
    trie = cgul_trie__new_with_indexing(&local);
    if (local) {
        goto out;
    }

    /* Insert randomly generated numeric keys. */
    main__insert_keys(&local, trie, keys, prng,
                      KEY_COUNT_MIN, KEY_COUNT_MAX,
                      KEY_LENGTH_MIN, KEY_LENGTH_MAX);
    if (local) {
        goto out;
    }

    /* Sort the keys vector. */
    cgul_vector__sort(
        &local, keys, (cgul_vector__compare_t)main__compare_keys);

    /* Test cgul_trie__find_at(). */
    main__test_find_at(&local, trie, keys);
    if (local) {
        goto out;
    }

    /* Test cgul_trie__find_rank(). */
    main__test_find_rank(&local, trie, keys);
    if (local) {
        goto out;
    }

    /* Test cgul_trie__remove_at().  (This removes all the keys from
     * the trie in random order.) */
    main__test_remove_at(&local, trie, prng);
    if (local) {
        goto out;
    }

 out:

    /* Handle exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        fprintf(stderr, "***  seed: %lu\n", (unsigned long)seed);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (trie) {
        /* The keys should be freed above. */
        cgul_trie__delete(trie);
    }
    if (prng) {
        cgul_random_lfg32__delete(prng);
    }
    if (keys) {
        cgul_vector__delete(keys);
    }

    return rv;
}
