#include "cgul_cxx.h"
#include <iostream>
#include <stdexcept>

using namespace std;

/**
 * Print all the keys in the trie <code>trie</code> that start with
 * the prefix <code>prefix</code>.  If an error occurs, an exception
 * is thrown.
 *
 * @param[in] trie <code>cgul_trie_cxx</code> instance
 * @param[in] prefix prefix
 */
static void
main__print_keys_by_prefix(cgul_trie_cxx& trie,
                           const char* prefix)
{
    unsigned long int i = 0;
    const char** keys = NULL;
    unsigned long int keys_size = 0;

    try {

        /* Iterate over all the keys that match the prefix. */
        if (trie.find_keys_by_prefix(prefix, &keys, &keys_size)) {
            for (i = 0 ; i < keys_size ; ++i) {
            printf("%s -> %s\n", prefix, keys[i]);
        }
    }

    } catch (...) {
        /* Clean up. */
        if (keys) {
            free(keys);
        }
        /* Rethrow. */
        throw;
    }

    /* Clean up. */
    if (keys) {
        free(keys);
    }

}

/**
 * Print all the values in the trie <code>trie</code> associated with
 * keys that start with the prefix <code>prefix</code>.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in] trie <code>cgul_trie_cxx</code> instance
 * @param[in] prefix prefix
 */
static void
main__print_values_by_prefix(cgul_trie_cxx& trie,
                             const char* prefix)
{
    unsigned long int i = 0;
    void** values = NULL;
    unsigned long int values_size = 0;

    try {

        /* Iterate over all the values that match the prefix. */
        if (trie.find_values_by_prefix(prefix, &values, &values_size)) {
            for (i = 0 ; i < values_size ; ++i) {
                printf("%s -> %p\n", prefix, values[i]);
            }
        }

    } catch (...) {
        /* Clean up. */
        if (values) {
            free(values);
        }
        /* Rethrow. */
        throw;
    }

    /* Clean up. */
    if (values) {
        free(values);
    }

}

/**
 * Print all the nodes in the trie <code>trie</code> associated with
 * keys that start with the prefix <code>prefix</code>.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in] trie <code>cgul_trie_cxx</code> instance
 * @param[in] prefix prefix
 */
static void
main__print_nodes_by_prefix(cgul_trie_cxx& trie,
                            const char* prefix)
{
    unsigned long int i = 0;
    const char* key = NULL;
    void* value = NULL;
    cgul_trie_node_cxx** nodes = NULL;
    unsigned long int nodes_size = 0;

    try {

        /* Iterate over all the nodes that match the prefix. */
        if (trie.find_nodes_by_prefix(prefix, &nodes, &nodes_size)) {
            for (i = 0 ; i < nodes_size ; ++i) {
                key = nodes[i]->get_key();
                value = nodes[i]->get_value();
                printf("%s -> (%s, %p)\n", prefix, key, value);
            }
        }

    } catch (...) {
        /* Clean up. */
        if (nodes) {
            free(nodes);
        }
        /* Rethrow. */
        throw;
    }

    /* Clean up. */
    if (nodes) {
        free(nodes);
    }

}

/**
 * Print the keys in the trie <code>trie</code> that is the longest
 * prefix of <code>prefix</code>.
 *
 * @param[in] trie <code>cgul_trie_cxx</code> instance
 * @param[in] prefix prefix
 */
static void
main__print_longest_prefix_of(cgul_trie_cxx& trie,
                              const char* prefix)
{
    const char* key = NULL;
    key = trie.find_longest_prefix_of(prefix);
    cgul_libc_cxx::printf("%s -> %s\n", prefix, key);
}

static int
main__fold_keys_cb(const char* key,
                   void* data)
{
    printf("%s\n", key ? key : "(null)");
    return 1;  /* Continue traversing. */
}

static int
main__fold_values_cb(void* value,
                     void* data)
{
    printf("%p\n", value);
    return 1;  /* Continue traversing. */
}

static int
main__fold_pairs_cb(const char* key,
                    void* value,
                    void* data)
{
    printf("%s -> %p\n", key, value);
    return 1;  /* Continue traversing. */
}

static int
main__traverse_cb(cgul_trie_cxx* trie,
                  cgul_trie_node_cxx* n,
                  void* data)
{
    const char* key = n->get_key();
    printf("%s\n", key ? key : "(null)");
    return 1;  /* Continue traversing. */
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    size_t i = 0;
    cgul_trie_cxx trie;
    cgul_trie_node_cxx* node = NULL;
    const char* ss[] = {
        "",
        "Hello",
        "Hello, World!",
        "Goodbye",
        "Goodbye, Cruel World!"
    };
    size_t ss_size = sizeof(ss) / sizeof(ss[0]);

    try {

        /* Test if the trie is empty.  Also, make sure the trie is empty
         * and front, back, oldest, and youngest are all NULL. */
        printf("===== Is Empty? (before)\n");
        printf("    is empty: %d\n", trie.is_empty());
        printf("       front: %p\n", trie.get_front());
        printf("        back: %p\n", trie.get_back());
        printf("      oldest: %p\n", trie.get_oldest());
        printf("    youngest: %p\n", trie.get_youngest());

        /* Insert keys into the trie. */
        for (i = 0 ; i < ss_size ; ++i) {
            char* tmp = cgul_libc_cxx::strdup(ss[i]);
            if (!trie.insert(tmp, NULL)) {
                free(tmp);
            }
        }

        printf("===== Is Empty? (after)\n");
        printf("    is empty: %d\n", trie.is_empty());
        printf("       front: %p\n", trie.get_front());
        printf("        back: %p\n", trie.get_back());
        printf("      oldest: %p\n", trie.get_oldest());
        printf("    youngest: %p\n", trie.get_youngest());

        /* Iterate over the keys in the trie in chronological order. */
        printf("===== Chronological Order\n");
        node = trie.get_oldest();
        for ( ; node ; node = node->get_younger()) {
            printf("-->%s<--\n", node->get_key());
        }

        /* Iterate over the keys in the trie in reverse chronological order. */
        printf("===== Reverse Chronological Order\n");
        node = trie.get_youngest();
        for ( ; node ; node = node->get_older()) {
            printf("-->%s<--\n", node->get_key());
        }

        /* Iterate over the keys in the trie in sort order. */
        printf("===== Sort Order\n");
        node = trie.get_front();
        for ( ; node ; node = node->get_next()) {
            printf("-->%s<--\n", node->get_key());
        }

        /* Iterate over the keys in the trie in reverse sort order. */
        printf("===== Reverse Sort Order\n");
        node = trie.get_back();
        for ( ; node ; node = node->get_prev()) {
            printf("-->%s<--\n", node->get_key());
        }

        /* Find or don't find the nodes for individual keys. */
        printf("===== trie.find() \n");
        for (i = 0 ; i < ss_size ; ++i) {
            node = trie.find(ss[i]);
            if (node) {
                printf("found: -->%s<--\n", node->get_key());
            }
        }
        node = trie.find("not in trie");
        if (node) {
            throw cgul_exception_cxx("Unexpectedly found key that "
                                     "should not be in trie");
        }

        /* Find keys by prefix. */
        printf("===== trie.find_keys_by_prefix()\n");
        main__print_keys_by_prefix(trie, "");
        main__print_keys_by_prefix(trie, "Hello");
        main__print_keys_by_prefix(trie, "Good");
        main__print_keys_by_prefix(trie, "foo");

        /* Find values by prefix. */
        printf("===== trie.find_values_by_prefix()\n");
        main__print_values_by_prefix(trie, "");
        main__print_values_by_prefix(trie, "Hello");
        main__print_values_by_prefix(trie, "Good");
        main__print_values_by_prefix(trie, "foo");

        /* Find nodes by prefix. */
        printf("===== cgul_trie__find_nodes_by_prefix()\n");
        main__print_nodes_by_prefix(trie, "");
        main__print_nodes_by_prefix(trie, "Hello");
        main__print_nodes_by_prefix(trie, "Good");
        main__print_nodes_by_prefix(trie, "foo");

        /* Find key that is the longest prefix of another string. */
        printf("===== cgul_trie__find_longest_prefix_of()\n");
        main__print_longest_prefix_of(trie, "foo");
        main__print_longest_prefix_of(trie, "Hello, Old Chap!");

        /* Left fold over the keys. */
        printf("===== cgul_trie__foldl_keys()\n");
        trie.foldl_keys(main__fold_keys_cb, NULL);

        /* Right fold over the keys. */
        printf("===== cgul_trie__foldr_keys()\n");
        trie.foldr_keys(main__fold_keys_cb, NULL);

        /* Left fold over the values. */
        printf("===== cgul_trie__foldl_values()\n");
        trie.foldl_values(main__fold_values_cb, NULL);

        /* Right fold over the values. */
        printf("===== cgul_trie__foldr_values()\n");
        trie.foldr_values(main__fold_values_cb, NULL);

        /* Left fold over the pairs. */
        printf("===== cgul_trie__foldl_pairs()\n");
        trie.foldl_pairs(main__fold_pairs_cb, NULL);

        /* Right fold over the pairs. */
        printf("===== cgul_trie__foldr_pairs()\n");
        trie.foldr_pairs(main__fold_pairs_cb, NULL);

        /* Traverse the nodes. */
        printf("===== cgul_trie__traverse()\n");
        trie.traverse(main__traverse_cb, NULL);

        /* Clear the trie.  (Note, you can pass free() in directly to
         * cgul_trie__clear().) */
        printf("===== cgul_trie__clear()\n");
        cgul_cache_cxx freer(cgul_cache_cxx::get_freer());
        trie.clear(&freer, NULL);

    } catch (const exception& e) {
        cerr << "*** Error: " << e.what() << endl;
        rv = 1;
    } catch (const cgul_exception_cxx& e) {
        cerr << "*** Error: " << e.get_message() << endl;
        rv = 1;
    }

    return rv;
}
