#include <cassert>
#include <iostream>
#include <stdexcept>
#include "cgul_cxx.h"

using namespace std;

/**
 * This function performs the lookup that converts
 * <code>varname</code> into its associated value.  This function then
 * validates the printf()-style conversion specification <code>cs</code>.
 * If <code>cs</code> is valid, it uses it to call
 * <code>dst->append_sprintf()</code> to append the value
 * associated with <code>varname</code> to the destination string
 * <code>dst</code>.  If <code>cs</code> is not valid, it simply
 * appends the value associated with <code>varname<code> to the
 * destination string <code>dst</code> without using a format string.
 * If an error occurs, an exception is thrown.
 *
 * @param[in] varname variable name
 * @param[in] cs conversion specification or <code>NULL</code>
 * @param[in,out] dst destination string
 * @param[in] data client data
 * @return whether the lookup was successful
 */
static int
do_lookup(const char* varname,
          cgul_convspec_cxx* cs,
          cgul_string_cxx* dst,
          cgul_stanza_cxx* stanza)
{
    int rv = 0;
    int use_format = 0;
    const char* value = NULL;
    cgul_string_cxx format;

    /* Do not assert "cs" because it can be NULL. */
    assert(varname);
    assert(dst);
    assert(stanza);

    /*
     * Perform the lookup.
     */
    value = stanza->lookup(varname);
    if (value == NULL) {
        /* Lookup failed. */
        goto out;
    }

    /*
     * Validate the conversion specification.
     */

    /* If no conversion specification or cgul_varsub_formatted()
     * detected a problem with it, skip the validation. */
    if (cs == NULL) {
        goto append;
    }

    /* All of our values are strings so a "%s" format string is required. */
    if (cs->get_cspec() != 's') {
        goto append;
    }

    /* "cs" appears to be ok.  Get the format string. */
    try {
        cs->serialize(format);
    } catch (...) {
        /* Ignore the exception. */
        goto append;        
    }

    /* Use the format_validated string. */
    use_format = 1;

 append:

    /* Perform the substitution by appending "value" to "dst". */
    if (use_format) {
        /* Append using the format string. */
        dst->append_sprintf(format.get_value(), value);
    } else {
        /* Something was wrong with the format string.  Append without it. */
        dst->append(value);
    }

    /* Success. */
    rv = 1;

 out:
    return rv;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    const char* src = NULL;
    cgul_string_cxx dst;
    cgul_crlf_file_cxx f;
    cgul_crlf_file_cxx crlf;
    cgul_stanza_cxx stanza;

    try {

        /* Read the mapping of variable names to values. */
        f.open_fname("main_varsub_mapping.txt");
        stanza.open_crlf_file(f);
        stanza.all_one_stanza(1);
        stanza.load_next();

        /* Open the text file. */
        crlf.open_fname("main_varsub_formatted_text.txt");

        /* Iterate over each line in the text file. */
        while ((src = crlf.get_line()) != NULL) {

            /* Clear dst. */
            dst.set_value("");

            /* Perform formatted variable substitution. */
            cgul_varsub_formatted_cxx(
                src,
                &dst,
                (cgul_varsub_formatted_cxx__lookup_t)do_lookup,
                &stanza);

            /* Print results. */
            cout << dst.get_value() << endl;
        }

    } catch (const exception& e) {
        cerr << "*** Error: " << e.what() << endl;
        rv = 1;
    } catch (const cgul_exception_cxx& e) {
        cerr << "*** Error: " << e.get_message() << endl;
        rv = 1;
    }

    return rv;
}
