#include "cgul.h"

/**
 * Load the tree such that each line of text in the file
 * <code>fname</code> is a key in the tree.  If an error occurs, an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] t tree
 * @param[in] fname file name
 */
static void
load_tree(cgul_exception_t* cex,
          cgul_rbtree_t t,
          const char* fname)
{
    int insert_rv = 0;
    const char* line = NULL;
    cgul_wstring_t ws = NULL;
    cgul_crlf_file_t crlf = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(fname);

    /* Allocate. */
    ws = cgul_wstring__new(cex);
    crlf = cgul_crlf_file__new_from_fname(cex,
                                          "tower_of_babel.txt",
                                          1 /* is_block_buffered */);
    if (*cex) {
        goto out;
    }

    /* Iterate over each line in the text file. */
    for (;;) {

        /* Get the next line. */
        line = cgul_crlf_file__get_line(cex, crlf);
        if (*cex) {
            goto out;
        }

        /* EOF */
        if (!line) {
            break;
        }

        /* Clear the temporary string. */
        cgul_wstring__set_value(cex, ws, CGUL_WCHAR__EMPTY_STRING);

        /* Convert the line from UTF-8 to UTF-32. */
        cgul_unicode__mbstowcs(cex, line, strlen(line), 0 /* skip bom */, ws);
        if (*cex) {
            goto out;
        }

        /* Insert the string into the red-black tree. */
        insert_rv = cgul_rbtree__insert(cex,
                                        t,
                                        cgul_wstring__get_value(cex, ws),
                                        NULL);
        if (*cex) {
            goto out;
        }
        if (!insert_rv) {
            fprintf(stderr, "*** Warning: duplicate key: \"%s\"\n", line);
        } else {
            /* The red-black tree now owns the key. */
            cgul_wstring__take_value(cex, ws);
        }
    }

 out:

    /* Clean up. */
    if (crlf) {
        cgul_crlf_file__delete(crlf);
    }
    if (ws) {
        cgul_wstring__delete(ws);
    }

    return;
}

/**
 * Print a wide-character string exposing the Unicode code points so
 * that the keys can be visually compared for correct ordering.
 *
 * @param[in] cex c-style exception
 * @param[in] ws wide-character string
 */
static void
print_key(cgul_exception_t* cex,
          cgul_wchar_t* ws)
{
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    for (i = 0 ; ws[i] ; ++i) {
        if (i) {
            printf(" ");
        }
        printf("%lu", (unsigned long)ws[i]);
    }
    puts("");

 out:
    return;
}

/**
 * Print the keys in the red-black tree <code>t</code> exposing the
 * Unicode code points so that the keys can be visually compared for
 * correct ordering.
 *
 * @param[in] cex c-style exception
 * @param[in] t tree
 */
static void
print_keys(cgul_exception_t* cex,
           cgul_rbtree_t t)
{
    cgul_wchar_t* ws = NULL;
    cgul_rbtree_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    /* Iterate over each key. */
    n = cgul_rbtree__get_front(cex, t);
    for ( ; n ; n = cgul_rbtree_node__get_next(cex, n)) {
        ws = cgul_rbtree_node__get_key(cex, n);
        print_key(cex, ws);
    }

 out:
    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_rbtree_t t = NULL;
    cgul_exception_t local = NULL;

    /* Allocate. */
    t = cgul_rbtree__new(&local, (cgul_rbtree__compare_t)cgul_wchar__wcscmp);
    if (local) {
        goto out;
    }

    /* Load the tree.  Each key in the tree will be a line from the
     * text file. */
    load_tree(&local, t, "tower_of_babel.txt");
    if (local) {
        goto out;
    }

    /* Print the lines to visually verify they are sorted. */
    print_keys(&local, t);
    if (local) {
        goto out;
    }

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (t) {
        cgul_rbtree__free_keys(t);
        cgul_rbtree__delete(t);
    }

    return rv;
}
